Package: seqsetvis
Type: Package
Title: Set Based Visualizations for Next-Gen Sequencing Data
Version: 1.32.0
Authors@R: person("Joseph R", "Boyd", email = "jrboyd@uvm.edu",
                  role = c("aut", "cre"), comment = c(ORCID="0000-0002-8969-9676"))
Description: seqsetvis enables the visualization and analysis of sets
        of genomic sites in next gen sequencing data. Although
        seqsetvis was designed for the comparison of mulitple ChIP-seq
        samples, this package is domain-agnostic and allows the
        processing of multiple genomic coordinate files (bed-like
        files) and signal files (bigwig files pileups from bam file).
        seqsetvis has multiple functions for fetching data from regions
        into a tidy format for analysis in data.table or tidyverse and
        visualization via ggplot2.
License: MIT + file LICENSE
Encoding: UTF-8
Suggests: BiocFileCache, BiocManager, BiocStyle, ChIPpeakAnno,
        GenomeInfoDb, covr, knitr, rmarkdown, testthat
Depends: R (>= 4.3), ggplot2
Imports: cowplot, data.table, eulerr, Seqinfo, GenomicAlignments,
        GenomicRanges, ggplotify, grDevices, grid, IRanges, limma,
        methods, pbapply, pbmcapply, png, RColorBrewer, Rsamtools,
        rtracklayer, S4Vectors, scales, stats, UpSetR
RoxygenNote: 7.3.3
Roxygen: list(markdown = TRUE)
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-30 06:29:48 UTC; root
biocViews: Software, ChIPSeq, MultipleComparison, Sequencing,
        Visualization
Config/pak/sysreqs: cmake make libbz2-dev liblzma-dev libpng-dev
        libuv1-dev libxml2-dev libssl-dev xz-utils zlib1g-dev
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 12:47:37 UTC
RemoteUrl: https://github.com/bioc/seqsetvis
RemoteRef: RELEASE_3_23
RemoteSha: 5c54cba0c5280d10b12fa2b088b441ada81b793c
Author: Joseph R Boyd [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-8969-9676>)
Maintainer: Joseph R Boyd <jrboyd@uvm.edu>
Built: R 4.6.0; ; 2026-04-30 07:22:52 UTC; unix

# Generated by roxygen2: do not edit by hand

export(add_cluster_annotation)
export(append_ynorm)
export(applyMovingAverage)
export(applySpline)
export(assemble_heatmap_cluster_bars)
export(calc_norm_factors)
export(centerAtMax)
export(centerFixedSizeGRanges)
export(centerGRangesAtMax)
export(clusteringKmeans)
export(clusteringKmeansNestedHclust)
export(col2hex)
export(collapse_gr)
export(convert_collapsed_coord)
export(copy_clust_info)
export(crossCorrByRle)
export(easyLoad_FUN)
export(easyLoad_IDRmerged)
export(easyLoad_bed)
export(easyLoad_broadPeak)
export(easyLoad_narrowPeak)
export(easyLoad_seacr)
export(expandCigar)
export(findMaxPos)
export(fragLen_calcStranded)
export(fragLen_fromMacs2Xls)
export(getReadLength)
export(get_mapped_reads)
export(ggellipse)
export(harmonize_seqlengths)
export(make_clustering_matrix)
export(merge_clusters)
export(prepare_fetch_GRanges)
export(prepare_fetch_GRanges_names)
export(prepare_fetch_GRanges_width)
export(quantileGRangesWidth)
export(reorder_clusters_hclust)
export(reorder_clusters_manual)
export(reorder_clusters_stepdown)
export(reverse_clusters)
export(safeBrew)
export(split_cluster)
export(ssvAnnotateSubjectGRanges)
export(ssvConsensusIntervalSets)
export(ssvFactorizeMembTable)
export(ssvFeatureBars)
export(ssvFeatureBinaryHeatmap)
export(ssvFeatureEuler)
export(ssvFeaturePie)
export(ssvFeatureUpset)
export(ssvFeatureVenn)
export(ssvFetchBam)
export(ssvFetchBamPE)
export(ssvFetchBamPE.RNA)
export(ssvFetchBigwig)
export(ssvFetchGRanges)
export(ssvFetchSignal)
export(ssvMakeMembTable)
export(ssvOverlapIntervalSets)
export(ssvSignalBandedQuantiles)
export(ssvSignalClustering)
export(ssvSignalHeatmap)
export(ssvSignalHeatmap.ClusterBars)
export(ssvSignalLineplot)
export(ssvSignalLineplotAgg)
export(ssvSignalScatterplot)
export(viewGRangesWinSample_dt)
export(viewGRangesWinSummary_dt)
export(within_clust_sort)
exportMethods(ssvAnnotateSubjectGRanges)
exportMethods(ssvMakeMembTable)
import(GenomicAlignments)
import(GenomicRanges)
import(Rsamtools)
import(S4Vectors)
import(Seqinfo)
import(UpSetR)
import(cowplot)
import(data.table, except = c(shift, first, second, last))
import(eulerr)
import(ggplot2)
import(ggplotify)
import(methods)
import(pbapply)
import(pbmcapply)
import(png)
import(rtracklayer)
importFrom(RColorBrewer,brewer.pal)
importFrom(RColorBrewer,brewer.pal.info)
importFrom(Rsamtools,ScanBamParam)
importFrom(Rsamtools,scanBamHeader)
importFrom(Seqinfo,seqlengths)
importFrom(Seqinfo,seqlevels)
importFrom(grDevices,col2rgb)
importFrom(grDevices,gray)
importFrom(grDevices,rainbow)
importFrom(grDevices,rgb)
importFrom(grid,rasterGrob)
importFrom(limma,vennCounts)
importFrom(pbapply,pblapply)
importFrom(pbmcapply,pbmclapply)
importFrom(scales,hue_pal)
importFrom(stats,dist)
importFrom(stats,hclust)
importFrom(stats,kmeans)
importFrom(stats,quantile)
importFrom(stats,spline)
importFrom(stats,weighted.mean)

[![Coverage Status](https://img.shields.io/codecov/c/github/jrboyd/seqsetvis/master.svg)](https://codecov.io/github/jrboyd/seqsetvis?branch=master)

See the [vignette](https://bioconductor.org/packages/release/bioc/vignettes/seqsetvis/inst/doc/seqsetvis_overview.html) for an overview of most common functions and usage examples.
# Installation

## From Bioconductor

```{r bioc install, eval=FALSE}
## try http:// if https:// URLs are not supported
if (!requireNamespace("BiocManager", quietly=TRUE))
    install.packages("BiocManager")
BiocManager::install("seqsetvis")
```

## From github
Install devtools if needed
```{r devtools check, eval=FALSE}
if(!require(devtools)){
    install.packages("devtools")    
}
```

Install *seqsetvis* from github
```{r install seqsetvis from github, eval=FALSE}
devtools::install_github("jrboyd/seqsetvis")
```

#' ssvFetchBam for paired-end ChIP-seq files.
#' Only concordant reads are considered, but this has been minimally tested, please verify.
#'
#' Iterates a character vector (ideally named) and calls
#' \code{ssvFetchBamPE.single} on each.  Appends grouping variable to each
#' resulting data.table and uses rbindlist to efficiently combine results
#'
#' #' In contrast to ssvFetchBam, extension of reads to estimated fragment size is
#' not an issue as each read pair represents a fragment of exact size.
#'
#' \code{ssvFetchBamPE} iteratively calls \code{fetchWindowedBam.single}. See
#' \code{\link{ssvFetchBamPE.single}} for more info.
#' @export
#' @param file_paths character vector of file_paths to load from. Alternatively,
#' file_paths can be a data.frame or data.table whose first column is a
#' character vector of paths and additial columns will be used as metadata.
#' @param qgr Set of GRanges to query.  For valid results the width of each
#'   interval should be identical and evenly divisible by \code{win_size}.
#' @param unique_names names to use in final data.table to designate source
#'   bigwig. Default is 'sample'
#' @param win_size The window size that evenly divides widths in \code{qgr}.
#' @param win_method character.  one of c("sample", "summary").  Determines if
#'   \code{\link{viewGRangesWinSample_dt}} or
#'   \code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
#'   qgr.
#' @param summary_FUN function.  only relevant if win_method is "summary".
#'   passed to \code{\link{viewGRangesWinSummary_dt}}.
#' @param fragLens never used by ssvFetchBamPE Ignore.
#' @param anchor character, one of c("center", "center_unstranded", "left",
#'   "left_unstranded")
#' @param names_variable The column name where unique_names are stored.
#' @param return_data.table logical. If TRUE the internal data.table is returned
#'   instead of GRanges.  Default is FALSE.
#' @param max_dupes numeric >= 1.  duplicate reads by strandd start position
#'   over this number are removed, Default is Inf.
#' @param n_cores integer number of cores to use.
#' @param n_region_splits integer number of splits to apply to qgr. The query
#'   GRanges will be split into this many roughly equal parts for increased
#'   parallelization. Default is 1, no split.
#' @param min_isize integer. Read pairs must have an isize greater than or equal to this value.  Default is 1.
#' @param max_isize integer. Read pairs must have an isize less than or equal to this value.  Default is Inf.
#' @param return_unprocessed boolean. if TRUE returns read alignment in data.table. Default is FALSE.
#' @param return_fragSizes boolean. if TRUE returns fragment sizes for all reads per region.
#' @param force_skip_centerFix boolean, if TRUE all query ranges will be
#' used "as is".  This is already the case by default if win_method == "summary"
#' but may have applications where win_method == "sample".
#' @param ... passed to Rsamtools::ScanBamParam()
#' Uses mc.cores option if not supplied.
#' @return A tidy formatted GRanges (or data.table if specified) containing
#'   fetched values.
#' @rawNamespace import(data.table, except = c(shift, first, second, last))
#' @details if \code{qgr} contains the range chr1:1-100 and \code{win_size} is
#'   10, values from positions chr1 5,15,25...85, and 95 will be retrieved from
#'   \code{bw_file}
#' @importFrom stats weighted.mean
#' @examples
#' if(Sys.info()['sysname'] != "Windows"){
#' library(GenomicRanges)
#' bam_f = system.file("extdata/Bcell_PE.mm10.bam",
#'     package = "seqsetvis", mustWork = TRUE)
#' bam_files = c("a" = bam_f, "b" = bam_f)
#' data("Bcell_peaks")
#' qgr = Bcell_peaks
#' bw_gr = ssvFetchBamPE(bam_files, qgr, win_size = 10)
#' bw_gr2 = ssvFetchBamPE(as.list(bam_files), qgr, win_size = 10)
#'
#' bw_dt = ssvFetchBamPE(bam_files, qgr, win_size = 10,
#'     return_data.table = TRUE)
#' }
ssvFetchBamPE = function(file_paths,
                         qgr,
                         unique_names = NULL,
                         win_size = 50,
                         win_method = c("sample", "summary")[1],
                         summary_FUN = stats::weighted.mean,
                         fragLens = "not_used",
                         anchor = c("left", "left_unstranded", "center",
                                    "center_unstranded")[3],
                         names_variable = "sample",
                         return_data.table = FALSE,
                         max_dupes = Inf,
                         n_cores = getOption("mc.cores", 1),
                         n_region_splits = 1,
                         min_isize = 1,
                         max_isize = Inf,
                         return_unprocessed = FALSE,
                         return_fragSizes = FALSE,
                         force_skip_centerFix = FALSE,
                         ...){
    load_bamPE = function(f, nam, qgr) {
        message("loading ", f, " ...")
        if(!file.exists(paste0(f, ".bai"))){
            warning("creating index for ", f)
            Rsamtools::indexBam(f)
        }
        dt = ssvFetchBamPE.single(
            bam_f = f,
            qgr = qgr,
            win_size = win_size,
            win_method = win_method,
            summary_FUN = summary_FUN,
            anchor = anchor,
            return_data.table = TRUE,
            max_dupes = max_dupes,
            min_isize = min_isize,
            max_isize = max_isize,
            return_unprocessed = return_unprocessed,
            return_fragSizes = return_fragSizes,
            force_skip_centerFix = force_skip_centerFix,
            ...)
        # dt[[names_variable]] = rep(nam, nrow(dt))
        message("finished loading ", nam, ".")
        dt
    }

    bdt = ssvFetchSignal(file_paths = file_paths,
                         qgr = qgr,
                         load_signal = load_bamPE,
                         unique_names = unique_names,
                         names_variable = names_variable,
                         win_size = win_size,
                         win_method = win_method,
                         return_data.table = TRUE,
                         n_cores = n_cores,
                         n_region_splits = n_region_splits,
                         force_skip_centerFix = force_skip_centerFix)


    if(!return_data.table & !return_unprocessed & !return_fragSizes){
        bdt = GRanges(bdt)
    }
    bdt
}

#' fetch a windowed version of a paired-end bam file, returns GRanges
#' In contrast to ssvFetchBam, extension of reads to estimated fragment size is
#' not an issue as each read pair represents a fragment of exact size.
#'
#'
#' @param bam_f character or BamFile to load
#' @param qgr GRanges regions to fetchs
#' @param win_size numeric >=1.  pileup grabbed every win_size bp for win_method
#'   sample.  If win_method is summary, this is the number of windows used
#'   (confusing, sorry).
#' @param win_method character.  one of c("sample", "summary").  Determines if
#'   \code{\link{viewGRangesWinSample_dt}} or
#'   \code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
#'   qgr.
#' @param summary_FUN function.  only relevant if win_method is "summary".
#'   passed to \code{\link{viewGRangesWinSummary_dt}}.
#' @param anchor character, one of c("center", "center_unstranded", "left",
#'   "left_unstranded")
#' @param return_data.table logical. If TRUE the internal data.table is returned
#'   instead of GRanges.  Default is FALSE.
#' @param max_dupes numeric >= 1.  duplicate reads by strandd start position
#'   over this number are removed, Default is Inf.
#' @param min_isize integer. Read pairs must have an isize greater than or equal to this value.  Default is 1.
#' @param max_isize integer. Read pairs must have an isize less than or equal to this value.  Default is Inf.
#' @param return_unprocessed boolean. if TRUE returns read alignment in data.table. Default is FALSE.
#' @param return_fragSizes boolean. if TRUE returns fragment sizes for all reads per region.
#' @param force_skip_centerFix boolean, if TRUE all query ranges will be
#' used "as is".  This is already the case by default if win_method == "summary"
#' but may have applications where win_method == "sample".
#' @param ... passed to Rsamtools::ScanBamParam()
#' @return tidy GRanges (or data.table if specified) with pileups from bam file.
#'   pileup is calculated only every win_size bp.
#' @importFrom stats weighted.mean
ssvFetchBamPE.single = function(bam_f,
                                qgr,
                                win_size = 50,
                                win_method = c("sample", "summary")[1],
                                summary_FUN = stats::weighted.mean,
                                anchor = c("left", "left_unstranded", "center",
                                           "center_unstranded")[3],
                                return_data.table = FALSE,
                                max_dupes = Inf,
                                min_isize = 1,
                                max_isize = Inf,
                                return_unprocessed = FALSE,
                                return_fragSizes = FALSE,
                                force_skip_centerFix = FALSE,
                                ...) {
    x = id = y = NULL
    stopifnot(is.character(win_method))
    stopifnot(length(win_method) == 1)
    stopifnot(is(qgr, "GRanges"))
    stopifnot(win_method %in% c("sample", "summary"))
    stopifnot(is.function(summary_FUN))
    stopifnot(anchor %in% c("left", "left_unstranded", "center",
                            "center_unstranded"))
    if(return_unprocessed){
        score_gr = fetchBamPE(bam_f,
                              qgr,
                              max_dupes,
                              min_isize,
                              max_isize,
                              return_unprocessed = return_unprocessed,
                              ...)
        score_gr = merge(score_gr, data.table(which_label = as.character(qgr), id = names(qgr)), by = "which_label")
        return(score_gr)
    }
    if(return_fragSizes){
        score_gr = fetchBamPE(bam_f,
                              qgr,
                              max_dupes,
                              min_isize,
                              max_isize,
                              return_fragSizes = return_fragSizes,
                              ...)
        score_gr = merge(score_gr, data.table(which_label = as.character(qgr), id = names(qgr)), by = "which_label")
        return(score_gr)
    }
    switch (
        win_method,
        sample = {
            qgr = prepare_fetch_GRanges_width(qgr, win_size, skip_centerFix = force_skip_centerFix)
            score_gr = fetchBamPE(bam_f,
                                  qgr,
                                  max_dupes,
                                  min_isize,
                                  max_isize,
                                  ...)
            out = viewGRangesWinSample_dt(score_gr,
                                          qgr,
                                          win_size,
                                          anchor = anchor)
        },
        summary = {
            score_gr = fetchBamPE(bam_f,
                                  qgr,
                                  max_dupes,
                                  min_isize,
                                  max_isize,
                                  ...)
            out = viewGRangesWinSummary_dt(score_gr, qgr, win_size,
                                           summary_FUN = summary_FUN,
                                           anchor = anchor)
        }
    )

    out = out[order(x)][order(id)][order(strand)]

    if(!return_data.table){
        out = GRanges(out)
    }
    return(out)
}

fetchBamPE = function(bam_f,
                      qgr,
                      max_dupes = Inf,
                      min_isize = 1,
                      max_isize = Inf,
                      return_unprocessed = FALSE,
                      return_fragSizes = FALSE,
                      ...
){
    isize = paired = qname = which_label = start_min = end_max = NULL #reserve bindings
    stopifnot(is.numeric(min_isize))
    stopifnot(is.numeric(max_isize))
    stopifnot(min_isize > 0)
    stopifnot(max_isize >= min_isize)
    sbgr = qgr
    strand(sbgr) = "*"
    sbParam = Rsamtools::ScanBamParam(
        which = sbgr,
        what = Rsamtools::scanBamWhat(),
        ...)

    bam_raw = Rsamtools::scanBam(bam_f, param = sbParam)

    bam_dt = lapply(bam_raw, function(x){
        data.table(seqnames = x$rname,
                   strand = x$strand,
                   start = x$pos,
                   width = x$qwidth,
                   cigar = x$cigar,
                   qname = x$qname,
                   flag = x$flag,
                   mapq = x$mapq,
                   mrnm = x$mrnm,
                   mpos = x$mpos,
                   isize = x$isize,
                   seq = as.character(x$seq),
                   qual = as.character(x$qual))
    })

    bam_dt = data.table::rbindlist(bam_dt,
                                   use.names = TRUE,
                                   idcol = "which_label")
    if(return_fragSizes){
        bam_dt = bam_dt[isize > 0]
        bam_dt[, c("start_min", "end_max") := tstrsplit(which_label, "[:-]", keep = 2:3)]
        bam_dt = bam_dt[start > start_min & start + width < end_max]
        return(bam_dt[, list(which_label, fragment_size = isize)])
    }
    if(return_unprocessed){
        return(bam_dt)
    }

    # ggplot(bam_dt[isize >= 0, ], aes(x = isize)) + geom_histogram() + facet_wrap("which_label")

    bam_dt = bam_dt[!is.na(width)]
    bam_dt[, paired := .N == 2, list(qname, which_label)]
    bam_dt = bam_dt[paired == TRUE]
    bam_dt = bam_dt[isize >= min_isize & isize <= max_isize]

    #prepare for GRanges conversion
    if(nrow(bam_dt) > 1){
        bam_dt = bam_dt[, list(
            which_label,
            seqnames = seqnames,
            strand = "*",
            start = start,
            width = isize,
            end = start + isize - 1L
        )]
    }else{
        bam_dt = bam_dt[, list(
            which_label,
            seqnames = seqnames,
            strand = character(),
            start = start,
            width = isize,
            end = start + isize - 1L
        )]
    }

    if(max_dupes < Inf){
        bam_dt = .rm_dupesPE(bam_dt, max_dupes)
    }

    ext_cov = coverage(split(GRanges(bam_dt), bam_dt$which_label))
    score_gr = GRanges(ext_cov)
    score_gr = harmonize_seqlengths(score_gr, bam_f)

    if(length(score_gr) == 0){
        score_gr = sbgr
        mcols(score_gr) = NULL
        score_gr$score = 0
    }

    score_gr
}

#' Remove duplicate paired-end reads based on start and end position.  This is
#' an over-simplification.  For better duplicate handling, duplicates must be
#' marked in bam and flag passed to fetchBamPE() ... for ScanBamParam
#'
#' flag = scanBamFlag(isDuplicate = FALSE)
#'
#' @param reads_dt data.table of reads as loaded by fetchBamPE
#' @param max_dupes maximum allowed positional duplicates
#'
#' @return reads_dt with duplicated reads over max_dupes removed
.rm_dupesPE = function(reads_dt, max_dupes){
    ndupe = which_label = NULL
    reads_dt[, ndupe := 1L]
    reads_dt[,
             ndupe := seq_len(.N),
             by = list(which_label, start, end)]
    reads_dt = reads_dt[ndupe <= max_dupes]
    reads_dt$ndupe = NULL
    reads_dt
}

testthat::context("ssvFetchBamPE")

library(seqsetvis)
library(GenomicRanges)
library(data.table)
library(testthat)
pe_file = system.file("extdata/Bcell_PE.mm10.bam", package = "seqsetvis", mustWork = TRUE)

data("Bcell_peaks")
qgr = Bcell_peaks

pe_raw = ssvFetchBamPE(
    pe_file,
    qgr,
    return_data.table = TRUE,
    win_size = 50,
    return_unprocessed = TRUE
)

exp_cn = c("which_label", "seqnames",
           "qname", "strand", "start",
           "width", "cigar", "isize", "id",
           "sample", "flag", "mapq",
           "mrnm", "mpos", "seq", "qual")

test_that("ssvFetchBamPE return_unprocessed", {
    expect_setequal(
        colnames(pe_raw),
        exp_cn
    )
})

pe_dt = ssvFetchBamPE(pe_file,
                      qgr,
                      return_data.table = TRUE,
                      win_size = 50)

test_that("ssvFetchBamPE normal", {
    expect_equal(
        colnames(pe_dt),
        c(
            "seqnames",
            "start",
            "end",
            "width",
            "strand",
            "id",
            "y",
            "x",
            "sample"
        )
    )
})

pe_raw = ssvFetchBamPE(
    pe_file,
    qgr,
    return_data.table = TRUE,
    win_size = 50,
    return_fragSizes = TRUE
)


test_that("viewGRangesWinSample_dt strand and position functions", {
    bam_score = seqsetvis:::fetchBamPE(pe_file, qgr = qgr)
    score_gr = bam_score
    window_size = 50
    anchor = "center"
    qgr_stranded = qgr
    GenomicRanges::strand(qgr_stranded) = c(rep("+", 2), rep("-", 2))

    b_dt_center = viewGRangesWinSample_dt(bam_score, qgr_stranded, 50, anchor = "center")

    b_dt_center_uns = viewGRangesWinSample_dt(bam_score, qgr_stranded, 50, anchor = "center_unstranded")

    b_dt_left = viewGRangesWinSample_dt(bam_score, qgr_stranded, 50, anchor = "left")

    b_dt_left_uns = viewGRangesWinSample_dt(bam_score, qgr_stranded, 50, anchor = "left_unstranded")

    b_dt = rbindlist(list(center = b_dt_center,
                          center_unstranded = b_dt_center_uns,
                          left = b_dt_left,
                          left_unstranded = b_dt_left_uns), use.names = TRUE, idcol = "group")
    b_dt[, facet_label := paste(id, strand)]
    cowplot::plot_grid(
        ggplot(b_dt[grepl("center", group)]) + geom_path(aes(x = x, y = y, color = group)) + facet_wrap("facet_label") + labs(title = "centered"),
        ggplot(b_dt[!grepl("center", group)]) + geom_path(aes(x = x, y = y, color = group)) + facet_wrap("facet_label") + labs(title = "left")

    )


    #verify stranded equal for + and not equal for -
    expect_true(all(b_dt_center[strand == "+"]$x == b_dt_center_uns[strand == "+"]$x))
    expect_true(all(!b_dt_center[strand == "-"]$x == b_dt_center_uns[strand == "-"]$x))
    expect_true(all(b_dt_left[strand == "+"]$x == b_dt_left_uns[strand == "+"]$x))
    expect_true(all(!b_dt_left[strand == "-"]$x == b_dt_left_uns[strand == "-"]$x))
    #verify center and left not equal
    expect_true(all(!b_dt_center$x == b_dt_left$x))
    expect_true(all(!b_dt_center_uns$x == b_dt_left_uns$x))
    #verify center and left ARE equal if minus min
    expect_true(all(b_dt_center$x - min(b_dt_center$x) == b_dt_left$x - min(b_dt_left$x)))
    expect_true(all(b_dt_center_uns$x - min(b_dt_center_uns$x) == b_dt_left_uns$x - min(b_dt_left_uns$x)))
})


# pe_raw[isize > 0, mean(isize), .(id)]
# pe_raw$id = factor(pe_raw$id, levels = qgr$name)
# ggplot(pe_raw[isize > 0, ], aes(x = isize)) + geom_histogram() + facet_wrap("id")
# plist_pe = lapply(levels(pe_raw$id), function(qid) {
#     ggplot(pe_raw[id == qid &
#                       !is.na(strand)], aes(x = pos)) +
#         geom_histogram(bins = 30) +
#         facet_wrap("strand", ncol = 1) +
#         labs(title = qid)
#
# })
# cowplot::plot_grid(plotlist = plist_pe)

#
# se_dt_base = ssvFetchBam(
#     pe_file,
#     qgr,
#     return_data.table = TRUE,
#     win_size = 50,
#     fragLens = NA
# )
# se_dt_base[, y := y / 2]
# se_dt = ssvFetchBam(
#     pe_file,
#     qgr,
#     return_data.table = TRUE,
#     target_strand = "both",
#     win_size = 50,
#     fragLens = NA,
#     flip_strand = TRUE
# )
# pe_dt = ssvFetchBamPE(pe_file, qgr, return_data.table = TRUE, win_size = 50)
# pe_dt$strand = "pe"
#
# dt = rbindlist(
#     list(
#         base = se_dt_base,
#         stranded = se_dt,
#         pe = pe_dt
#     ),
#     use.names = TRUE,
#     idcol = "group"
# )
# dt$id = factor(dt$id, levels = qgr$name)
# ggplot(dt, aes(x = x, y = y, color = strand)) + geom_path() + facet_wrap("id")#facet_grid("group~id")

# ssvFetchBam(pe_file, qgr, win_size = 50,
#             target_strand = "both", return_data.table = TRUE, flip_strand = TRUE)

# strand_dt1 = ssvFetchBam(pe_file, qgr, fragLens = NA, win_size = 50,
#                         target_strand = "both", return_data.table = TRUE, flip_strand = FALSE)
# p1 = ggplot(strand_dt1, aes(x = x, y = y, color = strand)) + geom_path() + facet_wrap("id")+
#     labs(title = "noflip NA") + scale_color_manual(values = c("+" = "red", "-" = "blue"))
# strand_dt2 = ssvFetchBam(pe_file, qgr, fragLens = 138, win_size = 50,
#                         target_strand = "both", return_data.table = TRUE, flip_strand = FALSE)
# p2 = ggplot(strand_dt2, aes(x = x, y = y, color = strand)) + geom_path() + facet_wrap("id")+
#     labs(title = "noflip 138") + scale_color_manual(values = c("+" = "red", "-" = "blue"))
#
# strand_dt3 = ssvFetchBam(pe_file, qgr, fragLens = NA, win_size = 50,
#                         target_strand = "both", return_data.table = TRUE, flip_strand = TRUE)
# p3 = ggplot(strand_dt3, aes(x = x, y = y, color = strand)) + geom_path() + facet_wrap("id")+
#     labs(title = "flip NA") + scale_color_manual(values = c("+" = "red", "-" = "blue"))
# strand_dt4 = ssvFetchBam(pe_file, qgr, fragLens = 138, win_size = 50,
#                         target_strand = "both", return_data.table = TRUE, flip_strand = TRUE)
#
# p4 = ggplot(strand_dt4, aes(x = x, y = y, color = strand)) + geom_path() + facet_wrap("id") +
#     labs(title = "flip 138") + scale_color_manual(values = c("+" = "red", "-" = "blue"))
# cowplot::plot_grid(p1, p2, p3, p4)
#
# plist_noflip = lapply(seq_along(qgr), function(i){
#     gr = qgr[i]
#     p = seqsetvis::fragLen_calcStranded(
#         pe_file,
#         gr,
#         flip_strand = FALSE,
#         test_fragLen = seq(0, 600, by = 50),
#         include_plot_in_output = TRUE
#     )[[2]]
#     p + labs(title = qgr$name[i])
# })
# plist_flip = lapply(seq_along(qgr), function(i){
#     gr = qgr[i]
#     res = seqsetvis::fragLen_calcStranded(
#         pe_file,
#         gr,
#         flip_strand = TRUE,
#         test_fragLen = seq(0, 600, by = 50),
#         include_plot_in_output = TRUE
#     )
#     p = res[[2]]
#     p + labs(title = qgr$name[i])
# })
#
# shift_dt = crossCorrByRle(pe_file, qgr, flip_strand = TRUE)
# ggplot(shift_dt, aes(x = shift, y = correlation)) + geom_path() + facet_wrap("id")

# cowplot::plot_grid(plotlist = plist_flip)
# cowplot::plot_grid(plotlist = plist_noflip)

testthat::context("force_skip_centerFix")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
library(data.table)

qgr = CTCF_in_10a_overlaps_gr[1:5]
k = unique(as.character(seqnames(qgr)))
seqlevels(qgr) = k
qgr = centerFixedSizeGRanges(qgr, 500)

exp_factor = 100
qgr = resize(qgr, seq_along(qgr)*exp_factor, fix = "center")
#bed used to intersect bam
# rtracklayer::export.bed(qgr, con = "ctcf_5.bed")
bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
bw_file = system.file("extdata/test_loading.bw", package = "seqsetvis", mustWork = TRUE)
win_size = 50


test_that("ssvFetchBam - force_skip_centerFix", {
    bam_dt = ssvFetchBam(bam_file, qgr,
                         win_size = win_size,
                         force_skip_centerFix = TRUE,
                         return_data.table = TRUE, fragLens = 200)
    size_dt = bam_dt[, .(region_width = diff(range(x)) + win_size), .(id)]

    is_exp = size_dt[, region_width / as.numeric(id)] == exp_factor
    expect_true(all(is_exp))
})

test_that("ssvFetchBigwig - force_skip_centerFix", {
    skip_on_os("windows")
    pos = c(20, 180, 210, 440, 520, 521)
    region_size = 30
    test_qgr = GRanges("chrTest", IRanges(pos+1, pos + region_size))
    test_qgr = resize(test_qgr, seq_along(test_qgr)*region_size, fix = "center")
    names(test_qgr) = seq_along(test_qgr)
    bw_dt = ssvFetchBigwig(bw_file, test_qgr,
                         win_size = 30,
                         force_skip_centerFix = TRUE,
                         return_data.table = TRUE)
    size_dt = bw_dt[, .(region_width = diff(range(x)) + 30), .(id)]

    is_exp = size_dt[, region_width / as.numeric(id)] == region_size
    expect_true(all(is_exp))
})

test_that("ssvFetchBamPE - force_skip_centerFix", {
    pe_file = system.file("extdata/Bcell_PE.mm10.bam", package = "seqsetvis", mustWork = TRUE)
    data("Bcell_peaks")
    qgrPE = Bcell_peaks
    qgrPE$name = seq_along(qgrPE)
    qgrPE = resize(qgrPE, seq_along(qgrPE)*exp_factor, fix = "center")
    bam_dt = ssvFetchBamPE(pe_file, qgrPE,
                         win_size = win_size,
                         force_skip_centerFix = TRUE,
                         return_data.table = TRUE)
    size_dt = bam_dt[, .(region_width = diff(range(x)) + win_size), .(id)]
    is_exp = size_dt[, region_width / as.numeric(id)] == exp_factor
    expect_true(all(is_exp))

})

test_that("ssvFetchGranges - force_skip_centerFix", {
    qgrGR = CTCF_in_10a_overlaps_gr[1:5]
    qgrGR = resize(qgrGR, seq_along(qgrGR)*exp_factor, fix = "center")
    gr_dt = ssvFetchGRanges(CTCF_in_10a_overlaps_gr[1:5], qgrGR,
                           win_size = win_size,
                           force_skip_centerFix = TRUE,
                           return_data.table = TRUE)
    size_dt = gr_dt[, .(region_width = diff(range(x)) + win_size), .(id)]
    is_exp = size_dt[, region_width / as.numeric(id)] == exp_factor
    expect_true(all(is_exp))
})

#' converts a valid r color name ("black", "red", "white", etc.) to a hex value
#' @export
#' @param color_name character. one or more r color names.
#' @importFrom grDevices col2rgb rgb
#' @return hex value of colors coded by colors()
#' @examples
#' col2hex(c("red", "green", "blue"))
#' col2hex(c("lightgray", "gray", "darkgray"))
col2hex = function(color_name) {
    cnames = names(color_name)
    stopifnot(is.character(color_name))
    out = grDevices::rgb(t(grDevices::col2rgb(color_name))/255)
    names(out) = cnames
    out
}

#' convert a list of sets, each list item should be a character vector
#' denoting items in sets
#' @param set_list a list of character vectors.  default names will be added if
#' missing
#' @return converts list of characters/numeric to membership table matrix
set_list2memb = function(set_list) {
    stopifnot(is.list(set_list))
    if (is.null(names(set_list))) {
        names(set_list) = paste0("set_", LETTERS[seq_along(set_list)])
    }
    rn = unique(unlist(set_list))
    if(any(is.na(rn))){
        warning("Removing NA values from input. Is this correct?")
        rn = rn[!is.na(rn)]
        set_list = lapply(set_list, function(x){
            x[!is.na(x)]
        })
    }
    cn = names(set_list)
    memb = matrix(FALSE, nrow = length(rn), ncol = length(cn))
    rownames(memb) = rn
    colnames(memb) = cn
    for (column in cn) {
        memb[set_list[[column]], column] = TRUE
    }
    return(memb)
}

#' safeBrew
#'
#' Allows RColorBrew to handle n values less than 3 and greater than 8 without
#' warnings and return expected number of colors.
#'
#' For convenience, instead of the number n requested, n may be a character or
#' factor vector and outputs will be appropriately named for use with
#' scale_color/fill_manual.
#'
#' Additionally, accepts pal as "gg", "ggplot", or "ggplot2" to reproduce
#' default ggplot colors in the same way.
#'
#' @export
#' @param n integer value of number of colors to make palette for. Alternatively
#'   a character or factor, in which case palette will be generated for each
#'   unique item or factor level repsectively.
#' @param pal palette recognized by RColorBrewer
#' @importFrom RColorBrewer brewer.pal brewer.pal.info
#' @importFrom scales hue_pal
#' @return a character vector of hex coded colors of length n from the color
#'   brewer palette pal. If n is supplied as character or factor, output will be
#'   named accordingly.
#' @examples
#' plot(1:2, rep(0, 2),  col = safeBrew(2, "dark2"), pch = 16, cex = 6)
#' plot(1:12, rep(0, 12),  col = safeBrew(12, "set1"), pch = 16, cex = 6)
#' plot(1:12, rep(0, 12),  col = safeBrew(12, "set2"), pch = 16, cex = 6)
#' plot(1:12, rep(0, 12),  col = safeBrew(12, "set3"), pch = 16, cex = 6)
safeBrew = function(n, pal = "Dark2"){
    if(is.logical(n)){
        n = as.character(n)
    }
    if(is.numeric(n)){
        n_lev = n
    }else if(is.character(n)){
        n_lev = length(unique(n))
    }else if(is.factor(n)){
        n_lev = length(levels(n))
    }else{
        stop("n must be one of numeric, character, or factor. was:", paste(class(n), collapse = ", "))
    }

    stopifnot(is.numeric(n_lev))
    stopifnot(is.character(pal))
    if(n_lev < 1) stop("n must be at least 1")
    pal_info = RColorBrewer::brewer.pal.info
    pal_info$brewName = rownames(pal_info)
    rownames(pal_info) = tolower(rownames(pal_info))
    pal = tolower(pal)
    if(!any(pal == rownames(pal_info))){
        if(pal %in% c("gg", "ggplot", "ggplot2")){
            cols = scales::hue_pal()(n_lev)
        }else{
            stop("Palette ", pal, " not a valid RColorBrewer palette, ",
                 "see RColorBrewer::brewer.pal.info")
        }
    }else{
        maxColors = pal_info[pal,]$maxcolors
        nbrew = min(max(n_lev, 3), maxColors)
        cols = RColorBrewer::brewer.pal(
            n = nbrew,
            name = pal_info[pal,]$brewName)[(seq_len(n_lev)-1) %% maxColors + 1]
    }

    if(is.character(n)){
        names(cols) = unique(n)
    }
    if(is.factor(n)){
        names(cols) = levels(n)
    }
    cols
}

# x: the vector
# n: the number of samples centered: if FALSE, then average
# current sample and previous (n-1) samples if TRUE, then average
# symmetrically in past and
# future. (If n is even, use one more sample from future.)
# from http://www.cookbook-r.com/Manipulating_data/Calculating_a_moving_average/
movingAverage <- function(y, n = 1, centered = TRUE) {

    if (centered) {
        before <- floor((n - 1)/2)
        after <- ceiling((n - 1)/2)
    } else {
        before <- n - 1
        after <- 0
    }

    # Track the sum and count of number of non-NA items
    s <- rep(0, length(y))
    count <- rep(0, length(y))

    # Add the centered data
    new <- y
    # Add to count list wherever there isn't a
    count <- count + (!is.na(new))
    # Now replace NA_s with 0_s and add to total
    new[is.na(new)] <- 0
    s <- s + new

    # Add the data from before
    i <- 1
    while (i <= before) {
        # This is the vector with offset values to add
        new <- c(rep(NA, i), y[seq_len(length(y) - i)])

        count <- count + (!is.na(new))
        new[is.na(new)] <- 0
        s <- s + new

        i <- i + 1
    }

    # Add the data from after
    i <- 1
    while (i <= after) {
        # This is the vector with offset values to add
        new <- c(y[(i + 1):length(y)], rep(NA, i))

        count <- count + (!is.na(new))
        new[is.na(new)] <- 0
        s <- s + new

        i <- i + 1
    }

    # return sum divided by count
    s/count
}

#' applyMovingAverage
#'
#' http://www.cookbook-r.com/Manipulating_data/Calculating_a_moving_average/
#'
#' @param dt a tidy data.table containing two-dimensional data
#' @param n the number of samples centered: if FALSE, then average
#' @param centered current sample and previous (n-1) samples if TRUE, then
#'   average symmetrically in past and future. (If n is even, use one more
#'   sample from future.)
#' @param x_ the variable name of the x-values
#' @param y_ the variable name of the y-values
#' @param by_ optionally, any variables that provide grouping to the data.
#' default is none. see details.
#' @param maFun a function that accepts x, y, and n as arguments and
#' returns a list of length 2 with named elements x and y.
#'
#' @return a newly derived data.table where a movingAverage has been applied.
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' agg_dt = CTCF_in_10a_profiles_dt[, list(y = mean(y)), by = list(sample, x)]
#' ggplot(agg_dt) +
#'     geom_line(aes(x = x, y = y, color = sample))
#'
#' ma_smooth = applyMovingAverage(agg_dt, n = 5,
#'     y_ = 'y', by_ = c('sample'))
#' ggplot(ma_smooth) +
#'     geom_line(aes(x = x, y = y, color = sample))
#'
#' ma_smooth$method = "moving_average"
#' agg_dt$method = "none"
#' ggplot(rbind(ma_smooth, agg_dt)) +
#'     geom_line(aes(x = x, y = y, color = method)) +
#'     facet_wrap(~sample)
applyMovingAverage = function(dt, n, centered = TRUE, x_ = "x", y_ = "y", by_ = c("id", "sample"),
                              maFun = movingAverage){
    output_GRanges = FALSE
    if(is(dt, "GRanges")){
        dt = as.data.table(dt)
        output_GRanges = TRUE
    }
    stopifnot(data.table::is.data.table(dt))
    stopifnot(is.character(x_), is.character(y_), is.character(by_))
    stopifnot(is.function(maFun))
    if (!any(x_ == colnames(dt))) {
        stop("applyMovingAverage : x_ (", x_,
             ") not found in colnames of input data.table")
    }
    if (!any(y_ == colnames(dt))) {
        stop("applyMovingAverage : y_ (", y_,
             ") not found in colnames of input data.table")
    }
    if (by_[1] != "" | length(by_) > 1)
        if (!all(by_ %in% colnames(dt))) {
            stop("applyMovingAverage : by_ (", by_,
                 ") not found in colnames of input data.table")
        }
    max_n = floor(length(unique(dt[[x_]]))/2)
    if(n > max_n){
        stop("n is too large to be meaningful. max allowed: ", max_n, ", n: ", n)
    }
    dt = dt[order(get(x_))]
    if(by_[1] != ""){
        for(.by_ in by_){
            dt = dt[order(get(.by_))]
        }
    }

    stopifnot(n > 1)
    dupe_x_within_by = suppressWarnings(
        any(dt[, any(duplicated(get(x_))), by = by_]$V1))
    if (dupe_x_within_by)
        warning("applyMovingAverage : Duplicate values of x_ (\"", x_,
                "\") exist within groups defined with by_ (\"", by_, "\"). ",
                "This Results in averages through the means of yvalues at",
                " duplicated xs.")
    extra_cols = setdiff(colnames(dt), c(x_, y_, by_))
    # sdt = dt[, list(n = floor(.N * n)), by = by_]
    sdt = dt[, list(y = maFun(y = get(y_), n = n, centered = centered), x= get(x_)), by = by_]
    if(x_ != "x"){
        data.table::setnames(sdt, "x", x_)
    }
    if(y_ != "y"){
        data.table::setnames(sdt, "y", y_)
    }

    #append extra cols
    for(cn in by_){
        stopifnot(dt[[cn]] == sdt[[cn]])
    }
    sdt = cbind(sdt, dt[, c(extra_cols), with = FALSE])
    if(output_GRanges){
        sdt = GRanges(sdt)
    }
    return(sdt)

}

#' ggellipse
#'
#' returns a ggplot with ellipses drawn using specified parameters
#' used by ssvFeatureVenn and ssvFeatureEuler
#'
#' uses eulerr's non-exported ellipse drawing coordinate function
#'
#' @param xcentres numeric x-coord of centers of ellipses
#' @param ycentres numeric y-coord of centers of ellipses, must have same
#' length as xcentres
#' @param r numeric radius1 of ellipse, must have length of 1 or match length
#' of xcentres
#' @param r2 numeric radius2 of ellipse, must have length of 1 or match length
#' of xcentres.  same as r by default.
#' @param phi numeric phi of ellipse, must have length of 1 or match length
#' of xcentres.  0 by default.
#' @param circle_colors character of rcolors or hex colors or NULL.  if
#' null safeBrew of Dark2 is used
#' @param group_names character/factor names of color/fill groups.  capital
#' letters by default.
#' @param line_alpha numeric value from 0 to 1. alpha of lines, 1 by default
#' @param fill_alpha numeric value from 0 to 1. alpha of fill, .3 by default.
#' @param line_width numeric > 0.  passed to size. 2 by default
#' @param n_points integer > 1.  number of points to approximate circle with.
#' 200 by default
#' @return a ggplot containing ellipses
#' @export
#' @examples
#' ggellipse(xcentres = c(1, 1, 2),
#'     ycentres = c(2, 1, 1),
#'     r = c(1, 2, 1))
#' ggellipse(xcentres = c(1, 1, 2),
#'     ycentres = c(2, 1, 1),
#'     r = c(1, 2, 1),
#'     fill_alpha = 0,
#'     group_names = paste("set", 1:3))
#'ggellipse(xcentres = c(1, 1, 2),
#'     ycentres = c(2, 1, 1),
#'     r = c(1, 2, 1),
#'     circle_colors = c("red", "orange", "yellow"),
#'     line_alpha = 0,
#'     group_names = paste("set", 1:3))
ggellipse = function(xcentres,
                     ycentres,
                     r,
                     r2 = r,
                     phi = rep(0, length(xcentres)),
                     circle_colors = NULL,
                     group_names = LETTERS[seq_along(xcentres)],
                     line_alpha = 1,
                     fill_alpha = .3,
                     line_width = 2,
                     n_points = 200){
    stopifnot(length(xcentres) == length(ycentres))
    n_circles = length(xcentres)
    stopifnot(is.numeric(xcentres), is.numeric(ycentres),
              is.numeric(r), is.numeric(phi))
    stopifnot(length(r) == 1 || length(r) == n_circles)
    if(length(r) == 1) r = rep(r, n_circles)
    stopifnot(length(r2) == 1 || length(r2) == n_circles)
    if(length(r2) == 1) r2 = rep(r2, n_circles)
    stopifnot(length(phi) == 1 || length(phi) == n_circles)
    if(length(phi) == 1) phi = rep(phi, n_circles)
    stopifnot(all(is.character(group_names) | is.factor(group_names)))
    stopifnot(length(group_names) == n_circles)
    stopifnot(is.numeric(line_alpha), is.numeric(fill_alpha),
              is.numeric(line_width))

    #ellipse coordinates
    # internal function from eulerr
    eulerr_ellipse = function (h, k, a, b = a, phi = 0, n = 200L) {
        theta <- seq.int(0, 2 * pi, length.out = n)
        m <- length(h)
        out <- vector("list", m)
        for (i in seq_along(h)) {
            out[[i]]$x <- h[i] + a[i] * cos(theta) * cos(phi[i]) -
                b[i] * sin(theta) * sin(phi[i])
            out[[i]]$y <- k[i] + b[i] * sin(theta) * cos(phi[i]) +
                a[i] * cos(theta) * sin(phi[i])
        }
        out
    }

    e <- eulerr_ellipse(xcentres, ycentres, r,
                        r2, phi, n_points)
    names(e) = group_names
    x = y = group = NULL #reserve data.table bindings
    ell_dt = lapply(e, function(ell)data.table::data.table(x = ell$x,
                                                           y = ell$y))
    ell_dt = data.table::rbindlist(ell_dt, use.names = TRUE, idcol = "group")
    if(is.factor(group_names)){
        ell_dt$group = factor(ell_dt$group, levels = levels(group_names))
    }else{
        ell_dt$group = factor(ell_dt$group, levels = group_names)
    }

    #check colors
    stopifnot(is.null(circle_colors) || all(is.character(circle_colors)))

    if (is.null(circle_colors)) {
        circle_colors = safeBrew(n_circles, "Dark2")
    } else {
        not_hex = substr(circle_colors, 0, 1) != "#"
        circle_colors[not_hex] = col2hex(circle_colors[not_hex])
    }
    is_hex = substr(circle_colors, 0, 1) == "#"
    stopifnot(all(is_hex))
    circle_colors = substr(circle_colors, 1, 7)
    stopifnot(all(nchar(circle_colors) == 7))
    stopifnot(length(circle_colors) == 1 || length(circle_colors) == n_circles)
    if (length(circle_colors) < n_circles)
        circle_colors <- rep(circle_colors, length.out = n_circles)

    # make scales
    names(circle_colors) = group_names

    ahex = substr(grDevices::rgb(red = 1, blue = 1, green = 1,
                                 alpha = fill_alpha), start = 8, stop = 9)
    fill_scale = paste0(circle_colors, ahex)
    names(fill_scale) = names(circle_colors)

    ahex = substr(grDevices::rgb(red = 1, blue = 1, green = 1,
                                 alpha = line_alpha), start = 8, stop = 9)
    line_scale = paste0(circle_colors, ahex)
    names(line_scale) = names(circle_colors)
    # make plot
    p = ggplot() +
        labs(fill = "", color = "") +
        theme_minimal() +
        theme(plot.background = element_blank(),
              axis.title = element_blank(),
              axis.text = element_blank(),
              axis.ticks = element_blank(),
              panel.grid = element_blank(),
              legend.position = "top") +
        coord_fixed()
    p = p +
        geom_polygon(data = ell_dt, aes(x = x,
                                        y = y,
                                        # col = NA,
                                        fill = group
        ), color = "#00000000") +
        geom_polygon(data = ell_dt, aes(x = x,
                                        y = y,
                                        # fill = NA,
                                        col = group,
                                        linewidth = line_width
        ), fill = "#00000000") +
        scale_alpha_identity() +
        guides(alpha = "none") +
        scale_fill_manual(values = fill_scale) +
        scale_color_manual(values = line_scale) +
        scale_linewidth_identity()
    p
}

#' ssv_mclapply
#'
#' @param X For pbsapply and pblapply, a vector (atomic or list) or an
#'   expressions vector (other objects including classed objects will be coerced
#'   by as.list.) For pbapply an array, including a matrix. For pbtapply an R
#'   object for which a split method exists. Typically vector-like, allowing
#'   subsetting with "[".
#' @param FUN The function to be applied to each element of X: see apply,
#'   sapply, and lapply. In the case of functions like +, '%*%', etc., the
#'   function name must be backquoted or quoted. If FUN is NULL, pbtapply
#'   returns a vector which can be used to subscript the multi-way array
#'   pbtapply normally produces.
#' @param mc.cores Number of cores to use for pbmclapply. Defaults to option
#'   mc.cores.
#' @param ... passed to pbapply::pblapply or pbmcapply::pbmclapply
#' @return result of either pblapply or pbmclapply
#'
#' @importFrom pbapply pblapply
#' @importFrom pbmcapply pbmclapply
ssv_mclapply = function(X, FUN, mc.cores = getOption("mc.cores", 1), ...){
    if(.Platform$OS.type == "windows" || mc.cores == 1) {
        pbapply::pblapply(X = X, FUN = FUN, ...)

    } else {
        pbmcapply::pbmclapply(X = X, FUN = FUN, mc.cores = mc.cores, ...)
    }
}


#' get_mapped_reads
#'
#' @param bam_files Path to 1 or more bam files.  Must be indexed.
#'
#' @return the total mapped reads in each bam file as a named numeric vector.
#' @export
#'
#' @examples
#' bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
#' get_mapped_reads(bam_file)
get_mapped_reads = function(bam_files){
    if(is.null(names(bam_files))){
        names(bam_files) = bam_files
    }
    .get_mapped_reads.single = function(f){
        stats = Rsamtools::idxstatsBam(f)
        sum(stats[,3])
    }
    vapply(bam_files, .get_mapped_reads.single, FUN.VALUE = c(1))
}

testthat::context("safeBrew")

library(seqsetvis)
library(testthat)

test_that("safeBrew default", {
    cols = safeBrew(21)
    expect_equal(length(cols), 21)
    expect_null(names(cols))
    expect_equal(cols[1], "#1B9E77")
    expect_equal(cols[1], cols[9])
    expect_equal(cols[1], cols[17])
})

test_that("safeBrew character n", {
    cols = safeBrew(LETTERS)
    expect_equal(length(cols), length(LETTERS))
    expect_equal(names(cols), LETTERS)
    expect_equal(unname(cols[1]), "#1B9E77")
    expect_equal(unname(cols[1]), unname(cols[9]))
    expect_equal(unname(cols[1]), unname(cols[17]))
})

test_that("safeBrew rev character n", {
    cols = safeBrew(rev(LETTERS))
    expect_equal(length(cols), length(LETTERS))
    expect_equal(names(cols), rev(LETTERS))
    expect_equal(unname(cols[1]), "#1B9E77")
    expect_equal(unname(cols[1]), unname(cols[9]))
    expect_equal(unname(cols[1]), unname(cols[17]))
})

test_that("safeBrew rev factor n", {
    fLET = factor(LETTERS, levels = rev(LETTERS))
    cols = safeBrew(fLET)
    expect_equal(length(cols), length(LETTERS))
    expect_equal(names(cols), rev(LETTERS))
    expect_equal(unname(cols[1]), "#1B9E77")
    expect_equal(unname(cols[1]), unname(cols[9]))
    expect_equal(unname(cols[1]), unname(cols[17]))
})

test_that("safeBrew rev factor n", {
    c.D = safeBrew(3, pal = "Dark2")
    c.d = safeBrew(3, pal = "dark2")
    c.s = safeBrew(3, pal = "set1")
    c.g1 = safeBrew(3, pal = "gg")
    c.g2 = safeBrew(3, pal = "ggplot")
    c.g3 = safeBrew(3, pal = "ggplot2")
    expect_equal(c.D[1], "#1B9E77")
    expect_equal(c.D, c.d)
    expect_equal(c.s[1], "#E41A1C")
    expect_equal(c.g1[1], "#F8766D")
    expect_equal(c.g1, c.g2)
    expect_equal(c.g1, c.g3)
})

# functions useful for fetching signal data regardless of source:
# bam, bigwig, etc.

#' signal loading framework
#'
#' Does nothing unless load_signal is overridden to carry out reading data from
#' file_paths (likely via the appropriate ssvFetch* function, ie.
#' \code{\link{ssvFetchBigwig}} or \code{\link{ssvFetchBam}}
#'
#' @param file_paths character vector of file_paths to load from. Alternatively,
#'   file_paths can be a data.frame or data.table whose first column is a
#'   character vector of paths and additial columns will be used as metadata.
#' @param qgr GRanges of intervals to return from each file
#' @param unique_names unique file ids for each file in file_paths.  Default is
#'   names of file_paths vector
#' @param names_variable character, variable name for column containing
#'   unique_names entries.  Default is "sample"
#' @param file_attribs optional data.frame/data.table with one row per item in
#' file paths.  Each column will be a variable added to final tidy output.
#' @param win_size numeric/integer window size resolution to load signal at.
#'   Default is 50.
#' @param win_method character.  one of c("sample", "summary").  Determines if
#'   \code{\link{viewGRangesWinSample_dt}} or
#'   \code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
#'   qgr.
#' @param return_data.table logical. If TRUE data.table is returned instead of
#'   GRanges, the default.
#' @param load_signal function taking f, nam, and qgr arguments.  f is from
#'   file_paths, nam is from unique_names, and qgr is qgr. See details.
#' @param n_cores integer number of cores to use. Uses mc.cores option if not
#'   supplied.
#' @param n_region_splits integer number of splits to apply to qgr. The query
#'   GRanges will be split into this many roughly equal parts for increased
#'   parallelization. Default is 1, no split.
#' @param force_skip_centerFix boolean, if TRUE all query ranges will be used
#'   "as is".  This is already the case by default if win_method == "summary"
#'   but may have applications where win_method == "sample".
#' @details load_signal is passed f, nam, and qgr and is executed in the
#'   environment where load_signal is defined. See \code{\link{ssvFetchBigwig}}
#'   and \code{\link{ssvFetchBam}} for examples.
#' @return A GRanges with values read from file_paths at intervals of win_size.
#'   Originating file is coded by unique_names and assigned to column of name
#'   names_variable.  Output is data.table is return_data.table is TRUE.
#' @export
#' @import pbmcapply
#' @examples
#' library(GenomicRanges)
#' data(CTCF_in_10a_overlaps_gr)
#' bam_f = system.file("extdata/test.bam",
#'     package = "seqsetvis", mustWork = TRUE)
#' bam_files = c("a" = bam_f, "b" = bam_f)
#' qgr = CTCF_in_10a_overlaps_gr[1:2]
#' qgr = resize(qgr, 500, "center")
#' load_bam = function(f, nam, qgr) {
#'     message("loading ", f, " ...")
#'     dt = seqsetvis:::ssvFetchBam.single(bam_f = f,
#'                       qgr = qgr,
#'                       win_size = 50,
#'                       fragLen = NULL,
#'                       target_strand = "*",
#'                       return_data.table = TRUE)
#'
#'     data.table::set(dt, j = "sample", value = nam)
#'     message("finished loading ", nam, ".")
#'     dt
#' }
#' ssvFetchSignal(bam_files, qgr, load_signal = load_bam)
ssvFetchSignal = function(file_paths,
                          qgr,
                          unique_names = NULL,
                          names_variable = "sample",
                          file_attribs = NULL,
                          win_size = 50,
                          win_method = c("sample", "summary")[1],
                          return_data.table = FALSE,
                          load_signal = function(f, nam, qgr) {
                              warning("nothing happened, ",
                                      "supply a function to",
                                      "load_signal parameter.")
                          },
                          n_cores = getOption("mc.cores", 1),
                          n_region_splits = 1,
                          force_skip_centerFix = FALSE) {



    tmp = .get_file_attribs(file_paths, file_attribs, names_variable)
    file_paths = tmp$file_paths
    file_attribs = tmp$file_attribs
    remove(tmp)
    unique_names = .get_unique_names(unique_names, file_paths, file_attribs, names_variable)
    file_attribs[[names_variable]] = unique_names
    if (is.null(file_attribs[[names_variable]])) {
        file_attribs[[names_variable]] = unique_names
    }
    if (is.null(names(file_paths))) {
        names(file_paths) = unique_names
    }
    stopifnot(is.character(file_paths))
    stopifnot(is(qgr, "GRanges"))
    stopifnot(is.character(unique_names))
    stopifnot(is.character(names_variable))
    stopifnot(is.numeric(win_size))
    if (any(duplicated(unique_names))) {
        stop("some unique_names are duplicated:\n",
             paste(collapse = "\n",
                   unique(unique_names[duplicated(unique_names)])), "\n",
             "If you haven't manually supplied uninque_names, this is derived from",
             " file paths supplied.  \nTry supplying unique_names manually if you ",
             "intended to load duplicate files.")
    }
    stopifnot(file.exists(file_paths))
    qgr = prepare_fetch_GRanges_width(qgr = qgr,
                                      win_size = win_size,
                                      target_size = NULL,
                                      skip_centerFix = win_method != "sample" | force_skip_centerFix)
    #}
    qgr = prepare_fetch_GRanges_names(qgr, include_id = TRUE)
    stopifnot(is.numeric(n_region_splits))
    if(n_region_splits > length(qgr)){
        n_region_splits = length(qgr)
    }
    n_region_splits = round(n_region_splits)
    stopifnot(n_region_splits >= 1)
    if (win_method == "summary") {
        if(any(win_size > width(qgr))){
            warning("Some query regions (", sum(win_size > width(qgr)), " of ", length(qgr), ") are smaller than the number of windows specified.\n",
                    "Consider reducing the number of windows or increasing the width of query regions.\n",
                    "In affected regions, the same base position will be represented at multiple values of x.")
        }
    }

    if(n_region_splits == 1){
        nam_load_signal = function(nam) {
            f = file_paths[nam]
            load_signal(f, nam, qgr)
        }
        if(n_cores == 1 | length(unique_names) == 1){
            bw_list = lapply(unique_names,
                             nam_load_signal)
        }else{
            bw_list = ssv_mclapply(unique_names,
                                   nam_load_signal,
                                   mc.cores = n_cores)
        }

    }else{
        split_qgr = split(qgr, ceiling(seq_along(qgr)/length(qgr)*n_region_splits))
        file_attribs = file_attribs[rep(seq_len(nrow(file_attribs)), each = n_region_splits),, drop = FALSE]

        task_df = expand.grid(seq_along(split_qgr), unique_names)
        colnames(task_df) = c("region_var", "file_var")
        rownames(task_df) = paste(task_df$file_var, task_df$region_var)
        nam_load_signal = function(nam) {
            f = file_paths[task_df[nam,]$file_var]
            sub_qgr = split_qgr[[task_df[nam,]$region_var]]
            load_signal(f, task_df[nam,]$file_var, sub_qgr)
        }
        if(n_cores == 1){
            bw_list = lapply(rownames(task_df),
                             nam_load_signal)
        }else{
            bw_list = ssv_mclapply(rownames(task_df),
                                   nam_load_signal,
                                   mc.cores = n_cores)
        }
    }
    for (i in seq_along(bw_list)) {
        if(truelength(bw_list[[i]]) < 200){#data.table specific check
            bw_list[[i]] = setalloccol(bw_list[[i]])
        }
        for (attrib in colnames(file_attribs)) {
            if(nrow(bw_list[[i]]) == 0){
                data.table::set(bw_list[[i]], j = attrib, value = file_attribs[[attrib]][0])
            }else{
                data.table::set(bw_list[[i]], j = attrib, value = file_attribs[[attrib]][i])
            }
        }
    }
    out = data.table::rbindlist(bw_list)
    data.table::set(out, j = names_variable, value = factor(out[[names_variable]], levels = unique_names))
    if (!return_data.table) {
        out = GRanges(out)
    }
    return(out)
}

#' Creates a named version of input GRanges using the same method seqsetvis uses internally to ensure consistency.
#'
#' If $id is set, that value is used as name and duplicates are checked for.
#'
#' @param qgr input GRanges object the set/check names on
#' @param include_id if TRUE, $id is retained. Default is FALSE.
#'
#' @return and named GRanges based on input qgr.
#' @export
#'
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' qgr = CTCF_in_10a_overlaps_gr
#' names(qgr) = NULL
#' #default is to paste "region_" and iteration along length of qgr
#' prepare_fetch_GRanges_names(qgr)
#' #id gets used is already set
#' qgr$id = paste0("peak_", rev(seq_along(qgr)), "_of_", length(qgr))
#' prepare_fetch_GRanges_names(qgr)
prepare_fetch_GRanges_names = function(qgr, include_id = FALSE){
    if(length(qgr) < 1) stop("length of query GRanges was 0!")
    if(is.null(qgr$id)){#need id
        if (!is.null(qgr$name)) {
            qgr$id = qgr$name
        }else if (!is.null(names(qgr))) {
            qgr$id = names(qgr)
        }else{
            qgr$id = paste0("region_", seq_along(qgr))
        }
    }
    if(any(duplicated(qgr$id))) stop("qgr$id must be unique. qgr$id is either directly supplied or taken from qgr$name or names(qgr).")
    names(qgr) = qgr$id
    if(!include_id){
        qgr$id = NULL
    }
    qgr
}

#' get a windowed sampling of score_gr
#'
#' This method is appropriate when all GRanges in qgr are identical width and
#' when it is practical to use a window_size smaller than features in genomic
#' signal.  For instance, when retrieving signal around peaks or promoters this
#' method maintains a fixed genomic scale across regions.  This allows meaingful
#' comparison of peak widths can be made.
#'
#' Summarizes score_gr by grabbing value of "score" every window_size bp.
#' Columns in output data.table are: standard GRanges columns: seqnames, start,
#' end, width, strand id - matched to names(score_gr). if names(score_gr) is
#' missing, added as 1:length(score_gr). y - value of score from score_gr. x -
#' relative bp position.
#'
#' @param score_gr GRanges with a "score" metadata column.
#' @param qgr regions to view by window.
#' @param window_size qgr will be represented by value from score_gr every
#'   window_size bp.
#' @param attrib_var character name of attribute to pull data from. Default is
#'   "score", compatible with with bigWigs or bam coverage.
#' @param fill_value numeric or character value to use where queried regions are
#'   empty.  Default is 0 and appropriate for both calculated coverage and
#'   bedgraph/bigwig like files.  Will automatically switch to "MISSING" if data
#'   is guessed to be qualitative.
#' @param anchor character. controls how x value is derived from position for
#'   each region in qgr.  0 may be the left side or center.  If not unstranded,
#'   x coordinates are flipped for (-) strand. One of c("center",
#'   "center_unstranded", "left", "left_unstranded"). Default is "center".
#' @return data.table that is GRanges compatible
#' @export
#' @import Seqinfo
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' bam_file = system.file("extdata/test.bam",
#'     package = "seqsetvis")
#' qgr = CTCF_in_10a_overlaps_gr[seq_len(5)]
#' qgr = GenomicRanges::resize(qgr, width = 500, fix = "center")
#' bam_gr = seqsetvis:::fetchBam(bam_file, qgr)
#' bam_dt = viewGRangesWinSample_dt(bam_gr, qgr, 50)
#'
#' if(Sys.info()['sysname'] != "Windows"){
#'     bw_file = system.file("extdata/MCF10A_CTCF_FE_random100.bw",
#'         package = "seqsetvis")
#'     bw_gr = rtracklayer::import.bw(bw_file, which = qgr)
#'     bw_dt = viewGRangesWinSample_dt(bw_gr, qgr, 50)
#' }
viewGRangesWinSample_dt = function(score_gr,
                                   qgr,
                                   window_size,
                                   attrib_var = "score",
                                   fill_value = 0,
                                   anchor = c("center", "center_unstranded",
                                              "left", "left_unstranded")[1]) {
    #reserve bindings for data.table
    x = id = NULL
    stopifnot(is(score_gr, "GRanges"))
    stopifnot(!is.null(mcols(score_gr)[[attrib_var]]))
    stopifnot(is(qgr, "GRanges"))
    if(is.null(qgr$id)) qgr = prepare_fetch_GRanges_names(qgr, include_id = TRUE)
    stopifnot(is.numeric(window_size))
    stopifnot(window_size >= 1)
    stopifnot(window_size %% 1 == 0)
    stopifnot(anchor %in% c("center", "center_unstranded",
                            "left", "left_unstranded"))
    windows = slidingWindows(qgr, width = window_size, step = window_size)

    # names(windows) = qgr$id
    # windows's handling of names seems to have changed and now every nest
    # GRanges has parent's name
    names(windows) = NULL
    windows = unlist(windows)
    windows$id = names(windows)
    windows = resize(windows, width = 1, fix = "center")
    olaps = suppressWarnings(data.table::as.data.table(
        findOverlaps(
            query = windows,
            subject = score_gr,
            ignore.strand = TRUE
        )
    ))
    # patch up missing/out of bound data with 0
    missing_idx = setdiff(seq_along(windows), olaps$queryHits)
    if (length(missing_idx) > 0) {
        olaps = rbind(
            olaps,
            data.table::data.table(
                queryHits = missing_idx,
                subjectHits = length(score_gr) + 1
            )
        )[order(queryHits)]
        suppressWarnings({
            patch_gr = GRanges(Seqinfo::seqlevels(score_gr)[1],
                               IRanges::IRanges(1, 1))
            mcols(patch_gr)[[attrib_var]] = fill_value
            Seqinfo::seqlevels(patch_gr) = Seqinfo::seqlevels(score_gr)
            score_gr = c(score_gr,
                         patch_gr)
        })
    }
    # set y and output windows = windows[olaps$queryHits]
    windows$y = fill_value
    windows[olaps$queryHits]$y = mcols(score_gr[olaps$subjectHits])[[attrib_var]]
    score_dt = data.table::as.data.table(windows)

    return(shift_anchor(score_dt, window_size, anchor))
}


#' Summarizes signal in bins.  The same number of bins per region in qgr is used
#' and widths can vary in qgr, in contrast to
#' \code{\link{viewGRangesWinSample_dt}} where width must be constant across
#' regions.
#'
#' This function is most appropriate where features are expected to vary greatly
#' in size and feature boundaries are important, ie. gene bodies, enhancers or
#' TADs.
#'
#' Columns in output data.table are: standard GRanges columns: seqnames, start,
#' end, width, strand id - matched to names(score_gr). if names(score_gr) is
#' missing, added as seq_along(score_gr). y - value of score from score_gr x -
#' relative bp position
#' @param score_gr GRanges with a "score" metadata column.
#' @param qgr regions to view by window.
#' @param n_tiles numeric >= 1, the number of tiles to use for every region in
#'   qgr.
#' @param attrib_var character name of attribute to pull data from. Default is
#'   "score", compatible with with bigWigs or bam coverage.
#' @param attrib_type one of NULL, qualitative or quantitative.  If NULL will
#'   attempt to guess by casting attrib_var attribute to character or factor.
#'   Default is NULL.
#' @param fill_value numeric or character value to use where queried regions are
#'   empty.  Default is 0 and appropriate for both calculated coverage and
#'   bedgraph/bigwig like files.  Will automatically switch to "MISSING" if data
#'   is guessed to be qualitative.
#' @param anchor character. controls how x value is derived from position for
#'   each region in qgr.  0 may be the left side or center.  If not unstranded,
#'   x coordinates are flipped for (-) strand. One of c("center",
#'   "center_unstranded", "left", "left_unstranded"). Default is "center".
#' @param summary_FUN function. used to aggregate score by tile.  must accept
#'   x=score and w=width numeric vectors as only arguments. default is
#'   weighted.mean.  limma::weighted.median is a good alternative.
#' @return data.table that is GRanges compatible
#' @importFrom Seqinfo seqlevels
#' @export
#' @importFrom stats weighted.mean
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' bam_file = system.file("extdata/test.bam",
#'     package = "seqsetvis")
#' qgr = CTCF_in_10a_overlaps_gr[1:5]
#' # unlike viewGRangesWinSample_dt, width is not fixed
#' # qgr = GenomicRanges::resize(qgr, width = 500, fix = "center")
#' bam_gr = seqsetvis:::fetchBam(bam_file, qgr)
#' bam_dt = viewGRangesWinSummary_dt(bam_gr, qgr, 50)
#'
#' if(Sys.info()['sysname'] != "Windows"){
#'     bw_file = system.file("extdata/MCF10A_CTCF_FE_random100.bw",
#'         package = "seqsetvis")
#'     bw_gr = rtracklayer::import.bw(bw_file, which = qgr)
#'     bw_dt = viewGRangesWinSummary_dt(bw_gr, qgr, 50)
#' }
viewGRangesWinSummary_dt = function (score_gr,
                                     qgr,
                                     n_tiles = 100,
                                     attrib_var = "score",
                                     attrib_type = NULL,
                                     fill_value = 0,
                                     anchor = c("center", "center_unstranded",
                                                "left", "left_unstranded")[1],
                                     summary_FUN = stats::weighted.mean) {
    #reserve bindings for data.table
    x = id = tile_start = tile_end = tile_id =
        tile_width = scored_width = tile_density = score_width = NULL
    stopifnot(is(score_gr, "GRanges"))
    stopifnot(!is.null(mcols(score_gr)[[attrib_var]]))
    stopifnot(is(qgr, "GRanges"))
    if(is.null(qgr$id)) qgr = prepare_fetch_GRanges_names(qgr, include_id = TRUE)
    stopifnot(is.numeric(n_tiles))
    stopifnot(n_tiles >= 1)
    stopifnot(n_tiles %% 1 == 0)
    stopifnot(anchor %in% c("center", "center_unstranded", "left",
                            "left_unstranded"))
    if(any(n_tiles > width(qgr))){
        qgr.sp = split(qgr, ifelse(n_tiles > width(qgr), "too_small", "ok"))
        tiles.ok = tile(qgr.sp$ok, n_tiles)
        tiles.too_small = tile(qgr.sp$too_small, width(qgr.sp$too_small))
        for(i in seq_along(tiles.too_small)){
            gr_i = tiles.too_small[[i]]
            idx = ceiling(seq_len(n_tiles)/n_tiles*length(gr_i))
            tiles.too_small[[i]] = gr_i[idx]
        }
        tiles = c(tiles.too_small, tiles.ok)
    }else{
        tiles = tile(qgr, n_tiles)
    }

    names(tiles) = NULL
    tiles = unlist(tiles)
    tiles$id = names(tiles)
    tiles$tile_id = seq_along(tiles)


    olaps = suppressWarnings(data.table::as.data.table(
        findOverlaps(
            query = tiles,
            subject = score_gr,
            ignore.strand = TRUE
        )
    ))
    #fill in gaps with zeroes
    missing_idx = setdiff(seq_along(tiles), olaps$queryHits)
    if (length(missing_idx) > 0) {
        olaps = rbind(
            olaps,
            data.table::data.table(
                queryHits = missing_idx,
                subjectHits = length(score_gr) + 1
            )
        )[order(queryHits)]
        suppressWarnings({
            patch_gr = GRanges(Seqinfo::seqlevels(score_gr)[1],
                               IRanges::IRanges(1, 1))
            mcols(patch_gr)[[attrib_var]] = NA
            Seqinfo::seqlevels(patch_gr) = Seqinfo::seqlevels(score_gr)
            score_gr = c(score_gr,
                         patch_gr
            )
        })
    }
    mcols(score_gr) = mcols(score_gr)[attrib_var]
    cov_dt = cbind(as.data.table(score_gr[olaps$subjectHits])[, -c(1, 4:5)],
                   as.data.table(tiles[olaps$queryHits])[, -c(1, 4:5)])
    colnames(cov_dt)[4:5] = c("tile_start", "tile_end")

    cov_dt[start == 1 &
               end == 1, c("start", "end") := list(tile_start, tile_end)]
    #trim score regions to fit in tiles so score weighting is accurate

    cov_dt[start < tile_start, start := tile_start]
    cov_dt[end > tile_end, end := tile_end]
    cov_dt[, score_width := end - start + 1]
    cov_dt[, tile_width := tile_end - tile_start + 1]

    if(all(is.na(cov_dt[[attrib_var]]))){
        cov_dt[[attrib_var]] = fill_value
    }
    data.table::set(cov_dt, i = which(is.na(cov_dt[[attrib_var]])), j = attrib_var, value = fill_value)

    if(is.null(attrib_type)){
        if(any(class(cov_dt[[attrib_var]]) %in% c("character", "factor"))){
            attrib_type = "qualitative"
            if(fill_value == 0){
                fill_value = "MISSING"
            }
        }else{
            attrib_type = "quantitative"
        }
    }

    # check for incompletely retrieved regions (zeroes omitted for instance)
    check_dt = cov_dt[, list(scored_width = sum(score_width)), by = list(tile_id, id, tile_width)]
    # check_dt$tile_width = width(tiles)
    #add a dummy interval of score zero to correct width
    repair_dt = check_dt[tile_width != scored_width, list(
        start = -1,
        end = -1,
        score = fill_value,
        tile_start = -1,
        tile_end = -1,
        id = id,
        tile_id = tile_id,
        score_width = tile_width - scored_width,
        tile_width
    )]
    setnames(repair_dt, "score", attrib_var)
    cov_dt = rbind(cov_dt, repair_dt)

    if(attrib_type == "qualitative"){
        density_dt = cov_dt[, list(tile_density = summary_FUN(score_width / tile_width, rep(1, length(width)))),
                            by = c("tile_id", "id", attrib_var)]
    }else if(attrib_type == "quantitative"){
        density_dt = cov_dt[, list(tile_density = summary_FUN(get(attrib_var), score_width)),
                            by = list(tile_id, id)]
    }else{
        stop("attrib_type must be one of qualitative or quantitative")
    }

    tiles = as.data.table(tiles)
    tiles[, x := (seq_len(n_tiles) - .5) / n_tiles, by = id]

    # density_dt[, x := (seq_len(n_tiles) - .5) / n_tiles, by = id]
    if (!all(tiles$id %in% density_dt$id))
        stop("something bad happened when merging density ",
             "data.table back to tiles GRanges.")

    if(attrib_type == "qualitative"){
        score_dt = merge(tiles, density_dt[, list(y = tile_density, id, tile_id, ATTRIB_VAR = get(attrib_var))], by = c("tile_id", "id"))
        setnames(score_dt, "ATTRIB_VAR", attrib_var)
        #score_dt[is.na(quality), quality := fill_value]
    }else{
        score_dt = merge(tiles, density_dt[, list(y = tile_density, id, tile_id)], by = c("tile_id", "id"))
        #score_dt[is.na(y), y := fill_value]
    }
    score_dt$tile_id = NULL
    #slightly different than summary,
    #x is already set and regions are already contiguous.
    #just need to flip x or center as needed.
    switch(
        anchor,
        center = {
            center_val = mean((seq_len(n_tiles) - .5) / n_tiles)
            score_dt[, x := x - center_val, by = id]
            score_dt[strand == "-", x := (-1 * x)]
        },
        center_unstranded = {
            center_val = mean((seq_len(n_tiles) - .5) / n_tiles)
            score_dt[, x := x - center_val, by = id]
        },
        left = {
            score_dt[strand == "-", x := (1 - 1 * x), by = id]
        },
        left_unstranded = {
            #do nothing
        }
    )

    #ensure x is rounded and easily flippable
    score_dt$x = round(score_dt$x, 9)

    score_dt
}

#' orients the relative position of x's zero value and
#' extends ranges to be contiguous
#'
#' @param score_dt data.table, GRanges() sufficient
#' @param window_size numeric, window size used to generate score_dt
#' @param anchor character, one of c("center", "center_unstranded",
#' "left", "left_unstranded")
#' @return score_dt with x values shifted appropriately and start and end
#' extended to make ranges contiguous
shift_anchor = function(score_dt, window_size, anchor) {
    x = id = NULL
    shift = round(window_size / 2)
    fudge = 1
    if(window_size == 1) fudge = 0
    switch(
        anchor,
        center = {
            # score_dt[, x := start - min(start) + shift, by = id]
            # score_dt[, x := x - round(mean(x)), by = id]
            # if(window_size %% 2 == 0){
            #     score_dt[strand == "-", x := -1 * x]
            # }else{
            #     score_dt[strand == "-", x := -1 * x - fudge]
            # }
            score_dt[, x := -1]
            score_dt[strand != "-", x := start - min(start) + shift,
                     by = id]
            #flip negative

            score_dt[strand == "-", x := -1 * (end - max(end) - shift),
                     by = id]
            score_dt[, x := x - round(mean(x)), by = id]

        },
        center_unstranded = {
            score_dt[, x := start - min(start) + shift, by = id]
            score_dt[, x := x - round(mean(x)), by = id]
        },
        left = {
            score_dt[, x := -1]
            score_dt[strand != "-", x := start - min(start) + shift,
                     by = id]
            #flip negative

            score_dt[strand == "-", x := -1 * (end - max(end) - shift),
                     by = id]

        },
        left_unstranded = {
            score_dt[, x := start - min(start) + shift, by = id]
        }
    )

    score_dt[, start := start - shift]
    score_dt[, end := end + window_size - shift - 1]
    return(score_dt)
}

#' prepares GRanges for windowed fetching.
#'
#' Deprecated and renamed as prepare_fetch_GRanges_width
#'
#' output GRanges parallels input with consistent width evenly divisible by
#' win_size.  Has warning if GRanges needed resizing, otherwise no warning
#' and input GRanges is returned unchanged.
#'
#' @param qgr GRanges to prepare
#' @param win_size numeric window size for fetch
#' @param min_quantile numeric value from 0 to 1. Lowest possible quantile value.  Only
#' relevant if target_size is not specified.
#' @param target_size numeric final width of qgr if known. Default of NULL
#' leads to quantile based determination of target_size.
#' @param skip_centerFix boolean, if FALSE (default) all regions will be resized
#' GenomicRanges::resize(x, w, fix = "center") to a uniform size based on
#' min_quantile to a width divisible by win_size.
#' @return GRanges, either identical to qgr or with suitable consistent width
#' applied.
#' @export
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' #use prepare_fetch_GRanges_width instead:
#' qgr = prepare_fetch_GRanges_width(CTCF_in_10a_overlaps_gr, win_size = 50)
#' #no warning if qgr is already valid for windowed fetching
#' prepare_fetch_GRanges_width(qgr, win_size = 50)
prepare_fetch_GRanges = function(qgr,
                                 win_size,
                                 min_quantile = .75,
                                 target_size = NULL,
                                 skip_centerFix = FALSE) {
    .Deprecated("prepare_fetch_GRanges_width")
    prepare_fetch_GRanges_width(qgr, win_size, min_quantile, target_size, skip_centerFix)
}

#' prepares GRanges for windowed fetching.
#'
#' output GRanges parallels input with consistent width evenly divisible by
#' win_size.  Has warning if GRanges needed resizing, otherwise no warning
#' and input GRanges is returned unchanged.
#'
#' @param qgr GRanges to prepare
#' @param win_size numeric window size for fetch
#' @param min_quantile numeric value from 0 to 1. Lowest possible quantile value.  Only
#' relevant if target_size is not specified.
#' @param target_size numeric final width of qgr if known. Default of NULL
#' leads to quantile based determination of target_size.
#' @param skip_centerFix boolean, if FALSE (default) all regions will be resized
#' GenomicRanges::resize(x, w, fix = "center") to a uniform size based on
#' min_quantile to a width divisible by win_size.
#' @return GRanges, either identical to qgr or with suitable consistent width
#' applied.
#' @export
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' qgr = prepare_fetch_GRanges_width(CTCF_in_10a_overlaps_gr, win_size = 50)
#' #no warning if qgr is already valid for windowed fetching
#' prepare_fetch_GRanges_width(qgr, win_size = 50)
prepare_fetch_GRanges_width = function(qgr,
                                       win_size,
                                       min_quantile = .75,
                                       target_size = NULL,
                                       skip_centerFix = FALSE) {
    if(length(qgr) < 1) stop("length of query GRanges was 0!")
    if (!skip_centerFix &&
        (length(unique(width(qgr))) > 1 || width(qgr)[1] %% win_size != 0)) {
        if (is.null(target_size)) {
            target_size = quantileGRangesWidth(
                qgr = qgr,
                min_quantile = min_quantile,
                win_size = win_size
            )
        }
        if (target_size %% win_size != 0) {
            stop(
                "target_size: ",
                target_size,
                " not evenly divisible by win_size: ",
                win_size
            )
        }

        qgr = centerFixedSizeGRanges(qgr, fixed_size = target_size)
        message(
            "widths of qgr were not ",
            "identical and evenly divisible by win_size.",
            "\nA fixed width of ",
            target_size,
            " was applied based on the data provided."
        )
    }
    if (any(start(qgr) < 1)) {
        warning("Some out of bounds GRanges had to be shifted back to start >= 1.")
        fix_shift = 1 - start(qgr)
        fix_shift = vapply(fix_shift, function(x)
            max(x, 0), FUN.VALUE = 1)
        qgr = IRanges::shift(qgr, fix_shift)
    }
    return(qgr)
}

#' Quantile width determination strategy
#'
#' Returns the lowest multiple of win_size greater than
#' min_quantile quantile of width(qgr)
#'
#' @param qgr GRanges to calculate quantile width for
#' @param min_quantile numeric value from 0 to 1. The minimum quantile of width in qgr
#' @param win_size numeric/integer >=1, returned value will be a multiple of
#' this
#' @return numeric that is >= min_quantile and evenly divisible by win_size
#' @export
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' gr = CTCF_in_10a_overlaps_gr
#' quantileGRangesWidth(gr)
#' quantileGRangesWidth(gr, min_quantile = .5, win_size = 100)
quantileGRangesWidth = function(qgr,
                                min_quantile = .75,
                                win_size = 1) {
    stopifnot(is(qgr, "GRanges"))
    stopifnot(is.numeric(min_quantile), is.numeric(win_size))
    stopifnot(min_quantile >= 0 && min_quantile <= 1)
    stopifnot(length(min_quantile) == 1 && length(win_size) == 1)
    stopifnot(win_size %% 1 == 0)
    stopifnot(win_size >= 1)
    qwidth = quantile(width(qgr), min_quantile)
    fwidth = ceiling(qwidth / win_size) * win_size
    return(fwidth)
}

.get_file_attribs = function(file_paths, file_attribs, names_variable){
    if(is.data.table(file_paths)){
        file_paths = as.data.frame(file_paths)
    }
    if(is.data.table(file_attribs)){
        file_attribs = as.data.frame(file_attribs)
    }
    if(is.null(file_attribs)){
        if (is.data.frame(file_paths)) {
            if (ncol(file_paths) == 1) {
                file_attribs = data.frame(matrix(
                    0, nrow = nrow(file_paths), ncol = 0
                ))
            } else{
                k = which(grepl("file", colnames(file_paths)))[1]
                if(is.na(k)) k = 1
                file_attribs = file_paths[,-k, drop = FALSE]
            }

        } else{
            #file_paths is assumed to be a character vector
            file_attribs = data.frame(data.frame(matrix(
                0, nrow = length(file_paths), ncol = 0
            )))
        }
    }
    if (is.data.frame(file_paths)) {
        if(any(grepl("file", colnames(file_paths)))){
            k = which(grepl("file", colnames(file_paths)))[1]
            file_paths = file_paths[[k]]
        }else{
            file_paths = file_paths[[1]]
        }
    }

    if (is.list(file_paths)) {
        file_paths = unlist(file_paths)
    }
    if (is.factor(file_paths)){
        file_paths = as.character(file_paths)
    }

    #sort file paths and attrib in the same way if name variable is a factor
    if(names_variable %in% colnames(file_attribs)){
        if(is.factor(file_attribs[[names_variable]])){
            o = order(file_attribs[[names_variable]])
            file_paths = file_paths[o]
            file_attribs = file_attribs[o,, drop = FALSE]
        }
    }
    return(list(file_paths = file_paths, file_attribs = file_attribs))
}

.get_unique_names = function(unique_names, file_paths, file_attribs, names_variable){
    if(is.null(unique_names)){
        if(!is.null(file_attribs[[names_variable]])){
            unique_names = file_attribs[[names_variable]]
        }else if (is.data.frame(file_paths) || is.data.table(file_paths)) {
            if(is.null(colnames(file_paths))){
                unique_names = file_paths[[1]]
            }else{
                if(any(grepl("file", colnames(file_paths)))){
                    k = which(grepl("file", colnames(file_paths)))[1]
                    unique_names = file_paths[[k]]
                }else{
                    unique_names = file_paths[[1]]
                }
            }
        }else{
            if(is.null(names(file_paths))){
                unique_names = file_paths
            }else{
                unique_names = names(file_paths)
            }

        }
    }
    if (is.factor(unique_names)){
        unique_names = as.character(unique_names)
        #using levels can cause unintentional reordering
        # unique_names = levels(droplevels(unique_names))
    }
    if(any(duplicated(unique_names))){
        stop("some unique_names are duplicated:\n",
             paste(collapse = "\n",
                   unique(unique_names[duplicated(unique_names)])), "\n",
             "If you haven't manually supplied uninque_names, this is derived from",
             " file paths supplied.  \nTry supplying unique_names manually if you ",
             "intended to load duplicate files.")
    }
    unique_names
}

testthat::context("Anchor")
# flipping viewGranges
library(seqsetvis)
library(GenomicRanges)
library(testthat)
library(data.table)
data(CTCF_in_10a_overlaps_gr)
qgr = CTCF_in_10a_overlaps_gr[1:5]
qgr = resize(qgr, 120)
strand(qgr) = c("+", "-", "-", "+", "-")
# qgr = centerFixedSizeGRanges(qgr, 500)
#bed used to intersect bam
# rtracklayer::export.bed(qgr, con = "ctcf_5.bed")
bam_file = system.file("extdata/test.bam", package = "seqsetvis")

fetchBam = seqsetvis:::fetchBam

bam_gr = fetchBam(bam_file, qgr)
# bams
bams = c("A" = bam_file, "B" = bam_file)

#sampling
test_that("viewGRangesWinSample_dt center", {
    exp_x = seq(-2, 1)*(30)+15
    dt = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 30, anchor = "center")
    expect_setequal(dt$x, exp_x)
    # dt$group = "stranded"
    dt_us = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 30, anchor = "center_unstranded")
    expect_setequal(dt_us$x, exp_x)
    # dt_us$group = "unstranded"
    # ssvSignalLineplot(rbind(dt, dt_us), sample_ = "id", color_ = "strand", group_ = "group")
    expect_true(all(dt[strand == "+"][order(x)]$y == dt_us[strand == "+"][order(x)]$y))
    expect_false(all(dt[strand == "-"][order(x)]$y == dt_us[strand == "-"][order(x)]$y))
    expect_false(all(dt$x > 0))
    expect_false(all(dt$x < 0))
    expect_false(all(dt_us$x > 0))
    expect_false(all(dt_us$x < 0))
})

test_that("viewGRangesWinSample_dt center odd window", {
    exp_x = seq(-2, 1)*30+15
    dt = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 30, anchor = "center")
    expect_setequal(dt$x, exp_x)
    # dt$group = "stranded"
    dt_us = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 30, anchor = "center_unstranded")
    expect_setequal(dt_us$x, exp_x)
})

test_that("viewGRangesWinSample_dt left", {
    exp_x = seq(-2, 1)*(30)+75
    dt = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 30, anchor = "left")
    expect_setequal(dt$x, exp_x)
    # dt$group = "stranded"
    dt_us = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 30, anchor = "left_unstranded")
    expect_setequal(dt_us$x, exp_x)
    # dt_us$group = "unstranded"
    # ssvSignalLineplot(rbind(dt, dt_us), sample_ = "id", color_ = "strand", group_ = "group")
    expect_equal(dt[strand == "+"][order(x)]$y, dt_us[strand == "+"][order(x)]$y)
    expect_failure(expect_equal(dt[strand == "-"][order(x)]$y, dt_us[strand == "-"][order(x)]$y))
    expect_true(all(dt$x > 0))
    expect_true(all(dt_us$x > 0))
})

test_that("viewGRangesWinSample_dt left odd window", {
    exp_x = (seq(-2, 0)+2)*40+20
    dt = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 40, anchor = "left")
    expect_setequal(dt$x, exp_x)
    # dt$group = "stranded"
    dt_us = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 40, anchor = "left_unstranded")
    expect_setequal(dt_us$x, exp_x)
})

test_that("ssvFetchBam anchors", {
    exp_30 = seq(-2, 1)*(30)+75
    dt_30_left = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 30, anchor = "left", fragLens = NA)
    dt_30_leftUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 30, anchor = "left_unstranded", fragLens = NA)
    dt_30_center = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 30, anchor = "center", fragLens = NA)
    dt_30_centerUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 30, anchor = "center_unstranded", fragLens = NA)

    expect_setequal(dt_30_left$x, exp_30)
    expect_setequal(dt_30_leftUn$x, exp_30)
    expect_setequal(dt_30_center$x, exp_30-60)
    expect_setequal(dt_30_centerUn$x, exp_30-60)

    exp_40 = (seq(-2, 0)+2)*40+20
    dt_40_left = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 40, anchor = "left", fragLens = NA)
    dt_40_leftUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 40, anchor = "left_unstranded", fragLens = NA)
    dt_40_center = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 40, anchor = "center", fragLens = NA)
    dt_40_centerUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 40, anchor = "center_unstranded", fragLens = NA)

    expect_setequal(dt_40_left$x, exp_40)
    expect_setequal(dt_40_leftUn$x, exp_40)
    expect_setequal(dt_40_center$x, exp_40-60)
    expect_setequal(dt_40_centerUn$x, exp_40-60)

    exp_5 = floor(seq(1, width(qgr[1])/5)*5-2.5)
    dt_5_left = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 5, anchor = "left", fragLens = NA)
    dt_5_leftUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 5, anchor = "left_unstranded", fragLens = NA)
    dt_5_center = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 5, anchor = "center", fragLens = NA)
    dt_5_centerUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 5, anchor = "center_unstranded", fragLens = NA)

    expect_setequal(dt_5_left$x, exp_5)
    expect_setequal(dt_5_leftUn$x, exp_5)
    expect_setequal(dt_5_center$x, exp_5-60)
    expect_setequal(dt_5_centerUn$x, exp_5-60)
})

test_that("ssvFetchBam anchors target_strand = 'both'", {
    exp_30 = seq(-2, 1)*(30)+75
    dt_30_left = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 30, anchor = "left", fragLens = NA, target_strand = "both")
    dt_30_leftUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 30, anchor = "left_unstranded", fragLens = NA, target_strand = "both")
    dt_30_center = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 30, anchor = "center", fragLens = NA, target_strand = "both")
    dt_30_centerUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 30, anchor = "center_unstranded", fragLens = NA, target_strand = "both")

    expect_setequal(dt_30_left$x, exp_30)
    expect_setequal(dt_30_leftUn$x, exp_30)
    expect_setequal(dt_30_center$x, exp_30-60)
    expect_setequal(dt_30_centerUn$x, exp_30-60)

    exp_40 = (seq(-2, 0)+2)*40+20
    dt_40_left = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 40, anchor = "left", fragLens = NA, target_strand = "both")
    dt_40_leftUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 40, anchor = "left_unstranded", fragLens = NA, target_strand = "both")
    dt_40_center = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 40, anchor = "center", fragLens = NA, target_strand = "both")
    dt_40_centerUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 40, anchor = "center_unstranded", fragLens = NA, target_strand = "both")

    expect_setequal(dt_40_left$x, exp_40)
    expect_setequal(dt_40_leftUn$x, exp_40)
    expect_setequal(dt_40_center$x, exp_40-60)
    expect_setequal(dt_40_centerUn$x, exp_40-60)

    exp_5 = floor(seq(1, width(qgr[1])/5)*5-2.5)
    dt_5_left = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 5, anchor = "left", fragLens = NA, target_strand = "both")
    dt_5_leftUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 5, anchor = "left_unstranded", fragLens = NA, target_strand = "both")
    dt_5_center = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 5, anchor = "center", fragLens = NA, target_strand = "both")
    dt_5_centerUn = ssvFetchBam(bam_file, qgr, win_method = "sample", win_size = 5, anchor = "center_unstranded", fragLens = NA, target_strand = "both")

    expect_setequal(dt_5_left$x, exp_5)
    expect_setequal(dt_5_leftUn$x, exp_5)
    expect_setequal(dt_5_center$x, exp_5-60)
    expect_setequal(dt_5_centerUn$x, exp_5-60)
})

#summary
test_that("viewGRangesWinSummary_dt center", {
    dt = viewGRangesWinSummary_dt(bam_gr, qgr, n_tiles = 10, anchor = "center")
    # dt$group = "stranded"
    dt_us = viewGRangesWinSummary_dt(bam_gr, qgr, n_tiles = 10, anchor = "center_unstranded")
    # dt_us$group = "unstranded"
    # ssvSignalLineplot(rbind(dt, dt_us), sample_ = "id", color_ = "strand", group_ = "group")
    expect_equal(dt[strand == "+"][order(x)]$y, dt_us[strand == "+"][order(x)]$y)
    expect_failure(expect_equal(dt[strand == "-"][order(x)]$y, dt_us[strand == "-"][order(x)]$y))
    expect_false(all(dt$x > 0))
    expect_false(all(dt$x < 0))
    expect_false(all(dt_us$x > 0))
    expect_false(all(dt_us$x < 0))
})

test_that("viewGRangesWinSummary_dt left", {
    dt = viewGRangesWinSummary_dt(bam_gr, qgr, n_tiles = 10, anchor = "left")
    # dt$group = "stranded"
    dt_us = viewGRangesWinSummary_dt(bam_gr, qgr, n_tiles = 10, anchor = "left_unstranded")
    # dt_us$group = "unstranded"
    # ssvSignalLineplot(rbind(dt, dt_us), sample_ = "id", color_ = "strand", group_ = "group")
    expect_equal(dt[strand == "+"][order(x)]$y, dt_us[strand == "+"][order(x)]$y)
    expect_failure(expect_equal(dt[strand == "-"][order(x)]$y, dt_us[strand == "-"][order(x)]$y))
    expect_true(all(dt$x > 0))
    expect_true(all(dt_us$x > 0))
})



test_that("ssvFetchBam anchor GRanges", {
    gr = ssvFetchBam(bams, qgr, win_size = 30, anchor = "center")
    gr_uns = ssvFetchBam(bams, qgr, win_size = 30, anchor = "center_unstranded")
    is_pos = as.character(strand(gr)) == "+"
    # plot(gr_uns$x[is_pos], gr$x[is_pos])
    # plot(gr_uns$x[!is_pos], gr$x[!is_pos])
    expect_equal(start(gr_uns)[is_pos], start(gr)[is_pos])
    expect_failure(expect_equal(start(gr_uns)[!is_pos], start(gr)[!is_pos]))
    expect_lt(min(gr$x), 0)
    expect_lt(min(gr_uns$x), 0)
})

test_that("ssvFetchBam anchor data.table", {
    dt = ssvFetchBam(bams, qgr, win_size = 30, anchor = "left",
                     return_data.table = TRUE)
    dt_uns = ssvFetchBam(bams, qgr, win_size = 30, anchor = "left_unstranded",
                         return_data.table = TRUE)

    is_pos = dt$strand == "+"
    # plot(dt_uns$x[is_pos], dt$x[is_pos])
    # plot(dt_uns$x[!is_pos], dt$x[!is_pos])
    expect_equal(dt_uns$start[is_pos], dt$start[is_pos])
    expect_failure(expect_equal(dt_uns$start[!is_pos], dt$start[!is_pos]))
    expect_gte(min(dt$x), 0)
    expect_gte(min(dt_uns$x), 0)

})

#bigwigs
bigwigs = dir(system.file("extdata", package = "seqsetvis"), pattern = "random100.bw$", full.names = TRUE)
test_that("ssvFetchBigwig anchor", {
    skip_on_os("windows")
    gr = ssvFetchBigwig(bigwigs, qgr, win_size = 30, anchor = "center")
    gr_uns = ssvFetchBigwig(bigwigs, qgr, win_size = 30, anchor = "center_unstranded")
    is_pos = as.character(strand(gr)) == "+"
    # plot(gr_uns$x[is_pos], gr$x[is_pos])
    # plot(gr_uns$x[!is_pos], gr$x[!is_pos])
    expect_equal(gr_uns$x[is_pos], gr$x[is_pos])
    expect_failure(expect_equal(gr_uns$x[!is_pos], gr$x[!is_pos]))
    expect_lt(min(gr$x), 0)
    expect_lt(min(gr_uns$x), 0)

    dt = ssvFetchBigwig(bigwigs, qgr, win_size = 30, anchor = "left", return_data.table = TRUE)
    dt_uns = ssvFetchBigwig(bigwigs, qgr, win_size = 30, anchor = "left_unstranded", return_data.table = TRUE)

    is_pos = dt$strand == "+"
    # plot(dt_uns$x[is_pos], dt$x[is_pos])
    # plot(dt_uns$x[!is_pos], dt$x[!is_pos])
    expect_equal(dt_uns$x[is_pos], dt$x[is_pos])
    expect_failure(expect_equal(dt_uns$x[!is_pos], dt$x[!is_pos]))
    expect_gte(min(dt$x), 0)
    expect_gte(min(dt_uns$x), 0)
})


# bams
bams = c("A" = bam_file, "B" = bam_file)

test_that("ssvFetchBam anchor - summary GRanges", {
    gr = ssvFetchBam(bams, qgr, win_size = 30, anchor = "center", win_method = "summary")
    gr_uns = ssvFetchBam(bams, qgr, win_size = 30, anchor = "center_unstranded", win_method = "summary")
    is_pos = as.character(strand(gr)) == "+"
    # plot(gr_uns$x[is_pos], gr$x[is_pos])
    # plot(gr_uns$x[!is_pos], gr$x[!is_pos])
    expect_equal(start(gr_uns)[is_pos], start(gr)[is_pos])
    expect_failure(expect_equal(start(gr_uns)[!is_pos], start(gr)[!is_pos]))
    expect_lt(min(gr$x), 0)
    expect_lt(min(gr_uns$x), 0)
})
test_that("ssvFetchBam anchor - summary data.table", {
    dt = ssvFetchBam(bams, qgr, win_size = 30, anchor = "left", return_data.table = TRUE, win_method = "summary")
    dt_uns = ssvFetchBam(bams, qgr, win_size = 30, anchor = "left_unstranded", return_data.table = TRUE, win_method = "summary")

    is_pos = dt$strand == "+"
    # plot(dt_uns$x[is_pos], dt$x[is_pos])
    # plot(dt_uns$x[!is_pos], dt$x[!is_pos])
    expect_equal(dt_uns$start[is_pos], dt$start[is_pos])
    expect_failure(expect_equal(dt_uns$start[!is_pos], dt$start[!is_pos]))
    expect_gte(min(dt$x), 0)
    expect_gte(min(dt_uns$x), 0)

})

#bigwigs
bigwigs = dir(system.file("extdata", package = "seqsetvis"), pattern = "random100.bw$", full.names = TRUE)
test_that("ssvFetchBigwig anchor - summary", {
    skip_on_os("windows")
    gr = ssvFetchBigwig(bigwigs, qgr, win_size = 30, anchor = "center", win_method = "summary")
    gr_uns = ssvFetchBigwig(bigwigs, qgr, win_size = 30, anchor = "center_unstranded", win_method = "summary")
    is_pos = as.character(strand(gr)) == "+"
    # plot(gr_uns$x[is_pos], gr$x[is_pos])
    # plot(gr_uns$x[!is_pos], gr$x[!is_pos])
    expect_equal(gr_uns$x[is_pos], gr$x[is_pos])
    expect_failure(expect_equal(gr_uns$x[!is_pos], gr$x[!is_pos]))
    expect_lt(min(gr$x), 0)
    expect_lt(min(gr_uns$x), 0)

    dt = ssvFetchBigwig(bigwigs, qgr, win_size = 30, anchor = "left", return_data.table = TRUE, win_method = "summary")
    dt_uns = ssvFetchBigwig(bigwigs, qgr, win_size = 30, anchor = "left_unstranded", return_data.table = TRUE, win_method = "summary")

    is_pos = dt$strand == "+"
    # plot(dt_uns$x[is_pos], dt$x[is_pos])
    # plot(dt_uns$x[!is_pos], dt$x[!is_pos])
    expect_equal(dt_uns$x[is_pos], dt$x[is_pos])
    expect_failure(expect_equal(dt_uns$x[!is_pos], dt$x[!is_pos]))
    expect_gte(min(dt$x), 0)
    expect_gte(min(dt_uns$x), 0)
})

qgr = CTCF_in_10a_overlaps_gr[1:5]
qgr = resize(qgr, 120)
strand(qgr) = c("+", "-", "-", "+", "-")

#bug target ssvFetchBam with stranded qgr and odd win_size causes inconsistent x values
test_that("bugfix - stranded qgr and odd win_size causes inconsistent x values - center(default)", {
    view_size = 30
    even_size = 10
    odd_size = 5
    qgr = resize(qgr, view_size, fix = "center")
    qdf = data.frame(file = bam_file, mark = "a")
    res_even = ssvFetchBam(qdf, win_size = even_size, qgr = qgr, fragLens = NA,
                           target_strand = "*", return_data.table = TRUE)
    # strand(qgr) = "+"
    res_odd = ssvFetchBam(qdf, win_size = odd_size, qgr = qgr, fragLens = NA,
                          target_strand = "*", return_data.table = TRUE)

    expect_equal(view_size / even_size, length(unique(res_even$x)))
    expect_equal(view_size / odd_size, length(unique(res_odd$x)))
})

test_that("bugfix - stranded qgr and odd win_size causes inconsistent x values - center(default) - summary", {
    view_size = 30
    even_size = 10
    odd_size = 5
    qgr = resize(qgr, view_size, fix = "center")
    qdf = data.frame(file = bam_file, mark = "a")
    res_even = ssvFetchBam(qdf, win_size = even_size, qgr = qgr, fragLens = NA,
                           win_method = "summary",
                           target_strand = "both", return_data.table = TRUE)
    # strand(qgr) = "+"
    res_odd = ssvFetchBam(qdf, win_size = odd_size, qgr = qgr, fragLens = NA,
                          win_method = "summary",
                          target_strand = "both", return_data.table = TRUE)

    expect_equal(even_size, length(unique(res_even$x)))
    expect_equal(odd_size, length(unique(res_odd$x)))
})


test_that("bugfix - stranded qgr and odd win_size causes inconsistent x values - left", {
    view_size = 30
    even_size = 10
    odd_size = 5
    qgr = resize(qgr, view_size, fix = "center")
    qdf = data.frame(file = bam_file, mark = "a")
    res_even = ssvFetchBam(qdf, win_size = even_size, qgr = qgr,
                           anchor = "left",
                           fragLens = NA,
                           target_strand = "both", return_data.table = TRUE)
    # strand(qgr) = "+"
    res_odd = ssvFetchBam(qdf, win_size = odd_size, qgr = qgr,
                          anchor = "left",
                          fragLens = NA,
                          target_strand = "both", return_data.table = TRUE)

    expect_equal(view_size / even_size, length(unique(res_even$x)))
    expect_equal(view_size / odd_size, length(unique(res_odd$x)))
})

test_that("bugfix - stranded qgr and odd win_size causes inconsistent x values - left - summary", {
    view_size = 30
    even_size = 10
    odd_size = 5
    qgr = resize(qgr, view_size, fix = "center")
    qdf = data.frame(file = bam_file, mark = "a")
    res_even = ssvFetchBam(qdf, win_size = even_size, qgr = qgr,
                           win_method = "summary",
                           anchor = "left",
                           fragLens = NA,
                           target_strand = "both", return_data.table = TRUE)
    # strand(qgr) = "+"
    res_odd = ssvFetchBam(qdf, win_size = odd_size, qgr = qgr,
                          win_method = "summary",
                          anchor = "left",
                          fragLens = NA,
                          target_strand = "both", return_data.table = TRUE)

    expect_equal(even_size, length(unique(res_even$x)))
    expect_equal(odd_size, length(unique(res_odd$x)))
})

#
test_that("ssvFetchBam bugfix - query GRanges size of 1 - loses strand sensitivity and x bounces around from 0 to 1", {
    view_size.sm = 1
    view_size.lg = 5
    win_size = 1
    qdf = data.frame(file = bam_file, mark = "a")
    qgr.sm = resize(qgr, view_size.sm, fix = "center")
    res_small = ssvFetchBam(qdf, win_size = win_size,
                            qgr = qgr.sm,
                            fragLens = NA,
                            target_strand = "+", return_data.table = TRUE)
    # strand(qgr) = "+"
    qgr.lg = resize(qgr, view_size.lg, fix = "center")
    res_large = ssvFetchBam(qdf, win_size = win_size,
                            qgr = qgr.lg,
                            fragLens = NA,
                            target_strand = "+", return_data.table = TRUE)


    setequal(res_small[x == 0]$y,
             res_large[x == 0]$y)

    expect_setequal(res_small[x == 0]$y, res_large[x == 0]$y)

    all(res_small[x == 0][order(strand)][order(x)]$y ==
            res_large[x == 0][order(strand)][order(x)]$y)

    expect_equal(res_small[x == 0][order(strand)][order(x)]$y,
                 res_large[x == 0][order(strand)][order(x)]$y)
})

testthat::context("ViewGrangesWin")
# flipping viewGranges
library(seqsetvis)
library(GenomicRanges)
library(testthat)
library(data.table)

qgr = CTCF_in_10a_overlaps_gr[1:5]
strand(qgr) = c("+", "-", "-", "+", "-")
# qgr = centerFixedSizeGRanges(qgr, 500)
#bed used to intersect bam
# rtracklayer::export.bed(qgr, con = "ctcf_5.bed")
bam_file = system.file("extdata/test.bam", package = "seqsetvis")

fetchBam = seqsetvis:::fetchBam

bam_gr = fetchBam(bam_file, qgr)

#test varying size regions
vgr = qgr
end(vgr) = end(vgr) + 0:4*100

test_that("viewGrangeWinSample_dt ids match input", {
    names(qgr) = paste0("peak_", seq_along(qgr))
    sample_dt = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 100)
    expect_equal(sort(unique(names(qgr))), sort(unique(sample_dt$id)))
})

test_that("viewGrangeWinSummary_dt ids match input", {
    names(qgr) = paste0("peak_", seq_along(qgr))
    summary_dt = viewGRangesWinSummary_dt(bam_gr, qgr, n_tiles = 100)
    expect_equal(sort(unique(names(qgr))), sort(unique(summary_dt$id)))
})

test_that("viewGrangeWinSample_dt unnamed qgr still creates id", {
    names(qgr) = NULL
    sample_dt = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 100)
    expect_true(!is.null(sample_dt$id))
})

test_that("viewGrangeWinSummary_dt unnamed qgr still creates id", {
    names(qgr) = NULL
    summary_dt = viewGRangesWinSummary_dt(bam_gr, qgr, n_tiles = 100)
    expect_true(!is.null(summary_dt$id))
})

test_that("viewGrangeWinSample_dt ids match input", {
    names(qgr) = paste0("peak_", seq_along(qgr))
    summary_dt = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 100)
    expect_equal(sort(unique(names(qgr))), sort(unique(summary_dt$id)))

    names(qgr) = seq_along(qgr)
    summary_dt = viewGRangesWinSample_dt(bam_gr, qgr, window_size = 100)
    expect_equal(sort(unique(names(qgr))), sort(unique(summary_dt$id)))
})

test_that("viewGrangeWinSummary_dt ids match input", {
    names(qgr) = paste0("peak_", seq_along(qgr))
    summary_dt = viewGRangesWinSummary_dt(bam_gr, qgr, n_tiles = 100)
    expect_equal(sort(unique(names(qgr))), sort(unique(summary_dt$id)))

    names(qgr) = seq_along(qgr)
    summary_dt = viewGRangesWinSummary_dt(bam_gr, qgr, n_tiles = 100)
    expect_equal(sort(unique(names(qgr))), sort(unique(summary_dt$id)))
})


test_that("viewGRangesWinSample_dt sizes vary, viewGRangesWinSummary_dt don't", {
    sample_dt = viewGRangesWinSample_dt(bam_gr, vgr, window_size = 100, anchor = "left")
    expect_gt(length(unique(sample_dt[, .N, by = id]$N)), 1)
    summary_dt = viewGRangesWinSummary_dt(bam_gr, vgr, n_tiles = 10, anchor = "left")
    summary_dt[, .N, by = id]
    expect_equal(length(unique(summary_dt[, .N, by = id]$N)), 1)
})


test_that("can fetch bam summary", {
    #sample is bp scale
    sample_dt = ssvFetchBam(bam_file, vgr, return_data.table = TRUE)
    expect_gte(max(abs(range(sample_dt$x))), 100)
    #summaries have range of x < .5 for center anchor
    summary_dt = ssvFetchBam(bam_file, vgr, win_method = "summary", return_data.table = TRUE)
    expect_lte(max(abs(range(summary_dt$x))), .5)
    summary_dt.med = ssvFetchBam(bam_file, vgr[5], win_method = "summary",
                                 summary_FUN = function(x, w){min(w)}, return_data.table = TRUE)
    expect_lte(max(abs(range(summary_dt.med$x))), .5)
    #summary fun did something
    expect_false(all(summary_dt$y == summary_dt.med$y))
    #win size did something
    summary_dt10 = ssvFetchBam(bam_file, vgr, win_method = "summary",
                               win_size = 10,
                               return_data.table = TRUE)
    expect_true(all(summary_dt10[, .N, by = id]$N == 10))
    expect_true(all(summary_dt[, .N, by = id]$N == 50))
})

test_that("can fetch bam summary", {
    bams = c("A" = bam_file, "B" = bam_file)
    bams_sample_dt = ssvFetchBam(bams, vgr)
    expect_gte(max(abs(range(bams_sample_dt$x))), 100)
    bams_summary_dt = ssvFetchBam(bams, vgr, win_method = "summary")
    expect_lte(max(abs(range(bams_summary_dt$x))), .5)
    bams_summary_dt10 = ssvFetchBam(bams, vgr, win_method = "summary",
                                    win_size = 10, return_data.table = TRUE)
    bams_summary_grFUN = ssvFetchBam(bams, vgr, win_method = "summary",
                                     summary_FUN = function(x, w){min(w)})
    expect_false(all(bams_summary_dt$y == bams_summary_grFUN$y))
    expect_true(all(bams_summary_dt10[, .N, by = id]$N == 10*length(bams)))
    expect_true(all(as.data.table(bams_summary_dt)[, .N, by = id]$N == 50*length(bams)))
})

bigwig_file = system.file("extdata/MCF10A_CTCF_FE_random100.bw", package = "seqsetvis", mustWork = TRUE)

test_that("can fetch bigwig summary", {
    skip_on_os("windows")
    #sample is bp scale
    sample_dt = ssvFetchBigwig(bigwig_file, vgr, return_data.table = TRUE)
    expect_gte(max(abs(range(sample_dt$x))), 100)
    #summaries have range of x < .5 for center anchor
    summary_dt = ssvFetchBigwig(bigwig_file, vgr, win_method = "summary", return_data.table = TRUE)
    expect_lte(max(abs(range(summary_dt$x))), .5)
    summary_dt.med = ssvFetchBigwig(bigwig_file, vgr[5], win_method = "summary",
                                    summary_FUN = function(x, w){min(w)}, return_data.table = TRUE)
    expect_lte(max(abs(range(summary_dt.med$x))), .5)
    #summary fun did something
    expect_false(all(summary_dt$y == summary_dt.med$y))
    #win size did something
    summary_dt10 = ssvFetchBigwig(bigwig_file, vgr, win_method = "summary",
                                  win_size = 10,
                                  return_data.table = TRUE)
    expect_true(all(summary_dt10[, .N, by = id]$N == 10))
    expect_true(all(summary_dt[, .N, by = id]$N == 50))
})

test_that("can fetch bigwig summary", {
    skip_on_os("windows")
    bigwigs = c("A" = bigwig_file, "B" = bigwig_file)
    bigwigs_sample_dt = ssvFetchBigwig(bigwigs, vgr)
    expect_gte(max(abs(range(bigwigs_sample_dt$x))), 100)
    bigwigs_summary_dt = ssvFetchBigwig(bigwigs, vgr, win_method = "summary")
    expect_lte(max(abs(range(bigwigs_summary_dt$x))), .5)
    bigwigs_summary_dt10 = ssvFetchBigwig(bigwigs, vgr, win_method = "summary",
                                          win_size = 10, return_data.table = TRUE)
    bigwigs_summary_dtFUN = ssvFetchBigwig(bigwigs, vgr, win_method = "summary",
                                           summary_FUN = function(x, w){min(w)})
    expect_false(all(bigwigs_summary_dt$y == bigwigs_summary_dtFUN$y))
    expect_true(all(bigwigs_summary_dt10[, .N, by = id]$N == 10*length(bigwigs)))
    expect_true(all(as.data.table(bigwigs_summary_dt)[, .N, by = id]$N == 50*length(bigwigs)))
})

# fetchBam and fetchBw params
# win_method
# summary_FUN

testthat::context("PrepareFetch")
# flipping viewGranges
library(seqsetvis)
library(GenomicRanges)
library(testthat)
library(data.table)

qgr = CTCF_in_10a_overlaps_gr[1:5]
start(qgr) = seq_len(5)
end(qgr) = 3*seq_len(5)
qgr = GenomicRanges::shift(qgr, -2)
strand(qgr) = c("+", "-", "-", "+", "-")

#sampling
test_that("prepare_fetch_GRanges start never less than 1", {
    expect_warning({
        tgr = prepare_fetch_GRanges(qgr, win_size = 5, target_size = 10)
    }, regexp = "Some out of bounds")
    expect_true(all(width(tgr) == 10))
    expect_true(all(start(tgr) >= 1))
})

#' collapse_gr
#'
#' collapse non-contiguous regions (i.e. exons) into a contiguous coordinate
#' starting at 1.  this is strand sensitive and intended for use with
#' all exons of a single gene.
#'
#' @param genome_gr a GRanges of regions on a single chromosome.  Regions
#' are intended to be non-contiguous and may even overlap.
#'
#' @return a new GRanges object with same mcols as input with all intervals
#' starting at 1 and no empty space between syntenic regions.
#' @export
#'
#' @examples
#' library(data.table)
#' library(GenomicRanges)
#' dev_dat = data.table(seqnames = "chrTest",
#'                      transcript_id = c(1, 1,  2, 2, 3, 3, 3),
#'                      start = c(5,  30,  8, 30, 2, 30, 40),
#'                      end = c(10, 35, 15, 38, 7, 35, 45),
#'                      strand = "+")
#'
#' genome_gr = GRanges(dev_dat)
#' collapse_gr(genome_gr)
#'
#' neg_gr = genome_gr
#' strand(neg_gr) = "-"
#' collapse_gr(neg_gr)
collapse_gr = function(genome_gr){
    stopifnot("GRanges" %in% class(genome_gr))
    stopifnot(length(genome_gr) > 0)
    stopifnot(length(unique(seqnames(genome_gr))) == 1)
    stopifnot(length(unique(strand(genome_gr))) == 1)
    if(as.character(strand(genome_gr[1])) == "-"){
        tmp = mcols(genome_gr)
        genome_gr = GRanges(seqnames(genome_gr),
                            IRanges::IRanges(max(end(genome_gr)) - end(genome_gr) + 1,
                        max(end(genome_gr))- start(genome_gr) + 1))
        mcols(genome_gr) = tmp

    }
    strand(genome_gr) = "*"
    tx_gr = genome_gr
    rm_gr = GenomicRanges::setdiff(range(genome_gr), genome_gr)
    for(i in rev(seq_along(rm_gr))){
        irm_gr = rm_gr[i]
        to_shift = start(tx_gr) > end(irm_gr)
        tx_gr[to_shift]  = GenomicRanges::shift(tx_gr[to_shift],
                                                shift = -width(irm_gr))
    }
    tx_gr = GenomicRanges::shift(tx_gr, -min(start(tx_gr)) + 1)
    return(tx_gr)
}

#' convert_collapsed_coord
#'
#' (preliminary implementation, sub-optimal)
#'
#' see \code{\link{collapse_gr}} for explanation of intended uses. this function
#' translates all values of x from original genomic coordinates to new
#' coordinate space created by \code{\link{collapse_gr}}.
#'
#'
#' @param genome_gr non-contiguous regions to collapse a la
#'   \code{\link{collapse_gr}}
#' @param x numeric, positions within genome_gr to convert to collapsed
#'   coordinates.
#'
#' @return numeric, positions of every value of x within collapse coordinates.
#' values outside of collapsed regions (an intron or outside range) will be NA.
#' @export
#'
#' @examples
#' library(data.table)
#' library(GenomicRanges)
#' dev_dat = data.table(seqnames = "chrTest",
#'                      transcript_id = c(1, 1,  2, 2, 3, 3, 3),
#'                      start = c(5,  30,  8, 30, 2, 30, 40),
#'                      end = c(10, 35, 15, 38, 7, 35, 45),
#'                      strand = "+")
#'
#' genome_gr = GRanges(dev_dat)
#' convert_collapsed_coord(genome_gr, start(genome_gr))
#' convert_collapsed_coord(genome_gr, end(genome_gr))
convert_collapsed_coord = function(genome_gr, x){
    stopifnot("GRanges" %in% class(genome_gr))
    stopifnot(length(genome_gr) > 0)
    stopifnot(length(unique(seqnames(genome_gr))) == 1)
    stopifnot(length(unique(strand(genome_gr))) == 1)
    if(as.character(strand(genome_gr[1])) == "-"){
        x = max(end(genome_gr)) - x + 1
        tmp = mcols(genome_gr)
        genome_gr = GRanges(seqnames(genome_gr),
                            IRanges::IRanges(max(end(genome_gr)) - end(genome_gr) + 1,
                                    max(end(genome_gr))- start(genome_gr) + 1))
        mcols(genome_gr) = tmp
    }
    genome_gr
    red_gr = reduce(genome_gr)
    xnew = vapply(x, function(xx){
        is_over = xx > end(red_gr)
        if(any(is_over)){
            is_in = max(which(is_over)) + 1
        }else{
            is_in = 1
        }

        xx - start(red_gr)[is_in] + sum(width(red_gr[is_over])) + 1
    }, FUN.VALUE = 1)


    is_miss = vapply(x, function(xx){
        !any(start(red_gr) <= xx & end(red_gr) >= xx)
    }, FUN.VALUE = TRUE)

    xnew[is_miss] = NA

    return(xnew)
}

#' chromHMM state segmentation in the MCF7 cell line
#'
#' @description Vignette data for seqsetvis was downloaded directly from GEO
#' series
#' \href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE57498}{GSE57498}.
#' This data is the state segmentation by chromHMM in the MCF7 cell line.
#' chromHMM creates a hidden markov model by integrating several ChIP-seq
#' samples, in this case: \itemize{
#' \item MCF7_H3K27ac_ChIP-Seq
#' \item MCF7_H3K27me3_ChIP-Seq
#' \item MCF7_H3K4me1_ChIP-Seq
#' \item MCF7_H3K4me3_ChIP-Seq
#' \item MCF7_RNApolIIp_ChIP-Seq
#' }
#'
#' @description Data from GEO series
#' \href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE57498}{GSE57498}
#' is from the publication
#' \href{https://www.ncbi.nlm.nih.gov/pubmed/24916973}{Taberlay PC et al. 2014}
#' @details Contains: \itemize{
#' \item \code{\link{chromHMM_demo_overlaps_gr}}
#' \item \code{\link{chromHMM_demo_bw_states_gr}}
#' \item \code{\link{chromHMM_demo_state_total_widths}}
#' \item \code{\link{chromHMM_demo_state_colors}}
#' \item \code{\link{chromHMM_demo_segmentation_url}}
#' \item \code{\link{chromHMM_demo_chain_url}}
#' }
#' @docType data
#' @keywords datasets
#' @name chromHMM_demo_data
NULL



#' URL to download hg19 MCF7 chromHMM segmentation
#'
#' @details file is gzipped bed with name, score, itemRgb and thick meta columns
#' @details part of \code{\link{chromHMM_demo_data}}
#'
#' @docType data
#' @keywords datasets
#' @name chromHMM_demo_segmentation_url
#' @format a character containing a URL
#' @return a character containing a URL
NULL

#' URL to download hg19ToHg38 liftover chain from UCSC
#'
#' @details file is gzipped .txt
#' @details part of \code{\link{chromHMM_demo_data}}
#'
#' @docType data
#' @keywords datasets
#' @name chromHMM_demo_chain_url
#' @format a character containing a URL
#' @return a character containing a URL
NULL

#' state name to total width mappings, hg38
#'
#' @details part of \code{\link{chromHMM_demo_data}}
#'
#' @docType data
#' @keywords datasets
#' @name chromHMM_demo_state_total_widths
#' @format named numeric of total widths per state
#' @return named numeric of total widths per state
NULL

#' original state name to color mappings stored in segmentation bed
#'
#' @details part of \code{\link{chromHMM_demo_data}}
#'
#' @docType data
#' @keywords datasets
#' @name chromHMM_demo_state_colors
#' @format a named character vector mapping states to hex colors
#' @return a named character vector mapping states to hex colors
NULL

#' overlap of MCF10A CTCF with MCF7 chromHMM states, hg38.
#'
#' @details part of \code{\link{chromHMM_demo_data}}
#' @details the result of ssvOverlapIntervalSets() on MCF10A CTCF peaks and
#'          MCF7 chromHMM states with use_first = TRUE
#' @details first (the MCF10A peaks) and no_hit columns have been removed
#' each remaining column represents MCF10A peaks overlapping with a state.
#'
#' @docType data
#' @keywords datasets
#' @name chromHMM_demo_overlaps_gr
#' @format a GRanges object of length 98
#' with 10 logical metadata columns, 1 per state.
#' @return a GRanges object of length 98
#' with 10 logical metadata columns, 1 per state.
NULL

#' MCF10A CTCF profiles at 20 windows per chromHMM state, hg38.
#'
#' @details part of \code{\link{chromHMM_demo_data}}
#' @details the result of ssvFetchBigwig() on the MCF10A_CTCF_FE.bw
#' near 20 randomly selected windows per chromHMM state.
#'
#' @docType data
#' @keywords datasets
#' @name chromHMM_demo_bw_states_gr
#' @format a GRanges object of length 4000
#' with 5 metadata columns sufficient for use with ggplot2
#' @return a GRanges object of length 4000
#' with 5 metadata columns sufficient for use with ggplot2
NULL

# plotting functions for region/feature set comparison go here
# all plotting functions accept a variety of objects as first argument and
# handle them via the S4  generic ssvMakeMembTable
# ssvMakeMembTable has implemented methods for:
#   - membership table
#       columns are sets
#       rows are members
#       values are logicals indicating membership
#   - list containing character vectors of set members
#   - list of GRanges

#' ssvFeatureVenn
#'
#' ggplot implementation of vennDiagram from limma package.  Currently limited
#' at 3 sets.  [ssvFeatureUpset] and [ssvFeatureBinaryHeatmap] are good options for
#' more than 3 sets. [ssvFeatureEuler] can work too but can take a very long time
#' to run for more than 5 or so.
#'
#' @export
#'
#' @param object will be passed to [ssvMakeMembTable] for
#' conversion to membership matrix
#' @param group_names useful if names weren't provided or were lost in
#' creating membership matrix
#' @param counts_txt_size font size for count numbers
#' @param counts_as_labels if TRUE, geom_label is used instead of geom_text.
#' can be easier to read.
#' @param show_outside_count if TRUE, items outside of all sets are counted
#' outside. can be confusing.
#' @param line_width uses size aesthetic to control line width of circles.
#' @param circle_colors colors to use for circle line colors. Uses Dark2 set
#' from RColorBrewer by default.
#' @param fill_alpha alpha value to use for fill, defaults to .3.
#' @param line_alpha numeric value from 0 to 1. Alpha value for circle line
#' @param counts_color character. single color to use for displaying counts
#' @param counts_as_percent if TRUE, convert counts to percentages in plots.
#' @param percentage_digits The number of digits to round percentages to, default is 1.
#' @param percentage_suffix The character to append to percentages, default is "%".
#' @param n_points integer.  number of points to approximate circle with.
#' default is 200.
#' @param return_data logical.  If TRUE, return value is no longer ggplot and
#' is instead the data used to generate that plot. Default is FALSE.
#' @return ggplot venn diagram
#' @import ggplot2
#' @import S4Vectors
#' @importFrom limma vennCounts
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' ssvFeatureVenn(list(1:3, 2:6))
#' ssvFeatureVenn(CTCF_in_10a_overlaps_gr)
#' ssvFeatureVenn(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
#'
#' ssvFeatureVenn(list(1:3, 2:6),
#'   counts_as_percent = TRUE,
#'   percentage_digits = 2)
#'
#' ssvFeatureVenn(list(1:3, 2:6),
#'   counts_as_percent = TRUE,
#'   percentage_digits = 0,
#'   percentage_suffix = " %",
#'   counts_txt_size = 12)
ssvFeatureVenn = function(object,
                          group_names = NULL,
                          counts_txt_size = 5,
                          counts_as_labels = FALSE,
                          show_outside_count = FALSE,
                          line_width = 3,
                          circle_colors = NULL,
                          fill_alpha = .3,
                          line_alpha = 1,
                          counts_color = NULL,
                          counts_as_percent = FALSE,
                          percentage_digits = 1,
                          percentage_suffix = "%",
                          n_points = 200,
                          return_data = FALSE) {
    size = group = x = y = label = NULL#declare binding for data.table

    object = ssvMakeMembTable(object)
    stopifnot(is.numeric(counts_txt_size), is.numeric(line_width),
              is.numeric(fill_alpha))
    stopifnot(is.logical(counts_as_labels), is.logical(show_outside_count))
    all(apply(object, 2, class) == "logical")
    object <- limma::vennCounts(object)
    set_counts <- object[, "Counts"]
    if(counts_as_percent){
        set_counts = set_counts / sum(set_counts)
        set_counts = paste0(round(set_counts * 100, digits = percentage_digits), percentage_suffix)
    }

    nsets <- ncol(object) - 1
    if (nsets > 3)
        stop("Can't plot Venn diagram for more than 3 sets")
    if (is.null(group_names))
        group_names <- factor(colnames(object)[seq_len(nsets)],
                              levels = colnames(object)[seq_len(nsets)])
    if (is.null(counts_color))
        counts_color <- grDevices::rgb(0,0,0)

    xcentres <- switch(nsets,
                       0,
                       c(-1, 1),
                       c(-1, 1, 0),
                       c(1, 2, 1, 2)
    )
    ycentres <- switch(nsets,
                       0,
                       c(0, 0),
                       c(1, 1, -2)/sqrt(3)
    )
    r <- 1.5
    p = ggellipse(xcentres = xcentres,
                  ycentres = ycentres,
                  r = r,
                  circle_colors = circle_colors,
                  group_names = group_names,
                  line_alpha =  line_alpha,
                  fill_alpha = fill_alpha,
                  line_width = line_width,
                  n_points = n_points)

    df_text <- switch(nsets,
                      {
                          df = data.frame(x = 0,
                                          y = 0,
                                          label = set_counts[-1],
                                          size = counts_txt_size,
                                          col = counts_color)
                          if (show_outside_count) {
                              df = rbind(df, data.frame(x = 2.3, y = -2.1,
                                                        label = set_counts[1],
                                                        size = counts_txt_size,
                                                        col = counts_color))
                          }
                          df
                      }, {
                          df = data.frame(x = c(1.5, -1.5, 0),
                                          y = c(0.1, 0.1, 0.1),
                                          label = set_counts[-1],
                                          size = counts_txt_size,
                                          col = counts_color)
                          if (show_outside_count) {
                              df = rbind(df, data.frame(x = 2.3,
                                                        y = -2.1,
                                                        label = set_counts[1],
                                                        size = counts_txt_size,
                                                        col = counts_color))
                          }
                          df
                      }, {
                          df = data.frame(x = c(0, 1.5, 0.75, -1.5,
                                                -0.75, 0, 0),
                                          y = c(-1.7, 1, -0.35, 1, -0.35,
                                                0.9, 0),
                                          label = set_counts[-1],
                                          size = counts_txt_size,
                                          col = counts_color)
                          if (show_outside_count) {
                              df = rbind(df, data.frame(x = 2.5, y = -3,
                                                        label = set_counts[1],
                                                        size = counts_txt_size,
                                                        col = counts_color))
                          }
                          df
                      })

    counts_method = ifelse(counts_as_labels, geom_label, geom_text)
    p = p + counts_method(data = df_text,
                          mapping = aes(x = x,
                                        y = y,
                                        label = label,
                                        size = size),
                          show.legend = FALSE)
    if(return_data){
        return(
            data.table(xcentres = xcentres,
                       ycentres = ycentres,
                       r = r,
                       r2 = r,
                       phi = rep(0, length(xcentres)),
                       group_names = group_names)
        )
    }
    return(p)
}


#' ssvFeatureUpset
#'
#' Uses the UpSetR package to create an [UpSetR::upset] plot of region overlaps.
#'
#' @param object will be passed to \code{\link{ssvMakeMembTable}} for conversion
#'   to membership matrix
#' @param return_UpSetR If TRUE, return the UpSetR object, The default is FALSE
#'   and results in a ggplotified version compatible with cowplot etc.
#' @param nsets Number of sets to look at
#' @param nintersects Number of intersections to plot. If set to NA, all
#'   intersections will be plotted.
#' @param order.by How the intersections in the matrix should be ordered by.
#'   Options include frequency (entered as "freq"), degree, or both in any
#'   order.
#' @param ... Additional parameters passed to \code{\link[UpSetR]{upset}} in the
#'   UpSetR  package.
#'
#' @return ggplot version of UpSetR plot
#' @export
#' @import UpSetR
#' @import ggplotify
#'
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' ssvFeatureUpset(list(1:3, 2:6))
#' ssvFeatureUpset(CTCF_in_10a_overlaps_gr)
#' ssvFeatureUpset(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
ssvFeatureUpset = function(object,
                           return_UpSetR = FALSE,
                           nsets = NULL,
                           nintersects = 15,
                           order.by = "freq",
                           ...){
    object = ssvMakeMembTable(object)
    up_df = as.matrix(object)
    up_df = ifelse(up_df, 1, 0)
    up_df = as.data.frame(up_df)
    if(is.null(nsets)) nsets = ncol(up_df)
    if(ncol(up_df) > 1){
        p_up = UpSetR::upset(up_df,
                             nsets = nsets,
                             nintersects = nintersects,
                             order.by = order.by,
                             ...)
    }else{
        p_up = ggplot() + theme_void() +
            labs(title = "cannot run UpSetR on only 1 set.")
    }
    if(return_UpSetR){
        return(p_up)
    }
    p = ggplotify::as.ggplot(p_up)
    p
}


#' Try to load a bed-like file and convert it to a GRanges object
#'
#' @export
#' @param object A membership table
#' @param line_width numeric, passed to size aesthetic to control line width
#' @param shape shape argument passed to eulerr::euler
#' @param n_points number of points to use for drawing ellipses, passed to
#' eulerr:::ellipse
#' @param fill_alpha numeric value from 0 to 1. Alpha value for circle fill
#' @param line_alpha numeric value from 0 to 1. Alpha value for circle line
#' @param circle_colors colors to choose from for circles.  passed to ggplot2
#' color scales.
#' @param return_data logical.  If TRUE, return value is no longer ggplot and
#' is instead the data used to generate that plot. Default is FALSE.
#' @return ggplot of venneuler results
#' @import ggplot2
#' @import eulerr
#' @import S4Vectors
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' ssvFeatureEuler(list(1:3, 2:6))
#' ssvFeatureEuler(CTCF_in_10a_overlaps_gr)
#' ssvFeatureEuler(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
ssvFeatureEuler = function(object,
                           line_width = 2,
                           shape = c("circle", "ellipse")[1],
                           n_points = 200,
                           fill_alpha = .3,
                           line_alpha = 1,
                           circle_colors = NULL,
                           return_data = FALSE) {
    x = y = group = NULL#declare binding for data.table
    object = ssvMakeMembTable(object)
    stopifnot(is.numeric(line_width),
              is.numeric(n_points),
              is.numeric(fill_alpha))
    stopifnot(is.character(shape))
    stopifnot(shape %in% c("circle", "ellipse"))
    cn = colnames(object)
    eu = eulerr::euler(object, shape = shape)
    dd = eu$ellipses

    h <- dd$h
    k <- dd$k
    a <- dd$a
    b <- dd$b
    phi <- dd$phi

    if(return_data){
        return(
            data.table(xcentres = h,
                       ycentres = k,
                       r = a,
                       r2 = b,
                       phi = phi,
                       group_names = cn)
        )
    }

    p = ggellipse(xcentres = h,
                  ycentres = k,
                  r = a,
                  r2 = b,
                  phi = phi,
                  circle_colors = circle_colors,
                  group_names = cn,
                  line_alpha =  line_alpha,
                  fill_alpha = fill_alpha,
                  line_width = line_width,
                  n_points = n_points)
    p
}


#' bar plots of set sizes
#' @export
#' @param object passed to ssvMakeMembTable for conversion to membership table
#' @param show_counts logical.  should counts be displayed at the center of each
#' bar. default is TRUE
#' @param bar_colors character. rcolor or hex colors. default of NULL
#' uses RColorBrewer Dark2. Will repeat to match number of samples.
#' @param counts_text_colors character. rcolor or hex colors. default of NULL
#' uses RColorBrewer Dark2. Will repeat to match number of samples.
#' @param return_data logical.  If TRUE, return value is no longer ggplot and
#' is instead the data used to generate that plot. Default is FALSE.
#' @param count_label_size Font size bar count labels. Default is 8.
#' @return ggplot of bar plot of set sizes
#' @import ggplot2
#' @import S4Vectors
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' ssvFeatureBars(list(1:3, 2:6))
#' ssvFeatureBars(CTCF_in_10a_overlaps_gr, count_label_size = 10)
#' ssvFeatureBars(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
ssvFeatureBars = function(object,
                          show_counts = TRUE,
                          bar_colors = NULL,
                          counts_text_colors = NULL,
                          return_data = FALSE,
                          count_label_size = 8) {
    group = count = NULL#declare binding for data.table
    object = ssvMakeMembTable(object)
    stopifnot(is.logical(show_counts))
    stopifnot(is.null(bar_colors) || all(is.character(bar_colors)))
    n_bars = ncol(object)
    #bar colors
    if (is.null(bar_colors)) {
        bar_colors = safeBrew(n_bars, "Dark2")
    } else {
        not_hex = substr(bar_colors, 0, 1) != "#"
        bar_colors[not_hex] = col2hex(bar_colors[not_hex])
    }
    if (length(bar_colors) < n_bars)
        bar_colors <- rep(bar_colors, length.out = n_bars)
    names(bar_colors) = colnames(object)
    #text colors
    if (is.null(counts_text_colors)) {
        counts_text_colors = rep("black", n_bars)
    } else {
        not_hex = substr(counts_text_colors, 0, 1) != "#"
        counts_text_colors[not_hex] = col2hex(counts_text_colors[not_hex])
    }
    if (length(counts_text_colors) < n_bars)
        counts_text_colors <- rep(counts_text_colors, length.out = n_bars)
    names(counts_text_colors) = colnames(object)

    hit_counts = colSums(object)
    hit_counts_df = data.frame(count = hit_counts,
                               group = factor(names(hit_counts),
                                              levels = names(hit_counts)))
    if(return_data){
        return(as.data.table(hit_counts_df))
    }
    p <- ggplot(hit_counts_df, aes(x = group, y = count, fill = group)) +
        labs(x = "") +
        geom_bar(stat = "identity") +
        theme_bw() +
        theme(axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5),
              panel.grid.major.x = element_blank(),
              panel.grid.minor.x = element_blank()) +
        scale_fill_manual(values = bar_colors)
    if(show_counts)
        p = p + annotate("text",
                         y = hit_counts/2,
                         x = seq_along(hit_counts),
                         label = hit_counts,
                         color = counts_text_colors,
                         size = count_label_size/ggplot2::.pt)
    return(p)
}

#' ssvFeaturePie
#'
#' Generate a ggplot pie plot of set sizes.
#'
#' @export
#' @param object object that ssvMakeMembTable can convert to logical matrix
#'   membership
#' @param slice_colors colors to use for pie slices
#' @param return_data logical.  If TRUE, return value is no longer ggplot and is
#'   instead the data used to generate that plot. Default is FALSE.
#' @import ggplot2
#' @import S4Vectors
#' @return ggplot pie graph of set sizes
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' ssvFeaturePie(list(1:3, 2:6))
#' ssvFeaturePie(CTCF_in_10a_overlaps_gr)
#' ssvFeaturePie(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
ssvFeaturePie = function(object, slice_colors = NULL, return_data = FALSE) {
    count = group = NULL#declare binding for data.table

    stopifnot(is.null(slice_colors) || all(is.character(slice_colors)))

    object = ssvMakeMembTable(object)
    hit_counts = colSums(object)
    hit_counts_df = data.frame(count = hit_counts,
                               group = factor(names(hit_counts),
                                              levels = names(hit_counts)))

    n_slices = ncol(object)
    if (is.null(slice_colors)) {
        slice_colors = safeBrew(n_slices, "Dark2")
    } else {
        not_hex = substr(slice_colors, 0, 1) != "#"
        slice_colors[not_hex] = col2hex(slice_colors[not_hex])
    }
    if (length(slice_colors) < n_slices)
        slice_colors <- rep(slice_colors, length.out = n_slices)
    names(slice_colors) = colnames(object)

    if(return_data){
        return(as.data.table(hit_counts_df))
    }

    p = ggplot(hit_counts_df, aes(x = "", y = count, fill = group)) +
        labs(x = "") +
        geom_bar(width = 1, stat = "identity") +
        # guides(x = "none") +
        theme(axis.text.x = element_blank(),
              axis.line = element_blank(),
              panel.background = element_blank(),
              axis.ticks = element_blank()) +
        coord_polar("y", start = 0) +
        scale_y_reverse() +
        scale_fill_manual(values = slice_colors)
    hc = rev(hit_counts)/sum(hit_counts)
    hc = c(0, cumsum(hc)[-length(hc)]) + hc/2
    p = p + annotate(geom = "text",
                     y = hc * sum(hit_counts),
                     x = 1.1,
                     label = rev(hit_counts_df$count))
    return(p)
}



#' ssvFeatureBinaryHeatmap
#'
#' Outputs a ggplot binary heatmap, where color indicates TRUE and the other
#' indicates FALSE in a membership table. The heatmap is sorted, TRUE at the
#' top, by column left to right. Changes to column order can reveal different
#' patterns.
#'
#' As a svg output, the final plot can be unwieldy. The default of
#' `raster_approximation` = TRUE is easier to work with, especially for larger
#' membership tables.
#'
#' @export
#' @param object passed to ssvMakeMembTable
#' @param raster_approximation If TRUE, instead of standard ggplot, write
#' temporary raster png image and redraw that as plot background. default is
#' FALSE
#' @param true_color character. rcolor or hex color used for TRUE values.
#' default is "black".
#' @param false_color character. rcolor or hex color used for TRUE values.
#' default is "#EFEFEF", a gray.
#' @param raster_width_min raster width will be minimum multiple of number of
#' columns over this number.  ignored if raster_approximation is FALSE.
#' @param raster_height_min raster height will be minimum multiple of number of
#' rows over this number  ignored if raster_approximation is FALSE
#' @param return_data logical.  If TRUE, return value is no longer ggplot and
#' is instead the data used to generate that plot. Default is TRUE
#' @rawNamespace import(data.table, except = c(shift, first, second, last))
#' @return ggplot using geom_tile of membership table sorted from left to right.
#' @import png
#' @import ggplot2
#' @importFrom grid rasterGrob
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' ssvFeatureBinaryHeatmap(list(1:3, 2:6))
#' # horizontal version
#' ssvFeatureBinaryHeatmap(list(1:3, 2:6)) + coord_flip() +
#'   theme(axis.text.x = element_blank(), axis.text.y = element_text())
#' ssvFeatureBinaryHeatmap(CTCF_in_10a_overlaps_gr)
#' ssvFeatureBinaryHeatmap(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
#' ssvFeatureBinaryHeatmap(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,3:2])
ssvFeatureBinaryHeatmap = function(object,
                                   raster_approximation = TRUE,
                                   true_color = "black",
                                   false_color = "#EFEFEF",
                                   raster_width_min = 1000,
                                   raster_height_min = 1000,
                                   return_data = FALSE) {
    groups = bool = value = NULL#declare binding for data.table
    object = ssvMakeMembTable(object)
    stopifnot(is.logical(raster_approximation))

    mat = ifelse(as.matrix(object), 1, 0)
    for (i in rev(seq_len(ncol(mat)))){
        mat = mat[order(mat[, i], decreasing = FALSE), , drop = FALSE]
    }
    hdt = data.table::as.data.table(cbind(mat, row = seq_len(nrow(mat))))
    hdt = data.table::melt(hdt, id.vars = "row", variable.name = "groups")

    if (raster_approximation) {
        stopifnot(is.numeric(raster_width_min), is.numeric(raster_height_min))
        dmat = as.matrix(
            data.table::dcast(
                hdt,
                value.var = "value",
                formula = rev(row) ~ groups))[, 1 + seq_len(ncol(object)),
                                              drop = FALSE]
        low_v = 1
        dmat = ifelse(dmat > low_v, low_v, dmat)
        # dmat = -1*(dmat - low_v)
        if (raster_width_min >= ncol(dmat)) {
            # expand cols as necessary to meet target
            raster_width_multiplier = ceiling(raster_width_min/ncol(dmat))
            comp_col = rep(seq_len(ncol(dmat)), each = raster_width_multiplier)
        } else {
            # evenly spaced sample down to target
            comp_col = round(seq_len(raster_width_min)/raster_width_min *
                                 ncol(dmat))
        }
        if (raster_height_min >= nrow(dmat)) {
            # expand rows as necessary to meet target
            raster_height_multiplier = ceiling(raster_height_min/nrow(dmat))
            comp_row = rep(seq_len(nrow(dmat)), each = raster_height_multiplier)
        } else {
            # evenly spaced sample down to target
            comp_row = round(seq_len(raster_height_min)/raster_height_min *
                                 nrow(dmat))
        }
        # dmat = (dmat * -.95 + .95)
        comp_mat = dmat[comp_row, comp_col]
        comp_ar = array(comp_mat, c(nrow(comp_mat), ncol(comp_mat), 3))
        comp_ar[, , 1] = ifelse(comp_mat == 1,
                                col2rgb(true_color)[1,1]/255,
                                col2rgb(false_color)[1,1]/255)
        comp_ar[, , 2] = ifelse(comp_mat == 1,
                                col2rgb(true_color)[2,1]/255,
                                col2rgb(false_color)[2,1]/255)
        comp_ar[, , 3] = ifelse(comp_mat == 1,
                                col2rgb(true_color)[3,1]/255,
                                col2rgb(false_color)[3,1]/255)
        png::writePNG(comp_ar, target = "tmp.png")
        png_grob = grid::rasterGrob(png::readPNG("tmp.png"),
                                    width = unit(1, "npc"),
                                    height = unit(1, "npc"),
                                    interpolate = FALSE)
        file.remove("tmp.png")
        p = ggplot(hdt) +
            geom_tile(aes(x = groups, y = row, fill = NA, col = NULL)) +
            scale_fill_manual(values = c(`TRUE` = "black",
                                         `FALSE` = "#EFEFEF")) +
            scale_alpha(0) +
            labs(fill = "", y = "") +
            guides(fill = "none") +
            theme(
                axis.line = element_blank(),
                axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5),
                panel.grid.major.x = element_blank(),
                panel.grid.minor.x = element_blank(),
                plot.background = element_blank(),
                panel.background = element_blank(),
                axis.text.y = element_blank(),
                axis.ticks.y = element_blank(),
                axis.ticks.x = element_blank(),
                axis.title = element_blank()
            ) +
            annotation_custom(png_grob, xmin = 0.5, xmax = ncol(dmat) + 0.5,
                              ymin = 0.5, ymax = nrow(dmat) + 0.5)
    } else {
        hdt[, `:=`(bool, value == 1)]
        p = ggplot(hdt) +
            geom_tile(aes(x = groups, y = row, fill = bool, col = NULL)) +
            scale_fill_manual(values = c(`TRUE` = true_color,
                                         `FALSE` = false_color)) +
            labs(fill = "", y = "") +
            guides(fill = "none") +
            theme(
                axis.line = element_blank(),
                axis.text.x = element_text(angle = 90, hjust = 1, vjust = 0.5),
                panel.grid.major.x = element_blank(),
                panel.grid.minor.x = element_blank(),
                plot.background = element_blank(),
                panel.background = element_blank(),
                axis.text.y = element_blank(),
                axis.ticks.y = element_blank(),
                axis.ticks.x = element_blank(),
                axis.title = element_blank()
            )
    }
    if(return_data){
        return(hdt)
    }
    return(p)
}

testthat::context("FeaturePlots")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
gr_a = GRanges("chr1", IRanges(1:7*10, 1:7*10+1))
gr_b = GRanges("chr1", IRanges(5:10*10, 5:10*10+1))
gr_c = GRanges("chr1", IRanges(8:10*10+5, 8:10*10+6))
setL = list("gr_a" = gr_a, "gr_b" = gr_b, "gr_c" = gr_c)

test_that("ssvFeatureVenn various set sizes, no outside counts. 1.", {
    p1 = ssvFeatureVenn(setL[1])
    p1
    expect_s3_class(p1, class = "ggplot")
})

test_that("ssvFeatureVenn various set sizes, no outside counts. 2.", {
    p2 = ssvFeatureVenn(setL[1:2])
    p2
    expect_s3_class(p2, class = "ggplot")
})

test_that("ssvFeatureVenn various set sizes, no outside counts. 3.", {
    p3 = ssvFeatureVenn(setL[1:3])
    p3
    expect_s3_class(p3, class = "ggplot")
})

test_that("ssvFeatureVenn various set sizes, with outside counts. 1.", {
    p1 = ssvFeatureVenn(setL[1], show_outside_count = TRUE)
    p1
    expect_s3_class(p1, class = "ggplot")
})

test_that("ssvFeatureVenn various set sizes, with outside counts. 2.", {
    p2 = ssvFeatureVenn(setL[1:2], show_outside_count = TRUE)
    p2
    expect_s3_class(p2, class = "ggplot")
})

test_that("ssvFeatureVenn various set sizes, with outside counts. 3.", {
    p3 = ssvFeatureVenn(setL[1:3], show_outside_count = TRUE)
    p3
    expect_s3_class(p3, class = "ggplot")
})

test_that("ssvFeatureVenn can't plot more than 3 sets.", {
    setL = list("gr_a" = gr_a, "gr_b" = gr_b, "gr_c" =
                    gr_c, "gr_d" = gr_c, "gr_e" = gr_a)
    expect_error(ssvFeatureVenn(setL[1:4]))
    expect_error(ssvFeatureVenn(setL[1:5]))
})

test_that("ssvFeatureVenn circle colors", {
    p1 = ssvFeatureVenn(setL, circle_colors = "blue")
    expect_s3_class(p1, class = "ggplot")
})

test_that("ssvFeatureEuler fill, no fill", {
    p1 = ssvFeatureEuler(setL)
    expect_s3_class(p1, class = "ggplot")

    p2 = ssvFeatureEuler(setL, fill_alpha = 0)
    expect_s3_class(p2, class = "ggplot")
})

test_that("ssvFeatureBinaryHeatmap no raster option", {
    p1 = ssvFeatureBinaryHeatmap(setL, raster_approximation = FALSE)
    expect_s3_class(p1, class = "ggplot")
})

test_that("ssvFeatureBinaryHeatmap downsample", {
    p1 = ssvFeatureBinaryHeatmap(setL,
                                 raster_approximation = TRUE,
                                 raster_width_min = 2)
    expect_s3_class(p1, class = "ggplot")

    p2 = ssvFeatureBinaryHeatmap(setL,
                                 raster_approximation = TRUE,
                                 raster_height_min = 3)
    expect_s3_class(p2, class = "ggplot")

    p3 = ssvFeatureBinaryHeatmap(
        setL,
        raster_approximation = TRUE,
        raster_height_min = 3,
        raster_width_min = 2
    )
    expect_s3_class(p3, class = "ggplot")

})

test_that("ssvFeatureVenn various other paramters don't throw error", {
    olap = ssvOverlapIntervalSets(list(
        "gr_a" = gr_a,
        "gr_b" = gr_b,
        "gr_c" = gr_c
    ))
    p = ssvFeatureVenn(
        olap,
        circle_colors = c("red", "blue", "green"),
        fill_alpha = .1,
        counts_txt_size = 10,
        show_outside_count = TRUE,
        counts_as_labels = TRUE
    )
    expect_s3_class(p, class = "ggplot")
    p = ssvFeatureVenn(
        olap,
        circle_colors = c("red", "blue", "green"),
        fill_alpha = 0,
        counts_txt_size = 10,
        show_outside_count = TRUE,
        counts_as_labels = TRUE
    )
    expect_s3_class(p, class = "ggplot")
})

test_that("col2hex", {
    string_colors = c("red", "blue", "green")
    rgb_colors = col2rgb(string_colors)
    hex_colors = c("#FF0000", "#0000FF", "#00FF00")
    expect_equal(col2hex(string_colors), hex_colors)
    expect_error(col2hex(rgb_colors))
    expect_error(col2hex("asdf"))
})

.default_multi_resolver_FUN = function(x, variable.name = NULL){
    if(is.character(x)){
        x = paste0(sort(x)[1], "-from_", length(x))
    }else if(is.numeric(x)){
        x = max(x)
    }else if(is.factor(x)){
        x = sort(x)[1]
    }else if(is.logical(x)){
        x = any(x)
    }
    x
}

.ssvAnnotateSubjectGRanges = function(annotation_source,
                                      subject_gr,
                                      annotation_name = NULL,
                                      multi_resolver_FUN = "default"){
    if(!is(annotation_source, "GRanges")){
        stop("GRanges expected as annotation_source.")
    }
    if(!is(subject_gr, "GRanges")){
        stop("GRanges expected as subject_gr.")
    }
    if(is.null(names(subject_gr))){
        names(subject_gr) = seq_along(subject_gr)
    }
    if(!is.null(annotation_name)){
        if(!is.character(annotation_name)){
            stop("character or NULL expected as annotation_name.")
        }
    }
    anno_olap = findOverlaps(query = annotation_source, subject = subject_gr)
    subject_hit_ids = names(subject_gr)[subjectHits(anno_olap)]
    anno_hit_mcols = mcols(annotation_source[queryHits(anno_olap)])

    df1 = cbind(DataFrame(id__ = subject_hit_ids), anno_hit_mcols)
    if(any(duplicated(df1$id__))){
        if(is.function(multi_resolver_FUN)){
            message("Resolving multiple overlapping annotatation_source items",
                    " with provided function...")
        }else if(multi_resolver_FUN == "default"){
            message("Resolving multiple overlapping annotatation_source items",
                    " with default function (override with multi_resolver_FUN)...")
            multi_resolver_FUN = .default_multi_resolver_FUN
        }else{
            stop('multi_resolver_FUN must be a user provided function or "default".')
        }
        resolve_grp = function(grp){
            resolve_var_name = function(var_name){
                multi_resolver_FUN(x = grp[[var_name]], variable.name = var_name)
            }
            cn_todo = colnames(grp[-1])
            names(cn_todo) = cn_todo
            lapply(cn_todo, resolve_var_name)
        }
        res = lapply(
            split(df1, df1$id__),
            resolve_grp
        )
        test_len = vapply(res, function(x){max(lengths(x))}, FUN.VALUE = 1)
        if(any(test_len > 1)){
            print(res[which(test_len > 1)[1]])
            stop("multi_resolver_FUN did not resolve to single values. ",
                 "The first problematic entry is printed above.")

        }
        res = lapply(res, as.data.frame)
        df1 = do.call(rbind, res)
        # df1 = DataFrame(df1)
        df1 = cbind(DataFrame(id__ = rownames(df1)), df1)

    }
    #append annotation_name if specified
    if(!is.null(annotation_name)){
        not_id = colnames(df1) != "id__"
        colnames(df1)[not_id] = paste0(colnames(df1)[not_id],
                                       ".", annotation_name)
    }

    df2 = mcols(subject_gr)
    df2$id__ = rownames(df2)
    #merge subject and annotation, restore rownames
    shared_cn = setdiff(intersect(colnames(df1), colnames(df2)), "id__")
    if(length(shared_cn) > 0){
        stop("Annotation shares colnames with subject_gr. ",
             "Try specifying annotation_name to avoid this.\n",
             "  Offending colnames:\n",
             paste(shared_cn, collapse = "\n"))
    }
    dfm = merge(df2, df1, by = "id__", all.x = TRUE)
    rownames(dfm) = dfm$id__
    dfm$id__ = NULL
    mcols(subject_gr) =  dfm[names(subject_gr),]
    subject_gr
}

#' ssvAnnotateSubjectGRanges
#'
#' @param annotation_source A single GRanges, a list of GRanges, or a
#'   GRangesList
#' @param subject_gr The base GRanges to add annotation mcols to.
#' @param annotation_name Optional name for single GRanges. Required for list
#'   inputs if list does not have names.
#' @param multi_resolver_FUN Optional function to resolve multiple overlapping
#'   annotation source regions per subject region. This function must accept 2
#'   arguments. `x` is the values in a single mcol attribute and `variable.name`
#'   is the name of variable. A single value must be returned or an error will
#'   be generated. The default of "default" can handle numeric, logical,
#'   character, and factor types.
#'
#' @return GRanges with the same regions as `subject_gr` but with addtional
#'   mcols added from annotation_source.
#' @export
#' @rdname ssvAnnotateSubjectGRanges
#'
#' @examples
#' library(GenomicRanges)
#' data(CTCF_in_10a_narrowPeak_grs)
#' np_grs = CTCF_in_10a_narrowPeak_grs
#' olap_gr = ssvOverlapIntervalSets(np_grs)
#' # annotating with a signle GRanges is OK
#' ssvAnnotateSubjectGRanges(np_grs$MCF10A_CTCF, olap_gr)
#' # provide a name if that's useful
#' ssvAnnotateSubjectGRanges(np_grs$MCF10A_CTCF, olap_gr,
#'   annotation_name = "MCF10A")
#' # a named list adds each annotation
#' ssvAnnotateSubjectGRanges(np_grs, olap_gr)
#' # overriding list names is an option
#' ssvAnnotateSubjectGRanges(np_grs, olap_gr, LETTERS[1:3])
#' # GRangeList are handled like a standard list
#' ssvAnnotateSubjectGRanges(GRangesList(np_grs), olap_gr, LETTERS[1:3])
setGeneric("ssvAnnotateSubjectGRanges",
           function(annotation_source,
                    subject_gr,
                    annotation_name = NULL,
                    multi_resolver_FUN = "default"){
               standardGeneric("ssvAnnotateSubjectGRanges")
           })

#' @rdname ssvAnnotateSubjectGRanges
#' @export
setMethod("ssvAnnotateSubjectGRanges",
          signature(annotation_source = "GRanges"), .ssvAnnotateSubjectGRanges)

#' @rdname ssvAnnotateSubjectGRanges
#' @export
setMethod("ssvAnnotateSubjectGRanges", signature(annotation_source = "list"),
          function(annotation_source, subject_gr,
                   annotation_name = NULL, multi_resolver_FUN = "default"){
              if(!all(vapply(annotation_source, is,
                             FUN.VALUE = TRUE, class2 = "GRanges"))){
                  stop("All items in annotation_source list must be GRanges.")
              }
              if(!is.null(annotation_name)){
                  if(!is.character(annotation_name)){
                      stop("character or NULL expected as annotation_name.")
                  }
                  if(!length(annotation_name) == length(annotation_source)){
                      stop("annotation_source and annotation_name ",
                           "must be same length.")
                  }
                  names(annotation_source) = annotation_name
              }
              if(is.null(names(annotation_source))){
                  stop("annotation_source must have names set.")
              }
              if(any(duplicated(names(annotation_source)))){
                  stop("names of annotation_source must be unique.")
              }
              for(name in names(annotation_source)){
                  subject_gr = ssvAnnotateSubjectGRanges(
                      annotation_source = annotation_source[[name]],
                      subject_gr = subject_gr,
                      annotation_name = name,
                      multi_resolver_FUN = multi_resolver_FUN
                  )
              }
              subject_gr
          })

#' @rdname ssvAnnotateSubjectGRanges
#' @export
setMethod("ssvAnnotateSubjectGRanges",
          signature(annotation_source = "GRangesList"),
          function(annotation_source, subject_gr,
                   annotation_name = NULL, multi_resolver_FUN = "default"){
              ssvAnnotateSubjectGRanges(
                  as.list(annotation_source),
                  subject_gr = subject_gr,
                  annotation_name = annotation_name,
                  multi_resolver_FUN = multi_resolver_FUN)
          })

#' Transforms set of GRanges to all have the same size.
#'
#' \code{centerFixedSizeGRanges} First calculates the central coordinate of each
#' GRange in \code{grs} and extends in both direction by half of
#' \code{fixed_size}
#' @export
#' @param grs Set of GRanges with incosistent and/or incorrect size
#' @param fixed_size The final width of each GRange returned.
#' @return Set of GRanges after resizing all input GRanges, either shortened
#' or lengthened as required to match \code{fixed_size}
#' @import GenomicRanges
#' @examples
#' library(GenomicRanges)
#' grs = GRanges("chr1", IRanges(1:10+100, 1:10*3+100))
#' centered_grs = centerFixedSizeGRanges(grs, 10)
#' width(centered_grs)
centerFixedSizeGRanges = function(grs, fixed_size = 2000) {
    stopifnot(is(grs, "GRanges"))
    stopifnot(is(fixed_size, "numeric"))
    stopifnot(fixed_size > 0)
    m = floor(start(grs) + width(grs)/2)
    ext = floor(fixed_size/2)
    start(grs) = m - ext
    end(grs) = m + fixed_size - ext - 1
    # resize isn't ideal - repeated applications accumulate rounding shifts
    # grs = GenomicRanges::resize(x = grs, width = fixed_size, fix = "center")
    return(grs)
}


#' applies a spline smoothing to a tidy data.table containing x and y values.
#'
#' \code{applySpline} Is intended for two-dimensional tidy data.tables, as
#' retured by \code{ssvFetchBigwig}
#' @export
#' @param dt a tidy data.table containing two-dimensional data
#' @param n the number of interpolation points to use per input point, see
#' \code{?spline}.  n must be > 1.
#' @param x_ the variable name of the x-values
#' @param y_ the variable name of the y-values
#' @param by_ optionally, any variables that provide grouping to the data.
#' default is none. see details.
#' @param splineFun a function that accepts x, y, and n as arguments and
#' returns a list of length 2 with named elements x and y.
#' \code{stats::spline} by default.
#' see \code{stats::spline} for details.
#'
#' @return a newly derived data.table that is \code{n} times longer than
#' original.
#'
#' @details by_ is quite powerful.  If \code{by_ = c('gene_id', 'sample_id')},
#' splines
#' will be calculated individually for each gene in each sample. alternatively
#' if \code{by_ = c('gene_id')}
#' @seealso \code{\link{ssvFetchBigwig}}
#' @importFrom stats spline
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' #data may be blockier than we'd like
#' ggplot(CTCF_in_10a_profiles_dt[, list(y = mean(y)), by = list(sample, x)]) +
#'     geom_line(aes(x = x, y = y, color = sample))
#'
#' #can be smoothed by applying a spline  (think twice about doing so,
#' #it may look prettier but may also be deceptive or misleading)
#'
#' splined_smooth = applySpline(CTCF_in_10a_profiles_dt, n = 10,
#'     y_ = 'y', by_ = c('id', 'sample'))
#' ggplot(splined_smooth[, list(y = mean(y)), by = list(sample, x)]) +
#'     geom_line(aes(x = x, y = y, color = sample))
applySpline = function(dt, n, x_ = "x", y_ = "y", by_ = c("id", "sample"),
                       splineFun = stats::spline) {
    output_GRanges = FALSE
    if(is(dt, "GRanges")){
        dt = data.table::as.data.table(dt)
        output_GRanges = TRUE
    }
    stopifnot(data.table::is.data.table(dt))
    stopifnot(is.character(x_), is.character(y_), is.character(by_))
    stopifnot(is.function(splineFun))
    if (!any(x_ == colnames(dt))) {
        stop("applySpline : x_ (", x_,
             ") not found in colnames of input data.table")
    }
    if (!any(y_ == colnames(dt))) {
        stop("applySpline : y_ (", y_,
             ") not found in colnames of input data.table")
    }
    if (by_[1] != "" | length(by_) > 1)
        if (!all(by_ %in% colnames(dt))) {
            stop("applySpline : by_ (", by_,
                 ") not found in colnames of input data.table")
        }
    dt = dt[order(get(x_))]
    if(by_[1] != ""){
        for(.by_ in by_){
            dt = dt[order(get(.by_))]
        }
    }

    stopifnot(n > 1)
    dupe_x_within_by = suppressWarnings(
        any(dt[, any(duplicated(get(x_))), by = by_]$V1))
    if (dupe_x_within_by)
        warning("applySpline : Duplicate values of x_ (\"", x_,
                "\") exist within groups defined with by_ (\"", by_, "\"). ",
                "This Results in splines through the means of yvalues at",
                " duplicated xs.")
    extra_cols = setdiff(colnames(dt), c(x_, y_, by_))
    # sdt = dt[, list(n = floor(.N * n)), by = by_]
    sdt = dt[, splineFun(x = get(x_), y = get(y_), n = floor(.N * n)), by = by_]
    colnames(sdt)[colnames(sdt) == "x"] = x_
    colnames(sdt)[colnames(sdt) == "y"] = y_

    #repair with columns dropped in by_ apply spline
    #each row will be duplicated n times
    if(length(extra_cols) > 0){
        if(n > 1){
            repair = dt[rep(seq_len(nrow(dt)), each = n),
                        c(extra_cols, by_[by_ != ""]), with = FALSE]
            sdt = cbind(sdt, repair)
        }else{
            # warning("")
            # repair = unique(dt[, c(extra_cols, by_, x_), with = FALSE])
            # repair = dt
            # sdt
            # merge(sdt, repair, by = by_)
            # unique(sdt[, by_, with = FALSE])
            # merge(sdt, repair, by = by_)
        }

    }

    k = colnames(dt) %in% colnames(sdt)
    sdt = sdt[, colnames(dt)[k], with = FALSE]
    if(output_GRanges){
        sdt = GRanges(sdt)
    }
    return(sdt)
}

#' centers profile of x and y.  default is to center by region but across all
#' samples.
#'
#' \code{centerAtMax} locates the coordinate x of the maximum in y and shifts x
#' such that it is zero at max y.
#' @export
#' @param dt data.table
#' @param x_ the variable name of the x-values. default is 'x'
#' @param y_ the variable name of the y-values default is 'y'
#' @param by_ optionally, any variables that provide grouping to the data.
#' default is none.  see details.
#' @param view_size the size in \code{x_} to consider for finding the max
#' of \code{y_}.
#' if length(view_size) == 1, range will be c(-view_size, view_size).
#' if length(view_size) > 1, range will be range(view_size).
#' default value of NULL uses complete range of x.
#' @param replace_x logical, default TRUE.
#' if TRUE x_ will be replaced with position relative to summit.
#' if FALSE x_ will be preserved and x_summitPosition added.
#' @param trim_to_valid valid \code{x_} values are those with a set \code{y_}
#' value in all \code{by_} combinations
#' @param check_by_dupes default assumption is that there should be on set of
#' x_ for a by_ instance.
#' if this is not the case and you want to disable warnings about set this
#' to FALSE.
#' @param x_precision numerical precision of x, default is 3.
#' @details character.  by_ controls at the level of the data centering is
#' applied.  If by_ is "" or NULL, a single max position will be determined
#' for the entire dataset.  If by is "id" (the default) then each region will be
#' centered individually across all samples.
#' @return
#' data.table with x (or xnew if replace_x is FALSE) shifted such that
#' x = 0 matches the maximum y-value define by by_ grouping
#' @examples
#' data(CTCF_in_10a_profiles_gr)
#' centerAtMax(CTCF_in_10a_profiles_gr, y_ = 'y', by_ = 'id',
#'   check_by_dupes = FALSE)
#' #it's a bit clearer what's happening with trimming disabled
#' #but results are less useful for heatmaps etc.
#' centerAtMax(CTCF_in_10a_profiles_gr, y_ = 'y', by_ = 'id',
#'   check_by_dupes = FALSE, trim_to_valid = FALSE)
#' #specify view_size to limit range of x values considered, prevents
#' #excessive data trimming.
#' centerAtMax(CTCF_in_10a_profiles_gr, y_ = 'y', view_size = 100, by_ = 'id',
#' check_by_dupes = FALSE)
centerAtMax = function(dt,
                       x_ = "x",
                       y_ = "y",
                       by_ = "id",
                       view_size = NULL,
                       trim_to_valid = TRUE,
                       check_by_dupes = TRUE,
                       x_precision = 3,
                       replace_x = TRUE) {

    ymax = xsummit = xnew = N = NULL  #reserve data.table variables
    output_GRanges = FALSE
    if(is(dt, "GRanges")){
        dt = data.table::as.data.table(dt)
        output_GRanges = TRUE
    }
    if (!data.table::is.data.table(dt)) {
        stop("dt must be of type data.table, was ", class(dt))
    }
    stopifnot(is.character(x_),
              is.character(y_),
              is.character(by_) || is.null(by_))
    stopifnot(is.numeric(view_size) || is.null(view_size))
    stopifnot(is.logical(trim_to_valid),
              is.logical(check_by_dupes),
              is.logical(replace_x))

    if (!any(x_ == colnames(dt))) {
        stop("centerAtMax : x_ (", x_,
             ") not found in colnames of input data.table")
    }
    if (!any(y_ == colnames(dt))) {
        stop("centerAtMax : y_ (", y_,
             ") not found in colnames of input data.table")
    }
    if(length(unique(dt[[x_]])) != length(unique(round(dt[[x_]], x_precision)))){
        stop(
            "centerAtMax : x_precision,",
            x_precision,
            " , is too low, how many decimal places is x defined to?"
        )
    }
    # check_by_dupes = FALSE
    if (is.null(by_)) {
        by_ = ""
        check_by_dupes = TRUE
    }
    if (all(by_ != ""))
        if (!any(by_ %in% colnames(dt))) {
            stop("centerAtMax : by_ (", by_,
                 ") not found in colnames of input data.table")
        }
    if (check_by_dupes) {
        dupe_x_within_by = suppressWarnings(any(dt[, any(duplicated(get(x_))),
                                                   by = by_]$V1))
        if (dupe_x_within_by)
            message("centerAtMax : duplicate values of x_ (", x_,
                    ") exist within groups defined with by_ (", by_,
                    ").\n    If this is the desired functionality, set",
                    "check_by_dupes <- FALSE to hide future messages. ",
                    "If no by_ grouping is intended set by_ <- \"\" as",
                    "well.")
    }
    dt = data.table::copy(dt)
    if (is.null(view_size)) {
        view_size = range(dt[[x_]])
    } else if (length(view_size) == 1) {
        view_size = c(-view_size, view_size)
    }
    view_size = range(view_size)
    #verify view_size
    k = dt[[x_]] >= min(view_size) & dt[[x_]] <= max(view_size)
    if(!any(k)){
        stop("view_size is too restrictive for position values. Postions have a range of ", paste(range(dt[[x_]]), collapse = " to "), " and view_size has a range of ", paste(range(view_size), collapse = " to "),
             ".\nFor summary data, view_sizes are typically fractional values less than 1 and for sample data view size should be in bp.")
    }

    closestToZero = function(x) {
        x[order(abs(x))][1]
    }
    dt[, `:=`(ymax, max(get(y_)[get(x_) <= max(view_size) &
                                    get(x_) >= min(view_size)])), by = by_]
    dt[, `:=`(xsummit, closestToZero(get(x_)[get(y_) == ymax])), by = by_]
    dt[, `:=`(xnew, get(x_) - xsummit)]
    dt$xnew = round(dt$xnew, x_precision)
    dt[, `:=`(ymax, NULL)]
    dt[, `:=`(xsummit, NULL)]
    if (trim_to_valid) {
        # valid values of x are those with values in all by_ defined grouping
        xcounts = dt[, .N, xnew]
        xcounts = xcounts[N == max(N)]
        dt = dt[xnew %in% xcounts$xnew]
    }
    if (replace_x) {
        data.table::set(dt, j = x_, value = dt$xnew)
        dt$xnew = NULL
    } else {
        colnames(dt)[colnames(dt) == "xnew"] = paste0(x_, "_summitPosition")
    }
    if(output_GRanges){
        dt = GRanges(dt)
    }

    return(dt)
}

#' Centers query GRanges at maximum signal in prof_dt.
#'
#' @param prof_dt a GRanges or data.table as returned by ssvFetch*.
#' @param qgr the GRanges used to query ssvFetch* as the qgr argument.
#' @param x_ positional variable.  Should almost always be the default, "x".
#' @param y_ the signal value variable.  Likely the default value of "y" but
#'   could be "y_norm" if append_ynorm was applied to data.
#' @param by_ region identifier variable. Should almost always be the default,
#'   "id".
#' @param width Desired width of final regions.  Default is 1.
#' @param view_size the size in \code{x_} to consider for finding the max
#' of \code{y_}.
#' if length(view_size) == 1, range will be c(-view_size, view_size).
#' if length(view_size) > 1, range will be range(view_size).
#' default value of NULL uses complete range of x.
#'
#' @return a GRanges with same mcols as qgr that has been centered based on
#'   signal in prof_dt and with regions of specified width.
#' @export
#'
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' data(CTCF_in_10a_profiles_gr)
#' data(CTCF_in_10a_profiles_dt)
#' centerGRangesAtMax(CTCF_in_10a_profiles_dt, CTCF_in_10a_overlaps_gr)
#' centerGRangesAtMax(CTCF_in_10a_profiles_gr, CTCF_in_10a_overlaps_gr)
#'
#' centerGRangesAtMax(CTCF_in_10a_profiles_gr, CTCF_in_10a_overlaps_gr, view_size = 100)
#'
centerGRangesAtMax = function(prof_dt, qgr, x_ = "x", y_ = "y", by_ = "id", width = 1, view_size = NULL){
    if(length(by_) > 1) stop("only by_ of length 1 supported.")
    if(is(prof_dt, "GRanges")){
        prof_dt = data.table::as.data.table(prof_dt)
    }
    cent_dt = centerAtMax(prof_dt, x_ = x_, y_ = y_, by_ = by_, check_by_dupes = FALSE, view_size = view_size)
    cent_dt = cent_dt[, .SD[which(get(x_) == min(abs(get(x_))))[1],], c(by_)]
    cent_gr = GenomicRanges::GRanges(cent_dt[, list(seqnames, start = (start + end)/2, end = (start + end)/2)])
    names(cent_gr) = cent_dt[[by_]]

    qgr = prepare_fetch_GRanges_names(qgr)
    GenomicRanges::strand(cent_gr) = GenomicRanges::strand(qgr[names(cent_gr)])
    GenomicRanges::mcols(cent_gr) = GenomicRanges::mcols(qgr[names(cent_gr)])
    cent_gr = cent_gr[names(qgr)]
    GenomicRanges::resize(cent_gr, width, fix = "center")
}

#' findMaxPos
#'
#' @param prof_dt a GRanges or data.table as returned by ssvFetch*.
#' @param qgr the GRanges used to query ssvFetch* as the qgr argument.
#' @param x_ positional variable.  Should almost always be the default, "x".
#' @param y_ the signal value variable.  Likely the default value of "y" but
#'   could be "y_norm" if append_ynorm was applied to data.
#' @param by_ region identifier variable. Should almost always be the default,
#'   "id".
#' @param width Desired width of final regions.  Default is 1.
#'
#' @return data.table of relative x position from center per id
#' @export
#'
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' data(CTCF_in_10a_profiles_gr)
#' data(CTCF_in_10a_profiles_dt)
#' findMaxPos(CTCF_in_10a_profiles_dt, CTCF_in_10a_overlaps_gr)
#' findMaxPos(CTCF_in_10a_profiles_gr, CTCF_in_10a_overlaps_gr)
findMaxPos = function(prof_dt, qgr, x_ = "x", y_ = "y", by_ = "id", width = 1){
    #binding for data.table
    x = y = id = NULL
    if(length(by_) > 1) stop("only by_ of length 1 supported.")
    if(is(prof_dt, "GRanges")){
        prof_dt = data.table::as.data.table(prof_dt)
    }
    prof_dt[, list(x = x[which.max(y)[1]]), list(id)]
}

testthat::context("CenterAtMax")
library(seqsetvis)
library(testthat)
library(data.table)
library(GenomicRanges)

data(CTCF_in_10a_profiles_dt)
data(CTCF_in_10a_profiles_gr)
data(CTCF_in_10a_overlaps_gr)


n = 8
xs = (1:n-5)*5
fun = function(x)(x^2)
xs = c(xs, xs, xs)
ys = -c(fun(xs + floor((seq_along(xs)-1) / n)*5)) + floor((seq_along(xs)-1) / n)*50
#setup data.table of parabolic curves, each transposed on the x-axis
test_dt = data.table::data.table(xvals = xs, yvals = ys, locus = rep(letters[1:3], each = n))
# ggplot(test_dt, aes(x = xvals, y = yvals, col = locus)) + geom_line() + geom_point()

test_dt2 = rbind(test_dt, test_dt)
test_dt2$sample = rep(1:2, each = nrow(test_dt))
test_dt2[sample == 2, xvals := xvals + 10]
test_dt2[sample == 2, yvals := yvals + 300]
# ggplot(test_dt2, aes(x = xvals, y = yvals, col = locus)) + geom_line() + geom_point() + facet_grid(sample ~ .)
#should set by
cm_dt_noBy = centerAtMax(test_dt, x_ = "xvals", y_ = "yvals", by_ = "", check_by_dupes = FALSE)
ggplot(cm_dt_noBy, aes(x = xvals, y = yvals, col = locus)) +
    geom_line() +
    geom_point() +
    geom_point(data = test_dt, aes(col = locus), alpha = .3) +
    geom_line(data = test_dt, aes(col = locus), alpha = .3) +
    labs(title = "With by_ unset, all profiles are shifted to the global maximum")
#proper use of by
cm_dt_wBy = centerAtMax(test_dt, x_ = "xvals", y_ = "yvals", by_ = "locus")
ggplot(cm_dt_wBy, aes(x = xvals, y = yvals, col = locus)) +
    geom_line() +
    geom_point() +
    geom_point(data = test_dt, aes(col = locus), alpha = .3) +
    geom_line(data = test_dt, aes(col = locus), alpha = .3) +
    labs(title = "With by_ set properly, all profiles are shifted to their individual maximum")
#effect of trimming
cm_dt_wBy_noTrim =centerAtMax(test_dt, x_ = "xvals", y_ = "yvals", by = "locus", trim_to_valid = FALSE)
ggplot(cm_dt_wBy, aes(x = xvals, y = yvals, col = locus)) +
    geom_line() +
    geom_point() +
    geom_point(data = cm_dt_wBy_noTrim, aes(col = locus), alpha = .3) +
    geom_line(data = cm_dt_wBy_noTrim, aes(col = locus), alpha = .3) +
    labs(title = "When trim_to_valid is TRUE, non-universal values of x are removed.")



doTests_centerAtMax = function(test_object, test_object2){
    test_that("centerAtMax warnings for by_ specification", {
        expect_message(centerAtMax(test_object, x_ = "xvals", y_ = "yvals", by_ = NULL), regexp = "centerAtMax")
        expect_failure(expect_message(centerAtMax(test_object, x_ = "xvals", y_ = "yvals", by_ = "locus")))
    })

    test_that("centerAtMax errors for unmatched variable name", {
        expect_error(
            centerAtMax(test_object),
            regexp = "centerAtMax")
        expect_error(
            centerAtMax(test_object, x_ = "xvals"),
            regexp = "centerAtMax")
        expect_error(
            centerAtMax(test_object, x_ = "xvals", y_ = "yvals", by_ = "blerp"),
            regexp = "centerAtMax")
    })

    test_that("centerAtMax trimming reduces ranges", {
        cm_noTrim = centerAtMax(test_object, x_ = "xvals", y_ = "yvals", by_ = "locus", trim_to_valid = FALSE)
        expect_equal(nrow(cm_noTrim), nrow(test_object))
        cm_wTrim = centerAtMax(test_object, x_ = "xvals", y_ = "yvals", by_ = "locus", trim_to_valid = TRUE)
        nr_test = nrow(as.data.frame(test_object))
        nr_wTrim = nrow(as.data.frame(cm_wTrim))
        nr_noTrim = nrow(as.data.frame(test_object))
        expect_lt(nr_wTrim, nr_test)
    })

    test_that("centerAtMax input must be data.table", {
        expect_error(centerAtMax(dt = data.frame(1:3), x_ = "xvals", y_ = "yvals", by_ = c("grp", "grp2")), regexp = "must be of type data.table")
        expect_error(centerAtMax(dt = matrix(1:3), x_ = "xvals", y_ = "yvals", by_ = c("grp", "grp2")), regexp = "must be of type data.table")
        expect_error(centerAtMax(dt = (1:3), x_ = "xvals", y_ = "yvals", by_ = c("grp", "grp2")), regexp = "must be of type data.table")
        expect_error(centerAtMax(dt = "(1:3)", x_ = "xvals", y_ = "yvals", by_ = c("grp", "grp2")), regexp = "must be of type data.table")
    })

    test_that("centerAtMax viewSize", {
        dt1 = centerAtMax(
            dt = test_object2,
            x_ = "xvals",
            y_ = "yvals",
            by_ = c("locus"),
            view_size = 5,
            trim_to_valid = FALSE,
            check_by_dupes = FALSE
        )
        dt2 = centerAtMax(
            dt = test_object2,
            x_ = "xvals",
            y_ = "yvals",
            by_ = c("locus"),
            view_size = 10,
            trim_to_valid = FALSE,
            check_by_dupes = FALSE
        )
        nr1 = nrow(as.data.frame(dt1))
        nr2 = nrow(as.data.frame(dt2))
        expect_equal(nr1, nr2)
        expect_true(!all(dt2$xvals == dt1$xvals))
    })

    test_that("centerAtMax viewSize", {
        dt1 = centerAtMax(
            dt = test_object2,
            x_ = "xvals",
            y_ = "yvals",
            by_ = c("locus"),
            check_by_dupes = FALSE,
            replace_x = FALSE
        )
        dt2 = centerAtMax(
            dt = test_object2,
            x_ = "xvals",
            y_ = "yvals",
            by_ = c("locus"),
            check_by_dupes = FALSE,
            replace_x = TRUE
        )

        dt1 = as.data.frame(dt1)
        dt2 = as.data.frame(dt2)

        expect_equal(colnames(dt1)[ncol(dt1)], "xvals_summitPosition")
        expect_equal(ncol(dt1), ncol(dt2)+1)
    })
}

doTests_centerAtMax(test_dt, test_dt2)
test_gr = GRanges(seqnames = "chrTest",
                  IRanges(test_dt$xvals+100, test_dt$xvals+104),
                  xvals = test_dt$xvals,
                  yvals = test_dt$yvals,
                  locus = test_dt$locus)
test_gr2 = GRanges(seqnames = "chrTest",
                  IRanges(test_dt2$xvals+100, test_dt2$xvals+104),
                  xvals = test_dt2$xvals,
                  yvals = test_dt2$yvals,
                  locus = test_dt2$locus,
                  sample = test_dt2$sample)
doTests_centerAtMax(test_object = test_gr, test_object2 = test_gr2)
doTests_centerAtMax(as.data.table(test_gr), as.data.table(test_gr2))

test_that("centerAtMax multiple by_ arguments", {
    pdt = copy(test_dt2)
    pdt$centered = "centered: no"
    cm = centerAtMax(pdt, x_ = "xvals", y_ = "yvals", by_ = c("sample", "locus"), trim_to_valid = FALSE)
    cm$centered = "centered: yes"
    pdt =  rbind(pdt, cm)
    p = ggplot(pdt, aes(x = xvals, y = yvals, col = locus)) +
        geom_line() +
        geom_point() +
        labs(title = "With compound by_, all profiles are shifted to their individual maximum") +
        facet_grid(sample ~ centered)
    maxpos = cm[, xvals[which.max(yvals)], by = c("sample", "locus")]$V1
    expect_true(all(maxpos == 0))
})

test_that("centerAtMax sample by_ argument", {
    pdt = copy(test_dt2)
    pdt$centered = "centered: no"
    cm = centerAtMax(pdt, x_ = "xvals", y_ = "yvals", by_ = c("sample"), trim_to_valid = FALSE, check_by_dupes = FALSE)
    cm$centered = "centered: yes"
    pdt =  rbind(pdt, cm)
    p = ggplot(pdt, aes(x = xvals, y = yvals, col = locus)) +
        geom_line() +
        geom_point() +
        labs(title = "With sample by_, profiles are shifted equally per locus",
             subtitle = "akin to per locus across samples, a normal use") +
        facet_grid(sample ~ centered)
    maxpos = cm[, xvals[which.max(yvals)], by = "sample"]$V1
    expect_true(all(maxpos == 0))
})

test_that("centerAtMax locus by_ argument", {
    pdt = copy(test_dt2)
    pdt$centered = "centered: no"
    cm = centerAtMax(pdt, x_ = "xvals", y_ = "yvals", by_ = c("locus"), trim_to_valid = FALSE, check_by_dupes = FALSE)
    cm$centered = "centered: yes"
    pdt =  rbind(pdt, cm)
    p = ggplot(pdt, aes(x = xvals, y = yvals, col = locus)) +
        geom_line() +
        geom_point() +
        labs(title = "With locus by_, all profiles are shifted equally per sample",
             subtitle = "akin to per sample across loci, a weird use") +
        facet_grid(sample ~ centered)
    maxpos = cm[, xvals[which.max(yvals)], by = "locus"]$V1
    expect_true(all(maxpos == 0))
})

test_that("centerGrangesAtMax", {
    res1 = centerGRangesAtMax(CTCF_in_10a_profiles_dt, CTCF_in_10a_overlaps_gr)
    expect_s4_class(res1, "GRanges")
    expect_equal(names(res1), names(CTCF_in_10a_overlaps_gr))
    expect_true(any(start(res1) != start(CTCF_in_10a_overlaps_gr)))

    res2 = centerGRangesAtMax(CTCF_in_10a_profiles_gr, CTCF_in_10a_overlaps_gr)
    expect_true(all(start(res1) == start(res2)))
})

testthat::context("parallel")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
library(data.table)

qgr = CTCF_in_10a_overlaps_gr[1:5]
qgr = centerFixedSizeGRanges(qgr, 500)
names(qgr) = NULL
#bed used to intersect bam
# rtracklayer::export.bed(qgr, con = "ctcf_5.bed")
bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
test_bw = system.file("extdata/test_loading.bw", package = "seqsetvis", mustWork = TRUE)

test_that("ssvFetchBam summary method correct bins", {
    skip_on_os("windows")
    test_qgr2 = qgr

    gr_sample_split1 = ssvFetchBam(bam_file, win_size = 5,
                                   fragLens = NA,
                                   qgr = test_qgr2, win_method = "summary",
                                   n_region_splits = 1)
    gr_sample_split2 = ssvFetchBam(bam_file, win_size = 5,
                                   fragLens = NA,
                                   qgr = test_qgr2, win_method = "summary",
                                   n_region_splits = 2)
    gr_sample_split3 = ssvFetchBam(bam_file, win_size = 5,
                                   fragLens = NA,
                                   qgr = test_qgr2, win_method = "summary",
                                   n_region_splits = 5)

    expect_equal(gr_sample_split1$id, gr_sample_split2$id)
    expect_equal(gr_sample_split1$x, gr_sample_split2$x)
    expect_equal(gr_sample_split1$y, gr_sample_split2$y)
    expect_equal(gr_sample_split1$sample, gr_sample_split2$sample)

    expect_equal(gr_sample_split1$id, gr_sample_split3$id)
    expect_equal(gr_sample_split1$x, gr_sample_split3$x)
    expect_equal(gr_sample_split1$y, gr_sample_split3$y)
    expect_equal(gr_sample_split1$sample, gr_sample_split3$sample)
})

test_that("ssvFetchBigwig works with character vector bw_files", {
    skip_on_os("windows")
    pos = c(20, 180, 210, 440, 520, 521)
    region_size = 30
    test_qgr2 = GRanges("chrTest", IRanges(pos+1, pos + region_size))
    bw_files = rep(test_bw, 3)
    names(bw_files) = paste0("bw_", 1:3)

    gr_sample_split1 = ssvFetchBigwig(bw_files, win_size = 5,
                                      qgr = test_qgr2, win_method = "summary",
                                      n_region_splits = 1, return_data.table = TRUE)
    gr_sample_split2 = ssvFetchBigwig(bw_files, win_size = 5,
                                      qgr = test_qgr2, win_method = "summary",
                                      n_region_splits = 2, return_data.table = TRUE)
    gr_sample_split3 = ssvFetchBigwig(bw_files, win_size = 5,
                                      qgr = test_qgr2, win_method = "summary",
                                      n_region_splits = 5, return_data.table = TRUE)

    expect_equal(gr_sample_split1$id, gr_sample_split2$id)
    expect_equal(gr_sample_split1$x, gr_sample_split2$x)
    expect_equal(gr_sample_split1$y, gr_sample_split2$y)
    expect_equal(gr_sample_split1$sample, gr_sample_split2$sample)

    expect_equal(gr_sample_split1$id, gr_sample_split3$id)
    expect_equal(gr_sample_split1$x, gr_sample_split3$x)
    expect_equal(gr_sample_split1$y, gr_sample_split3$y)
    expect_equal(gr_sample_split1$sample, gr_sample_split3$sample)
})

testthat::context("StrandFlipping")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
library(data.table)

qgr = CTCF_in_10a_overlaps_gr[1:5]
qgr = centerFixedSizeGRanges(qgr, 500)

#bed used to intersect bam
# rtracklayer::export.bed(qgr, con = "ctcf_5.bed")
bam_file = system.file("extdata/test.bam", package = "seqsetvis")
xls_file = system.file("extdata/test_peaks.xls", package = "seqsetvis")

test_that("ssvFetchBam strand flipping both, + and -", {
    # skip_on_os("windows")
    test_gr = qgr
    strand(test_gr) = rep("+", 5)
    test_res = ssvFetchBam(bam_file, win_size = 20,
                           fragLens = NA, target_strand = "both",
                           qgr = test_gr, return_data.table = TRUE)

    flip_gr = qgr
    strand(flip_gr) = rep("-", 5)
    flip_res = ssvFetchBam(bam_file, win_size = 20,
                           fragLens = NA, target_strand = "both",
                           qgr = flip_gr, return_data.table = TRUE)

    # ggplot(test_res, aes(x = x, y = y, color = strand)) + geom_path() + facet_wrap("id")
    # ggplot(flip_res, aes(x = x, y = y, color = strand)) + geom_path() + facet_wrap("id")

    expect_equal(test_res[id == 1 & strand == "+"]$y,
                 rev(flip_res[id == 1 & strand == "-"]$y))
    expect_equal(test_res[id == 2 & strand == "+"]$y,
                 rev(flip_res[id == 2 & strand == "-"]$y))
    expect_equal(test_res[id == 3 & strand == "+"]$y,
                 rev(flip_res[id == 3 & strand == "-"]$y))

    expect_equal(test_res[id == 1 & strand == "-"]$y,
                 rev(flip_res[id == 1 & strand == "+"]$y))
    expect_equal(test_res[id == 2 & strand == "-"]$y,
                 rev(flip_res[id == 2 & strand == "+"]$y))
    expect_equal(test_res[id == 3 & strand == "-"]$y,
                 rev(flip_res[id == 3 & strand == "+"]$y))
})

test_that("ssvFetchBam strand flipping both, * and -", {
    # skip_on_os("windows")
    test_gr = qgr
    strand(test_gr) = rep("*", 5)
    test_res = ssvFetchBam(bam_file, win_size = 20,
                           fragLens = NA, target_strand = "both",
                           qgr = test_gr, return_data.table = TRUE)

    flip_gr = qgr
    strand(flip_gr) = rep("-", 5)
    flip_res = ssvFetchBam(bam_file, win_size = 20,
                           fragLens = NA, target_strand = "both",
                           qgr = flip_gr, return_data.table = TRUE)

    expect_equal(test_res[id == 1 & strand == "+"]$y,
                 rev(flip_res[id == 1 & strand == "-"]$y))
    expect_equal(test_res[id == 2 & strand == "+"]$y,
                 rev(flip_res[id == 2 & strand == "-"]$y))
    expect_equal(test_res[id == 3 & strand == "+"]$y,
                 rev(flip_res[id == 3 & strand == "-"]$y))

    expect_equal(test_res[id == 1 & strand == "-"]$y,
                 rev(flip_res[id == 1 & strand == "+"]$y))
    expect_equal(test_res[id == 2 & strand == "-"]$y,
                 rev(flip_res[id == 2 & strand == "+"]$y))
    expect_equal(test_res[id == 3 & strand == "-"]$y,
                 rev(flip_res[id == 3 & strand == "+"]$y))
})

test_that("ssvFetchBam strand flipping +, + and -", {
    # skip_on_os("windows")
    test_gr = qgr
    strand(test_gr) = rep("+", 5)
    test_res = ssvFetchBam(bam_file, win_size = 20,
                           fragLens = NA, target_strand = "+",
                           qgr = test_gr[1], return_data.table = TRUE)

    flip_gr = qgr
    strand(flip_gr) = rep("-", 5)
    flip_res = ssvFetchBam(bam_file, win_size = 20,
                           fragLens = NA, target_strand = "-",
                           qgr = flip_gr[1], return_data.table = TRUE)

    expect_equal(test_res$y, rev(flip_res$y))
})

test_that("ssvFetchBam strand flipping fragLen", {
    #flip_strand precedes fragLen extenstion
    test_gr = qgr
    strand(test_gr) = rep("+", 5)
    test_res = ssvFetchBam(bam_file, win_size = 20,
                           fragLens = 200, target_strand = "both",
                           qgr = test_gr[1], return_data.table = TRUE)

    flip_gr = qgr
    strand(flip_gr) = rep("+", 5)
    flip_res = ssvFetchBam(bam_file, win_size = 20,
                           fragLens = 200, target_strand = "both",
                           qgr = flip_gr[1], return_data.table = TRUE, flip_strand = TRUE)
    test_res$group = "normal"
    flip_res$group = "flip"
    #this is much more obvious in a plot
    # ggplot(rbind(test_res, flip_res), aes(x = x, y = y, color = strand)) + geom_path() + facet_wrap("group", ncol = 1)
    test_cor = cor(dcast(test_res, "x~strand", value.var = "y")[,-1])[1,2]
    flip_cor = cor(dcast(flip_res, "x~strand", value.var = "y")[,-1])[1,2]
    expect_true(abs(test_cor - flip_cor) < .1)

    expect_false(all(test_res$y == rev(flip_res$y)))
})

testthat::context("FetchGranges")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
library(data.table)

qgr = CTCF_in_10a_overlaps_gr[1:10]
qgr = centerFixedSizeGRanges(qgr, 500)
#bed used to intersect bam


test_that("ssvFetchGRanges win_size", {
    res50 = ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, qgr, win_size = 50)
    expect_equal(300, length(res50))
    res10 = ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, qgr, win_size = 50/5)
    expect_equal(300*5, length(res10))
})


test_that("ssvFetchGRanges attribs", {
    res10 = ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, qgr, win_size = 10, file_attribs = data.frame("a" = 1:3, "bc" = 4:6))
    expect_true(all(res10$a %in% 1:3))
    expect_true(all(res10$bc %in% 4:6))
})

test_that("ssvFetchGRanges win_method", {
    res5sum = ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, qgr, win_size = 5, win_method = "summary")
    expect_equal(length(qgr)*length(CTCF_in_10a_narrowPeak_grs)*5, length(res5sum))
    expect_true(!"tile_id" %in% colnames(mcols(res5sum)))
    expect_gte(min(res5sum$x), -.5)
    expect_lte(max(res5sum$x), .5)
})


test_that("ssvFetchGRanges anchor", {
    res5left = ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, qgr, win_size = 5,
                              win_method = "summary", anchor = "left")
    expect_equal(length(qgr)*length(CTCF_in_10a_narrowPeak_grs)*5, length(res5left))
    expect_gte(min(res5left$x), 0)
    expect_lte(max(res5left$x), 1)
})

test_that("ssvFetchGRanges target_strand", {
    res50star = ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, qgr,
                              target_strand = "*", return_data.table = TRUE)
    expect_equal(max(res50star$y), 1)
    res50minus = ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, qgr,
                               target_strand = "-", return_data.table = TRUE)
    expect_equal(max(res50minus$y), 0)
    res50plus = ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, qgr,
                               target_strand = "+", return_data.table = TRUE)
    expect_equal(max(res50plus$y), 0)
    res50both = ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, qgr,
                               target_strand = "both", return_data.table = TRUE)
    expect_equal(max(res50both$y), 0)

    np_gr = CTCF_in_10a_narrowPeak_grs
    strand(np_gr$MCF10A_CTCF[1:3]) = "+"
    strand(np_gr$MCF10AT1_CTCF[4:6]) = "-"

    res50minus = ssvFetchGRanges(np_gr, qgr,
                                 target_strand = "-", return_data.table = TRUE)
    expect_equal(max(res50minus$y), 1)
    res50plus = ssvFetchGRanges(np_gr, qgr,
                                target_strand = "+", return_data.table = TRUE)
    expect_equal(max(res50plus$y), 1)
    res50both = ssvFetchGRanges(np_gr, qgr,
                                target_strand = "both", return_data.table = TRUE)
    expect_equal(max(res50both$y), 1)

    expect_equal(nrow(res50both), nrow(res50minus)*2)
})

test_that("ssvFetchGRanges qualitative sample", {
    qual_gr = CTCF_in_10a_narrowPeak_grs
    qual_gr = lapply(qual_gr, function(x){
        x$quality = "low"
        x[x$qValue > 150]$quality = "high"
        x[x$qValue > 250]$quality = "very high"
        x
    })

    dt = ssvFetchGRanges(qual_gr, qgr = qgr,
                         attrib_var = "quality",
                         fill_value = "no peak",
                         return_data.table = TRUE)
    dt[, .N, sample]
    expect_equal(nrow(dt[is.na(y), .N, sample]), 0)
    expect_equal(length(table(dt$y)), 4)
})

test_that("ssvFetchGRanges qualitative summary", {
    qual_gr = CTCF_in_10a_narrowPeak_grs
    qual_gr = lapply(qual_gr, function(x){
        x$quality = "low"
        x[x$qValue > 150]$quality = "high"
        x[x$qValue > 250]$quality = "very high"
        x
    })

    dt = ssvFetchGRanges(qual_gr, qgr = qgr,
                         win_method = "summary",
                         attrib_var = "quality",
                         fill_value = "no peak",
                         return_data.table = TRUE)
    expect_true(!"tile_id" %in% colnames(dt))
    dt[, .N, sample]
    dt$quality = factor(dt$quality, levels = c("no peak", "low", "high", "very high"))
    # ggplot(dt, aes(x = x, y = y, group = id)) + geom_path() + facet_grid(quality~sample)
    ggplot(dt, aes(x = x, y = id, fill = y, group = id)) + geom_raster() + facet_grid(quality~sample)
    expect_equal(nrow(dt[is.na(y), .N, sample]), 0)

    expect_true(all(c("quality", "y") %in% colnames(dt)))
})

test_that("ssvFetchGRanges target_strand = 'both' - coverage", {
    qual_gr = CTCF_in_10a_narrowPeak_grs
    qual_gr = lapply(qual_gr, function(x){
        x$quality = "low"
        x[x$qValue > 150]$quality = "high"
        x[x$qValue > 250]$quality = "very high"
        strand(x) = "+"
        x
    })

    dt = ssvFetchGRanges(qual_gr, qgr = qgr,
                         win_method = "summary",
                         target_strand = "both",
                         fill_value = 0,
                         #attrib_var = "quality",
                         #fill_value = "no peak",
                         return_data.table = TRUE)
    expect_equal(max(dt$y), 1)
    expect_equal(min(dt$y), 0)

    dt = ssvFetchGRanges(qual_gr, qgr = qgr,
                         win_method = "sample",
                         target_strand = "both", fill_value = 0,
                         #attrib_var = "quality",
                         #fill_value = "no peak",
                         return_data.table = TRUE)
    expect_equal(max(dt$y), 1)
    expect_equal(min(dt$y), 0)
})

test_that("ssvFetchGRanges target_strand = 'both' - quality", {
    qual_gr = CTCF_in_10a_narrowPeak_grs
    qual_gr = lapply(qual_gr, function(x){
        x$quality = "low"
        x[x$qValue > 150]$quality = "high"
        x[x$qValue > 250]$quality = "very high"
        strand(x) = "+"
        x
    })

    dt = ssvFetchGRanges(qual_gr, qgr = qgr,
                         win_method = "summary",
                         target_strand = "both",
                         attrib_var = "quality",
                         fill_value = "no peak",
                         return_data.table = TRUE)

    expect_equal(table(dt[strand == "+"]$quality) %>% length, 4)
    expect_equal(table(dt[strand == "-"]$quality) %>% length, 1)

    dt = ssvFetchGRanges(qual_gr, qgr = qgr,
                         win_method = "sample",
                         target_strand = "both",
                         attrib_var = "quality",
                         fill_value = "no peak",
                         return_data.table = TRUE)
    expect_equal(table(dt[strand == "+"]$y) %>% length, 4)
    expect_equal(table(dt[strand == "-"]$y) %>% length, 1)
})

testthat::context("Splines")
library(testthat)
library(seqsetvis)
library(GenomicRanges)
library(data.table)
xs = (0:5-1.5)*5
fun = function(x)x^2
test_dt2 = data.table::data.table(xvals = c(xs, xs-5, xs - 10), yvals = c(fun(xs) - 100, fun(xs-5)+60 , fun(xs - 10)+150 ), grp = rep(letters[1:3], each = length(xs)))
test_gr2 = GRanges(test_dt2[, .(seqnames = "chrTest", start = ceiling(xvals - min(xvals) - 2.5), end = ceiling(xvals - min(xvals) + 2), xvals, yvals), by = .(grp)])
test_dt3 = as.data.table(test_gr2)
#
# ggplot(test_dt2, aes(x = xvals, y = yvals)) + geom_point() + stat_summary(fun.y = mean, geom="line")
# ggplot(test_dt2, aes(x = xvals, y = yvals, col = grp)) + geom_line() + geom_point()
#
# #should set by
# sp_dt_noBy =applySpline(test_dt2, x_ = "xvals", y_ = "yvals", n = 4)
# ggplot(sp_dt_noBy, aes(x = xvals, y = yvals)) + geom_point() + stat_summary(fun.y = mean, geom="line") +
#   geom_point(data = test_dt2, aes(col = grp)) + stat_summary(data = test_dt2, aes(col = grp), fun.y = mean, geom="line")
# #proper use of by
# sp_dt_wBy =applySpline(test_dt2, x_ = "xvals", y_ = "yvals", n = 4, by = "grp")
# ggplot(sp_dt_wBy, aes(x = xvals, y = yvals, col = grp)) + geom_point() + stat_summary(fun.y = mean, geom="line") +
#   geom_point(data = test_dt2, aes(col = grp), alpha = .4) + stat_summary(data = test_dt2, aes(col = grp), fun.y = mean, geom="line", alpha = .4)

doTest_applySpline = function(test_object){

    test_that("applySpline returned class same as input", {
        original_class = class(test_object)
        sp_dt = applySpline(test_object, x_ = "xvals", y_ = "yvals", by_ = "grp", n = 4)
        expect_equal(class(sp_dt), original_class)
    })

    test_that("applySpline returned colnames equal", {
        original_colnames = colnames(as.data.frame(test_object))
        sp_dt = applySpline(test_object, x_ = "xvals", y_ = "yvals", by_ = "grp", n = 4)
        res_colnames = colnames(as.data.frame(sp_dt))
        expect_equal(res_colnames, original_colnames)
    })

    test_that("applySpline warnings for by_ specification", {
        expect_warning(applySpline(test_object, x_ = "xvals", y_ = "yvals", n = 4, by_ = ""), regexp = "applySpline")
        expect_failure(expect_warning(applySpline(test_object, x_ = "xvals", y_ = "yvals", n = 4, by_ = "grp")))
    })

    test_that("applySpline errors for unmatched variable name", {
        expect_error(
            applySpline(test_object),
            regexp = "applySpline")
        expect_error(
            applySpline(test_object, x_ = "xvals"),
            regexp = "applySpline")
        expect_error(
            applySpline(test_object, x_ = "xvals", y_ = "yvals", by_ = "blerp"),
            regexp = "applySpline")
    })

    test_that("applySpline n*length in result", {
        sp_dt4 = applySpline(test_object, x_ = "xvals", y_ = "yvals", by_ = "grp", n = 4)
        expect_equal(nrow(as.data.frame(test_object))*4, nrow(as.data.frame(sp_dt4)))
        sp_dt10 = applySpline(test_object, x_ = "xvals", y_ = "yvals", by_ = "grp", n = 10)
        expect_equal(nrow(as.data.frame(test_object))*10, nrow(as.data.frame(sp_dt10)))
    })

    test_that("applySpline no warning from multiple by_", {
        test_object$grp2  = test_object$grp
        expect_failure(expect_warning(applySpline(dt = test_object, x_ = "xvals", y_ = "yvals", by_ = c("grp", "grp2"), n = 4)))
    })

    test_that("applySpline error if not data.table or GRanges", {
        expect_error(applySpline(dt = data.frame(1:3), x_ = "xvals", y_ = "yvals", by_ = c("grp", "grp2"), n = 4), regexp = "data.table::is.data.table\\(dt\\) is not TRUE")
        expect_error(applySpline(dt = matrix(1:3), x_ = "xvals", y_ = "yvals", by_ = c("grp", "grp2"), n = 4), regexp = "data.table::is.data.table\\(dt\\) is not TRUE")
        expect_error(applySpline(dt = (1:3), x_ = "xvals", y_ = "yvals", by_ = c("grp", "grp2"), n = 4), regexp = "data.table::is.data.table\\(dt\\) is not TRUE")
        expect_error(applySpline(dt = "(1:3)", x_ = "xvals", y_ = "yvals", by_ = c("grp", "grp2"), n = 4), regexp = "data.table::is.data.table\\(dt\\) is not TRUE")
    })

    test_that("applySpline n < 1 throws caught error", {
        expect_error(applySpline(test_object, x_ = "xvals", y_ = "yvals", by_ = "grp", n = .9), regexp = "n >")
    })

    test_that("applySpline order doesn't matter", {
        r1 = sample(seq_len(nrow(as.data.frame(test_object))))
        r2 = sample(seq_len(nrow(as.data.frame(test_object))))
        t1 = test_object[r1]
        t2 = test_object[r2]
        # if(class(test_object)[1] == "GRanges"){
        #
        # }else{
        #
        # }
        sp_o = applySpline(test_object, x_ = "xvals", y_ = "yvals", by_ = "grp", n = 3)
        sp_1 = applySpline(t1, x_ = "xvals", y_ = "yvals", by_ = "grp", n = 3)
        sp_2 = applySpline(t2, x_ = "xvals", y_ = "yvals", by_ = "grp", n = 3)

        sp_o = sp_o[order(sp_o$xvals)]
        sp_o = sp_o[order(sp_o$grp)]

        sp_1 = sp_1[order(sp_1$xvals)]
        sp_1 = sp_1[order(sp_1$grp)]

        sp_2 = sp_2[order(sp_2$xvals)]
        sp_2 = sp_2[order(sp_2$grp)]

        expect_equal(sp_o$yvals, sp_1$yvals)
        expect_equal(as.data.frame(sp_o)$start, as.data.frame(sp_1)$start)

        expect_equal(sp_o$yvals, sp_2$yvals)
        expect_equal(as.data.frame(sp_o)$start, as.data.frame(sp_2)$start)
    })

}

#dt = copy(test_object); x_ = "xvals"; y_ = "yvals"; by_ = "grp"; n = 4; splineFun = stats::spline

doTest_applySpline(test_dt2)
# doTest_applySpline(test_object = test_gr2)
doTest_applySpline(test_dt3)

testthat::context("CenterFixedSizeGRanges")
library(seqsetvis)
library(testthat)
library(GenomicRanges)

a = GRanges("chr1", IRanges(1:7*10, 1:7*10+1:7 + 10))
b = GRanges("chr1", IRanges(5:10*10, 5:10*10+1:6 + 8))

test_that("centerFixedSizeGRanges final width equals fixed_size", {
    expect_equal(width(centerFixedSizeGRanges(a, 2)), rep(2, length(a)))
    expect_equal(width(centerFixedSizeGRanges(a, 5)), rep(5, length(a)))
    expect_equal(width(centerFixedSizeGRanges(a, 80)), rep(80, length(a)))
    expect_equal(width(centerFixedSizeGRanges(b, 1)), rep(1, length(b)))
    expect_equal(width(centerFixedSizeGRanges(b, 7)), rep(7, length(b)))
    expect_equal(width(centerFixedSizeGRanges(b, 50)), rep(50, length(b)))
})
test_that("centerFixedSizeGRanges size shifts are centered", {
    a_larger = centerFixedSizeGRanges(grs = a, fixed_size = max(width(a)) + 2)
    expect_true(all(start(a_larger) < start(a)))
    expect_true(all(end(a_larger) > end(a)))
    a_smaller = centerFixedSizeGRanges(grs = a_larger, fixed_size = min(width(a)) - 2)
    expect_true(all(start(a_larger) < start(a_smaller)))
    expect_true(all(end(a_larger) > end(a_smaller)))
})
test_that("centerFixedSizeGRanges size sifts are reversible", {
    #primarily concerened about impacts of rounding
    a4 = centerFixedSizeGRanges(grs = a, fixed_size = 4)
    a5 = centerFixedSizeGRanges(grs = a, fixed_size = 5)
    a6 = centerFixedSizeGRanges(grs = a, fixed_size = 6)
    a7 = centerFixedSizeGRanges(grs = a, fixed_size = 7)
    a9 = centerFixedSizeGRanges(grs = a, fixed_size = 9)
    #derivations
    a7_from5 = centerFixedSizeGRanges(grs = a5, fixed_size = 7)
    a4_from5 = centerFixedSizeGRanges(grs = a5, fixed_size = 4)
    a6_from4 = centerFixedSizeGRanges(grs = a4, fixed_size = 6)
    a9_from4 = centerFixedSizeGRanges(grs = a4, fixed_size = 9)
    a7_from4 = centerFixedSizeGRanges(grs = a4, fixed_size = 7)
    a5_from4 = centerFixedSizeGRanges(grs = a4, fixed_size = 5)
    #reversals
    a5_from7rev = centerFixedSizeGRanges(grs = a7_from5, fixed_size = 5)
    a5_from4rev = centerFixedSizeGRanges(grs = a4_from5, fixed_size = 5)
    a4_from6rev = centerFixedSizeGRanges(grs = a6_from4, fixed_size = 4)
    a4_from9rev = centerFixedSizeGRanges(grs = a9_from4, fixed_size = 4)
    #recover even from even
    expect_true(all(a4 == a4_from6rev))
    expect_true(all(a6 == a6_from4))

    #recover odd from even
    expect_true(all(a5 == a5_from4rev))
    expect_true(all(a9 == a9_from4))#fail
    expect_true(all(a7 == a7_from4))
    expect_true(all(a5 == a5_from4))

    #recover odd from odd
    expect_true(all(a5 == a5_from7rev))
    expect_true(all(a7 == a7_from5))

    #recover even from odd
    expect_true(all(a4 == a4_from5))
    expect_true(all(a4 == a4_from9rev))
})

#' parse fragLen from MACS2 output
#' @param macs2xls_file character.  an xls file output by MACS2 to parse
#' frag length from
#' @return numeric fragment length
#' @export
#' @examples
#' xls_file = system.file("extdata/test_peaks.xls",
#'     package = "seqsetvis")
#' fragLen_fromMacs2Xls(xls_file)
fragLen_fromMacs2Xls = function(macs2xls_file){
    stopifnot(is.character(macs2xls_file))
    stopifnot(file.exists(macs2xls_file))
    str = utils::read.table(macs2xls_file,
                            nrows = 30,
                            comment.char = "",
                            sep = "\n",
                            stringsAsFactors = FALSE)$V1
    d = str[grepl(" d = ", str)]

    ts = str[grepl(" tag size is determined as ", str)]
    ts = sub("# tag size is determined as ", "", ts)
    ts = sub(" bps", "", ts)
    ts = as.numeric(ts)
    d = as.numeric(strsplit(d, " d = ")[[1]][2])
    return(d)
}

#' calculate fragLen from a bam file for specified regions
#' @param bam_f character or BamFile. bam file to read from.
#' .bai index file must be in same directory
#' @param qgr GRanges.  used as which for ScanBamParam. Can be NULL if it's
#' REALLY important to load the entire bam, force_no_which = TRUE also required.
#' @param n_regions numeric (integer) it's generally overkill to pull all
#' regions at this stage and will slow calculation down.  Default is 100.
#' @param include_plot_in_output if TRUE ouptut is a list of fragLen and a
#' ggplot showing values considered by calculation. Default is FALSE.
#' @param test_fragLen numeric.  The set of fragment lenghts to gather
#' strand cross correlation for.
#' @param flip_strand boolean. if TRUE strands that reads align to are swapped.
#' This is typically only necessary if there was a mismatch between library
#' chemistry and aligner settings. Default is FALSE.
#' @param ... passed to Rsamtools::ScanBamParam, can't be which or what.
#' @return numeric fragment length
#' @import Rsamtools
#' @export
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' bam_file = system.file("extdata/test.bam",
#'     package = "seqsetvis")
#' qgr = CTCF_in_10a_overlaps_gr[1:5]
#' fragLen_calcStranded(bam_file, qgr)
#' #if plot is included, a list is returned, item 2 is the plot
#' fragLen_calcStranded(bam_file, qgr,
#'   include_plot_in_output = TRUE)[[2]]
fragLen_calcStranded = function(bam_f,
                                qgr,
                                n_regions = 100,
                                include_plot_in_output = FALSE,
                                test_fragLen = seq(100, 400, 5),
                                flip_strand = FALSE,
                                ...){

    x = y = N = correlation = id = NULL #reserve bindings for data.table
    if(n_regions < length(qgr)){
        qgr = sample(qgr, n_regions)
    }
    cc_res = crossCorrByRle(
        bam_f,
        qgr,
        max_dupes = 1,
        fragment_sizes = test_fragLen,
        flip_strand = flip_strand,
        ...
    )
    fragLen = cc_res[, list(shift = shift[which.max(correlation)]),
                     by = list(id)][, mean(shift)]
    fragLen = round(fragLen)
    if(!include_plot_in_output){
        return(fragLen)
    }else{
        p_res = cc_res[, list(correlation = mean(correlation)), by = list(shift)]
        pdt = data.table(x = p_res$shift, raw = p_res$correlation,
                         moving_average = movingAverage(p_res$correlation))
        fragLen = pdt$x[which.max(pdt$moving_average)]
        pdt = data.table::melt(pdt, id.vars = "x",
                               variable.name = "transform", value.name = "y")
        p = ggplot(pdt) +
            labs(x = "Fragment Length", y = "strand cross correlation",
                 title = paste(basename(bam_f)),
                 subtitle = paste("Fragment Length determined by",
                                  "strand cross correlation maximization",
                                  sep = "\n")) +
            geom_line(aes(x = x, y = y, color = transform)) +
            # scale_color_manual(values = c(raw = "black",
            #                               moving_average = "red")) +
            annotate("line",
                     x = rep(fragLen, 2),
                     y = range(pdt$y),
                     color = "green") +
            annotate("label",
                     x = fragLen,
                     y = mean(range(pdt$y)),
                     label = fragLen,
                     color = "black")
        return(list(fragLen = fragLen, plot = p))
    }
}

#' Calculate cross correlation by using shiftApply on read coverage Rle
#'
#' @param bam_file character. Path to .bam file, must have index at .bam.bai.
#' @param query_gr GRanges.  Regions to calculate cross correlation for.
#' @param max_dupes integer.  Duplicate reads above this value will be removed.
#' @param fragment_sizes integer.  fragment size range to search for maximum
#'   correlation.
#' @param read_length integer. Any values outside fragment_range that must be
#'   searched.  If not supplied will be determined from bam_file.  Set as NA
#'   to disable this behavior.
#' @param flip_strand boolean. if TRUE strands that reads align to are swapped.
#' This is typically only necessary if there was a mismatch between library
#' chemistry and aligner settings. Default is FALSE.
#' @param ... arguments passed to ScanBamParam
#' @return named list of results
#' @export
#' @import GenomicRanges GenomicAlignments pbapply S4Vectors
#' @importFrom Rsamtools ScanBamParam
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' bam_f = system.file("extdata/test.bam",
#'     package = "seqsetvis", mustWork = TRUE)
#' query_gr = CTCF_in_10a_overlaps_gr[1:2]
#' crossCorrByRle(bam_f, query_gr[1:2], fragment_sizes = seq(50, 300, 50))
crossCorrByRle = function(bam_file,
                          query_gr,
                          max_dupes = 1,
                          fragment_sizes = 50:300,
                          read_length = NULL,
                          flip_strand = FALSE,
                          ...){
    rn = NULL # reserve for data.table
    if(is.null(query_gr$id)){
        if(is.null(names(query_gr))){
            query_gr$id = paste0("peak_", seq_along(query_gr))
        }else{
            query_gr$id = names(query_gr)
        }
    }else{
        if(is.null(names(query_gr))){
            names(query_gr) = query_gr$id
        }else{
            #both names() and $id are set, leave it alone
        }

    }
    q_widths = apply(cbind(width(query_gr), max(fragment_sizes)*3), 1, max)
    query_gr = resize(query_gr, q_widths, fix = "center")
    names(query_gr) = query_gr$id
    # query_gr = resize(query_gr, 500, fix = "center")

    query_gr = harmonize_seqlengths(query_gr, bam_file)

    Param <- Rsamtools::ScanBamParam(
        which=query_gr,
        what=c("flag","mapq"), ...)
    if(!file.exists(paste0(bam_file, ".bai"))){
        warning("creating index for ", bam_file)
        Rsamtools::indexBam(bam_file)
    }
    temp <- GenomicAlignments::readGAlignments(bam_file,param=Param)
    if(flip_strand){
        strand(temp) = ifelse(as.character(strand(temp)) == "+", "-", "+")
    }
    rl = getReadLength(bam_file, query_gr)
    if(is.null(read_length)){
        read_length = rl
    }
    if(is.na(read_length)){
        read_length = numeric()
    }
    fragment_sizes = sort(union(read_length, fragment_sizes))

    PosCoverage <- coverage(GenomicRanges::shift(GRanges(temp[strand(temp)=="+"])), -rl)
    PosCoverage = PosCoverage[query_gr]
    names(PosCoverage) = query_gr$id

    NegCoverage <- coverage(GRanges(temp[strand(temp)=="-"]))
    NegCoverage = NegCoverage[query_gr]
    names(NegCoverage) = query_gr$id
    # ShiftMatCor = vapply(as.list(seq_along(query_gr)), FUN.VALUE = as.list(seq(fragment_sizes)),
    #                      function(i){
    #     ShiftsCorTemp <- S4Vectors::shiftApply(fragment_sizes,
    #                                            PosCoverage[[i]],
    #                                            NegCoverage[[i]],
    #                                            cor, simplify = FALSE,
    #                                            verbose = FALSE)
    # })

    ShiftMatCor = vapply((seq_along(query_gr)), FUN.VALUE = numeric(length(fragment_sizes)),
                         function(i){
                             ShiftsCorTemp <- S4Vectors::shiftApply(fragment_sizes,
                                                                    PosCoverage[[i]],
                                                                    NegCoverage[[i]],
                                                                    cor, simplify = FALSE,
                                                                    verbose = FALSE)
                             unlist(ShiftsCorTemp)
                         })
    #necessary due to singleton query_gr or shift not resulting in matrix
    # ShiftMatCor = matrix(unlist(ShiftMatCor),
    #                      byrow = FALSE,
    #                      nrow = length(fragment_sizes),
    #                      ncol = length(query_gr))
    if(length(fragment_sizes) == 1){
        ShiftMatCor = matrix(ShiftMatCor, nrow = 1)
    }
    ShiftMatCor[is.nan(ShiftMatCor)] = 0

    colnames(ShiftMatCor) = query_gr$id
    rownames(ShiftMatCor) = fragment_sizes
    shift_dt = as.data.table(ShiftMatCor, keep.rownames = TRUE)
    shift_dt[, shift := as.numeric(rn)]
    shift_dt$rn = NULL
    shift_dt = melt(shift_dt, id.vars = "shift",
                    variable.name = "id", value.name = "correlation")
    return(shift_dt)
}

#' generic for methods to convert various objects to a logical matrix indicating
#' membership of items (rows) in sets (columns)
#' @export
#' @param object the object to convert. Supported types: list (of character
#' or GRanges), GRanges with membership table metadata, GRangesList,
#' data.frame/matrix/DataFrame of membership table
#' @rdname ssvMakeMembTable-methods
#' @exportMethod ssvMakeMembTable
#' @import methods
#' @return a logical matrix indicating membership of items (rows) in sets
#' (columns)
setGeneric("ssvMakeMembTable", function(object){
    standardGeneric("ssvMakeMembTable")
})


#' list of character vectors input
#' @export
#' @rdname ssvMakeMembTable-methods
#' @aliases ssvMakeMembTable,list-method
#' @import GenomicRanges
#' @import methods
#' @examples
#' char_list = list(letters[1:3], letters[2:4])
#' ssvMakeMembTable(char_list)
#' library(GenomicRanges)
#' gr_list = list(GRanges("chr1", IRanges(1:3*2, 1:3*2)),
#'     GRanges("chr1", IRanges(2:4*2, 2:4*2)))
#' ssvMakeMembTable(gr_list)
setMethod("ssvMakeMembTable", signature(object = "list"), function(object){
    #GRanges are a special case
    if (all(vapply(object, class, "character") == "GRanges")) {
        return(ssvMakeMembTable(GRangesList(object)))
    }
    if (any(vapply(object, class, "character") != "character")) {
        # warning("Converting all non-character items to characters.")
        object = lapply(object, as.character)
    }
    if (all(vapply(object, class, "character") == "character")) {
        char_object = object
        object = set_list2memb(char_object)
    } else {
        #This may not be reachable anymore but I'm leaving as a backstop
        stop("Can't handle list of non-character classes as object: ",
             paste(vapply(object, class, "character"), collapse = ", "))
    }
    return(object)
})

#' GRangesList input
#' @export
#' @rdname ssvMakeMembTable-methods
#' @aliases ssvMakeMembTable,GRangesList-method
#' @import GenomicRanges
#' @examples
#' library(GenomicRanges)
#' gr_list = list(GRanges("chr1", IRanges(1:3*2, 1:3*2)),
#'     GRanges("chr1", IRanges(2:4*2, 2:4*2)))
#' ssvMakeMembTable(GRangesList(gr_list))
setMethod("ssvMakeMembTable",
          signature(object = "GRangesList"), function(object){
              GRlist_object = object
              ssvMakeMembTable(ssvOverlapIntervalSets(GRlist_object))
          })



#' GRanges with mcols input
#' @export
#' @rdname ssvMakeMembTable-methods
#' @aliases ssvMakeMembTable,GRanges-method
#' @import GenomicRanges
#' @examples
#' gr = GRanges("chr1", IRanges(1:3*2, 1:3*2))
#' gr$set_a = c(TRUE, TRUE, FALSE)
#' gr$set_b = c(FALSE, TRUE, TRUE)
#' ssvMakeMembTable(gr)
setMethod("ssvMakeMembTable", signature(object = "GRanges"), function(object){
    mc_object = GenomicRanges::mcols(object)
    rownames(mc_object) = names(object)
    ssvMakeMembTable(mc_object)
})


#' DataFrame input
#' @export
#' @rdname ssvMakeMembTable-methods
#' @aliases ssvMakeMembTable,DataFrame-method
#' @import GenomicRanges
#' @examples
#' gr = GRanges("chr1", IRanges(1:3*2, 1:3*2))
#' gr$set_a = c(TRUE, TRUE, FALSE)
#' gr$set_b = c(FALSE, TRUE, TRUE)
#' ssvMakeMembTable(mcols(gr))
setMethod("ssvMakeMembTable", signature(object = "DataFrame"), function(object){
    DF_object = object
    df_object = as.data.frame(DF_object)
    ssvMakeMembTable(df_object)
})

#' matrix of logicals, membership table
#' @export
#' @rdname ssvMakeMembTable-methods
#' @aliases ssvMakeMembTable,matrix-method
#' @examples
#' memb_mat =  matrix(c(TRUE, TRUE, FALSE, FALSE, TRUE, FALSE, TRUE, FALSE),
#'     ncol = 2, byrow = FALSE)
#' ssvMakeMembTable(memb_mat)
setMethod("ssvMakeMembTable", signature(object = "matrix"), function(object){
    mat_object = object
    df_object = as.data.frame(mat_object)
    ssvMakeMembTable(df_object)
})

#' data.frame input, final output
#' The final method for all inputs, checks column names and returns logical
#' matrix
#' @export
#' @rdname ssvMakeMembTable-methods
#' @aliases ssvMakeMembTable,data.frame-method
#' @examples
#' memb_df = data.frame(a = c(TRUE, TRUE, FALSE, FALSE),
#'     b = c(TRUE, FALSE, TRUE, FALSE))
#' ssvMakeMembTable(memb_df)
setMethod("ssvMakeMembTable",
          signature(object = "data.frame"), function(object){
              col_is_logical = vapply(X = seq(ncol(object)),
                                   FUN = function(i)is.logical(object[,i]),
                                   FUN.VALUE = TRUE)
              if(!all(col_is_logical)){
                  message("Non-logical class columns have been dropped.")
                  object = object[, col_is_logical, drop = FALSE]
              }

              if (is.null(colnames(object))) {
                  colnames(object) = paste0("set_", LETTERS[seq_len(ncol(object))])
              }
              if(all(colnames(object) == paste0("V", seq_len(ncol(object))))){
                  colnames(object) = paste0("set_", LETTERS[seq_len(ncol(object))])
              }
              mat = as.matrix(object)
              rownames(mat) = rownames(object)
              return(mat)
          })

#' Convert any object accepted by ssvMakeMembTable to a factor
#' To avoid ambiguity,
#'
#' see \code{\link{ssvMakeMembTable}}
#'
#' @param object a valid object for conversion to a membership table and then
#' factor
#' @return a 2 column ("id" and "group") data.frame.
#' "id" is factor of item names if any or simply order of items.
#' "group" is a factor of set combinations
#' @export
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' ssvFactorizeMembTable(CTCF_in_10a_overlaps_gr)
#' ssvFactorizeMembTable(list(1:4, 2:3, 4:6))
ssvFactorizeMembTable = function(object){
    memb = ssvMakeMembTable(object)
    id_lev = names(rev(sort(rowSums(memb))))
    group_lev = names(rev(sort(colSums(memb))))
    grps = apply(memb, 1, function(x){
        paste(names(x)[x], collapse = " & ")
    })
    grp_counts = table(grps)
    grp_counts = rev(sort(grp_counts))
    grp_df = data.frame(id = names(grps), group = grps)
    rownames(grp_df) = NULL
    grp_df$id = factor(grp_df$id, levels = id_lev)
    grp_df$group = factor(grp_df$group, levels = names(grp_counts))
    return(grp_df)
}

testthat::context("FactorizeMembTable")
library(seqsetvis)

test_that("ssvFactorizeMembTable expected output format for GRanges", {
    gr = CTCF_in_10a_overlaps_gr
    names(gr) = paste("peak", seq_along(gr))
    fac = ssvFactorizeMembTable(gr)
    expect_true(is.data.frame(fac))
    expect_equal(colnames(fac), c("id", "group"))
    expect_equal(nrow(fac), length(CTCF_in_10a_overlaps_gr))
    expect_true(is.factor(fac$group))
    expect_true(is.factor(fac$id))
})

test_that("ssvFactorizeMembTable expected output values for GRanges", {
    gr = CTCF_in_10a_overlaps_gr
    fac = ssvFactorizeMembTable(gr)
    expect_equal(as.character(fac$group[2]), "MCF10A_CTCF")
    expect_equal(as.character(fac$group[5]), "MCF10A_CTCF & MCF10AT1_CTCF")
    expect_equal(as.character(fac$group[10]),  "MCF10A_CTCF & MCF10AT1_CTCF & MCF10CA1_CTCF")
})

test_that("ssvFactorizeMembTable expected output values for list", {
    setL = list(1:3, 2:3, c(3:5))
    setL = lapply(setL, function(x){
        LETTERS[x]
    })
    fac = ssvFactorizeMembTable(setL)
    expect_equal(nrow(fac), 5)
    expect_equal(as.character(fac$group[1]), "set_A")
    expect_equal(as.character(fac$group[2]),  "set_A & set_B")
    expect_equal(as.character(fac$group[3]),  "set_A & set_B & set_C")
    expect_equal(as.character(fac$group[4]),  "set_C")
    expect_equal(as.character(fac$group[5]),  "set_C")
})

#' easy awesome peak set vis
#' TESTING
#' seqsetvis allows you to...
#'
#' 2 steps
#' \code{\link{ssvOverlapIntervalSets}}.
#' \code{\link{ssvFetchBigwig}}.
#' Otherwise refer to the vignettes to see
"_PACKAGE"

#' CTCF ChIP-seq in breast cancer cell lines
#'
#' @description Vignette data for seqsetvis was downloaded directly from GEO
#' series
#' \href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE98551}{GSE98551}.
#' This data is CTCF ChIP-seq from a model of breast cancer progression
#' derived from the MCF10A cell line.
#'
#' @description Data from GEO series
#' \href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE98551}{GSE98551}
#' is from the publication
#' \href{https://www.ncbi.nlm.nih.gov/pubmed/28504305}{Fritz AJ et al. 2018}
#' @details Contains: \itemize{
#' \item \code{\link{CTCF_in_10a_overlaps_gr}}
#' \item \code{\link{CTCF_in_10a_profiles_dt}}
#' \item \code{\link{CTCF_in_10a_bigWig_urls}}
#' \item \code{\link{CTCF_in_10a_narrowPeak_urls}}
#' }
#' @docType data
#' @keywords datasets
#' @name CTCF_in_10a_data
NULL

#' Profiles for 100 randomly selected regions from overlapping CTCF peaks in
#' 10a cell ChIP-seq
#' Results from CTCF_in_10a_overlaps_gr
#'
#' A tidy GRanges at window size 50 bp within 350 bp of peak center
#'  The variables are as follows:
#' @details part of \code{\link{CTCF_in_10a_data}}
#' \enumerate{
#'   \item id. unique identifier
#'   \item y. fold-enrichment over input.
#'   \item x. bp relative to center
#'   \item sample. name of originating sample
#' }
#'
#' @docType data
#' @keywords datasets
#' @name CTCF_in_10a_profiles_gr
#' @format A tidy GRanges of 2100 rows and 4 metadata columns
NULL

#' Profiles for 100 randomly selected regions from overlapping CTCF peaks in
#' 10a cell ChIP-seq
#' Results from fetching bigwigs with CTCF_in_10a_overlaps_gr.
#'
#' A tidy data.table at window size 50 bp within 350 bp of peak center
#'  The variables are as follows:
#' @details part of \code{\link{CTCF_in_10a_data}}
#' \enumerate{
#'   \item seqnames. chromosome for GRanges compatibility
#'   \item start. start of interval
#'   \item end. end of interval
#'   \item width. width of interval
#'   \item strand. leftover from GRanges.
#'   \item id. unique identifier
#'   \item y. fold-enrichment over input.
#'   \item x. bp relative to center
#'   \item sample. name of originating sample
#' }
#'
#' @docType data
#' @keywords datasets
#' @name CTCF_in_10a_profiles_dt
#' @format A tidy data.table of 2100 rows and 9 columns
NULL

#' 100 randomly selected regions from overlapping CTCF peaks in 10a cell
#' ChIP-seq
#'
#' MACS2 narrowPeak calls on pooled biological replicates at pval 1e-5 and
#' then 0.05 IDR filtered. IDR cutoffs determined  by comparing top 150,000
#' pvalue sorted peak in replicates.
#'
#' See  GEO series GSE98551 for details.
#' @details part of \code{\link{CTCF_in_10a_data}}
#' @docType data
#' @keywords datasets
#' @name CTCF_in_10a_overlaps_gr
#' @format GenomicRanges with 3 metadata columns of membership table
NULL

#' FTP URL path for vignette data.
#'
#' FE bigWig tracks for CTCF ChIP-seq in a MCF10A progression model.
#' See GEO series GSE98551 for details.
#' @details part of \code{\link{CTCF_in_10a_data}}
#' @docType data
#' @keywords datasets
#' @name CTCF_in_10a_bigWig_urls
#' @format named character vector of length 3
NULL

#' FTP URL path for vignette data.
#' from
#'
#' macs2 peak calls for CTCF ChIP-seq in a MCF10A progression model.
#' See  GEO series GSE98551 for details.
#' @details part of \code{\link{CTCF_in_10a_data}}
#' @docType data
#' @keywords datasets
#' @name CTCF_in_10a_narrowPeak_urls
#' @format named character vector of length 3
NULL

#' list of GRanges that results in 100 random subset when overlapped
#'
#' @details part of \code{\link{CTCF_in_10a_data}}
#' @docType data
#' @keywords datasets
#' @name CTCF_in_10a_narrowPeak_grs
#' @format named list of GRanges of length 3
#' @return named list of GRanges of length 3
NULL

#' 4 random peaks for paired-end data
#'
#' matches \code{system.file("extdata/Bcell_PE.mm10.bam", package = "seqsetvis")}
#'
#' @details this is included only for testing ssvFetchBamPE functions.
#' @docType data
#' @keywords datasets
#' @name Bcell_peaks
#' @format GRanges length 4
#' @return GRanges length 4
NULL

#' 4 random peaks for single-end data and 4 control regions 30kb downstream from
#' each peak.
#'
#' matches \code{system.file("extdata/test_peaks.bam", package =
#' "seqsetvis")}
#'
#' @details this is included only for testing ssvFetchBam functions.
#' @docType data
#' @keywords datasets
#' @name test_peaks
#' @format GRanges length 8
#' @return GRanges length 8
NULL

#' plot profiles from bigwigs
#' @export
#' @param bw_data a GRanges or data.table of bigwig signal.
#' As returned from \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}
#' @param y_ the variable name in bw_data for y axis in plot
#' @param x_ the variable name in bw_data for x axis in plot
#' @param by_ the variable name in bw_data to facet on
#' @param hsv_reverse logical, should color scale be reversed? default FALSE
#' @param hsv_saturation numeric value from 0 to 1. Saturation for color scale. default 1
#' @param hsv_value numeric value from 0 to 1. Value for color scale. default 1
#' @param hsv_grayscale logical, if TRUE gray() is used instead of rainbow(). default FALSE
#' @param hsv_hue_min numeric [0, hsv_hue_max) hue min of color scale
#' @param hsv_hue_max numeric (hsv_hue_min, 1] hue max of color scale
#' @param hsv_symmetric if TRUE, colorscale is symmetrical, default FALSE.
#' @param n_quantile number of evenly size quantile bins
#' @param quantile_min the lowest quantile start
#' @param quantile_max the highest quantile end
#' @param return_data logical.  If TRUE, return value is no longer ggplot and
#' is instead the data used to generate that plot. Default is FALSE.
#' @return ggplot object using ribbon plots to show quantile distributions
#' @import ggplot2
#' @importFrom grDevices gray rainbow
#' @importFrom stats quantile
#' @rawNamespace import(data.table, except = c(shift, first, second, last))
#' @examples
#' data(CTCF_in_10a_profiles_gr)
#' #rainbow colors
#' qgr = CTCF_in_10a_profiles_gr
#' ssvSignalBandedQuantiles(qgr)
#' #grayscale
#' ssvSignalBandedQuantiles(qgr, hsv_grayscale = TRUE,
#'     hsv_symmetric = TRUE, hsv_reverse = TRUE)
#' #using "by_" per sample
#' ssvSignalBandedQuantiles(qgr, hsv_grayscale = TRUE,
#'     hsv_symmetric = TRUE, hsv_reverse = TRUE, by_ = "sample")
#' #adding spline smoothing
#' splined = applySpline(qgr, n = 10,
#'     by_ = c("id", "sample"))
#' ssvSignalBandedQuantiles(splined, n_quantile = 50,
#'     quantile_min = .25, quantile_max = .75,
#'     hsv_symmetric = TRUE, hsv_reverse = TRUE, by_ = "sample")
ssvSignalBandedQuantiles = function(bw_data, y_ = "y", x_ = "x", by_ = "fake",
                                    hsv_reverse = FALSE,
                                    hsv_saturation = 1, hsv_value = 1,
                                    hsv_grayscale = FALSE,
                                    hsv_hue_min = 0, hsv_hue_max = 0.7, hsv_symmetric = FALSE,
                                    n_quantile = 18, quantile_min = 0.05, quantile_max = 0.95,
                                    return_data = FALSE
) {
    #binding for data.table
    x = NULL
    if(is(bw_data, "GRanges")){
        bw_data = data.table::as.data.table(bw_data)
    }
    stopifnot(data.table::is.data.table(bw_data))
    stopifnot(is.character(x_), is.character(y_), is.character(by_))
    stopifnot(x_ %in% colnames(bw_data), y_ %in% colnames(bw_data))
    stopifnot(by_ %in% colnames(bw_data) || by_ == "fake")
    stopifnot(is.logical(hsv_reverse), is.logical(hsv_grayscale),
              is.logical(hsv_symmetric))
    num_args = c(hsv_saturation, hsv_value, hsv_hue_min,
                 hsv_hue_max, quantile_min, quantile_max)
    stopifnot(all(is.numeric(num_args)))
    stopifnot(all(num_args >= 0), all(num_args <= 1))
    stopifnot(n_quantile > 2)

    variable = value = V1 = low = high = low_q = high_q = NULL
    q_range = rn = q_num = q_low = q_high = NULL #declare binding for data.table
    q2do = 0:n_quantile/n_quantile
    q2do = round(quantile_min + q2do * (quantile_max - quantile_min), digits = 3)

    if(by_ == "fake"){
        bw_data[[by_]] = TRUE
        todo = TRUE
    }else{
        todo = unique(bw_data[, get(by_)])
    }

    calc_quantiles = function(td){#, x_, y_, by_, q2do){
        dt = bw_data[get(by_) == td, list(qs = list(list(stats::quantile(get(y_), q2do)))), by = x_]
        qmat = matrix(unlist(dt$qs), nrow = nrow(dt), ncol = length(q2do), byrow = TRUE)
        colnames(qmat) = paste0(q2do*100, "%")
        dt = cbind(dt, qmat)
        dt$qs = NULL
        dtm = data.table::melt(dt, id.vars = x_)
        data.table::setkey(dtm, variable)
        q2do_str = paste0(q2do * 100, "%")
        dt_low = dtm[q2do_str[-length(q2do_str)], list(get(x_), low_q = variable, low = value)]
        dt_high = dtm[q2do_str[-1], list(get(x_), high_q = variable, high = value)]
        dt_c = cbind(dt_low, dt_high[, -1])
        dt_c = dt_c[, list(V1, low, high, q_range = paste0(sub("%", "", low_q), "-", high_q))]
        q_o = unique(dt_c$q_range)
        dt_c$q_range = factor(dt_c$q_range, levels = q_o)

        dt_c$rn = seq_len(nrow(dt_c))
        dt_c[, `:=`(c("q_low", "q_high"), data.table::tstrsplit(sub("%", "", q_range), split = "-", keep = seq_len(2))), by = rn]
        dt_c[, `:=`(q_num, (as.numeric(q_low) + as.numeric(q_high))/2)]
        dt_c[, `:=`(c("rn", "q_low", "q_high"), NULL)]
        dt_c$facet_group = td
        return(dt_c)
    }

    all_q = lapply(todo, calc_quantiles)
    plot_dt = data.table::rbindlist(all_q)
    q_o = unique(plot_dt$q_range)
    if(hsv_symmetric){
        ncol = ceiling(length(q_o) / 2)
        if(hsv_grayscale){
            gray_vals = hsv_hue_min + (hsv_hue_max - hsv_hue_min)*((seq_len(ncol) - 1)/(ncol - 1))
            cols = grDevices::gray(gray_vals)
        }else{
            cols = grDevices::rainbow(ncol, start = hsv_hue_min, end = hsv_hue_max, s = hsv_saturation, v = hsv_value)
        }
        if(hsv_reverse) cols = rev(cols)
        ci = c(seq_len(ncol), rev(seq_len(ncol) - ifelse(length(q_o) %% 2 == 0, 0, 1)))
        cols = cols[ci]
    }else{
        ncol = length(q_o)
        if(hsv_grayscale){
            gray_vals = hsv_hue_min + (hsv_hue_max - hsv_hue_min)*((seq_len(ncol) - 1)/(ncol - 1))
            cols = grDevices::gray(gray_vals)
        }else{
            cols = grDevices::rainbow(ncol, start = hsv_hue_min, end = hsv_hue_max, s = hsv_saturation, v = hsv_value)
        }

        if(hsv_reverse) cols = rev(cols)
    }

    names(cols) = q_o

    #copy factor level order if any
    if(is.factor(bw_data[, get(by_)])){
        plot_dt$facet_group = factor(plot_dt$facet_group, levels = levels(bw_data[, get(by_)]))
    }

    colnames(plot_dt)[1] = "x"
    if(return_data){
        return(plot_dt)
    }

    p = ggplot(plot_dt) +
        geom_ribbon(aes(x = x, ymin = low, ymax = high, fill = q_range, color = q_range)) +
        labs(fill = "quantile band",
             y = y_,
             x = "bp",
             title = "Enrichment Profiles",
             subtitle = "aggregated by quantile range") +
        scale_fill_manual(values = cols) +
        scale_color_manual(breaks = q2do, values = cols)
    if(length(todo) > 1){
        p = p +
            facet_grid(facet_group ~ .) +
            theme(strip.text.y = element_text(size = 7))
    }
    p
}

#' maps signal from 2 sample profiles to the x and y axis. axes are standard or "volcano" min XY vs fold-change Y/X
#' @export
#' @param bw_data a GRanges or data.table of bigwig signal.
#' As returned from \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}
#' @param x_name sample name to map to x-axis, must be stored in variable specified in \code{xy_variable}
#' @param y_name sample name to map to y-axis, must be stored in variable specified in \code{xy_variable}
#' @param color_table data.frame with 2 columns, one of which must be named "group" and gets mapped to color.
#' The other column must be the same as by_ parameter and is used for merging.
#' @param value_variable variable name that stores numeric values for plotting, default is "y"
#' @param xy_variable variable name that stores sample, must contain entires for \code{x_name} and \code{y_name}
#' @param value_function a function to apply to \code{value_variable} in all combintations of \code{by_} per \code{x_name} and \code{y_name}
#' @param by_ variables that store individual measurement ids
#' @param plot_type standard or volcano, default is "standard"
#' @param show_help if TRUE overlay labels to aid plot interpretation, default is FALSE
#' @param fixed_coords if TRUE coordinate system is 1:1 ratio, default is TRUE
#' @param return_data logical.  If TRUE, return value is no longer ggplot and
#' is instead the data used to generate that plot. Default is FALSE.
#' @import ggplot2
#' @return ggplot of points comparing signal from 2 samples
#' @examples
#' data(CTCF_in_10a_profiles_gr)
#' ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
#'     x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF")
#' ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
#'     x_name = "MCF10A_CTCF", y_name = "MCF10CA1_CTCF")
#'
#' ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
#'     x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF",
#'     value_function = median) + labs(title = "median FE in regions")
#'
#' ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
#'     x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF",
#'     plot_type = "volcano")
#'
#' ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
#'     x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF",
#'     plot_type = "volcano", show_help = TRUE)
ssvSignalScatterplot = function(bw_data, x_name, y_name,
                                color_table = NULL,
                                value_variable = "y", xy_variable = "sample",
                                value_function = max,
                                by_ = "id",
                                plot_type = c("standard", "volcano")[1],
                                show_help = FALSE, fixed_coords = TRUE,
                                return_data = FALSE){
    xval = yval = xvolcano = id = yvolcano = group = NULL #declare binding for data.table
    if(is(bw_data, "GRanges")){
        bw_data = data.table::as.data.table(bw_data)
    }
    stopifnot(data.table::is.data.table(bw_data))
    stopifnot(xy_variable %in% colnames(bw_data))
    if(!any(x_name == bw_data[[xy_variable]])){
        stop(x_name, "not found in", xy_variable, "variable of data.table")
    }
    if(!any(y_name == bw_data[[xy_variable]])){
        stop(y_name, "not found in", xy_variable, "variable of data.table")
    }
    stopifnot(is.character(x_name), is.character(y_name),
              is.character(xy_variable), is.character(by_),
              is.character(plot_type))
    stopifnot(plot_type %in% c("standard", "volcano"))
    stopifnot(is.logical(show_help), is.logical(fixed_coords))
    stopifnot(is.function(value_function))

    plot_dt = merge(bw_data[get(xy_variable) == x_name, list(xval = value_function(get(value_variable))), by = by_],
                    bw_data[get(xy_variable) == y_name, list(yval = value_function(get(value_variable))), by = by_])
    if(!is.null(color_table)){
        plot_dt = merge(plot_dt, color_table, by = by_)
    }
    # if(is.null(plotting_group)){
    #     plot_dt$plotting_group = factor("none")
    # }else{
    #     plot_dt = merge(plot_dt, plotting_group)
    # }
    if(plot_type == "standard"){
        if(fixed_coords){
            xlim = c(0, plot_dt[, max(xval, yval)])
            ylim = xlim
        }else{
            xlim = c(0, plot_dt[, max(xval)])
            ylim = c(0, plot_dt[, max(yval)])
        }

        p = ggplot(plot_dt, aes(x = xval, y = yval))
        if(is.null(color_table)){
            p = p + geom_point(mapping = aes(alpha = 1, size = 1.5))
        }else{
            p = p + geom_point(mapping = aes(alpha = 1, size = 1.5, color = group))
        }
        p = p +
            scale_alpha_identity() +
            scale_size_identity() +
            guides(alpha = "none", size = "none") +
            labs(x = x_name, y = y_name, title = paste(value_variable, "aggregated per", paste(by_, collapse = ", "))) +
            ylim(ylim) + xlim(xlim)

        if(show_help){
            xpos1 = .2 * max(xlim)
            xpos2 = .8 * max(xlim)
            ypos1 = .2 * max(ylim)
            ypos2 = .8 * max(ylim)
            p = p + annotate("segment", x = 0, xend = 0, y = ypos1, yend = ypos2, arrow = arrow())
            p = p + annotate("label", x = 0, y = mean(xlim), label = gsub(" ", "\n", paste(y_name, "binding")), hjust = 0)
            p = p + annotate("segment", x = xpos1, xend = xpos2, y = 0, yend = 0, arrow = arrow())
            p = p + annotate("label", x = mean(xlim), y = 0, label = paste(x_name, "binding"), vjust = 0)
            p = p + annotate("label", x = 0, y = 0, label = "no\nbinding", hjust = 0, vjust = 0)
            p = p + annotate("segment", x = xpos1, xend = xpos2, y = ypos1, yend = ypos2, arrow = arrow())
            p = p + annotate("label", x = mean(xlim), y = mean(ylim), label = gsub(" ", "\n", paste("both binding")))
        }
    }else if(plot_type == "volcano"){
        plot_dt[, xvolcano := log2(max(yval, 1) / max(xval, 1)), by = by_]
        plot_dt[, yvolcano := log2(max(min(yval, xval), 1)), by = by_]
        xmax = plot_dt[, max(abs(c(xvolcano)))]
        lim = c(-xmax, xmax)
        p = ggplot(plot_dt, aes(x = xvolcano, y = yvolcano))
        if(is.null(color_table)){
            p = p + geom_point(mapping = aes(alpha = 1, size = 1.5))
        }else{
            p = p + geom_point(mapping = aes(alpha = 1, size = 1.5, color = group))
        }
        p = p +
            scale_size_identity() +
            scale_alpha_identity() +
            guides(alpha = "none", size = "none") +
            labs(x = paste("log2 fold-change of", y_name, "over", x_name),
                 y = paste("log2 min of", y_name, "and", x_name), title = "Max FE in regions") +
            xlim(lim)
        if(show_help){
            pos1 = .2 * max(lim)
            pos2 = .8 * max(lim)
            p = p + annotate("segment", y = 1, yend = 1, x = pos1, xend = pos2, arrow = arrow())
            p = p + annotate("label", y = 1, x = max(lim)/2, label = gsub(" ", "\n", paste(y_name, "binding")), vjust = 0)
            p = p + annotate("segment", y = 1, yend = 1, x = -pos1, xend = -pos2, arrow = arrow())
            p = p + annotate("label", y = 1, x = -max(lim)/2, label = gsub(" ", "\n", paste(x_name, "binding")), vjust = 0)
            p = p + annotate("label", x = 0, y = 1, label = "no\nbinding", hjust = .5, vjust = 0)
            ylim = range(plot_dt$yvolcano)
            ypos1 = 1 + (max(ylim) - min(ylim)) * .2
            ypos2 = 1 + (max(ylim) - min(ylim)) * .8
            p = p + annotate("segment", x = 0, xend = 0, y = ypos1, yend = ypos2, arrow = arrow())
            p = p + annotate("label", x = mean(lim), y = mean(ylim), label = gsub(" ", "\n", paste("both binding")))
            p
        }
    }
    # if(is.null(plotting_group)){
    #     p = p + guides(color = "none")
    # }
    if(return_data){
        return(plot_dt)
    }
    if(fixed_coords) p = p + coord_fixed()
    p
}

make_facet_str = function(facet_rows = character(), facet_columns = character()){
    facet_rows = facet_rows[facet_rows != ""]
    facet_columns = facet_columns[facet_columns != ""]
    if(length(facet_rows) > 0){
        row_str = paste0("`", paste(facet_rows, collapse = "`+`"), "`")
    }else{
        row_str = "."
    }
    if(length(facet_columns) > 0){
        col_str = paste0("`", paste(facet_columns, collapse = "`+`"), "`")
    }else{
        col_str = "."
    }

    paste0(row_str, "~", col_str)
}

#' make_clustering_matrix
#'
#' Create a wide matrix from a tidy data.table more suitable for clustering
#' methods
#'
#' @param tidy_dt the tidy data.table to covert to a wide matrix.  Must have
#'   entries for variables specified by row_, column_, fill_, and facet_.
#' @param row_ variable name mapped to row, likely peak id or gene name for ngs
#'   data
#' @param column_ varaible mapped to column, likely bp position for ngs data
#' @param fill_ numeric variable to map to fill
#' @param facet_ variable name to facet horizontally by
#' @param max_rows for speed rows are sampled to 500 by default, use Inf to plot
#'   full data
#' @param max_cols for speed columns are sampled to 100 by default, use Inf to
#'   plot full data
#' @param clustering_col_min numeric minimum for col range considered when
#'   clustering, default in -Inf
#' @param clustering_col_max numeric maximum for col range considered when
#'   clustering, default in Inf
#' @param dcast_fill value to supply to dcast fill argument. default is NA.
#' @param fun.aggregate Function to aggregate when multiple values present for
#'   facet_, row_, and column_. The
#'   function should accept a single vector argument or be a character string
#'   naming such a function.
#' @export
#' @return A wide matrix version of input tidy data.table
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' mat = make_clustering_matrix(CTCF_in_10a_profiles_dt)
#' mat[1:5, 1:5]
make_clustering_matrix = function(tidy_dt,
                                  row_ = "id",
                                  column_ = "x",
                                  fill_ = "y",
                                  facet_ = "sample",
                                  max_rows = 500,
                                  max_cols = 100,
                                  clustering_col_min = -Inf,
                                  clustering_col_max = Inf,
                                  dcast_fill = NA,
                                  fun.aggregate = "mean"){
    raw_nc = length(unique(tidy_dt[[column_]]))
    if(raw_nc > max_cols){
        new_scale = (seq_len(max_cols)-1) / (max_cols - 1)
        old_scale = (seq_len(raw_nc)-1) / (raw_nc - 1)
        kept = vapply(new_scale, function(v){
            which.min(abs(v - old_scale))
        }, 1)
        kept = sort(unique(tidy_dt[[column_]]))[kept]
        tidy_dt = tidy_dt[get(column_) %in% kept]
        message(raw_nc - max_cols,
                " columns were discarded according to max_cols: ",
                max_cols)
    }
    row_ids = unique(tidy_dt[[row_]])
    raw_nr = length(row_ids)
    if(raw_nr > max_rows){
        row_ids = sample(row_ids, max_rows)
        tidy_dt = tidy_dt[get(row_) %in% row_ids]
        message(raw_nr - max_rows,
                " rows were discarded according to max_rows: ",
                max_rows)
    }

    # dc_formula = paste(row_, "~", paste(c(facet_, column_), collapse = " + "))
    dc_formula = make_facet_str(facet_rows = row_, facet_columns = c(facet_, column_))
    if(is.character(fun.aggregate)){
        fun.aggregate = get(fun.aggregate)
    }
    if(is.numeric(tidy_dt[[column_]])){
        dc_dt = data.table::dcast(tidy_dt[get(column_) > clustering_col_min &
                                              get(column_) < clustering_col_max],
                                  formula = dc_formula,
                                  value.var = fill_,
                                  fill = dcast_fill,
                                  fun.aggregate = fun.aggregate)
    }else{
        dc_dt = data.table::dcast(tidy_dt,
                                  formula = dc_formula,
                                  value.var = fill_,
                                  fill = dcast_fill,
                                  fun.aggregate = fun.aggregate)
    }

    dc_mat = as.matrix(dc_dt[,-1])
    rownames(dc_mat) = dc_dt[[row_]]
    dc_mat
}

#' add_cluster_annotation
#'
#' adds rectangle boxes proportional to cluster sizes of heatmap with optional
#' labels.
#'
#' @param cluster_ids Vector of cluster ids for each item in heatmap. Should be
#'   sorted by plot order for heatmap.
#' @param p Optionally an existing ggplot to add annotation to.
#' @param xleft left side of cluster annotation rectangles. Default is 0.
#' @param xright right side of cluster annotation rectangles. Default is 1.
#' @param rect_colors colors of rectangle fill, repeat to match number of
#'   clusters. Default is c("black", "gray").
#' @param text_colors colors of text, repeat to match number of clusters.
#'   Default is reverse of rect_colors.
#' @param show_labels logical, shoud rectangles be labelled with cluster
#'   identity.  Default is TRUE.
#' @param label_angle angle to add clusters labels at.  Default is 0, which is
#'   horizontal.
#' @param row_ variable name mapped to row, likely id or gene name for ngs data. Default is "id" and works with ssvFetch* outputs.
#' @param cluster_ variable name to use for cluster info. Default is "cluster_id".
#'
#' @return A ggplot with cluster annotations added.
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' #simplest uses
#' add_cluster_annotation(factor(c(rep("A", 3), "B")))
#' p = ggplot() + coord_cartesian(xlim = c(0,10))
#' add_cluster_annotation(factor(c(rep("A", 3), "B")), p)
#'
#' #intended use with ssvSignalHeatmap
#' clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 3)
#' assign_dt = unique(clust_dt[, .(id, cluster_id)])[order(id)]
#' p_heat = ssvSignalHeatmap(clust_dt, show_cluster_bars = FALSE)
#' add_cluster_annotation(assign_dt$cluster_id, p_heat,
#'   xleft = -500, xright = -360, rect_colors = rainbow(3), text_colors = "gray")
#'
#' #when colors are named, the names are used rather that just the order
#' rect_colors = safeBrew(assign_dt$cluster_id)
#' text_colors = safeBrew(assign_dt$cluster_id, "greys")
#' p_clusters = add_cluster_annotation(assign_dt$cluster_id,
#'   rect_colors = rect_colors, text_colors = text_colors)
#' #specialized use as plot outside of heatmap
#' p1 = assemble_heatmap_cluster_bars(plots = list(p_clusters, p_heat), rel_widths = c(1, 3))
#'
#' #when colors are named, the names are used rather that just the order
#' #these plots will be identical even though order of colors changes.
#' rect_colors = rect_colors[c(2, 3, 1)]
#' text_colors = text_colors[c(3, 1, 2)]
#' p_clusters = add_cluster_annotation(assign_dt$cluster_id,
#'   rect_colors = rect_colors, text_colors = text_colors)
#' #specialized use as plot outside of heatmap
#' p2 = assemble_heatmap_cluster_bars(plots = list(p_clusters, p_heat), rel_widths = c(1, 3))
#'
#' cowplot::plot_grid(p1, p2, ncol = 1)
#'
add_cluster_annotation = function(cluster_ids, p = NULL,
                                  xleft = 0, xright = 1,
                                  rect_colors = c("black", "gray"),
                                  text_colors = rev(rect_colors),
                                  show_labels = TRUE,
                                  label_angle = 0,
                                  row_ = "id",
                                  cluster_ = "cluster_id"){
    if(is.data.table(cluster_ids)){
        assign_dt = unique(cluster_ids[, c(row_, cluster_), with = FALSE])[order(get(row_))]
        cluster_ids = assign_dt[[cluster_]]

    }
    if(is.character(cluster_ids)) cluster_ids = factor(cluster_ids, levels = unique(cluster_ids))
    if(is.numeric(cluster_ids)) cluster_ids = factor(cluster_ids)
    stopifnot(is.factor(cluster_ids))
    if(is.null(p)){
        tmp_df = data.frame(facet= "")
        p = ggplot(tmp_df) +
            coord_cartesian(xlim = c(xleft, xright), ylim = c(0, length(cluster_ids)), expand = FALSE) +
            theme_void() +
            facet_grid(.~facet)
    }
    ends = cumsum(rev(table(cluster_ids)))
    starts = c(1, ends[-length(ends)] + 1)
    starts = starts - .5
    names(starts) = names(ends)
    ends = ends + .5


    df_rects = data.frame(xmin = xleft,
                          xmax = xright,
                          ymin = starts,
                          ymax = ends)
    if(setequal(names(rect_colors), rownames(df_rects))){
        df_rects$fill = rect_colors[rownames(df_rects)]
    }else{
        df_rects$fill = rect_colors[seq_len(nrow(df_rects))%%length(rect_colors)+1]
    }
    if(setequal(names(text_colors), rownames(df_rects))){
        df_rects$color = text_colors[rownames(df_rects)]
    }else{
        df_rects$color = text_colors[seq_len(nrow(df_rects))%%length(text_colors)+1]
    }


    df_rects = df_rects[rev(seq_len(nrow(df_rects))),]
    cluster_labels = levels(cluster_ids)
    for(i in seq_len(nrow(df_rects))){
        p = p + annotate("rect",
                         xmin = df_rects$xmin[i],
                         xmax = df_rects$xmax[i],
                         ymin= df_rects$ymin[i],
                         ymax = df_rects$ymax[i],
                         fill = df_rects$fill[i])
        if(show_labels){
            p = p + annotate("text",
                             x = mean(c(df_rects$xmin[i], df_rects$xmax[i])),
                             y = mean(c(df_rects$ymin[i], df_rects$ymax[i])),
                             label = cluster_labels[i],
                             color = df_rects$color[i],
                             angle = label_angle)
        }
    }
    p
}

#' heatmap style representation of membership table. instead of clustering, each
#' column is sorted starting from the left.
#'
#' See \code{\link{ssvSignalHeatmap.ClusterBars}} for an alternative with more
#' control over where the cluster bars appear.
#'
#' @export
#' @param bw_data a GRanges or data.table of bigwig signal. As returned from
#'   \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}
#' @param nclust number of clusters
#' @param perform_clustering should clustering be done? default is auto. auto
#'   considers if row_ has been ordered by being a factor and if cluster_ is a
#'   numeric.
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param column_ varaible mapped to column, likely bp position for ngs data.
#'   Default is "x" and works with ssvFetch* output.
#' @param fill_ numeric variable to map to fill. Default is "y" and works with
#'   ssvFetch* output.
#' @param facet_ variable name to facet horizontally by. Default is "sample" and
#'   works with ssvFetch* output. Set to "" if data is not facetted.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param max_rows for speed rows are sampled to 500 by default, use Inf to plot
#'   full data
#' @param max_cols for speed columns are sampled to 100 by default, use Inf to
#'   plot full data
#' @param fill_limits limits for fill legend.  values will be cropped to this
#'   range if set.  Default of NULL uses natural range of fill_.
#' @param clustering_col_min numeric minimum for col range considered when
#'   clustering, default in -Inf
#' @param clustering_col_max numeric maximum for col range considered when
#'   clustering, default in Inf
#' @param within_order_strategy one of "hclust" or "sort".  if hclust,
#'   hierarchical clustering will be used. if sort, a simple decreasing sort of
#'   rosSums.
#' @param dcast_fill value to supply to dcast fill argument. default is NA.
#' @param return_data logical.  If TRUE, return value is no longer ggplot and is
#'   instead the data used to generate that plot. Default is FALSE.
#' @param show_cluster_bars if TRUE, show bars indicating cluster membership.
#' @param rect_colors colors of rectangle fill, repeat to match number of
#'   clusters. Default is c("black", "gray").
#' @param text_colors colors of text, repeat to match number of clusters.
#'   Default is reverse of rect_colors.
#' @param show_labels logical, shoud rectangles be labelled with cluster
#'   identity.  Default is TRUE.
#' @param label_angle angle to add clusters labels at.  Default is 0, which is
#'   horizontal.
#' @param fun.aggregate Function to aggregate when multiple values present for
#'   facet_, row_, and column_. Affects both clustering and plotting. The
#'   function should accept a single vector argument or be a character string
#'   naming such a function.
#' @import ggplot2
#' @return ggplot heatmap of signal profiles, facetted by sample
#' @examples
#' data(CTCF_in_10a_profiles_gr)
#'
#' #the simplest use
#' ssvSignalHeatmap(CTCF_in_10a_profiles_gr)
#' ssvSignalHeatmap(CTCF_in_10a_profiles_gr, show_cluster_bars = FALSE)
#'
#' #clustering can be done manually beforehand
#' clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_gr, nclust = 3)
#' ssvSignalHeatmap(clust_dt)
#'
#' ssvSignalHeatmap(clust_dt, max_rows = 20, max_cols = 7)
#'
#' # aggregation occurs when facet_ is shared by multiple samples
#' prof_gr = CTCF_in_10a_profiles_gr
#' prof_gr$mark = "CTCF"
#'
#' # by default, aggregation is done by mean
#' ssvSignalHeatmap(prof_gr, facet_ = "mark")
#'
#' # Using fun.aggregate we can control aggregation behavior
#' # 'fun.aggregate = mean' preserves the default behavior
#' ssvSignalHeatmap(prof_gr, facet_ = "mark",
#'   fun.aggregate = mean)
#'
#'
#' # maximum and minimum may be useful to plot
#' ssvSignalHeatmap(prof_gr, facet_ = "mark",
#'   fun.aggregate = max)
#' ssvSignalHeatmap(prof_gr, facet_ = "mark",
#'   fun.aggregate = min)
#'
#' # How many samples have at least a signal value of 10 at each position?
#' ssvSignalHeatmap(prof_gr, facet_ = "mark",
#'   fun.aggregate = function(x)sum(as.numeric(x > 10)))
ssvSignalHeatmap = function(bw_data,
                            nclust = 6,
                            perform_clustering = c("auto", "yes", "no")[1],
                            row_ = "id",
                            column_ = "x",
                            fill_ = "y",
                            facet_ = "sample",
                            cluster_ = "cluster_id",
                            max_rows = 500,
                            max_cols = 100,
                            fill_limits = NULL,
                            clustering_col_min = -Inf,
                            clustering_col_max = Inf,
                            within_order_strategy = c("hclust", "sort")[2],
                            dcast_fill = NA,
                            return_data = FALSE,
                            show_cluster_bars = TRUE,
                            rect_colors = c("black", "gray"),
                            text_colors = rev(rect_colors),
                            show_labels = TRUE,
                            label_angle = 0,
                            fun.aggregate = "mean"){
    id = xbp = x = to_disp = y = hit = val = y = y_gap = cluster_id = NULL#declare binding for data.table
    if(is(bw_data, "GRanges")){
        bw_data = data.table::as.data.table(bw_data)
    }
    stopifnot(is.data.table(bw_data))
    stopifnot(is.numeric(nclust) || nclust < 2)
    stopifnot(is.character(row_), is.character(column_), is.character(fill_),
              is.character(facet_), is.character(cluster_))
    stopifnot(row_ %in% colnames(bw_data), column_ %in% colnames(bw_data),
              fill_ %in% colnames(bw_data))
    stopifnot(facet_ %in% colnames(bw_data) || facet_ == "")
    stopifnot(is.numeric(max_rows), is.numeric(max_cols),
              is.numeric(clustering_col_min), is.numeric(clustering_col_max))
    stopifnot(is.null(fill_limits) || (is.numeric(fill_limits) && length(fill_limits) == 2))
    #determine if user wants clustering
    do_cluster = perform_clustering == "yes"
    if(perform_clustering == "auto"){
        if(is.factor(bw_data[[row_]]) & is.factor(bw_data[[cluster_]])){
            do_cluster = FALSE
        }else{
            do_cluster = TRUE
        }
    }

    raw_nc = length(unique(bw_data[[column_]]))
    if(raw_nc > max_cols){
        new_scale = (seq_len(max_cols)-1) / (max_cols - 1)
        old_scale = (seq_len(raw_nc)-1) / (raw_nc - 1)
        kept = vapply(new_scale, function(v){
            which.min(abs(v - old_scale))
        }, 1)
        kept = sort(unique(bw_data[[column_]]))[kept]
        bw_data = bw_data[get(column_) %in% kept]
        message(raw_nc - max_cols,
                " columns were discarded according to max_cols: ",
                max_cols)
    }
    row_ids = unique(bw_data[[row_]])
    raw_nr = length(row_ids)
    if(raw_nr > max_rows){
        row_ids = sample(row_ids, max_rows)
        bw_data = bw_data[get(row_) %in% row_ids]
        message(raw_nr - max_rows,
                " rows were discarded according to max_rows: ",
                max_rows)
    }

    if(do_cluster){
        plot_dt = ssvSignalClustering(bw_data = bw_data,
                                      nclust = nclust,
                                      row_ = row_,
                                      column_ = column_,
                                      fill_ = fill_,
                                      facet_ = facet_,
                                      cluster_ = cluster_,
                                      max_rows = max_rows,
                                      max_cols = max_cols,
                                      clustering_col_min = clustering_col_min,
                                      clustering_col_max = clustering_col_max,
                                      within_order_strategy = within_order_strategy,
                                      dcast_fill = dcast_fill,
                                      fun.aggregate = fun.aggregate)
    }else{
        plot_dt = bw_data
    }
    if(return_data){
        return(plot_dt)
    }
    message("making plot...")
    scale_floor = .1
    scale_vals = c(0, scale_floor + 0:10/10*(1 - scale_floor))
    plot_dt[[row_]] = factor(plot_dt[[row_]], levels = rev(levels(plot_dt[[row_]])))

    .base_raster_plot = function(){
        #ensym and !! are necessary to reproduce aes_string behavior using tidy ideom
        column_ = ensym(column_)
        row_ = ensym(row_)
        fill_ = ensym(fill_)
        ggplot(plot_dt) +
            geom_raster(aes(x = !!column_, y = !!row_, fill = !!fill_)) +
            theme(axis.line = element_blank(),
                  axis.text.y = element_blank(),
                  axis.ticks.y = element_blank(),
                  panel.background = element_blank())
    }
    if(is.numeric(plot_dt[[fill_]])){
        if(!is.null(fill_limits)){
            ymax = max(fill_limits)
            ymin = min(fill_limits)
            if(any(plot_dt[[fill_]] > ymax)) plot_dt[get(fill_) > ymax][[fill_]] = ymax
            if(any(plot_dt[[fill_]] < ymin)) plot_dt[get(fill_) < ymin][[fill_]] = ymin

            p = .base_raster_plot() +
                scale_fill_distiller(type = "div", palette = "Spectral",
                                     values = scale_vals, limits = c(ymin, ymax))
        }else{
            p = .base_raster_plot() +
                scale_fill_distiller(type = "div", palette = "Spectral",
                                     values = scale_vals)
        }
    }else{
        p = .base_raster_plot()
    }

    if(facet_ != ""){
        p = p +  facet_grid(paste(". ~", facet_))
    }

    if(is.numeric(plot_dt[, get(column_)])){
        xs = sort(unique(as.numeric(plot_dt[, get(column_)])), decreasing = FALSE)
        if(length(xs) == 1) xs = rep(xs, 2)
        xleft = min(xs) - (xs[2] - xs[1])/2
        xright = max(xs) + (xs[2] - xs[1])/2
        if(xleft < 0 & xright > 0){
            xbr = c(xleft, 0, xright)
        }else{
            xbr = c(xleft, xright)
        }
        p = p + scale_x_continuous(breaks = xbr)
    }else{
        xs = length(unique(plot_dt[, get(column_)]))
        xs = seq_len(xs)
    }

    p = p + theme(axis.text.x  = element_text(angle = 90, hjust = 1, vjust = .5))

    rclust = plot_dt[, list(cluster_id = unique(get(cluster_))), by = get(row_)]

    xfactor = diff(range(xs))
    xfloor = min(xs) - xfactor/20
    xleft = xfloor - .12*xfactor
    xright = xfloor - .03*xfactor
    if(show_cluster_bars){
        p = add_cluster_annotation(p,
                                   cluster_ids = rclust$cluster_id,
                                   xleft = xleft,
                                   xright = xright)
    }


    p
}

#' heatmap style representation of membership table. instead of clustering, each
#' column is sorted starting from the left.
#'
#' Compared to ssvSignalHeatmap, cluster_bars are displayed on the left once
#' instead of for each facet
#'
#' @export
#'
#' @param bw_data a GRanges or data.table of bigwig signal. As returned from
#'   \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}
#' @param nclust number of clusters
#' @param perform_clustering should clustering be done? default is auto. auto
#'   considers if row_ has been ordered by being a factor and if cluster_ is a
#'   numeric.
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param column_ varaible mapped to column, likely bp position for ngs data.
#'   Default is "x" and works with ssvFetch* output.
#' @param fill_ numeric variable to map to fill. Default is "y" and works with
#'   ssvFetch* output.
#' @param facet_ variable name to facet horizontally by. Default is "sample" and
#'   works with ssvFetch* output. Set to "" if data is not facetted.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param FUN_format_heatmap optional function to modify main ggplot (labels,
#'   themes, scales, etc.).  Take a ggplot and returns a ggplot. Default is
#'   NULL.
#' @param max_rows for speed rows are sampled to 500 by default, use Inf to plot
#'   full data
#' @param max_cols for speed columns are sampled to 100 by default, use Inf to
#'   plot full data
#' @param fill_limits limits for fill legend.  values will be cropped to this
#'   range if set.  Default of NULL uses natural range of fill_.
#' @param clustering_col_min numeric minimum for col range considered when
#'   clustering, default in -Inf
#' @param clustering_col_max numeric maximum for col range considered when
#'   clustering, default in Inf
#' @param within_order_strategy one of "hclust" or "sort".  if hclust,
#'   hierarchical clustering will be used. if sort, a simple decreasing sort of
#'   rosSums.
#' @param dcast_fill value to supply to dcast fill argument. default is NA.
#' @param return_data logical.  If TRUE, return value is no longer ggplot and is
#'   instead the data used to generate that plot. Default is FALSE.
#' @param return_unassembled_plots logical. If TRUE, return list of heatmap and
#'   cluster-bar ggplots.  Can be customized and passed to
#'   \code{\link{assemble_heatmap_cluster_bars}}
#' @param rel_widths numeric of length 2.  Passed to cowplot::plot_grid. Default
#'   is c(1, 9).
#' @param ... addtional arguments passed to cowplot::plot_grid
#' @param rect_colors colors of rectangle fill, repeat to match number of
#'   clusters. Default is c("black", "gray").
#' @param text_colors colors of text, repeat to match number of clusters.
#'   Default is reverse of rect_colors.
#' @param show_labels logical, shoud rectangles be labelled with cluster
#'   identity.  Default is TRUE.
#' @param label_angle angle to add clusters labels at.  Default is 0, which is
#'   horizontal.
#' @param fun.aggregate Function to aggregate when multiple values present for
#'   facet_, row_, and column_. Affects both clustering and plotting. The
#'   function should accept a single vector argument or be a character string
#'   naming such a function.
#' @import ggplot2
#' @return ggplot heatmap of signal profiles, facetted by sample
#' @examples
#' data(CTCF_in_10a_profiles_gr)
#'
#' #the simplest use
#' ssvSignalHeatmap.ClusterBars(CTCF_in_10a_profiles_gr)
#' ssvSignalHeatmap.ClusterBars(CTCF_in_10a_profiles_gr, rel_widths = c(1, 5))
#'
#' #clustering can be done manually beforehand
#' clust_dt = ssvSignalClustering(data.table::as.data.table(CTCF_in_10a_profiles_gr), nclust = 3)
#' ssvSignalHeatmap.ClusterBars(clust_dt)
#'
#' # aggregation occurs when facet_ is shared by multiple samples
#' prof_gr = CTCF_in_10a_profiles_gr
#' prof_gr$mark = "CTCF"
#'
#' # by default, aggregation is done by mean
#' ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark")
#'
#' # Using fun.aggregate we can control aggregation behavior
#' # 'fun.aggregate = mean' preserves the default behavior
#' ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark",
#'   fun.aggregate = mean)
#'
#'
#' # maximum and minimum may be useful to plot
#' ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark",
#'   fun.aggregate = max)
#' ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark",
#'   fun.aggregate = min)
#'
#' # How many samples have at least a signal value of 10 at each position?
#' ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark",
#'   fun.aggregate = function(x)sum(as.numeric(x > 10)))
ssvSignalHeatmap.ClusterBars = function(bw_data,
                                        nclust = 6,
                                        perform_clustering = c("auto", "yes", "no")[1],
                                        row_ = "id",
                                        column_ = "x",
                                        fill_ = "y",
                                        facet_ = "sample",
                                        cluster_ = "cluster_id",
                                        FUN_format_heatmap = NULL,
                                        max_rows = 500,
                                        max_cols = 100,
                                        fill_limits = NULL,
                                        clustering_col_min = -Inf,
                                        clustering_col_max = Inf,
                                        within_order_strategy = c("hclust", "sort")[2],
                                        dcast_fill = NA,
                                        return_data = FALSE,
                                        return_unassembled_plots = FALSE,
                                        rel_widths = c(1, 9),
                                        #add_an
                                        rect_colors = c("black", "gray"),
                                        text_colors = rev(rect_colors),
                                        show_labels = TRUE,
                                        label_angle = 0,
                                        fun.aggregate = "mean",
                                        ...){
    clust_dt = ssvSignalHeatmap(
        bw_data = bw_data,
        nclust = nclust,
        row_ = row_,
        column_ = column_,
        fill_ = fill_,
        facet_ = facet_,
        cluster_ = cluster_,
        max_rows = max_rows,
        max_cols = max_cols,
        clustering_col_min = clustering_col_min,
        clustering_col_max = clustering_col_max,
        within_order_strategy = within_order_strategy,
        dcast_fill = dcast_fill,
        return_data = TRUE,
        fun.aggregate = fun.aggregate
    )
    if(return_data) return(clust_dt)

    assign_dt = unique(clust_dt[, c(row_, cluster_), with = FALSE])

    p_heatmap = ssvSignalHeatmap(
        bw_data = clust_dt,
        nclust = nclust,
        perform_clustering = FALSE,
        row_ = row_,
        column_ = column_,
        fill_ = fill_,
        facet_ = facet_,
        cluster_ = cluster_,
        max_rows = max_rows,
        max_cols = max_cols,
        fill_limits = fill_limits,
        clustering_col_min = clustering_col_min,
        clustering_col_max = clustering_col_max,
        within_order_strategy = within_order_strategy,
        dcast_fill = dcast_fill,
        return_data = FALSE,
        show_cluster_bars = FALSE
    ) +
        labs(y = "") +
        coord_cartesian(ylim = c(0, nrow(assign_dt))+.5, expand = FALSE)

    if(!is.null(FUN_format_heatmap)){
        p_heatmap = FUN_format_heatmap(p_heatmap)
    }

    tmp_df = data.frame(facet= "")
    p_cluster_bar = ggplot(tmp_df) +
        coord_cartesian(xlim = c(0, 1), ylim = c(0, nrow(assign_dt))+.5, expand = FALSE) +
        # theme_void() +
        theme(panel.grid = element_blank(),
              panel.background = element_blank(),
              strip.background = element_blank(),
              axis.text.x = element_blank(),
              axis.text.y = element_blank(),
              axis.ticks.x = element_blank(),
              axis.ticks.y = element_blank()) +
        labs(x = "", y = "") +
        facet_grid(.~facet)

    p_cluster_bar = add_cluster_annotation(
        cluster_ids = assign_dt[[cluster_]],
        p = p_cluster_bar,
        xleft = 0,
        xright = 1,
        rect_colors = rect_colors,
        text_colors = text_colors,
        show_labels = show_labels,
        label_angle = label_angle)
    plots = list(cluster_bars = p_cluster_bar, heatmap = p_heatmap)
    if(return_unassembled_plots){
        return(plots)
    }
    assemble_heatmap_cluster_bars(plots, rel_widths = rel_widths, ...)
}

#' assemble_heatmap_cluster_bars
#'
#' @param plots list of plots as returned from ssvSignalHeatmap.ClusterBars when return_unassembled_plots = TRUE
#' @param ... arguments passed to cowplot::plot_grid
#'
#' @return A grob produced by cowplot::plot_grid
#' @export
#' @import cowplot
#'
#' @examples
#' data(CTCF_in_10a_profiles_gr)
#' plots = ssvSignalHeatmap.ClusterBars(CTCF_in_10a_profiles_gr, return_unassembled_plots = TRUE)
#' assemble_heatmap_cluster_bars(plots)
assemble_heatmap_cluster_bars = function(plots, ...){
    # plots = list(heatmap = p_heatmap, cluster_bars = p_cluster_bar)
    grobs = sync_height(plots)
    cowplot::plot_grid(plotlist = grobs, ..., nrow = 1)
}

sync_height = function(my_plots, sync_width = FALSE){
    stopifnot(is(my_plots, "list"))
    is_ok = vapply(my_plots, function(x){
        "ggplot" %in% class(x) | "grob" %in% class(x)
    }, FUN.VALUE = TRUE)
    stopifnot(all(is_ok))
    my_grobs = lapply(my_plots, function(x){
        if(grid::is.grob(x)){
            x
        }else{
            ggplotGrob(x)
        }
    })

    if(sync_width){
        val = "widths"
    }else{
        val = "heights"
    }
    dim_sizes = lapply(my_grobs, function(gt){
        gt[[val]]
    })
    max_dim_size = dim_sizes[[1]]
    if(length(dim_sizes) > 1){
        for(i in seq(2, length(dim_sizes))){
            max_dim_size = grid::unit.pmax(max_dim_size, dim_sizes[[i]])
        }
    }
    for(j in seq_along(my_grobs)){
        my_grobs[[j]][[val]] = max_dim_size
    }
    my_grobs
}

#' construct line type plots where each region in each sample is represented
#' @export
#' @param bw_data a GRanges or data.table of bigwig signal.
#' As returned from \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}
#' @param x_ variable name mapped to x aesthetic, x by default.
#' @param y_ variable name mapped to y aesthetic, y by default.
#' @param color_ variable name mapped to color aesthetic, sample by default.
#' @param sample_ variable name, along with region_ used to group and
#' facet by default, change group_ or facet_ to override.
#' @param region_ variable name, along with sample_ used to group and facet
#' by default, change group_ or facet_ to override.
#' @param group_ group aesthetic keeps lines of geom_path from mis-connecting.
#' auto_grp by default which combines sample_ and region_.
#' probably shouldn't change.
#' @param line_alpha alpha value for lines. default is 1.
#' @param facet_ facetting divides up plots.
#' auto_facet by default which combines sample_ and region_.
#' if overriding facet_method with facet_grid, make sure to include ~ between
#' two variables, ie. "a~b", ".~b", "a~."
#' @param facet_method ggplot2 facetting method or wrapper for same,
#' facet_wrap by default.
#' @param spline_n if not NULL, applySpline will be called with n = spline_n.
#' default is NULL.
#' @param return_data logical.  If TRUE, return value is no longer ggplot and
#' is instead the data used to generate that plot. Default is FALSE.
#' @import ggplot2
#' @return ggplot of signal potentially facetted by region and sample
#' @examples
#' data(CTCF_in_10a_profiles_gr)
#'
#' bw_gr = CTCF_in_10a_profiles_gr
#' ssvSignalLineplot(subset(bw_gr, bw_gr$id %in% seq_len(3)), facet_ = "sample")
#' ssvSignalLineplot(subset(bw_gr, bw_gr$id %in% seq_len(3)),
#'     facet_ = "sample~.",
#'     facet_method = facet_grid)
#' ssvSignalLineplot(subset(bw_gr, bw_gr$id %in% seq_len(3)),
#'     facet_ = paste("sample", "~", "id"), facet_method = facet_grid)
#' ssvSignalLineplot(subset(bw_gr, bw_gr$id %in% seq_len(3)))
#' ssvSignalLineplot(subset(bw_gr, bw_gr$id %in% seq_len(3)), facet_ = "id")
#' ssvSignalLineplot(subset(bw_gr, bw_gr$id %in% seq_len(3)),
#'     facet_ = "id", spline_n = 10)
ssvSignalLineplot = function(bw_data, x_ = "x", y_ = "y", color_ = "sample",
                             sample_ = "sample", region_ = "id",
                             group_ = "auto_grp", line_alpha = 1,
                             facet_ = "auto_facet",
                             facet_method = facet_wrap, spline_n = NULL,
                             return_data = FALSE){
    auto_grp = auto_facet = NULL
    if(is(bw_data, "GRanges")){
        bw_data = data.table::as.data.table(bw_data)
    }
    stopifnot(is.data.table(bw_data))
    stopifnot(is.character(x_), is.character(y_), is.character(color_),
              is.character(sample_), is.character(region_),
              is.character(group_), is.character(facet_))
    stopifnot(x_ %in% colnames(bw_data), y_ %in% colnames(bw_data),
              color_ %in% colnames(bw_data), sample_ %in% colnames(bw_data),
              region_ %in% colnames(bw_data))
    stopifnot(group_ %in% colnames(bw_data) || group_ == "auto_grp")
    facet_names = strsplit(facet_, " ?[~+] ?")[[1]]
    facet_names = facet_names[facet_names != "."]
    stopifnot(all(facet_names %in% colnames(bw_data)) || facet_ == "auto_facet")

    stopifnot(is.function(facet_method))
    stopifnot(is.numeric(spline_n) || is.null(spline_n))
    bw_data[,auto_grp := paste(get(sample_), get(region_))]
    bw_data[,auto_facet := paste(get(sample_), get(region_))]
    if(!is.null(spline_n)){
        plot_dt = applySpline(bw_data, n = spline_n, x_ = x_, y_ = y_,
                              by_ = unique(c(group_, sample_, region_, facet_)))
    }else{
        plot_dt = bw_data
    }
    if(return_data){
        return(plot_dt)
    }
    #ensym and !! are necessary to reproduce aes_string behavior using tidy ideom
    x_ = ensym(x_)
    y_ = ensym(y_)
    color_ = ensym(color_)
    group_ = ensym(group_)
    ggplot(plot_dt) + geom_path(aes(x = !!x_,
                                           y = !!y_,
                                           col = !!color_,
                                           group = !!group_),
                                alpha = line_alpha) +
        facet_method(facet_)
}

#' aggregate line signals in a single line plot
#' @export
#' @param bw_data a GRanges or data.table of bigwig signal.
#' As returned from \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}
#' @param x_ variable name mapped to x aesthetic, x by default.
#' @param y_ variable name mapped to y aesthetic, y by default.
#' @param sample_ variable name, along with region_ used to group by default,
#' @param color_ variable name mapped to color aesthetic, sample_ by default.
#' change group_ to override.
#' @param group_ group aesthetic keeps lines of geom_path from mis-connecting.
#' Most useful if you need to supply a
#' variable to later facet upon. Defaults to value of sample_.
#' @param agg_fun the aggregation function to apply by sample_ and x_,
#' default is mean
#' @param spline_n if not NULL, applySpline will be called with n = spline_n.
#' default is NULL.
#' @param return_data logical.  If TRUE, return value is no longer ggplot and
#' is instead the data used to generate that plot. Default is FALSE.
#' @import ggplot2
#' @return ggplot of signal aggregated with agg_fun() by sample.
#' @examples
#' data(CTCF_in_10a_profiles_gr)
#' bw_gr = CTCF_in_10a_profiles_gr
#' ssvSignalLineplotAgg(bw_gr) +
#'     labs(title = "agg regions by sample.")
#' ssvSignalLineplotAgg(CTCF_in_10a_profiles_gr, spline_n = 10) +
#'     labs(title = "agg regions by sample, with spline smoothing.")
#' ssvSignalLineplotAgg(subset(bw_gr, bw_gr$id %in% seq_len(10)),
#'     sample_ = "id", color_ = "id") +
#'     labs(title = "agg samples by region id (weird)")
#' ssvSignalLineplotAgg(subset(bw_gr, bw_gr$id %in% seq_len(10)), sample_ = "id",
#'     color_ = "id", spline_n = 10) +
#'     labs(title = "agg samples by region id (weird), with spline smoothing")
ssvSignalLineplotAgg = function(bw_data,
                                x_ = "x",
                                y_ = "y",
                                sample_ = "sample",
                                color_ = sample_,
                                group_ = sample_,
                                agg_fun = mean,
                                spline_n = NULL,
                                return_data = FALSE){
    group_var = y = NULL # reserve for data.table
    if(is(bw_data, "GRanges")){
        bw_data = data.table::as.data.table(bw_data)
    }
    stopifnot(is.data.table(bw_data))
    stopifnot(is.character(x_), is.character(y_),
              is.character(sample_), is.character(color_),
              is.character(group_))
    stopifnot(x_ %in% colnames(bw_data), y_ %in% colnames(bw_data),
              color_ %in% colnames(bw_data), sample_ %in% colnames(bw_data))
    stopifnot(all(group_ %in% colnames(bw_data)) || group_ == "auto_grp")

    stopifnot(is.function(agg_fun))
    stopifnot(is.numeric(spline_n) || is.null(spline_n))

    agg_dt = bw_data[, list(y = agg_fun(get(y_))),
                     by = c(unique(c(group_, color_, sample_, x_)))]

    if(!is.null(spline_n)){
        plot_dt = applySpline(agg_dt, n = spline_n, x_ = x_, y_ = "y",
                              by_ = c(group_, sample_))
    }else{
        plot_dt = agg_dt
    }
    plot_dt = plot_dt[order(get(x_))]
    plot_dt[, group_var := paste(mget(unique(c(group_, sample_, color_))), collapse = "-"), by = seq_len(nrow(plot_dt))]
    if(return_data){
        return(plot_dt)
    }

    #ensym and !! are necessary to reproduce aes_string behavior using tidy ideom
    x_ = ensym(x_)
    color_ = ensym(color_)
    ggplot(plot_dt) + geom_path(aes(x = !!x_,
                                    y = y,
                                    col = !!color_,
                                    group = group_var)) +
        labs(y = y_)
}

testthat::context("SignalClustering")

#new feature - ssvSignalClustering accepts memb_table
#new feature - ssvSignalHeatmap.ClusterBars displays cluster bars once on the left instead of in each facet
#new feature - ssvSignalClustering accepts centroids instead of nclust

library(seqsetvis)
library(testthat)
library(GenomicRanges)

test_that("ssvSignalClustering max rows", {
    clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, max_rows = 10)
    expect_equal(nrow(clust_dt[!is.na(id)]), 420)
    expect_equal(nrow(clust_dt[is.na(id)]), 0)

    library(data.table)
    prof_dt = copy(CTCF_in_10a_profiles_dt)
    setnames(prof_dt, "id", "new_id")

    clust_dt2 = ssvSignalClustering(prof_dt, max_rows = 10, row_ = "new_id")
    expect_equal(nrow(clust_dt2[!is.na(new_id)]), 420)
    expect_equal(nrow(clust_dt2[is.na(new_id)]), 0)
})

test_that("ssvSignalHeatmap.ClusterBars", {
    p_heat = ssvSignalHeatmap.ClusterBars(CTCF_in_10a_profiles_dt)
    class(p_heat)
    expect_s3_class(p_heat, "ggplot")

    p_list = ssvSignalHeatmap.ClusterBars(CTCF_in_10a_profiles_dt, return_unassembled_plots = TRUE)
    class(p_list)
    class(p_list[[1]])
    expect_is(p_list, "list")
    expect_s3_class(p_list[[1]], "ggplot")
    expect_s3_class(p_list[[2]], "ggplot")
})

test_that("ssvSignalHeatmap fill_limits", {
    p_heat = ssvSignalHeatmap(CTCF_in_10a_profiles_dt, fill_limits = c(20, 60))
    expect_equal(max(p_heat$data$y), 60)
    expect_equal(min(p_heat$data$y), 20)

    p_heat2 = ssvSignalHeatmap.ClusterBars(CTCF_in_10a_profiles_dt, fill_limits = c(20, 60), return_unassembled_plots = TRUE)[[2]]
    expect_equal(max(p_heat2$data$y), 60)
    expect_equal(min(p_heat2$data$y), 20)
})

test_that("ssvSignalHeatmap memb_table", {
    memb_table = ssvMakeMembTable(CTCF_in_10a_narrowPeak_grs)
    setequal(rownames(memb_table), CTCF_in_10a_profiles_dt$id)
    clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, memb_table = memb_table)
    p_heat = ssvSignalHeatmap(clust_dt, show_cluster_bars = FALSE)
    expect_s3_class(p_heat, "ggplot")

    p_heat.anno = add_cluster_annotation(p = p_heat, cluster_ids = clust_dt, xleft = -400, xright = -360, show_labels = FALSE, rect_colors = safeBrew(3))
    expect_s3_class(p_heat.anno, "ggplot")
})

test_that("ssvSignalHeatmap centroids from tidy", {
    centroids.tidy = CTCF_in_10a_profiles_dt[id %in% 1:3]
    data.table::setnames(centroids.tidy, "id", "cluster_id")

    cent_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, k_centroids = centroids.tidy)
    p_heat = ssvSignalHeatmap(cent_dt, show_cluster_bars = TRUE)
    expect_s3_class(p_heat, "ggplot")
})

test_that("ssvSignalHeatmap centroids from wide", {
    centroids.tidy = CTCF_in_10a_profiles_dt[id %in% 1:3]
    data.table::setnames(centroids.tidy, "id", "cluster_id")

    wide_mat = make_clustering_matrix(centroids.tidy, row_ = "cluster_id")

    cent_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, k_centroids = wide_mat)
    p_heat = ssvSignalHeatmap(cent_dt, show_cluster_bars = TRUE)
    expect_s3_class(p_heat, "ggplot")
})

testthat::context("SignalPlots")
#all ssvSignalPlots functions should accept GRanges in addition to data.table
# as originally implemented
library(seqsetvis)
library(testthat)
library(GenomicRanges)

doTest_ssvSignalPlots = function(test_object){

    test_that("ssvSignalBandedQuantiles works with proper inputs", {
        p_noBy = ssvSignalBandedQuantiles(test_object)
        p_wBy = ssvSignalBandedQuantiles(test_object, by_ = "sample")
        expect_s3_class(p_noBy, "ggplot")
        expect_s3_class(p_wBy, "ggplot")
    })

    test_that("ssvSignalBandedQuantiles different hsv setting", {
        p_rev = ssvSignalBandedQuantiles(test_object,
                                         by_ = "sample", hsv_reverse = TRUE)
        expect_s3_class(p_rev, "ggplot")

        p_symm = ssvSignalBandedQuantiles(test_object,
                                          by_ = "sample", hsv_symmetric = TRUE)
        expect_s3_class(p_symm, "ggplot")

        p_symmgray = ssvSignalBandedQuantiles(
            test_object,
            by_ = "sample",
            hsv_symmetric = TRUE,
            hsv_grayscale = TRUE
        )
        expect_s3_class(p_symmgray, "ggplot")

        p_unsymmgray = ssvSignalBandedQuantiles(
            test_object,
            by_ = "sample",
            hsv_symmetric = FALSE,
            hsv_grayscale = TRUE
        )
        expect_s3_class(p_unsymmgray, "ggplot")

        p_symmrev = ssvSignalBandedQuantiles(
            test_object,
            by_ = "sample",
            hsv_symmetric = TRUE,
            hsv_reverse = TRUE
        )
        expect_s3_class(p_symmrev, "ggplot")

        p_symmrev = ssvSignalBandedQuantiles(
            test_object,
            by_ = "sample",
            hsv_symmetric = TRUE,
            hsv_reverse = TRUE
        )
        expect_s3_class(p_symmrev, "ggplot")
    })


    test_that("ssvSignalScatterplot works with basic inputs", {
        p1 = ssvSignalScatterplot(test_object,
                                  x_name = "MCF10A_CTCF",
                                  y_name = "MCF10AT1_CTCF")
        expect_s3_class(p1, "ggplot")
        p2 = ssvSignalScatterplot(
            test_object,
            x_name = "MCF10A_CTCF",
            y_name = "MCF10AT1_CTCF",
            plot_type = "volcano"
        )
        expect_s3_class(p2, "ggplot")
    })

    test_that("ssvSignalScatterplot throws correct error for bad x_name or y_name",
              {
                  expect_error(
                      ssvSignalScatterplot(
                          test_object,
                          x_name = "badx",
                          y_name = "MCF10AT1_CTCF",
                          value_variable = "x"
                      ),
                      "badx"
                  )
                  expect_error(
                      ssvSignalScatterplot(
                          test_object,
                          x_name = "MCF10A_CTCF",
                          y_name = "bady",
                          value_variable = "x"
                      )
                  )
              })

    test_that("ssvSignalScatterplot works with other inputs", {
        p1 = ssvSignalScatterplot(
            test_object,
            x_name = "MCF10A_CTCF",
            y_name = "MCF10AT1_CTCF",
            value_variable = "x"
        )
        expect_s3_class(p1, "ggplot")
        p2 = ssvSignalScatterplot(
            test_object,
            x_name = "MCF10A_CTCF",
            y_name = "MCF10CA1_CTCF",
            value_function = median
        )
        expect_s3_class(p2, "ggplot")
        p3 = ssvSignalScatterplot(
            test_object,
            x_name = "1",
            y_name = "2",
            value_function = median,
            by_ = "sample",
            xy_variable = "id"
        )
        memb = data.frame(id = unique(test_object$id),  group = letters[1:5])
        p4 = ssvSignalScatterplot(
            bw_data = test_object,
            x_name = "MCF10A_CTCF",
            y_name = "MCF10A_CTCF",
            color_table = memb
        )
        expect_s3_class(p4, "ggplot")

        p5 = ssvSignalScatterplot(
            bw_data = test_object,
            x_name = "MCF10A_CTCF",
            y_name = "MCF10A_CTCF",
            fixed_coords = FALSE,
            color_table = memb
        )
        expect_s3_class(p5, "ggplot")
    })

    test_that("ssvSignalScatterplot works with help enabled inputs", {
        p1 = ssvSignalScatterplot(
            test_object,
            x_name = "MCF10A_CTCF",
            y_name = "MCF10AT1_CTCF",
            value_variable = "x",
            show_help = TRUE
        )
        expect_s3_class(p1, "ggplot")

        p1v = ssvSignalScatterplot(
            test_object,
            x_name = "MCF10A_CTCF",
            y_name = "MCF10AT1_CTCF",
            value_variable = "x",
            show_help = TRUE,
            plot_type = "volcano"
        )
        expect_s3_class(p1v, "ggplot")

        p2 = ssvSignalScatterplot(
            test_object,
            x_name = "MCF10A_CTCF",
            y_name = "MCF10CA1_CTCF",
            value_function = median,
            show_help = TRUE
        )
        expect_s3_class(p2, "ggplot")

        p3 = ssvSignalScatterplot(
            test_object,
            x_name = "1",
            y_name = "2",
            value_function = median,
            by_ = "sample",
            xy_variable = "id",
            show_help = TRUE
        )

        memb = data.frame(id = unique(test_object$id),  group = letters[1:5])
        p4 = ssvSignalScatterplot(
            bw_data = test_object,
            x_name = "MCF10A_CTCF",
            y_name = "MCF10A_CTCF",
            color_table = memb,
            show_help = TRUE
        )
        expect_s3_class(p4, "ggplot")
    })

    test_that("ssvSignalHeatmap works with minimal inputs", {
        hidden = capture_output({
            p = ssvSignalHeatmap(test_object, nclust = 5)
        })
        expect_s3_class(p, "ggplot")
    })

    test_that("ssvSignalHeatmap works with maxCols and maxRows", {
        expect_message({
            p_max_rows = ssvSignalHeatmap(test_object, nclust = 2, max_rows = 8)
        }, regexp = "92 rows were discarded")
        expect_s3_class(p_max_rows, "ggplot")

        expect_message({
            p_max_cols = ssvSignalHeatmap(test_object, nclust = 2, max_cols = 2)
        }, regexp = "12 columns were discarded")
        expect_s3_class(p_max_cols, "ggplot")
    })

    test_that("ssvSignalHeatmap works with manual clustering", {
        hidden = capture_output({
            clust_dt = ssvSignalClustering(test_object, nclust = 2)
            p1 = ssvSignalHeatmap(clust_dt)
            p2 = ssvSignalHeatmap(clust_dt, perform_clustering = "no")
        })
        expect_s3_class(p1, "ggplot")
        expect_s3_class(p2, "ggplot")
    })

    test_that("ssvSignalLineplot", {
        #from examples
        sub_to = test_object[test_object$id %in% 1:3]
        test_plots = list(
            ssvSignalLineplot(sub_to, facet_ = "sample"),
            ssvSignalLineplot(sub_to,
                              facet_ = "sample~.",
                              facet_method = facet_grid),
            ssvSignalLineplot(
                sub_to,
                facet_ = paste("sample", "~", "id"),
                facet_method = facet_grid
            ),
            ssvSignalLineplot(sub_to),
            ssvSignalLineplot(sub_to, facet_ = "id"),
            ssvSignalLineplot(sub_to,
                              facet_ = "id", spline_n = 10)
        )
        lapply(test_plots, function(p1) {
            expect_s3_class(p1, "ggplot")
        })
    })

    test_that("ssvSignalLineplotAgg", {
        #from examples
        test_plots = list(
            ssvSignalLineplotAgg(test_object) +
                labs(title = "agg regions by sample."),
            ssvSignalLineplotAgg(test_object, spline_n = 10) +
                labs(title = "agg regions by sample, with spline smoothing."),
            ssvSignalLineplotAgg(test_object[test_object$id %in% 1:10],
                                 sample_ = "id", color_ = "id") +
                labs(title = "agg samples by region id (weird)"),
            ssvSignalLineplotAgg(
                test_object[test_object$id %in% 1:10],
                sample_ = "id",
                color_ = "id",
                spline_n = 10
            ) +
                labs(title = paste("agg samples by region id",
                                   "(weird), with spline smoothing"))
        )
        lapply(test_plots, function(p1) {
            expect_s3_class(p1, "ggplot")
        })
    })
}

test_object = CTCF_in_10a_profiles_dt
test_object$sample = factor(test_object$sample)
res_dt = test_object
res_gr = GRanges(res_dt)

doTest_ssvSignalPlots(test_object = res_dt)
doTest_ssvSignalPlots(test_object = res_gr)

test_that("ssvSignalHeatmap 1 clust", {
    expect_failure(expect_error(ssvSignalHeatmap(res_dt, nclust = 1)))
})

test_that("ssvSignalHeatmap too many clust", {
    t_dt = res_dt[id %in% unique(id)[1:5]]
    expect_failure(expect_error(ssvSignalHeatmap(t_dt, nclust = 5)))
    expect_failure(expect_error(ssvSignalHeatmap(t_dt, nclust = 6)))
})

#' calc_norm_factors
#'
#' Calculate normalization factors in a two step process:
#'
#' 1) summarize every region for each sample (default summary function is max)
#'
#' 2) caclulate a value to cap each sample to based on regions (default is 95th
#' quantile).
#'
#' The uderlying assumption here is that meaningful enrichment is present at the
#' majority of regions provided.  If prevalence varies by a specific factor, say
#' ChIP-seq targets with different characteristics - ie. when analyzing TSSes
#' for H3K4me3 and an infrequent transcription factor it is more appropriate to
#' specify appropriate quantile cutoffs per factor.
#'
#' @param full_dt a data.table, as returned by ssvFetch*(..., return_data.table.
#'   = TRUE)
#' @param value_ character, attribute in full_dt to normalzie.
#' @param cap_value_ character, new attribute name specifying values to cap to.
#' @param by1 character vector, specifies attributes relevant to step 1.
#' @param by2 character vector, specifies attributes relevant to step 1 and 2.
#' @param aggFUN1 function called on value_ with by = c(by1, by2) in step 1.
#' @param aggFUN2 function called on result of aggFUN1 with by = by2 in step 2.
#'
#' @return data.table mapping by2 to cap_value_.
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' calc_norm_factors(CTCF_in_10a_profiles_dt)
#' calc_norm_factors(CTCF_in_10a_profiles_dt,
#'   aggFUN1 = mean, aggFUN2 = function(x)quantile(x, .5))
calc_norm_factors = function(full_dt,
                             value_ = "y",
                             cap_value_ = "y_cap_value",
                             by1 = "id",
                             by2 = "sample",
                             aggFUN1 = max,
                             aggFUN2 = function(x)quantile(x, .95)
){
    value_summary_ = NULL #reserve binding for data.table
    stopifnot(data.table::is.data.table(full_dt))
    stopifnot(value_ %in% colnames(full_dt))
    stopifnot(by1 %in% colnames(full_dt))
    stopifnot(by2 %in% colnames(full_dt))
    cap_dt = full_dt[, list(value_summary_ = aggFUN1(get(value_))), c(by1, by2)][, list(value_capped_ = aggFUN2(value_summary_)), c(by2)]
    data.table::setnames(cap_dt, "value_capped_", cap_value_)
    cap_dt[]
}


#' append_ynorm
#'
#' see \code{\link{calc_norm_factors}} for normalization details.
#'
#' @param full_dt a data.table, as returned by ssvFetch*(..., return_data.table
#'   = TRUE).
#' @param value_ character, attribute in full_dt to normalzie.
#' @param cap_value_ character, new attribute name specifying values to cap to.
#' @param norm_value_ character, new attribute name specifying normalized values.
#' @param by1 character vector, specifies attributes relevant to step 1.
#' @param by2 character vector, specifies attributes relevant to step 1 and 2.
#' @param aggFUN1 function called on value_ with by = c(by1, by2) in step 1.
#' @param aggFUN2 function called on result of aggFUN1 with by = by2 in step 2.
#' @param cap_dt optionally, provide user generated by2 to cap_value_ mapping
#' @param do_not_cap if TRUE, normalized values are not capped to 1. Default is FALSE.
#' @param do_not_scaleTo1 if TRUE, normalized values are not scaled to 1. Default is FALSE.
#' @param force_append if TRUE, any previous cap_value or norm_value is overridden. Default is FALSE.
#'
#' @return data.table, full_dt with cap_value_ and norm_value_ values appended.
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' append_ynorm(CTCF_in_10a_profiles_dt)
#' append_ynorm(CTCF_in_10a_profiles_dt,
#'   aggFUN1 = mean, aggFUN2 = function(x)quantile(x, .5))
append_ynorm = function(full_dt,
                        value_ = "y",
                        cap_value_ = "y_cap_value",
                        norm_value_ = "y_norm",
                        by1 = "id",
                        by2 = "sample",
                        aggFUN1 = max,
                        aggFUN2 = function(x)quantile(x, .95),
                        cap_dt = NULL,
                        do_not_cap = FALSE,
                        do_not_scaleTo1 = FALSE,
                        force_append = FALSE
){
    if(do_not_cap && do_not_scaleTo1){
        warning("With do_not_cap and do_not_scaleTo1, only cap_value will be appended and normalized value will be unchanged. This is not likely what you want.")
    }
    stopifnot(data.table::is.data.table(full_dt))
    stopifnot(value_ %in% colnames(full_dt))
    stopifnot(by1 %in% colnames(full_dt))
    stopifnot(by2 %in% colnames(full_dt))
    if(force_append){
        full_dt[[cap_value_]] = NULL
        full_dt[[norm_value_]] = NULL
    }
    stopifnot(!cap_value_ %in% colnames(full_dt))
    stopifnot(!norm_value_ %in% colnames(full_dt))
    if(is.null(cap_dt)){
        cap_dt = calc_norm_factors(full_dt, value_, cap_value_, by1, by2, aggFUN1, aggFUN2)
    }else{
        stopifnot(cap_value_ %in% colnames(cap_dt))
        stopifnot(by2 %in% colnames(cap_dt))
    }

    full_dt = merge(full_dt, cap_dt, by = by2)
    data.table::set(full_dt, NULL, norm_value_, full_dt[[value_]])
    if(!do_not_cap){
        full_dt[get(norm_value_) > get(cap_value_), (norm_value_) := get(cap_value_)]
    }
    if(!do_not_scaleTo1){
        full_dt[, (norm_value_) := get((norm_value_)) / get(cap_value_)]
    }
    full_dt[]
}

testthat::context("normalization")

library(seqsetvis)
library(testthat)

test_that("append_ynorm defaults", {
    #these should all work cleanly
    norm_dt = append_ynorm(CTCF_in_10a_profiles_dt)
    expect_equal(sort(colnames(norm_dt)), sort(c(colnames(CTCF_in_10a_profiles_dt), "y_cap_value", "y_norm")))
})

test_that("append_ynorm manual names", {
    #these should all work cleanly
    dt = CTCF_in_10a_profiles_dt
    dt[, ytest := y]
    dt$y = NULL
    norm_dt = append_ynorm(dt, value_ = "ytest", cap_value_ = "captest", norm_value_ = "normtest")
    expect_equal(sort(colnames(norm_dt)), sort(c(colnames(dt), "captest", "normtest")))
})

test_that("append_ynorm manual cap_dt", {
    norm_dt1 = append_ynorm(CTCF_in_10a_profiles_dt)

    cap_dt = calc_norm_factors(CTCF_in_10a_profiles_dt)
    cap_dt[, y_cap_value := y_cap_value * 10]

    norm_dt2 = append_ynorm(CTCF_in_10a_profiles_dt, cap_dt = cap_dt)

    expect_true(all(norm_dt2$y_norm < norm_dt1$y_norm | norm_dt2$y_norm == 0))
})


test_that("calc_norm_factors defaults", {
    cap_dt = calc_norm_factors(CTCF_in_10a_profiles_dt)
    expect_equal(sort(colnames(cap_dt)), sort(c("sample", "y_cap_value")))
})

test_that("calc_norm_factors manual names", {
    cap_dt = calc_norm_factors(CTCF_in_10a_profiles_dt, cap_value_ = "captest", by2 = "seqnames")
    expect_equal(sort(colnames(cap_dt)), sort(c("seqnames", "captest")))
})

#' Fetch coverage values for a list of GRanges.
#'
#' \code{ssvFetchGRanges} Gets coverage values for each region of the query
#' GRanges (\code{qgr}). Values correspond to the center of each window of size
#' \code{win_size}.  A tidy formatted data.table object is returned suitable for
#' plotting using ggplots.
#'
#' @param grs a list of GRanges for which to calculate coverage.
#' @param qgr Set of GRanges to query.  For valid results the width of each
#'   interval should be identical and evenly divisible by \code{win_size}.
#' @param file_attribs data.frame (1 row per item in grs) containing attributes
#'   to append to results.
#' @param unique_names The column name where unique_names are stored. Default is
#'   'sample'
#' @param names_variable The column name where unique_names are stored. Default
#'   is 'sample'
#' @param win_size The window size that evenly divides widths in \code{qgr}.
#' @param win_method character.  one of c("sample", "summary").  Determines if
#'   \code{\link{viewGRangesWinSample_dt}} or
#'   \code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
#'   qgr.
#' @param summary_FUN function.  only relevant if win_method is "summary".
#'   passed to \code{\link{viewGRangesWinSummary_dt}}.
#' @param target_strand character. if one of "+" or "-", reads are filtered to
#'   match. ignored if any other value.
#' @param use_coverage boolean or NULL, if TRUE, query regions are scored by the
#'   number of intervals overlapping.  Default of NULL checks if attrib_var is
#'   "score" and uses coverage if so.
#' @param attrib_var character, column in mcols of GRanges to pull values from.
#'   Default of "score" is compatible with internal coverage calculation or
#'   bedgraph-like files.
#' @param fill_value numeric or character value to use where queried regions are
#'   empty.  Default is 0 and appropriate for both calculated coverage and
#'   bedgraph/bigwig like files.  Will automatically switch to "MISSING" if data
#'   is guessed to be qualitative.
#' @param anchor character, one of c("center", "center_unstranded", "left",
#'   "left_unstranded")
#' @param return_data.table logical. If TRUE the internal data.table is returned
#'   instead of GRanges.  Default is FALSE.
#' @param n_cores integer number of cores to use. Uses mc.cores option if not
#'   supplied.
#' @param force_skip_centerFix boolean, if TRUE all query ranges will be
#' used "as is".  This is already the case by default if win_method == "summary"
#' but may have applications where win_method == "sample".
#' @return A tidy formatted GRanges (or data.table if specified) containing
#'   fetched values.
#' @export
#' @examples
#' data(CTCF_in_10a_narrowPeak_grs)
#' data(CTCF_in_10a_overlaps_gr)
#' ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, CTCF_in_10a_overlaps_gr, win_size = 200)
ssvFetchGRanges = function(grs,
                           qgr,
                           file_attribs = data.frame(matrix(0, nrow = length(grs), ncol = 0)),
                           unique_names = names(grs),
                           names_variable = "sample",
                           win_size = 50,
                           win_method = c("sample", "summary")[1],
                           summary_FUN = function(x, w)max(x),
                           target_strand = c("*", "+", "-", "both")[1],
                           use_coverage = NULL,
                           attrib_var = "score",
                           fill_value = 0,
                           anchor = c("left", "left_unstranded", "center",
                                      "center_unstranded")[3],
                           return_data.table = FALSE,
                           n_cores = getOption("mc.cores", 1),
                           force_skip_centerFix = FALSE){
    if(!is.list(grs)){
        if(is(grs, "GRangesList")){
            grs = as.list(grs)
        }else{
            grs = list(grs)
        }
    }
    all_gr = all(vapply(grs, function(x){
        "GRanges" %in% class(x)
    }, FUN.VALUE = TRUE))
    stopifnot(all_gr)
    if(is.null(unique_names)){
        unique_names = paste("regions", LETTERS[seq_along(grs)])
    }
    if(is.null(file_attribs[[names_variable]])){
        file_attribs[[names_variable]] = unique_names
    }
    if(is.null(use_coverage)){
        use_coverage = attrib_var == "score"
    }
    # gr_dt = lapply(grs, function(x){
    gr_dt = parallel::mclapply(grs, mc.cores = n_cores, function(x){
        if(use_coverage){
            fetchGRanges_as_coverage(x, qgr, win_size, win_method,
                                     summary_FUN, target_strand,
                                     anchor, fill_value = fill_value,
                                     force_skip_centerFix = force_skip_centerFix)
        }else{
            fetchGRanges_by_attrib_var(x, qgr, win_size, win_method,
                                       summary_FUN, target_strand,
                                       anchor, fill_value = fill_value,
                                       attrib_var = attrib_var,
                                       force_skip_centerFix = force_skip_centerFix)
        }
    })
    for(i in seq_along(gr_dt)){
        for(attrib in colnames(file_attribs)){
            gr_dt[[i]][[attrib]] = file_attribs[[attrib]][i]
        }

    }
    gr_dt = rbindlist(gr_dt)#, use.names = TRUE, idcol = names_variable)
    if(!return_data.table){
        gr_dt = GRanges(gr_dt)
    }
    gr_dt
}

# ssvFetchGRanges = function(grs,
#                            qgr,
#                            file_attribs = data.frame(matrix(0, nrow = length(grs), ncol = 0)),
#                            unique_names = names(grs),
#                            names_variable = "sample",
#                            win_size = 50,
#                            win_method = c("sample", "summary")[1],
#                            summary_FUN = function(x, w)max(x),
#                            target_strand = c("*", "+", "-", "both")[1],
#                            use_coverage = NULL,
#                            attrib_var = "score",
#                            anchor = c("left", "left_unstranded", "center",
#                                       "center_unstranded")[3],
#                            return_data.table = FALSE,
#                            n_cores = getOption("mc.cores", 1)){
fetchGRanges_as_coverage = function(x, qgr, win_size, win_method, summary_FUN,
                                    target_strand, anchor,
                                    fill_value = 0, force_skip_centerFix = FALSE){
    switch(win_method,
           sample = {
               qgr = prepare_fetch_GRanges_width(qgr,
                                           win_size,
                                           skip_centerFix = force_skip_centerFix)
               if(target_strand == "both"){
                   pos_gr = GRanges(coverage(subset(x, strand == "+")))
                   neg_gr = GRanges(coverage(subset(x, strand == "-")))
                   pos_dt = viewGRangesWinSample_dt(
                       pos_gr, qgr,
                       win_size,
                       attrib_var = "score",
                       anchor = anchor)
                   neg_dt = viewGRangesWinSample_dt(
                       neg_gr, qgr,
                       win_size,
                       attrib_var = "score",
                       anchor = anchor)
                   pos_dt[, strand := "+"]
                   neg_dt[, strand := "-"]
                   out = rbind(
                       pos_dt,
                       neg_dt
                   )
               }else{
                   if(target_strand == "*"){
                       score_gr = GRanges(coverage(x))
                   }else{
                       score_gr = GRanges(coverage(subset(x, strand == target_strand)))
                   }

                   out = viewGRangesWinSample_dt(
                       score_gr,
                       qgr,
                       win_size,
                       attrib_var = "score",
                       anchor = anchor)
                   out[, strand := target_strand]
               }
               out
           },
           summary = {
               if(target_strand == "both"){
                   pos_gr = GRanges(coverage(subset(x, strand == "+")))
                   neg_gr = GRanges(coverage(subset(x, strand == "-")))
                   pos_dt = viewGRangesWinSummary_dt(
                       pos_gr,
                       qgr,
                       win_size,
                       summary_FUN = summary_FUN,
                       attrib_var = "score",
                       fill_value = fill_value,
                       anchor = anchor)

                   neg_dt = viewGRangesWinSummary_dt(
                       neg_gr,
                       qgr,
                       win_size,
                       attrib_var = "score",
                       fill_value = fill_value,
                       summary_FUN = summary_FUN,
                       anchor = anchor)
                   pos_dt[, strand := "+"]
                   neg_dt[, strand := "-"]
                   out = rbind(
                       pos_dt,
                       neg_dt
                   )
               }else{
                   if(target_strand == "*"){
                       score_gr = GRanges(coverage(x))
                   }else{
                       score_gr = GRanges(coverage(subset(x, strand == target_strand)))
                   }
                   out = viewGRangesWinSummary_dt(
                       score_gr,
                       qgr,
                       win_size,
                       attrib_var = "score",
                       fill_value = fill_value,
                       summary_FUN = summary_FUN,
                       anchor = anchor)
                   out[, strand := target_strand]
               }
               out
           })
}

fetchGRanges_by_attrib_var = function(x, qgr, win_size, win_method, summary_FUN,
                                      target_strand, anchor,
                                      fill_value = "MISSING", attrib_var = "name",
                                      force_skip_centerFix = FALSE){
    switch(win_method,
           sample = {
               qgr = prepare_fetch_GRanges_width(qgr, win_size,
                                           skip_centerFix = force_skip_centerFix)
               if(target_strand == "both"){
                   pos_gr = subset(x, strand == "+")
                   neg_gr = subset(x, strand == "-")
                   pos_dt = viewGRangesWinSample_dt(
                       pos_gr,
                       qgr,
                       win_size,
                       attrib_var = attrib_var,
                       fill_value = fill_value,
                       anchor = anchor
                   )
                   neg_dt = viewGRangesWinSample_dt(
                       neg_gr,
                       qgr,
                       win_size,
                       attrib_var = attrib_var,
                       fill_value = fill_value,
                       anchor = anchor
                   )
                   pos_dt[, strand := "+"]
                   neg_dt[, strand := "-"]
                   out = rbind(
                       pos_dt,
                       neg_dt
                   )
               }else{
                   if(target_strand != "*"){
                       x = subset(x, strand == target_strand)
                   }

                   out = viewGRangesWinSample_dt(
                       x,
                       qgr,
                       win_size,
                       attrib_var = attrib_var,
                       fill_value = fill_value,
                       anchor = anchor)
                   out[, strand := target_strand]
               }
               out
           },
           summary = {
               if(target_strand == "both"){
                   pos_gr = subset(x, strand == "+")
                   neg_gr = subset(x, strand == "-")
                   pos_dt = viewGRangesWinSummary_dt(
                       pos_gr,
                       qgr,
                       win_size,
                       attrib_var = attrib_var,
                       fill_value = fill_value,
                       summary_FUN = summary_FUN,
                       anchor = anchor
                   )

                   neg_dt = viewGRangesWinSummary_dt(
                       neg_gr,
                       qgr,
                       win_size,
                       attrib_var = attrib_var,
                       fill_value = fill_value,
                       summary_FUN = summary_FUN,
                       anchor = anchor
                   )
                   pos_dt[, strand := "+"]
                   neg_dt[, strand := "-"]
                   out = rbind(
                       pos_dt,
                       neg_dt
                   )
               }else{
                   if(target_strand != "*"){
                       x = subset(x, strand == target_strand)
                   }
                   out = viewGRangesWinSummary_dt(
                       x,
                       qgr,
                       win_size,
                       attrib_var = attrib_var,
                       fill_value = fill_value,
                       summary_FUN = summary_FUN,
                       anchor = anchor
                   )
                   out[, strand := target_strand]
               }
               out
           })
}

#' easyLoad_FUN takes a character vector of file paths run an arbitrary function defined in load_FUN
#'
#' @param file_paths character vector of paths to narrowPeak files.  If named,
#' those names will be used in output unless overriden by providing file_names.
#' @param load_FUN Arbitrary function that takes at least a file path as argument.  May take other arguments that should be set in call to easyLoad_FUN.
#' @param file_names character vector of names for output list.  If not NULL
#' will override any existing names for file_paths.  Default is NULL.
#' @param n_cores number of cores to use, uses mc.cores option if set or 1.
#' @param ... extra parameters passed to load_FUN
#'
#' @return a named list of results from load_FUN
#' @export
#'
#' @examples
#' bed_f = system.file("extdata/test_loading.bed",
#'     package = "seqsetvis", mustWork = TRUE)
#' easyLoad_bed(bed_f, "my_bed")
easyLoad_FUN = function(file_paths, load_FUN,
                        file_names = NULL,
                        n_cores = getOption("mc.cores", 1),
                        ...){
    if (is.factor(file_paths)) {
        file_paths = as.character(file_paths)
    }
    stopifnot(is.character(file_paths))
    stopifnot(all(file.exists(file_paths) |
                      grepl("://", file_paths)))
    #check names of file-paths
    if (is.null(names(file_paths))) {
        names(file_paths) = basename(file_paths)
    }
    #override names with file_names if not NULL
    if (!is.null(file_names)) {
        if (is.factor(file_names)) {
            file_names = as.character(file_names)
        }
        stopifnot(is.character(file_paths))
        stopifnot(length(file_paths) == length(file_names))
        names(file_paths) = file_names
    }
    if(n_cores == 1){
        grs <- lapply(
            file_paths, load_FUN, ...)
    }else{
        grs = ssv_mclapply(
            file_paths,
            load_FUN,
            mc.cores = n_cores, ...)
    }
    grs
}

#' easyLoad_IDRmerged loads "overlapped-peaks.txt" from IDR.
#'
#' @param file_paths character vector of paths to narrowPeak files.  If named,
#' those names will be used in output unless overriden by providing file_names.
#' @param file_names character vector of names for output list.  If not NULL
#' will override any existing names for file_paths.  Default is NULL.
#' @param n_cores number of cores to use, uses mc.cores option if set or 1.
#' @param max_idr maximum IDR value allowed
#'
#' @return named list of GRanges
#' @export
#'
#' @examples
#' idr_file = system.file("extdata/test_idr.overlapped-peaks.txt",
#'     package = "seqsetvis", mustWork = TRUE)
#' easyLoad_IDRmerged(idr_file)
#' easyLoad_IDRmerged(idr_file, max_idr = .01)
easyLoad_IDRmerged = function(file_paths,
                              file_names = NULL,
                              n_cores = getOption("mc.cores", 1),
                              max_idr = .05){
    #bindings for data.table
    chr1 = start1 = start2 = stop1 = stop2 = sig.value1 = sig.value2 = idr.local = IDR = NULL
    load_idr = function(f, max_idr = .05){
        idr_dt = suppressWarnings({data.table::fread(f)[IDR <= max_idr]})
        GenomicRanges::GRanges(idr_dt[, list(seqnames = chr1, start = pmin(start1, start2), end = pmax(stop1, stop2), sig.value1, sig.value2, idr.local, IDR)])
    }
    easyLoad_FUN(file_paths, load_idr, file_names, n_cores, max_idr = max_idr)
}

#functions that load a list of files as a list of GRanges

#' easyLoad_bed takes a character vector of file paths to bed plus files and
#' returning named list of GRanges.
#'
#' Mainly a utility function for loading MACS2 narrowPeak and broadPeak.
#'
#' @param file_paths character vector of paths to narrowPeak files.  If named,
#' those names will be used in output unless overriden by providing file_names.
#' @param file_names character vector of names for output list.  If not NULL
#' will override any existing names for file_paths.  Default is NULL.
#' @param extraCols named character vector of classes.  passed to
#' rtracklayer::import for format = "BED". default is character().
#' @param n_cores number of cores to use, uses mc.cores option if set or 1.
#' @return a named list of GRanges loaded from file_paths
#' @import rtracklayer
#' @export
#' @examples
#' bed_f = system.file("extdata/test_loading.bed",
#'     package = "seqsetvis", mustWork = TRUE)
#' easyLoad_bed(bed_f, "my_bed")
easyLoad_bed = function(file_paths,
                        file_names = NULL,
                        extraCols = character(),
                        n_cores = getOption("mc.cores", 1)) {
    load_bed = function(f){
        gr = rtracklayer::import(f, format = "BED",
                                 extraCols = extraCols)
        missed = setdiff(c("name", "score", names(extraCols)), colnames(mcols(gr)))
        for(m in missed){
            if(m == "name"){
                if(length(gr) == 0){
                    mcols(gr)[[m]] = character()
                }else{
                    mcols(gr)[[m]] = paste0("DEFAULT_", seq_along(gr))
                }
            }else if(m == "score"){
                mcols(gr)[[m]] = rep(0, length(gr))
            }else{
                default_value = switch(extraCols[m],
                             numeric = {
                                 0L
                             },
                             character = {
                                 "DEFAULT"
                             },
                             integer = {
                                 0
                             })
                mcols(gr)[[m]] = rep(default_value, length(gr))
            }

        }
        gr
    }

    easyLoad_FUN(file_paths,
                 load_bed,
                 file_names = file_names,
                 n_cores = n_cores)
}

#' easyLoad_narrowPeak takes a character vector of file paths to narrowPeak
#' files from MACS2 and returns a named list of GRanges.
#' @param file_paths character vector of paths to narrowPeak files.  If named,
#' those names will be used in output unless overriden by providing file_names.
#' @param file_names character vector of names for output list.  If not NULL
#' will override any existing names for file_paths.  Default is NULL.
#' @param n_cores number of cores to use, uses mc.cores option if set or 1.
#' @return a named list of GRanges loaded from file_paths
#' @import rtracklayer
#' @export
#' @examples
#' np_f = system.file("extdata/test_loading.narrowPeak",
#'     package = "seqsetvis", mustWork = TRUE)
#' easyLoad_narrowPeak(np_f, "my_narrowPeak")
easyLoad_narrowPeak = function(file_paths,
                               file_names = NULL,
                               n_cores = getOption("mc.cores", 1)) {
    #from: https://charlesjb.github.io/How_to_import_narrowPeak/
    extraCols_narrowPeak <-
        c(
            signalValue = "numeric",
            pValue = "numeric",
            qValue = "numeric",
            relSummit = "integer"
        )
    easyLoad_bed(file_paths = file_paths,
                 file_names = file_names,
                 extraCols = extraCols_narrowPeak,
                 n_cores = n_cores)
}

#' easyLoad_broadPeak takes a character vector of file paths to narrowPeak
#' files from MACS2 and returns a named list of GRanges.
#' @param file_paths character vector of paths to narrowPeak files.  If named,
#' those names will be used in output unless overriden by providing file_names.
#' @param file_names character vector of names for output list.  If not NULL
#' will override any existing names for file_paths.  Default is NULL.
#' @param n_cores number of cores to use, uses mc.cores option if set or 1.
#' @return a named list of GRanges loaded from file_paths
#' @import rtracklayer
#' @export
#' @examples
#' bp_f = system.file("extdata/test_loading.broadPeak",
#'     package = "seqsetvis", mustWork = TRUE)
#' easyLoad_broadPeak(bp_f, "my_broadPeak")
easyLoad_broadPeak = function(file_paths,
                              file_names = NULL,
                              n_cores = getOption("mc.cores", 1)) {
    #from: https://charlesjb.github.io/How_to_import_narrowPeak/
    extraCols_broadPeak <-
        c(signalValue = "numeric",
          pValue = "numeric",
          qValue = "numeric")
    easyLoad_bed(file_paths = file_paths,
                 file_names = file_names,
                 extraCols = extraCols_broadPeak,
                 n_cores = n_cores)
}

#' easyLoad_seacr takes a character vector of file paths to seacr output bed files and returns a named list of GRanges.
#' @param file_paths character vector of paths to seacr bed files.  If named,
#' those names will be used in output unless overriden by providing file_names.
#' @param file_names character vector of names for output list.  If not NULL
#' will override any existing names for file_paths.  Default is NULL.
#' @param n_cores number of cores to use, uses mc.cores option if set or 1.
#' @return a named list of GRanges loaded from file_paths
#' @import rtracklayer
#' @export
#' @examples
#' bed_f = system.file("extdata/test_loading.seacr.bed",
#'     package = "seqsetvis", mustWork = TRUE)
#' easyLoad_seacr(bed_f, "my_seacr")
easyLoad_seacr = function(file_paths,
                          file_names = NULL,
                          n_cores = getOption("mc.cores", 1)){
    extraCols_seacr = c(
        total_signal = "numeric",
        max_signal = "numeric",
        max_signal_region = "character"
    )
    easyLoad_bed(file_paths = file_paths,
                 file_names = file_names,
                 extraCols = extraCols_seacr,
                 n_cores = n_cores)
}

testthat::context("EasyLoad")
library(seqsetvis)
library(testthat)
library(GenomicRanges)


fail_tests = function(met_nam){
    met = get(met_nam)

    test_that(paste(met_nam, "error on non-character inputs"), {
        exp = "is.character\\(file_paths\\) is not TRUE"
        expect_error(met(1), exp)
        expect_error(met(1L), exp)
        expect_error(met(mean), exp)
        expect_error(met(list()), exp)
    })

    test_that(paste(met_nam, "error on non-existant file"), {
        exp = "all\\(file.exists.+is not TRUE"
        bad_file = "not a file.anywhere.ever.beddybye"
        expect_error(met(bad_file), exp)
        expect_error(met(bad_file), exp)
        expect_error(met(bad_file), exp)
        expect_error(met(bad_file), exp)
    })
}

for(met_nam in c("easyLoad_bed",
                 "easyLoad_broadPeak",
                 "easyLoad_narrowPeak")){
    fail_tests(met_nam)
}

test_dir = system.file("extdata/", package = "seqsetvis")
test_files = c("test_loading.narrowPeak", "test_loading.broadPeak", "test_loading.bed")
test_files = paste0(test_dir, "/", test_files)
stopifnot(all(file.exists(test_files)))

test_that("easyLoad_narrowPeak works, 1 file", {
    tgr = easyLoad_narrowPeak(test_files[1])
    expect_is(tgr, "list")
    expect_is(tgr[[1]], "GRanges")
    expect_equal(length(tgr), 1)
    expect_equal(names(tgr), "test_loading.narrowPeak")
    expect_equal(ncol(mcols(tgr[[1]])), 6)
})

test_that("easyLoad_narrowPeak works, 3 files and names", {
    tgr = easyLoad_narrowPeak(rep(test_files[1],3), file_names = LETTERS[1:3])
    expect_is(tgr, "list")
    expect_is(tgr[[1]], "GRanges")
    expect_equal(length(tgr), 3)
    expect_equal(names(tgr), LETTERS[1:3])
    expect_equal(ncol(mcols(tgr[[1]])), 6)
})

test_that("easyLoad_narrowPeak works, 1 file", {
    tgr = easyLoad_broadPeak(test_files[2])
    expect_is(tgr, "list")
    expect_is(tgr[[1]], "GRanges")
    expect_equal(length(tgr), 1)
    expect_equal(names(tgr), "test_loading.broadPeak")
    expect_equal(ncol(mcols(tgr[[1]])), 5)
})

test_that("easyLoad_narrowPeak works, 3 files and names", {
    tgr = easyLoad_broadPeak(rep(test_files[2],3), file_names = LETTERS[1:3])
    expect_is(tgr, "list")
    expect_is(tgr[[1]], "GRanges")
    expect_equal(length(tgr), 3)
    expect_equal(names(tgr), LETTERS[1:3])
    expect_equal(ncol(mcols(tgr[[1]])), 5)
})

test_that("easyLoad_bed works, 1 file", {
    tgr = easyLoad_bed(test_files[3])
    expect_is(tgr, "list")
    expect_is(tgr[[1]], "GRanges")
    expect_equal(length(tgr), 1)
    expect_equal(names(tgr), "test_loading.bed")
    expect_equal(ncol(mcols(tgr[[1]])), 2)
})

test_that("easyLoad_bed works, 3 files and names", {
    tgr = easyLoad_bed(rep(test_files[3],3), file_names = LETTERS[1:3])
    expect_is(tgr, "list")
    expect_is(tgr[[1]], "GRanges")
    expect_equal(length(tgr), 3)
    expect_equal(names(tgr), LETTERS[1:3])
    expect_equal(ncol(mcols(tgr[[1]])), 2)
})

test_that("easyLoad_bed empty file", {
    empty_bed_file = dir(test_dir, pattern = "empty.bed", full.names = TRUE)
    bed_empty = easyLoad_bed(empty_bed_file)[[1]]
    bed_normal = easyLoad_bed(test_files[3])[[1]]
    expect_is(bed_empty, "GRanges")
    expect_equal(length(bed_empty), 0)
    expect_equal(
        colnames(GenomicRanges::mcols(bed_empty)),
        colnames(GenomicRanges::mcols(bed_normal))
    )

})

test_that("easyLoad_bed empty file", {
    empty_np_file = dir(test_dir, pattern = "empty.narrowPeak", full.names = TRUE)
    np_empty = easyLoad_narrowPeak(empty_np_file)[[1]]
    np_normal = easyLoad_narrowPeak(test_files[1])[[1]]
    expect_is(np_empty, "GRanges")
    expect_equal(length(np_empty), 0)
    expect_equal(
        colnames(GenomicRanges::mcols(np_empty)),
        colnames(GenomicRanges::mcols(np_normal))
    )
})

#' Iterates a character vector (ideally named) and calls
#' \code{ssvFetchBigwig.single} on each.  Appends grouping variable to each
#' resulting data.table and uses rbindlist to efficiently combine results.
#'
#' \code{ssvFetchBigwig} iteratively calls \code{fetchWindowedBigwig.single}.
#' See \code{\link{ssvFetchBigwig.single}} for more info.
#' @export
#' @param file_paths character vector of file_paths to load from. Alternatively,
#' file_paths can be a data.frame or data.table whose first column is a
#' character vector of paths and additial columns will be used as metadata.
#' @param qgr Set of GRanges to query.  For valid results the width of each
#' interval should be identical and evenly divisible by \code{win_size}.
#' @param unique_names names to use in final data.table to designate source
#' bigwig.
#' @param names_variable The column name where unique_names are stored.
#' Default is 'sample'
#' @param win_size The window size that evenly divides widths in \code{qgr}.
#' @param win_method character.  one of c("sample", "summary").  Determines
#' if \code{\link{viewGRangesWinSample_dt}} or
#' \code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
#' qgr.
#' @param summary_FUN function.  only relevant if win_method is "summary".
#' passed to \code{\link{viewGRangesWinSummary_dt}}.
#' @param fragLens never used by ssvFetchBigwig. Ignore.
#' @param anchor character, one of c("center", "center_unstranded",
#' "left", "left_unstranded")
#' @param return_data.table logical. If TRUE the internal data.table is
#' returned instead of GRanges.  Default is FALSE.
#' @param n_cores integer number of cores to use.
#' Uses mc.cores option if not supplied.
#' @param n_region_splits integer number of splits to apply to qgr. The query
#'   GRanges will be split into this many roughly equal parts for increased
#'   parallelization. Default is 1, no split.
#' @param force_skip_centerFix boolean, if TRUE all query ranges will be
#' used "as is".  This is already the case by default if win_method == "summary"
#' but may have applications where win_method == "sample".
#' @return A tidy formatted GRanges (or data.table if specified) containing
#' fetched values.
#' @rawNamespace import(data.table, except = c(shift, first, second, last))
#' @details if \code{qgr} contains the range chr1:1-100 and \code{win_size} is
#' 10, values from positions chr1 5,15,25...85, and 95 will be
#' retrieved from \code{bw_file}
#' @examples
#' if(Sys.info()['sysname'] != "Windows"){
#' library(GenomicRanges)
#' bw_f = system.file("extdata/test_loading.bw",
#'     package = "seqsetvis", mustWork = TRUE)
#' bw_files = c("a" = bw_f, "b" = bw_f)
#' qgr = GRanges("chrTest", IRanges(1, 30))
#' bw_gr = ssvFetchBigwig(bw_files, qgr, win_size = 10)
#' bw_gr2 = ssvFetchBigwig(as.list(bw_files), qgr, win_size = 10)
#'
#' bw_dt = ssvFetchBigwig(bw_files, qgr, win_size = 10,
#'     return_data.table = TRUE)
#' }
ssvFetchBigwig = function(file_paths,
                          qgr,
                          unique_names = NULL,
                          names_variable = "sample",
                          win_size = 50,
                          win_method = c("sample", "summary")[1],
                          summary_FUN = stats::weighted.mean,
                          fragLens = "not_used",
                          anchor = c("left", "left_unstranded", "center",
                                     "center_unstranded")[3],
                          return_data.table = FALSE,
                          n_cores = getOption("mc.cores", 1),
                          n_region_splits = 1,
                          force_skip_centerFix = FALSE) {

    load_bw = function(f, nam, qgr) {
        message("loading ", f, " ...")
        dt = ssvFetchBigwig.single(bw_file = f,
                                   qgr = qgr,
                                   win_size = win_size,
                                   win_method = win_method,
                                   summary_FUN = summary_FUN,
                                   anchor = anchor,
                                   return_data.table = TRUE,
                                   force_skip_centerFix = force_skip_centerFix)
        # dt[[names_variable]] = nam
        message("finished loading ", nam, ".")
        dt
    }

    ssvFetchSignal(file_paths = file_paths,
                   qgr = qgr,
                   unique_names = unique_names,
                   names_variable = names_variable,
                   win_size = win_size,
                   win_method = win_method,
                   return_data.table = return_data.table,
                   load_signal = load_bw,
                   n_cores = n_cores,
                   n_region_splits = n_region_splits,
                   force_skip_centerFix = force_skip_centerFix)
}

#' Fetch values from a bigwig appropriate for heatmaps etc.
#'
#' \code{ssvFetchBigwig.single} Gets values for each region of the query
#' GRanges (\code{qgr}). Values correspond to the center of each window of
#' size \code{win_size}.  A tidy formatted data.table
#' object is returned suitable for plotting using ggplots.
#' @param bw_file The character vector path to bigwig files to read from.
#' @param qgr Set of GRanges to query.  For valid results the width of each
#' interval should be identical and evenly divisible by \code{win_size}.
#' @param win_size The window size that evenly divides widths in \code{qgr}.
#' @param win_method character.  one of c("sample", "summary").  Determines
#' if \code{\link{viewGRangesWinSample_dt}} or
#' \code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
#' qgr.
#' @param summary_FUN function.  only relevant if win_method is "summary".
#' passed to \code{\link{viewGRangesWinSummary_dt}}.
#' @param anchor character, one of c("center", "center_unstranded",
#' "left", "left_unstranded")
#' @param return_data.table logical. If TRUE the internal data.table is
#' returned instead of GRanges.  Default is FALSE.
#' @param force_skip_centerFix boolean, if TRUE all query ranges will be
#' used "as is".  This is already the case by default if win_method == "summary"
#' but may have applications where win_method == "sample".
#' @return A GRanges (or data.table if specified) containing fetched values.
#' @rawNamespace import(data.table, except = c(shift, first, second, last))
#' @details if \code{qgr} contains the range chr1:1-100 and \code{win_size} is
#' 10, values from positions chr1 5,15,25...85, and 95 will be retrieved
#' from \code{bw_file}
ssvFetchBigwig.single = function(bw_file,
                                 qgr,
                                 win_size = 50,
                                 win_method = c("sample", "summary")[1],
                                 summary_FUN = stats::weighted.mean,
                                 anchor = c("left", "left_unstranded", "center",
                                            "center_unstranded")[3],
                                 return_data.table = FALSE,
                                 force_skip_centerFix = FALSE) {
    stopifnot(is.character(bw_file))
    stopifnot(is(qgr, "GRanges"))
    stopifnot(is.numeric(win_size))
    switch (win_method,
            sample = {
                qgr = prepare_fetch_GRanges_width(qgr, win_size, skip_centerFix = force_skip_centerFix)
                fetch_gr = qgr
                names(fetch_gr) = NULL
                score_gr = rtracklayer::import.bw(bw_file, which = fetch_gr)
                out = viewGRangesWinSample_dt(score_gr, qgr, win_size, anchor = anchor)[]
            },
            summary = {
                fetch_gr = qgr
                names(fetch_gr) = NULL
                score_gr = rtracklayer::import.bw(bw_file, which = fetch_gr)
                out = viewGRangesWinSummary_dt(score_gr, qgr, win_size,
                                               summary_FUN = summary_FUN,
                                               anchor = anchor)[]
            }
    )

    if(!return_data.table){
        out = GRanges(out)
    }
    return(out)
}

testthat::context("FetchBw_dt")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
# library(rtracklayer)
test_bw = system.file("extdata/test_loading.bw", package = "seqsetvis", mustWork = TRUE)
pos = c(20, 180, 210, 440, 520, 521)
region_size = 30
test_qgr = GRanges("chrTest", IRanges(pos+1, pos + region_size))
exp_colnames = c("seqnames", "start", "end", "width", "strand", "id", "y", "x")

ssvFetchBigwig.single = seqsetvis:::ssvFetchBigwig.single

test_that("ssvFetchBigwig.single return expected for valid even win_size", {
    skip_on_os("windows")
    #these should all work cleanly
    for(win in c(2, 6, 10, 30)){
        bw_dt = ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = test_qgr, return_data.table = TRUE)
        expect_is(bw_dt, "data.table")
        expect_equal(colnames(bw_dt), exp_colnames)
        for(tid in unique(bw_dt$id)){
            test_dt = bw_dt[id == tid]
            expect_equal(nrow(test_dt), region_size / win) #expected number of regions
            bw_gr = GRanges(test_dt)
            expect_equal(length(findOverlaps(bw_gr, bw_gr)), region_size / win, info = "each returned GRange should only intersect itself.")
        }
    }
})

test_that("ssvFetchBigwig.single return expected for valid odd win_size", {
    skip_on_os("windows")
    for(win in c(1, 3, 5, 15)){
        bw_dt = ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = test_qgr, return_data.table = TRUE)
        expect_is(bw_dt, "data.table")
        expect_equal(colnames(bw_dt), exp_colnames)
        for(tid in unique(bw_dt$id)){
            test_dt = bw_dt[id == tid]
            expect_equal(nrow(test_dt), region_size / win) #expected number of regions
            bw_gr = GRanges(test_dt)
            expect_equal(length(findOverlaps(bw_gr, bw_gr)), region_size / win, info = "each returned GRange should only intersect itself.")
        }
    }
})

test_that("ssvFetchBigwig.single use GRanges names as id", {
    skip_on_os("windows")
    for(win in c(1, 3, 5, 15)){
        qgr = test_qgr
        names(qgr) = paste0("myNames_", seq_along(qgr))
        bw_dt = ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = qgr, return_data.table = TRUE)
        expect_is(bw_dt, "data.table")
        expect_equal(colnames(bw_dt), exp_colnames)
        expect_true(all(grepl("myNames", bw_dt$id)))
        for(tid in unique(bw_dt$id)){
            test_dt = bw_dt[id == tid]
            expect_equal(nrow(test_dt), region_size / win) #expected number of regions
            bw_gr = GRanges(test_dt)
            expect_equal(length(findOverlaps(bw_gr, bw_gr)), region_size / win, info = "each returned GRange should only intersect itself.")
        }
    }
})

test_that("ssvFetchBigwig.single patches missing values", {
    skip_on_os("windows")
    for(win in c(1, 5, 20)){
        qgr = GRanges("chrTest", IRanges(1, 2000))
        bw_dt = ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = qgr, return_data.table = TRUE)
        expect_is(bw_dt, "data.table")
        expect_equal(colnames(bw_dt), exp_colnames)
    }
})

test_that("ssvFetchBigwig.single throws message if widths aren't divisble by win_size", {
    skip_on_os("windows")
    mix_width_gr = test_qgr
    end(mix_width_gr) =  end(mix_width_gr) + seq_along(mix_width_gr)
    for(win in c(7, 31)){
        expect_message(regexp = "widths of qgr were not identical and evenly divisible by win_size", {
            ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = mix_width_gr, return_data.table = TRUE)
        })
    }
})



test_that("ssvFetchBigwig.single throws message if widths vary", {
    skip_on_os("windows")
    mix_width_gr = test_qgr
    end(mix_width_gr) =  end(mix_width_gr) + seq_along(mix_width_gr)*3
    for(win in c(1, 3)){
        expect_message(regexp = "widths of qgr were not identical and evenly divisible by win_size", {
            ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = mix_width_gr, return_data.table = TRUE)
        })
    }
})

test_that("ssvFetchBigwig works with character vector bw_files", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    names(bw_files) = paste0("bw_", 1:3)
    hidden = capture_output({res = ssvFetchBigwig(file_paths = bw_files,
                                                           win_size = 3,
                                                           qgr = test_qgr, return_data.table = TRUE)})
    expect_s3_class(res, "data.table")
    expect_equal(colnames(res), c(exp_colnames, "sample"))
})

test_that("ssvFetchBigwig works with list bw_files", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    names(bw_files) = paste0("bw_", 1:3)
    bw_files = as.list(bw_files)
    hidden = capture_output({res = ssvFetchBigwig(file_paths = bw_files,
                                                           win_size = 3,
                                                           qgr = test_qgr, return_data.table = TRUE)})
    expect_s3_class(res, "data.table")
    expect_equal(colnames(res), c(exp_colnames, "sample"))
})

test_that("ssvFetchBigwig can set variable name", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    names(bw_files) = paste0("bw_", 1:3)
    hidden = capture_output({res = ssvFetchBigwig(file_paths = bw_files,
                                                           win_size = 3,
                                                           qgr = test_qgr,
                                                           names_variable = "group", return_data.table = TRUE)})
    expect_s3_class(res, "data.table")
    expect_equal(colnames(res), c(exp_colnames, "group"))
})

test_that("ssvFetchBigwig duplicate names throws error", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    expect_error(
        ssvFetchBigwig(file_paths = bw_files,
                                win_size = 3,
                                qgr = test_qgr, return_data.table = TRUE)
    )
})

testthat::context("FetchBw")
#tests seqsetvis::ssvFetchBigwig.single and seqsetvis::ssvFetchBigwig
#originally data.table version of each which may explain some strangeness
library(seqsetvis)
library(testthat)
library(GenomicRanges)
# library(rtracklayer)
test_bw = system.file("extdata/test_loading.bw", package = "seqsetvis", mustWork = TRUE)
pos = c(20, 180, 210, 440, 520, 521)
region_size = 30
test_qgr = GRanges("chrTest", IRanges(pos+1, pos + region_size))
exp_colnames = c("seqnames", "start", "end", "width", "strand", "id", "y", "x")[6:8]

ssvFetchBigwig.single = seqsetvis:::ssvFetchBigwig.single

test_that("ssvFetchBigwig.single return expected for valid even win_size", {
    skip_on_os("windows")
    #these should all work cleanly
    for(win in c(2, 6, 10, 30)){
        bw_gr = ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = test_qgr)
        expect_is(bw_gr, "GRanges")
        expect_equal(colnames(mcols(bw_gr)), exp_colnames)
        for(tid in unique(bw_gr$id)){
            test_gr = bw_gr[bw_gr$id == tid]
            expect_equal(length(test_gr), region_size / win) #expected number of regions
            expect_equal(length(findOverlaps(test_gr, test_gr)), region_size / win, info = "each returned GRange should only intersect itself.")
        }
    }
})

test_that("ssvFetchBigwig.single return expected for valid odd win_size", {
    skip_on_os("windows")
    for(win in c(1, 3, 5, 15)){
        bw_gr = ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = test_qgr)
        expect_is(bw_gr, "GRanges")
        expect_equal(colnames(mcols(bw_gr)), exp_colnames)
        for(tid in unique(bw_gr$id)){
            test_gr = bw_gr[bw_gr$id == tid]
            expect_equal(length(test_gr), region_size / win) #expected number of regions
            expect_equal(length(findOverlaps(test_gr, test_gr)), region_size / win, info = "each returned GRange should only intersect itself.")
        }
    }
})

test_that("ssvFetchBigwig.single use GRanges names as id", {
    skip_on_os("windows")
    for(win in c(1, 3, 5, 15)){
        qgr = test_qgr
        names(qgr) = paste0("myNames_", seq_along(qgr))
        bw_gr = ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = qgr)
        expect_is(bw_gr, "GRanges")
        expect_equal(colnames(mcols(bw_gr)), exp_colnames)
        expect_true(all(grepl("myNames", bw_gr$id)))
        for(tid in unique(bw_gr$id)){
            test_gr = bw_gr[bw_gr$id == tid]
            expect_equal(length(test_gr), region_size / win) #expected number of regions
            expect_equal(length(findOverlaps(test_gr, test_gr)), region_size / win, info = "each returned GRange should only intersect itself.")
        }
    }
})

test_that("ssvFetchBigwig.single patches missing values", {
    skip_on_os("windows")
    for(win in c(1, 5, 20)){
        qgr = GRanges("chrTest", IRanges(1, 2000))
        bw_gr = ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = qgr)
        expect_is(bw_gr, "GRanges")
        expect_equal(colnames(mcols(bw_gr)), exp_colnames)
    }
})

test_that("ssvFetchBigwig.single throws message if widths aren't divisble by win_size", {
    skip_on_os("windows")
    mix_width_gr = test_qgr
    end(mix_width_gr) =  end(mix_width_gr) + seq_along(mix_width_gr)
    for(win in c(7, 31)){
        expect_message(regexp = "widths of qgr were not identical and evenly divisible by win_size", {
            ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = mix_width_gr)
        })
    }
})



test_that("ssvFetchBigwig.single throws warning if widths vary", {
    skip_on_os("windows")
    mix_width_gr = test_qgr
    end(mix_width_gr) =  end(mix_width_gr) + seq_along(mix_width_gr)*3
    for(win in c(1, 3)){
        expect_message(regexp = "widths of qgr were not identical and evenly divisible by win_size", {
            ssvFetchBigwig.single(bw_file = test_bw, win_size = win, qgr = mix_width_gr)
        })
    }
})

test_that("ssvFetchBigwig works with character vector bw_files", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    names(bw_files) = paste0("bw_", 1:3)
    hidden = capture_output({res = ssvFetchBigwig(file_paths = bw_files,
                                                  win_size = 3,
                                                  qgr = test_qgr)})
    expect_is(res, "GRanges")
    expect_equal(colnames(mcols(res)), c(exp_colnames, "sample"))
})

test_that("ssvFetchBigwig works with list bw_files", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    names(bw_files) = paste0("bw_", 1:3)
    bw_files = as.list(bw_files)
    hidden = capture_output({res = ssvFetchBigwig(file_paths = bw_files,
                                                  win_size = 3,
                                                  qgr = test_qgr)})
    expect_is(res, "GRanges")
    expect_equal(colnames(mcols(res)), c(exp_colnames, "sample"))
})

test_that("ssvFetchBigwig query GRanges output id set", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    names(bw_files) = paste0("bw_", 1:3)
    test_gr = test_qgr
    gr_sample = ssvFetchBigwig(bw_files, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true("id" %in% colnames(gr_sample))

    names(test_gr) = NULL
    gr_sample = ssvFetchBigwig(bw_files, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true("id" %in% colnames(gr_sample))

    test_gr$id = seq_along(test_gr) #when id set, id in output is missing
    gr_sample = ssvFetchBigwig(bw_files, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true("id" %in% colnames(gr_sample))

    test_gr = test_qgr
    test_gr$id = seq_along(test_gr) #when id set, id in output is missing
    gr_sample = ssvFetchBigwig(bw_files, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true("id" %in% colnames(gr_sample))

})

test_that("ssvFetchBam query GRanges $name gets used", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    names(bw_files) = paste0("bw_", 1:3)
    test_gr = test_qgr
    test_gr$name = paste("peak", seq_along(test_gr))
    gr_sample = ssvFetchBigwig(bw_files, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true(all(unique(gr_sample$id) == unique(test_gr$name)))
})

test_that("ssvFetchBigwig can set variable name", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    names(bw_files) = paste0("bw_", 1:3)
    hidden = capture_output({res = ssvFetchBigwig(file_paths = bw_files,
                                                  win_size = 3,
                                                  qgr = test_qgr,
                                                  names_variable = "group")})
    expect_is(res, "GRanges")
    expect_equal(colnames(mcols(res)), c(exp_colnames, "group"))
})

test_that("ssvFetchBigwig duplicate names throws error", {
    skip_on_os("windows")
    bw_files = rep(test_bw, 3)
    expect_error(
        ssvFetchBigwig(file_paths = bw_files,
                       win_size = 3,
                       qgr = test_qgr)
    )
})

test_that("ssvFetchBigwig sample method correct bins", {
    skip_on_os("windows")
    test_qgr2 = test_qgr[c(1,2,5)]
    gr_sample = ssvFetchBigwig(test_bw, win_size = 3, qgr = test_qgr2)
    #for non-overlapping, all reduce granges width should be equal to input qgr
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    #for non-overlapping, all granges width should be equal to win_size
    expect_true(all(width(gr_sample) == 3))

    #same as above but width are auto adjusted to be divisible by win_size
    gr_sample = ssvFetchBigwig(test_bw, win_size = 4, qgr = test_qgr2)
    expect_true(all(width(reduce(gr_sample)) == 32))
    expect_true(all(width(gr_sample) == 4))
})

test_that("ssvFetchBigwig summary method correct bins", {
    skip_on_os("windows")
    test_qgr2 = test_qgr[c(1,2,5)]
    gr_sample = ssvFetchBigwig(test_bw, win_size = 3, qgr = test_qgr2, win_method = "summary")
    #for non-overlapping, all reduce granges width should be equal to input qgr
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    #for non-overlapping, all granges width should be equal to win_size
    expect_true(all(width(gr_sample) == 30 / 3))

    #same as above but width are auto adjusted to be divisible by win_size
    gr_sample = ssvFetchBigwig(test_bw, win_size = 4, qgr = test_qgr2, win_method = "summary")
    expect_true(all(width(reduce(gr_sample)) == 30))
    expect_true(all(width(gr_sample) %in% 7:8))
})

test_that("ssvFetchBigwig test_bw as data.frame/table", {
    skip_on_os("windows")
    gr_sample = ssvFetchBigwig(
        data.frame(rep(test_bw,3),
                   colors = c("red", "green", "blue"),
                   sample = c("10a", "10b", "10c")),
        qgr = test_qgr,
        win_size = 5,
        return_data.table = TRUE)
    expect_true(all(levels(gr_sample$colors) %in%
                        c("red", "green", "blue")))
    expect_true(all(levels(gr_sample$sample) %in%
                        c("10a", "10b", "10c")))
})

test_that("ssvFetchBigwig single column data.table", {
    skip_on_os("windows")
    qdt = data.table(file = test_bw)
    res = ssvFetchBigwig(qdt,
                      qgr = test_qgr[1],
                      win_size = 5,
                      return_data.table = TRUE,
                      win_method = "summary")
    expect_equal(levels(res$sample)[1], test_bw)
})

# providng names as a factor is causing mismatch between names and output
testthat::context("bugfix factor names")
library(testthat)
library(seqsetvis)
library(GenomicRanges)
# library(rtracklayer)
test_in = system.file("extdata", package = "seqsetvis", mustWork = TRUE)
bw_files = dir(test_in, pattern = "FE.+.bw$", full.names = TRUE)
test_df = data.frame(file = bw_files, sample = sub("_.+", "", basename(bw_files)))

data("CTCF_in_10a_overlaps_gr")
test_qgr = CTCF_in_10a_overlaps_gr

test_df.char = test_df
prof_dt.char = ssvFetchBigwig(test_df.char, test_qgr, return_data.table = TRUE)
prof_dt.char = prof_dt.char[, .(y = mean(y)), .(x, sample)]

test_df.fact = test_df
test_df.fact$sample = factor(test_df.fact$sample, levels = rev(test_df.fact$sample))

prof_dt.fact = ssvFetchBigwig(test_df.fact, test_qgr, return_data.table = TRUE)

prof_dt.fact = prof_dt.fact[, .(y = mean(y)), .(x, sample)]

prof_dt.char = prof_dt.char[order(x)][order(as.character(sample))]
prof_dt.fact = prof_dt.fact[order(x)][order(as.character(sample))]

ggplot() +
    geom_path(data = prof_dt.char, aes(x = x, y = y, group = sample), color = "blue") +
    geom_path(data = prof_dt.fact, aes(x = x, y = y+2, group = sample), color = "orange") +
    facet_wrap(~sample)

test_that("sample should be a factor in both profiles", {
    #for character sample query, order of levels in output should match input order
    expect_is(test_df.char$sample, "character")
    expect_is(prof_dt.char$sample, "factor")
    expect_equal(test_df.char$sample , levels(prof_dt.char$sample))

    #
    expect_is(test_df.fact$sample, "factor")
    expect_is(prof_dt.fact$sample, "factor")
    expect_equal(levels(test_df.fact$sample) , levels(prof_dt.fact$sample))

})

test_that("fragLen_calcStranded parameters matter", {
    expect_equal(prof_dt.char$y , prof_dt.fact$y)
})

#' Intersect a list of GRanges to create a single GRanges object of
#' merged ranges including metadata describing overlaps per input GRanges
#'
#' @export
#' @param grs A list of GRanges
#' @param ext An integer specifying how far to extend ranges before merging.
#' in effect, ranges withing 2*ext of one another will be joined during the
#' merge
#' @param use_first A logical.  If True, instead of merging all grs, only use
#' first and add metadata logicals for others.
#' @param preserve_mcols Controls carrying forward mcols metadata from input
#'   list of GRanges. If TRUE, all mcols will be carried forward with the item
#'   name appended. If a character vector, only those attributes will be carried
#'   and all must be present in all GRanges. The default of FALSE will carry
#'   nothing forward and only membership table will be generated.
#'   [ssvAnnotateSubjectGRanges] is used internally.
#' @param ... arguments passed to IRanges::findOverlaps, i.e. maxgap, minoverlap, type, select, invert.
#' @return GRanges with metadata columns describing overlap of input grs.
#' @examples
#' library(GenomicRanges)
#' a = GRanges("chr1", IRanges(1:7*10, 1:7*10))
#' b = GRanges("chr1", IRanges(5:10*10, 5:10*10))
#' ssvOverlapIntervalSets(list(a, b))
#' @import GenomicRanges
ssvOverlapIntervalSets = function(grs, ext = 0, use_first = FALSE, preserve_mcols = FALSE, ...){
  queryHits = NULL
  if(is(grs, "GRangesList")){
    grs = as.list(grs)
  }
  if(!is.list(grs)){
    stop("grs must be a List or GRangesList")
  }
  if(is.null(names(grs))){
    names(grs) = paste0("set_", LETTERS[seq_along(grs)])
  }
  if(use_first){
    base_gr = grs[[1]]
    mcols(base_gr) = NULL
  }else{
    grs_nometa = lapply(grs, function(x){mcols(x) = NULL; x})
    base_gr = reduce(unlist(GRangesList(grs_nometa)))
    start(base_gr) = start(base_gr) - ext
    end(base_gr) = end(base_gr) + ext
    base_gr = reduce(base_gr)
    start(base_gr) = start(base_gr) + ext
    end(base_gr) = end(base_gr) - ext
  }
  suppressWarnings({
    for(i in seq_along(grs)){
      nam = names(grs)[i]
      mcols(base_gr)[[nam]] = FALSE
      olaps = findOverlaps(base_gr, grs[[i]], ...)
      mcols(base_gr)[[nam]][queryHits(olaps)] = TRUE
    }
    if(use_first){
      base_gr$no_hit = apply(mcols(base_gr), 1, function(x)all(!x))
    }
  })
  names(base_gr) = seq_along(base_gr)
  if(preserve_mcols){
      base_gr = ssvAnnotateSubjectGRanges(grs, base_gr)
  }
  return(base_gr)
}

#' Intersect a list of GRanges to create a single GRanges object of merged
#' ranges including metadata describing overlaps per input GRanges.
#'
#' In constrast to ssvOverlapIntervalSets, only regions where a consensus of
#' input grs are present are preserved and annotated.
#'
#' @param grs A list of GRanges
#' @param ext An integer specifying how far to extend ranges before merging. in
#'   effect, ranges withing 2*ext of one another will be joined during the merge
#' @param min_number An integer number specifying the absloute minimum of input
#'   grs that must overlap for a site to be considered consensus.
#' @param min_fraction A numeric between 0 and 1 specifying the fraction of grs
#'   that must overlap to be considered consensus.
#' @param preserve_mcols Controls carrying forward mcols metadata from input
#'   list of GRanges. If TRUE, all mcols will be carried forward with the item
#'   name appended. If a character vector, only those attributes will be carried
#'   and all must be present in all GRanges. The default of FALSE will carry
#'   nothing forward and only membership table will be generated.
#'   [ssvAnnotateSubjectGRanges] is used internally.
#' @param ... arguments passed to IRanges::findOverlaps, i.e. maxgap, minoverlap, type, select, invert.
#' @details Only the most stringent of min_number or min_fraction will be
#'   applied.
#' @return GRanges with metadata columns describing consensus overlap of input
#'   grs.
#' @importFrom Seqinfo seqlengths
#' @export
#'
#' @examples
#' library(GenomicRanges)
#' a = GRanges("chr1", IRanges(1:7*10, 1:7*10))
#' b = GRanges("chr1", IRanges(5:10*10, 5:10*10))
#' ssvConsensusIntervalSets(list(a, b))
ssvConsensusIntervalSets = function(grs, ext = 0, min_number = 2, min_fraction = .5, preserve_mcols = FALSE, ...){
  if(is(grs, "GRangesList")){
    grs = as.list(grs)
  }
  if(!is.list(grs)){
    stop("grs must be a List or GRangesList")
  }
  if(is.null(names(grs))){
    names(grs) = paste0("set_", LETTERS[seq_along(grs)])
  }
  stopifnot(min_fraction >= 0 & min_fraction <= 1)
  stopifnot(min_number >= 0 & min_number <= length(grs))
  min_by_fraction = ceiling(length(grs) * min_fraction)
  min_consensus = 0
  if(min_by_fraction > min_number){
    message("based on min_fraction of ", min_fraction, " consensus threshold is ", min_by_fraction, ".")
    min_consensus = min_by_fraction
  }else{
    message("min_number of ", min_number, " used for consensus threshold.")
    min_consensus = min_number
  }

  if(min_consensus > length(grs)){
    stop("Consensus threshold of ", min_consensus, " is too high for length of input grs, ", length(grs))
  }

  grs_ext = lapply(grs, function(gr){
    resize(gr, width(gr)+2*ext, fix = "center")
  })

  grs_cov = GRanges(coverage(unlist(GRangesList(grs_ext))) >= max(min_consensus, 1))
  grs_cov = subset(grs_cov, score == TRUE)
  grs_cov$score = NULL
  for(nam in names(grs)){
    gr = grs[[nam]]
    olaps = findOverlaps(grs_cov, gr, ...)
    mcols(grs_cov)[[nam]] = FALSE
    mcols(grs_cov)[[nam]][queryHits(olaps)] = TRUE
  }
  Seqinfo::seqlengths(grs_cov) = NA
  if(preserve_mcols){
      grs_cov = ssvAnnotateSubjectGRanges(annotation_source = grs, subject_gr = grs_cov)
  }
  grs_cov
}

testthat::context("FeaturePlots")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
a = GRanges("chr1", IRanges(1:7*10, 1:7*10+1))
b = GRanges("chr1", IRanges(5:10*10, 5:10*10+1))
c = GRanges("chr1", IRanges(8:10*10+5, 8:10*10+6))

#test all plotting functions
method_names = c("ssvFeatureVenn", "ssvFeatureEuler", "ssvFeatureBars", "ssvFeaturePie", "ssvFeatureBinaryHeatmap", "ssvFeatureUpset")
for(met_name in method_names){
  if(!exists(met_name)){
    warning("function", met_name, "couldn't be found! not tested.")
    next
  }
  met = get(met_name)
  if(!is.function(met)){
    warning("function", met_name, "wasn't a function! not tested.")
    next
  }


  test_that(paste(met_name, "accepts ssvOverlapIntervalSets output."), {

    olap = ssvOverlapIntervalSets(list("a" = a, "b" = b, "c" = c))
    p = met(olap)
    p
    expect_s3_class(p, class = "ggplot")
  })

  test_that(paste(met_name, "accepts various membership table types."), {
    olap = ssvOverlapIntervalSets(list("a" = a, "b" = b, "c" = c))
    DF = mcols(olap) #DataFrame
    p = met(DF)
    p
    expect_s3_class(p, class = "ggplot")
    df = as.data.frame(DF) #data.frame
    p = met(df)
    p
    expect_s3_class(p, class = "ggplot")
    mat = as.matrix(df) #matrix
    p = met(mat)
    p
    expect_s3_class(p, class = "ggplot")
  })

  test_that(paste(met_name, "accepts various membership table types. missing names."), {
    olap = ssvOverlapIntervalSets(list("a" = a, "b" = b, "c" = c))
    DF = mcols(olap) #DataFrame

    df = as.data.frame(DF) #data.frame
    colnames(df) = NULL
    p = met(df)
    p
    expect_s3_class(p, class = "ggplot")
    mat = as.matrix(df) #matrix
    p = met(mat)
    p
    expect_s3_class(p, class = "ggplot")
  })

  test_that(paste(met_name, "accepts set lists. names present and missing"), {
    olap = ssvOverlapIntervalSets(list("a" = a, "b" = b, "c" = c))
    DF = mcols(olap) #DataFrame
    df = as.data.frame(DF) #data.frame
    rownames(df) = 1:nrow(df)
    sets = apply(df, 2, function(x)rownames(df)[x])
    p = met(sets)
    p
    expect_s3_class(p, class = "ggplot")
    names(sets) = NULL #missing names
    p = met(sets)
    p
    expect_s3_class(p, class = "ggplot")
  })

}

testthat::context("ssvAnnotateSubjectGRanges")
library(seqsetvis)
library(GenomicRanges)
library(testthat)

data(CTCF_in_10a_narrowPeak_grs)
np_grs = CTCF_in_10a_narrowPeak_grs
olap_gr = ssvOverlapIntervalSets(np_grs)

# get_exp = function(x){
#     message('c(\n"', paste(x, collapse = '", \n"'), '"\n)'    )
# }
# get_exp(colnames(mcols(test_gr)))

test_that("ssvAnnotateSubjectGRanges - single GR no names", {
    test_gr = ssvAnnotateSubjectGRanges(np_grs$MCF10A_CTCF, olap_gr)
    expect_equal(colnames(mcols(test_gr)),
                 c("MCF10A_CTCF", "MCF10AT1_CTCF", "MCF10CA1_CTCF",
                   "name", "score", "signalValue", "pValue", "qValue", "peak"))
})

test_that("ssvAnnotateSubjectGRanges - single GR with name", {
    test_gr = ssvAnnotateSubjectGRanges(np_grs$MCF10A_CTCF, olap_gr, annotation_name = "MCF10A")
    expect_equal(colnames(mcols(test_gr)),
                 c("MCF10A_CTCF", "MCF10AT1_CTCF",
                   "MCF10CA1_CTCF", "name.MCF10A",
                   "score.MCF10A", "signalValue.MCF10A",
                   "pValue.MCF10A", "qValue.MCF10A", "peak.MCF10A"))
})

test_that("ssvAnnotateSubjectGRanges - list of GRs", {
    test_gr = ssvAnnotateSubjectGRanges(np_grs, olap_gr)
    expect_equal(colnames(mcols(test_gr)),
                 c(
                     "MCF10A_CTCF",
                     "MCF10AT1_CTCF",
                     "MCF10CA1_CTCF",
                     "name.MCF10A_CTCF",
                     "score.MCF10A_CTCF",
                     "signalValue.MCF10A_CTCF",
                     "pValue.MCF10A_CTCF",
                     "qValue.MCF10A_CTCF",
                     "peak.MCF10A_CTCF",
                     "name.MCF10AT1_CTCF",
                     "score.MCF10AT1_CTCF",
                     "signalValue.MCF10AT1_CTCF",
                     "pValue.MCF10AT1_CTCF",
                     "qValue.MCF10AT1_CTCF",
                     "peak.MCF10AT1_CTCF",
                     "name.MCF10CA1_CTCF",
                     "score.MCF10CA1_CTCF",
                     "signalValue.MCF10CA1_CTCF",
                     "pValue.MCF10CA1_CTCF",
                     "qValue.MCF10CA1_CTCF",
                     "peak.MCF10CA1_CTCF"
                 ))
})

test_that("ssvAnnotateSubjectGRanges - list of GRs - override names", {
    test_gr = ssvAnnotateSubjectGRanges(np_grs, olap_gr, LETTERS[1:3])
    expect_equal(colnames(mcols(test_gr)),
                 c(
                     "MCF10A_CTCF",
                     "MCF10AT1_CTCF",
                     "MCF10CA1_CTCF",
                     "name.A",
                     "score.A",
                     "signalValue.A",
                     "pValue.A",
                     "qValue.A",
                     "peak.A",
                     "name.B",
                     "score.B",
                     "signalValue.B",
                     "pValue.B",
                     "qValue.B",
                     "peak.B",
                     "name.C",
                     "score.C",
                     "signalValue.C",
                     "pValue.C",
                     "qValue.C",
                     "peak.C"
                 ))
})

test_that("ssvAnnotateSubjectGRanges - GRangesList - override names", {
    test_gr = ssvAnnotateSubjectGRanges(GRangesList(np_grs), olap_gr, LETTERS[1:3])
    expect_equal(colnames(mcols(test_gr)),
                 c(
                     "MCF10A_CTCF",
                     "MCF10AT1_CTCF",
                     "MCF10CA1_CTCF",
                     "name.A",
                     "score.A",
                     "signalValue.A",
                     "pValue.A",
                     "qValue.A",
                     "peak.A",
                     "name.B",
                     "score.B",
                     "signalValue.B",
                     "pValue.B",
                     "qValue.B",
                     "peak.B",
                     "name.C",
                     "score.C",
                     "signalValue.C",
                     "pValue.C",
                     "qValue.C",
                     "peak.C"
                 ))
})

test_that("ssvAnnotateSubjectGRanges - list of GRs - error no names", {
    bad_grs = np_grs
    names(bad_grs) = NULL
    expect_error(ssvAnnotateSubjectGRanges(bad_grs, olap_gr), "annotation_source must have names set.")
})

test_that("ssvOverlapIntervalSets and ssvConsensusIntervalSets - message", {
    #when np_grs overlap multi-multi it's not clear which metadata should be preserved.
    expect_message({
        olap_anno_gr = ssvOverlapIntervalSets(np_grs, preserve_mcols = TRUE, ext = 50000000)
    },
    "Resolving multiple overlapping annotatation_source items with default function"
    )

    expect_message({
        cons_anno_gr = ssvConsensusIntervalSets(np_grs, preserve_mcols = TRUE, ext = 500000)
    },
    "Resolving multiple overlapping annotatation_source items with default function"
    )

    expect_message({
        anno_gr = ssvAnnotateSubjectGRanges(annotation_source = np_grs$MCF10A_CTCF, subject_gr = olap_anno_gr[1])
    },
    "Resolving multiple overlapping annotatation_source items with default function"
    )


})

test_that("ssvAnnotateSubjectGranges - multiple overlapping annotatation_source", {
    test_multi_resolver_FUN = function(x, variable.name = NULL){
        if(is.character(x)){
            x = "a"
        }else if(is.numeric(x)){
            x = 1
        }else if(is.factor(x)){
            x = factor("a")
        }else if(is.logical(x)){
            x = TRUE
        }
        x
    }

    olap_anno_gr = ssvOverlapIntervalSets(np_grs, preserve_mcols = TRUE, ext = 50000000)
    mcols(olap_anno_gr) = NULL

    anno_multi = ssvAnnotateSubjectGRanges(annotation_source = np_grs, subject_gr = olap_anno_gr[1])
    expect_equal(length(anno_multi), 1)
    expect_equal(anno_multi$score.MCF10A_CTCF, 3683)
    expect_equal(anno_multi$name.MCF10A_CTCF, "MCF10A_CTCF_pooled_peak_1190-from_11")

    anno_test = ssvAnnotateSubjectGRanges(annotation_source = np_grs, subject_gr = olap_anno_gr[1], multi_resolver_FUN = test_multi_resolver_FUN)
    expect_equal(length(anno_test), 1)
    expect_equal(anno_test$score.MCF10A_CTCF, 1)
    expect_equal(anno_test$name.MCF10A_CTCF, "a")

})

testthat::context("OverlapIntervalSets")
library(seqsetvis)
library(GenomicRanges)
library(testthat)
gr_a = GRanges("chr1", IRanges(1:7*10, 1:7*10+1))
gr_b = GRanges("chr1", IRanges(5:10*10, 5:10*10+1))
gr_c = GRanges("chr1", IRanges(8:10*10+5, 8:10*10+6))

test_that("ssvOverlapIntervalSets grs input are valid. single GR fails.", {
  expect_error(ssvOverlapIntervalSets(gr_a))
})

test_that("ssvOverlapIntervalSets grs input are valid. single numeric fails.", {
    expect_error(ssvOverlapIntervalSets(1))
})

test_that("ssvOverlapIntervalSets grs input are valid. GRangesList OK.", {
    expect_s4_class(ssvOverlapIntervalSets(GRangesList("gr_a" = gr_a, "gr_b" = gr_b)), class = "GRanges")
})

test_that("ssvOverlapIntervalSets grs input are valid. List of GRanges OK.", {
    expect_s4_class(ssvOverlapIntervalSets(list("gr_a" = gr_a, "gr_b" = gr_b)), class = "GRanges")
})

test_that("ssvOverlapIntervalSets grs input are valid. GRangesList OK.", {
    expect_s4_class(ssvOverlapIntervalSets(GRangesList(list("gr_a" = gr_a, "gr_b" = gr_b))), class = "GRanges")
})

test_that("ssvOverlapIntervalSets grs input are valid. List of GRanges with non-matching mcols OK.", {
    gr_amod = gr_a
    gr_amod$score = "scored"
    expect_s4_class(ssvOverlapIntervalSets(list("gr_a" = gr_amod, "gr_b" = gr_b)), class = "GRanges")
})

test_that("ssvOverlapIntervalSets useFirst.", {
    ol = ssvOverlapIntervalSets(list("gr_a" = gr_a, "gr_b" = gr_b), use_first = TRUE)
    expect_s4_class(ol, class = "GRanges")
    expect_equal(length(ol), length(gr_a))
})

test_that("ssvOverlapIntervalSets ext reduces number of results", {
    ol_ext1 = ssvOverlapIntervalSets(list("gr_a" = gr_a, "gr_b" = gr_b), ext = 1)
    ol_ext10 = ssvOverlapIntervalSets(list("gr_a" = gr_a, "gr_b" = gr_b), ext = 10)
    expect_equal(length(ol_ext1), 10)
    expect_equal(length(ol_ext10), 1)
})


#consensus
test_that("ssvConsensusIntervalSets grs input are valid. single GR fails.", {
  expect_error(ssvConsensusIntervalSets(gr_a))
})

test_that("ssvConsensusIntervalSets grs input are valid. single numeric fails.", {
  expect_error(ssvConsensusIntervalSets(1))
})

test_that("ssvConsensusIntervalSets grs input are valid. GRangesList OK.", {
  expect_s4_class(ssvConsensusIntervalSets(GRangesList("gr_a" = gr_a, "gr_b" = gr_b)), class = "GRanges")
})

test_that("ssvConsensusIntervalSets grs input are valid. List of GRanges OK.", {
  expect_s4_class(ssvConsensusIntervalSets(list("gr_a" = gr_a, "gr_b" = gr_b)), class = "GRanges")
})

test_that("ssvConsensusIntervalSets grs input are valid. GRangesList OK.", {
  expect_s4_class(ssvConsensusIntervalSets(GRangesList(list("gr_a" = gr_a, "gr_b" = gr_b))), class = "GRanges")
})

test_that("ssvConsensusIntervalSets grs input are valid. List of GRanges with non-matching mcols OK.", {
  gr_amod = gr_a
  gr_amod$score = "scored"
  expect_s4_class(ssvConsensusIntervalSets(list("gr_a" = gr_amod, "gr_b" = gr_b)), class = "GRanges")
})

test_that("ssvConsensusIntervalSets useFirst.", {
  ol = ssvConsensusIntervalSets(list("gr_a" = gr_a, "gr_b" = gr_b), min_number = 2)
  expect_equal(length(ol), 3)
})

test_that("ssvConsensusIntervalSets ext reduces number of results", {
  ol_ext1 = ssvConsensusIntervalSets(list("gr_a" = gr_a, "gr_b" = gr_b), ext = 1)
  ol_ext10 = ssvConsensusIntervalSets(list("gr_a" = gr_a, "gr_b" = gr_b), ext = 10)
  expect_equal(length(ol_ext1), 3)
  expect_equal(length(ol_ext10), 1)
})

#' Iterates a character vector (ideally named) and calls
#' \code{ssvFetchBam.single} on each.  Appends grouping variable to each
#' resulting data.table and uses rbindlist to efficiently combine results
#'
#' \code{ssvFetchBam} iteratively calls \code{fetchWindowedBam.single}. See
#' \code{\link{ssvFetchBam.single}} for more info.
#' @export
#' @param file_paths character vector of file_paths to load from. Alternatively,
#' file_paths can be a data.frame or data.table whose first column is a
#' character vector of paths and additial columns will be used as metadata.
#' @param qgr Set of GRanges to query.  For valid results the width of each
#'   interval should be identical and evenly divisible by \code{win_size}.
#' @param unique_names names to use in final data.table to designate source
#'   bigwig. Default is 'sample'
#' @param win_size The window size that evenly divides widths in \code{qgr}.
#' @param win_method character.  one of c("sample", "summary").  Determines if
#'   \code{\link{viewGRangesWinSample_dt}} or
#'   \code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
#'   qgr.
#' @param summary_FUN function.  only relevant if win_method is "summary".
#'   passed to \code{\link{viewGRangesWinSummary_dt}}.
#' @param fragLens numeric. The fragment length to use to extend reads.  The
#'   default value "auto" causes an automatic calculation from 100 regions in
#'   qgr.  NA causes no extension of reads to fragment size.
#' @param target_strand character. One of c("*", "+", "-"). Controls filtering
#'   of reads by strand.  Default of "*" combines both strands.
#' @param flip_strand boolean. if TRUE strands are flipped.
#' @param anchor character, one of c("center", "center_unstranded", "left",
#'   "left_unstranded")
#' @param names_variable The column name where unique_names are stored.
#' @param file_attribs optional data.frame/data.table with one row per item in
#' file paths.  Each column will be a variable added to final tidy output.
#' @param return_data.table logical. If TRUE the internal data.table is returned
#'   instead of GRanges.  Default is FALSE.
#' @param max_dupes numeric >= 1.  duplicate reads by strandd start position
#'   over this number are removed, Default is Inf.
#' @param splice_strategy character, one of c("none", "ignore", "add", "only",
#'   "splice_count"). Default is "none" and spliced alignment are asssumed not
#'   present. fragLen will be forced to be NA for any other value.  "ignore" will
#'   not count spliced regions.  add" counts spliced regions along with others,
#'   "only" will only count spliced regions and ignore others.
#' @param n_cores integer number of cores to use.
#' Uses mc.cores option if not supplied.
#' @param n_region_splits integer number of splits to apply to qgr. The query
#'   GRanges will be split into this many roughly equal parts for increased
#'   parallelization. Default is 1, no split.
#' @param return_unprocessed boolean. if TRUE returns read alignment in data.table. Default is FALSE.
#' @param force_skip_centerFix boolean, if TRUE all query ranges will be
#' used "as is".  This is already the case by default if win_method == "summary"
#' but may have applications where win_method == "sample".
#' @param ... passed to Rsamtools::ScanBamParam()
#' @return A tidy formatted GRanges (or data.table if specified) containing
#'   fetched values.
#' @rawNamespace import(data.table, except = c(shift, first, second, last))
#' @details if \code{qgr} contains the range chr1:1-100 and \code{win_size} is
#'   10, values from positions chr1 5,15,25...85, and 95 will be retrieved from
#'   \code{bw_file}
#' @importFrom stats weighted.mean
#' @importFrom pbapply pblapply
#' @examples
#' if(Sys.info()['sysname'] != "Windows"){
#' data(CTCF_in_10a_overlaps_gr)
#' library(GenomicRanges)
#' bam_f = system.file("extdata/test.bam",
#'     package = "seqsetvis", mustWork = TRUE)
#' bam_files = c("a" = bam_f, "b" = bam_f)
#' qgr = CTCF_in_10a_overlaps_gr[1:5]
#' bw_gr = ssvFetchBam(bam_files, qgr, win_size = 10)
#' bw_gr2 = ssvFetchBam(as.list(bam_files), qgr, win_size = 10)
#'
#' bw_dt = ssvFetchBam(bam_files, qgr, win_size = 10,
#'     return_data.table = TRUE)
#' }
ssvFetchBam = function(file_paths,
                       qgr,
                       unique_names = NULL,
                       names_variable = "sample",
                       file_attribs = NULL,
                       win_size = 50,
                       win_method = c("sample", "summary")[1],
                       summary_FUN = stats::weighted.mean,
                       fragLens = "auto",
                       target_strand = c("*", "+", "-", "both")[1],
                       flip_strand = FALSE,
                       anchor = c("left", "left_unstranded", "center",
                                  "center_unstranded")[3],
                       return_data.table = FALSE,
                       max_dupes = Inf,
                       splice_strategy = c("none", "ignore", "add",
                                           "only", "splice_count")[1],
                       n_cores = getOption("mc.cores", 1),
                       n_region_splits = 1,
                       return_unprocessed = FALSE,
                       force_skip_centerFix = FALSE,
                       ...){
    stopifnot(all(is.character(fragLens) |
                      is.numeric(fragLens) |
                      is.na(fragLens)))
    exp_fragLen = ifelse(is.data.frame(file_paths) || is.data.table(file_paths), nrow(file_paths), length(file_paths))
    stopifnot(length(fragLens) == 1 || length(fragLens) == exp_fragLen)
    if(splice_strategy != "none") fragLens = NA
    if(return_unprocessed) fragLens = NA
    if(length(fragLens) == 1){
        if (is.data.frame(file_paths) || is.data.table(file_paths)) {
            fragLens = rep(fragLens[1], nrow(file_paths))
        }else{
            fragLens = rep(fragLens[1], length(file_paths))
        }
    }
    tmp = .get_file_attribs(file_paths, file_attribs, names_variable)
    file_paths = tmp$file_paths
    file_attribs = tmp$file_attribs
    remove(tmp)
    unique_names = .get_unique_names(unique_names, file_paths, file_attribs, names_variable)

    names(fragLens) = unique_names
    names(file_paths) = unique_names

    if(any(fragLens == 'auto', na.rm = TRUE)){
        message("Calculating average fragment length for ", sum(fragLens == "auto", na.rm = TRUE), " bam file(s).")
        calc_fl = ssv_mclapply(names(fragLens), function(nam){
            if(!is.na(fragLens[nam]) && fragLens[nam] == "auto"){
                fl = fragLen_calcStranded(file_paths[nam], qgr, flip_strand = flip_strand)
                message("fragLen for ", basename(nam), " was calculated as: ", fl)
            }else{
                fl = fragLens[nam]
            }
            fl
        })

        calc_fl = unlist(calc_fl)
        names(calc_fl) = names(fragLens)
        fragLens = calc_fl
    }

    fragLens = as.numeric(fragLens)
    names(fragLens) = unique_names

    load_bam = function(f, nam, qgr) {
        message("loading ", nam, " ...")
        if(!file.exists(paste0(f, ".bai"))){
            warning("creating index for ", f)
            Rsamtools::indexBam(f)
        }
        fl = fragLens[nam]
        if(!is.na(fl))
            if(fl == "auto"){
                fl = NULL
            }
        dt = ssvFetchBam.single(bam_f = f,
                                qgr = qgr,
                                win_size = win_size,
                                win_method = win_method,
                                summary_FUN = summary_FUN,
                                fragLen = fl,
                                target_strand = target_strand,
                                anchor = anchor,
                                return_data.table = TRUE,
                                max_dupes = max_dupes,
                                splice_strategy = splice_strategy,
                                flip_strand = flip_strand,
                                return_unprocessed = return_unprocessed,
                                force_skip_centerFix = force_skip_centerFix,
                                ...)
        # dt[[names_variable]] = rep(nam, nrow(dt))
        message("finished loading ", nam, ".")
        dt
    }

    bdt = ssvFetchSignal(file_paths = file_paths,
                         qgr = qgr,
                         load_signal = load_bam,
                         unique_names = unique_names,
                         names_variable = names_variable,
                         file_attribs = file_attribs,
                         win_size = win_size,
                         win_method = win_method,
                         return_data.table = TRUE,
                         n_cores = n_cores,
                         n_region_splits = n_region_splits,
                         force_skip_centerFix = force_skip_centerFix)

    if(!return_data.table & !return_unprocessed){
        bdt = GRanges(bdt)
    }
    bdt
}

#' fetch a windowed version of a bam file, returns GRanges
#'
#' @param bam_f character or BamFile to load
#' @param qgr GRanges regions to fetchs
#' @param win_size numeric >=1.  pileup grabbed every win_size bp for win_method
#'   sample.  If win_method is summary, this is the number of windows used
#'   (confusing, sorry).
#' @param win_method character.  one of c("sample", "summary").  Determines if
#'   \code{\link{viewGRangesWinSample_dt}} or
#'   \code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
#'   qgr.
#' @param summary_FUN function.  only relevant if win_method is "summary".
#'   passed to \code{\link{viewGRangesWinSummary_dt}}.
#' @param fragLen numeric, NULL, or NA.  if numeric, supplied value is used. if
#'   NULL, value is calculated with fragLen_calcStranded if NA, raw bam pileup
#'   with no cross strand shift is returned.
#' @param target_strand character. if one of "+" or "-", reads are filtered
#'   accordingly. ignored if any other value.
#' @param anchor character, one of c("center", "center_unstranded", "left",
#'   "left_unstranded")
#' @param return_data.table logical. If TRUE the internal data.table is returned
#'   instead of GRanges.  Default is FALSE.
#' @param max_dupes numeric >= 1.  duplicate reads by strandd start position
#'   over this number are removed, Default is Inf.
#' @param splice_strategy character, one of c("none", "ignore", "add", "only",
#'   "splice_count"). Default is "none" and spliced alignment are asssumed not
#'   present. fragLen must be NA for any other value to be valid.  "ignore" will
#'   not count spliced regions.  add" counts spliced regions along with others,
#'   "only" will only count spliced regions and ignore others.
#' @param flip_strand if TRUE, strand alignment is flipped prior to fragLen
#'   extension. Default is FALSE.
#' @param return_unprocessed boolean. if TRUE returns read alignment in data.table. Default is FALSE.
#' @param force_skip_centerFix boolean, if TRUE all query ranges will be
#' used "as is".  This is already the case by default if win_method == "summary"
#' but may have applications where win_method == "sample".
#' @param ... passed to Rsamtools::ScanBamParam()
#' @return tidy GRanges (or data.table if specified) with pileups from bam file.
#'   pileup is calculated only every win_size bp.
#' @importFrom stats weighted.mean
ssvFetchBam.single = function(bam_f,
                              qgr,
                              win_size = 50,
                              win_method = c("sample", "summary")[1],
                              summary_FUN = stats::weighted.mean,
                              fragLen = NULL,
                              target_strand = c("*", "+", "-", "both")[1],
                              anchor = c("left", "left_unstranded", "center",
                                         "center_unstranded")[3],
                              return_data.table = FALSE,
                              max_dupes = Inf,
                              splice_strategy = c("none", "ignore", "add",
                                                  "only", "splice_count")[1],
                              flip_strand = FALSE,
                              return_unprocessed = FALSE,
                              force_skip_centerFix = FALSE,
                              ...) {
    x = id = y = NULL #binding for data.table
    stopifnot(is.character(win_method))
    stopifnot(length(win_method) == 1)
    stopifnot(is(qgr, "GRanges"))
    stopifnot(win_method %in% c("sample", "summary"))
    stopifnot(is.function(summary_FUN))
    stopifnot(target_strand %in% c("*", "+", "-", "both"))
    stopifnot(anchor %in% c("left", "left_unstranded", "center",
                            "center_unstranded"))
    stopifnot(splice_strategy %in% c("none", "ignore", "add",
                                     "only", "splice_count"))
    if(return_unprocessed){
        score_gr = fetchBam(bam_f,
                            qgr,
                            fragLen = NA,
                            target_strand = "*",
                            max_dupes,
                            splice_strategy = "none",
                            return_unprocessed = return_unprocessed,
                            ...)
        tmp = qgr
        strand(tmp) = "*"
        score_gr = merge(score_gr, data.table(which_label = as.character(tmp), id = names(qgr)), by = "which_label")
        return(score_gr)
    }
    if(splice_strategy == "splice_count"){
        splice_res = list()
        if(target_strand == "*"){
            splice_res[["star"]] = fetchBam(bam_f,
                                            qgr,
                                            fragLen = NA,
                                            target_strand = target_strand,
                                            max_dupes, splice_strategy,
                                            flip_strand = flip_strand,
                                            ...)
        }else{
            if(target_strand == "both" | target_strand == "+"){
                splice_res[["pos"]] = fetchBam(bam_f,
                                               qgr,
                                               fragLen = NA,
                                               target_strand = "+",
                                               max_dupes, splice_strategy,
                                               flip_strand = flip_strand,
                                               ...)
            }
            if(target_strand == "both" | target_strand == "-"){
                splice_res[["neg"]] = fetchBam(bam_f,
                                               qgr,
                                               fragLen = NA,
                                               target_strand = "-",
                                               max_dupes, splice_strategy,
                                               flip_strand = flip_strand,
                                               ...)
            }
        }
        return(rbindlist(splice_res))
    }

    #splitting by qgr strand is necessary for strand sensitive fetch when
    #features on opposing strands overlap
    strands_TODO = unique(as.character(strand(qgr)))
    out_l = list()
    for(s in strands_TODO){
        strand_qgr = subset(qgr, strand == s)
        switch (
            win_method,
            sample = {
                strand_qgr = prepare_fetch_GRanges_width(strand_qgr, win_size, skip_centerFix = force_skip_centerFix)
                if(target_strand == "both"){
                    pos_gr = fetchBam(bam_f, strand_qgr, fragLen, "+",
                                      max_dupes, splice_strategy,
                                      flip_strand = flip_strand, ...)
                    neg_gr = fetchBam(bam_f, strand_qgr, fragLen, "-",
                                      max_dupes, splice_strategy,
                                      flip_strand = flip_strand, ...)
                    pos_dt = viewGRangesWinSample_dt(pos_gr, strand_qgr,
                                                     win_size, anchor = anchor)
                    neg_dt = viewGRangesWinSample_dt(neg_gr, strand_qgr,
                                                     win_size, anchor = anchor)
                    pos_dt[, strand := "+"]
                    neg_dt[, strand := "-"]
                    out = rbind(
                        pos_dt,
                        neg_dt
                    )
                }else{
                    score_gr = fetchBam(bam_f, strand_qgr, fragLen, target_strand,
                                        max_dupes, splice_strategy,
                                        flip_strand = flip_strand, ...)
                    out = viewGRangesWinSample_dt(score_gr, strand_qgr,
                                                  win_size, anchor = anchor)
                    out[, strand := target_strand]
                }


            },
            summary = {
                if(target_strand == "both"){
                    pos_gr = fetchBam(bam_f, strand_qgr, fragLen, "+",
                                      max_dupes, splice_strategy,
                                      flip_strand = flip_strand, ...)
                    neg_gr = fetchBam(bam_f, strand_qgr, fragLen, "-",
                                      max_dupes, splice_strategy,
                                      flip_strand = flip_strand, ...)
                    pos_dt = viewGRangesWinSummary_dt(pos_gr, strand_qgr, win_size,
                                                      summary_FUN = summary_FUN,
                                                      anchor = anchor)

                    neg_dt = viewGRangesWinSummary_dt(neg_gr, strand_qgr, win_size,
                                                      summary_FUN = summary_FUN,
                                                      anchor = anchor)
                    pos_dt[, strand := "+"]
                    neg_dt[, strand := "-"]
                    out = rbind(
                        pos_dt,
                        neg_dt
                    )
                }else{
                    score_gr = fetchBam(bam_f, strand_qgr, fragLen, target_strand,
                                        max_dupes, splice_strategy,
                                        flip_strand = flip_strand, ...)
                    out = viewGRangesWinSummary_dt(score_gr, strand_qgr, win_size,
                                                   summary_FUN = summary_FUN,
                                                   anchor = anchor)
                    out[, strand := target_strand]
                }
            }
        )
        out_l[[s]] = out
    }
    out = rbindlist(out_l)
    out = out[order(x)][order(id)][order(strand)]

    if(!return_data.table){
        out = GRanges(out)
    }
    return(out)
}

#' fetch a bam file pileup with the ability to consider read extension to
#' fragment size (fragLen)
#' @param bam_f character or BamFile to load
#' @param qgr GRanges regions to fetchs
#' @param fragLen numeric, NULL, or NA.  if numeric, supplied value is used. if
#'   NULL, value is calculated with fragLen_calcStranded (default) if NA, raw
#'   bam pileup with no cross strand shift is returned.
#' @param target_strand character. if one of "+" or "-", reads are filtered to
#'   match. ignored if any other value.
#' @param max_dupes numeric >= 1.  duplicate reads by strandd start position
#'   over this number are removed, Default is Inf.
#' @param splice_strategy character, one of c("none", "ignore", "add", "only").
#'   Default is "none" and split read alignments are asssumed not present.
#'   fragLen must be NA for any other value to be valid.  "ignore" will not
#'   count spliced regions.  "add" counts spliced regions along with others,
#'   "only" will only count spliced regions and ignore others.
#' @param flip_strand if TRUE, strand alignment is flipped prior to fragLen
#'   extension. Default is FALSE.
#' @param return_unprocessed boolean. if TRUE returns read alignment in data.table. Default is FALSE.
#' @param ... passed to ScanBamParam(), can't be which or what.
#' @return GRanges containing tag pileup values in score meta column.  tags are
#'   optionally extended to fragment length (fragLen) prior to pile up.
fetchBam = function(bam_f,
                    qgr,
                    fragLen = NULL,
                    target_strand = c("*", "+", "-")[1],
                    max_dupes = Inf,
                    splice_strategy = c("none", "ignore", "add",
                                        "only", "splice_count")[1],
                    flip_strand = FALSE,
                    return_unprocessed = FALSE,
                    ...){
    which_label = NULL #reserve binding
    stopifnot(is.numeric(max_dupes))
    stopifnot(max_dupes >= 1)
    if(target_strand == "both") target_strand = "*"
    stopifnot(target_strand %in% c("*", "+", "-"))
    if(!is.na(fragLen) && splice_strategy != "none"){
        stop("fragLen must be NA if splice_strategy is not 'none'.")
    }
    if( ! splice_strategy %in% c("none", "ignore", "add",
                                 "only", "splice_count")){
        stop('splice_strategy must be one of: "none", "ignore", "add", "only"')
    }
    qgr = harmonize_seqlengths(qgr, bam_f)
    if(is.null(fragLen)){
        fragLen = fragLen_calcStranded(bam_f, qgr, flip_strand = flip_strand)
        message("fragLen was calculated as: ", fragLen)
    }
    if(!is.na(fragLen)){
        stopifnot(is.numeric(fragLen))
        stopifnot(fragLen %% 1 == 0)
        stopifnot(fragLen >= 1)
    }
    sbgr = qgr
    strand(sbgr) = "*"

    if(return_unprocessed){
        sbParam = Rsamtools::ScanBamParam(
            which = sbgr,
            what = Rsamtools::scanBamWhat(),
            ...)
        bam_raw = Rsamtools::scanBam(bam_f, param = sbParam)
        bam_dt = lapply(bam_raw, function(x){
            data.table(seqnames = x$rname,
                       strand = x$strand,
                       start = x$pos,
                       width = x$qwidth,
                       cigar = x$cigar,
                       qname = x$qname,
                       flag = x$flag,
                       mapq = x$mapq,
                       mrnm = x$mrnm,
                       mpos = x$mpos,
                       isize = x$isize,
                       seq = as.character(x$seq),
                       qual = as.character(x$qual))
        })
        bam_dt = data.table::rbindlist(bam_dt,
                                       use.names = TRUE,
                                       idcol = "which_label")
        return(bam_dt)
    }else{
        sbParam = Rsamtools::ScanBamParam(
            which = sbgr,
            what = c("rname", "strand", "pos", "qwidth", "cigar"),
            ...)
        bam_raw = Rsamtools::scanBam(bam_f, param = sbParam)
        bam_dt = lapply(bam_raw, function(x){
            data.table(seqnames = x$rname, strand = x$strand,
                       start = x$pos, width = x$qwidth, cigar = x$cigar)
        })
        bam_dt = data.table::rbindlist(bam_dt,
                                       use.names = TRUE,
                                       idcol = "which_label")
    }


    bam_dt = bam_dt[!is.na(width)]
    #as.character for GRanges does a couple annoying things. for GRanges
    #of width 1, end is omitted.  Appending strand also not desirable here.
    # want chr1:17508106-17508106 not chr1:17508106:-
    gr_as.character = function(gr){
        if(length(gr) == 0){
            character()
        }else{
            paste0(seqnames(gr), ":", start(gr), "-", end(gr))
        }
    }
    # it does not make sense to use query gr strand in reference to splice count
    if(splice_strategy == "splice_count"){
        toflip = character()
    }else{
        toflip = gr_as.character(subset(qgr, strand == "-"))
    }

    if(target_strand %in% c("+", "-")){
        if(flip_strand){
            bam_dt[, strand := ifelse(strand == "+", "-", "+")]
        }
        bam_dt = bam_dt[strand == target_strand & !(which_label %in% toflip) |
                            strand != target_strand & (which_label %in% toflip)]
    }

    bam_dt[, end := start + width - 1L]

    if(max_dupes < Inf){
        bam_dt = .rm_dupes(bam_dt, max_dupes)
    }
    if(is.na(fragLen)){
        bam_dt = switch(
            splice_strategy,
            none = {bam_dt},
            ignore = {.expand_cigar_dt(bam_dt)},
            add = {.expand_cigar_dt(bam_dt,
                                    op_2count = c("M", "D", "=", "X", "N"))},
            only = {.expand_cigar_dt(bam_dt, op_2count = c("N"))},
            splice_count = {.expand_cigar_dt(bam_dt, op_2count = c("N"))}
        )
    }else{
        # bam_dt[, end := start + width - 1L]
        if(flip_strand){
            bam_dt[strand == "-", end := start + as.integer(fragLen) - 1L]
            bam_dt[strand == "+", start := end - as.integer(fragLen) + 1L]
        }else{
            bam_dt[strand == "+", end := start + as.integer(fragLen) - 1L]
            bam_dt[strand == "-", start := end - as.integer(fragLen) + 1L]
        }

    }
    bam_dt$strand = target_strand
    if(splice_strategy == "splice_count"){
        sp_dt = bam_dt[, .N,
                       by = list(which_label, seqnames, start, end, strand)]
        return(sp_dt)
    }
    ext_cov = coverage(split(GRanges(bam_dt), bam_dt$which_label))
    score_gr = GRanges(ext_cov)
    score_gr = harmonize_seqlengths(score_gr, bam_f)

    if(length(score_gr) == 0){
        score_gr = sbgr
        mcols(score_gr) = NULL
        score_gr$score = 0
    }
    # if(target_strand == "+"){
    #     strand(score_gr) = "+"
    # }
    # if(target_strand == "-"){
    #     strand(score_gr) = "-"
    # }

    return(score_gr)
}

#' harmonize_seqlengths
#'
#' ensures compatibility between seqlength of gr and bam_file based on header
#'
#' @param query_gr GRanges, object to harmonize seqlengths for
#' @param bam_file character, a path to a valid bam file
#' @param force_fix Logical, if TRUE incompatible seqnames are removed from the query_gr.  Default is FALSE.
#'
#' @return GRanges with seqlengths matching bam_file
#' @importFrom Seqinfo seqlengths
#' @importFrom Rsamtools scanBamHeader
#' @export
#' @examples
#' library(GenomicRanges)
#' query_gr = GRanges("chr1", IRanges(1, 100))
#' #seqlengths has not been set
#' seqlengths(query_gr)
#' bam = system.file("extdata/test.bam", package = "seqsetvis")
#' gr2 = harmonize_seqlengths(query_gr, bam)
#' #seqlengths now set
#' seqlengths(gr2)
harmonize_seqlengths = function(query_gr, bam_file, force_fix = FALSE){
    chr_lengths = Rsamtools::scanBamHeader(bam_file)[[1]]$targets
    if(!all(as.character(seqnames(query_gr)) %in% names(chr_lengths))){
        if(force_fix){
            len.orig = length(query_gr)
            query_gr = subset(query_gr, seqnames %in% names(chr_lengths))
            len.new = length(query_gr)
            message(len.orig - len.new, " regions were removed due to force_fix = TRUE where seqnames were missing from bam file.")
        }else{
            stop("There are chromosomes present in query GRanges (",
                 paste(setdiff(as.character(seqnames(query_gr)), names(chr_lengths)), collapse = ", "),
                 ") not present in bam file (", bam_file, ")!\n",
                 "If you want to proceed, run harmonize_seqlengths with force_fix = TRUE on query_gr to remove incompatible seqnames and retry.")
        }
    }
    to_remove =setdiff(names(Seqinfo::seqlengths(query_gr)), unique(as.character(seqnames(query_gr))))
    if(length(to_remove) > 0){
        query_gr = GenomeInfoDb::dropSeqlevels(query_gr, to_remove)
    }
    Seqinfo::seqlengths(query_gr) =
        chr_lengths[names(Seqinfo::seqlengths(query_gr))]
    too_long = end(query_gr) >
        Seqinfo::seqlengths(query_gr)[as.character(seqnames(query_gr))]
    if(any(too_long)){
        message(sum(too_long),
                " region shifted for extending beyond seqlengths")
        fix_gr = query_gr[too_long]
        shift_by = -(end(fix_gr) - Seqinfo::seqlengths(fix_gr)[
            as.character(seqnames(fix_gr))])
        query_gr[too_long] = GenomicRanges::shift(fix_gr, shift_by)
    }
    too_short = start(query_gr) < 1
    if(any(too_short)){
        message(sum(too_short),
                " region shifted for starting before seqlengths")
        fix_gr = query_gr[too_short]
        shift_by = 1 - start(fix_gr)
        query_gr[too_short] = GenomicRanges::shift(fix_gr, shift_by)
    }
    query_gr
}

#' determine the most common read length for input bam_file.  uses 50 randomly
#' selected regions from query_gr.  If fewer than 20 reads are present, loads
#' all of query_gr.
#'
#' @param bam_file indexed bam file
#' @param query_gr GRanges to read from bam file
#' @export
#' @return numeric of most common read length.
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' qgr = CTCF_in_10a_overlaps_gr[1:5]
#' bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
#' getReadLength(bam_file, qgr)
getReadLength = function(bam_file,
                         query_gr){
    Param <- Rsamtools::ScanBamParam(which=sample(query_gr,
                                                  min(50, length(query_gr))),
                                     what=c("flag","mapq"))
    temp <- GenomicAlignments::readGAlignments(bam_file,param=Param)
    if(length(temp) < 20){
        Param <- Rsamtools::ScanBamParam(which=sample(query_gr),
                                         what=c("flag","mapq"))
        temp <- GenomicAlignments::readGAlignments(bam_file,param=Param)
    }
    if(length(temp) == 0){
        warning("No reads could be found in query_gr to determine read length. Setting to 180.",
                "Please verify that query_gr is appropriate for bam_file: ", bam_file)
        readlength = 180
    }else{
        readlength=as.numeric(names(sort(table(width(temp)), decreasing = TRUE))[1])
    }
    readlength
}


#' Expand cigar codes to GRanges
#'
#' see \href{https://samtools.github.io/hts-specs/SAMv1.pdf}{sam specs} for
#' cigar details
#'
#'
#' @param cigar_dt data.table with 5 required named columns in any order.
#'   c("which_label", "seqnames", "strand", "start", "cigar")
#' @param op_2count Cigar codes to count. Default is alignment (M), deletion
#'   (D), match (=), and mismatch (X).  Other useful codes may be skipped
#'   regions for RNA splicing (N).  The locations of any insterions (I) or
#'   clipping/padding (S, H, or P) will be a single bp immediately before the
#'   interval.
#' @param return_data.table if TRUE, a data.table is returned, else a GRanges.
#' Default is FALSE.
#' @export
#' @return data.table with cigar entries expanded
#' @examples
#' data(CTCF_in_10a_overlaps_gr)
#' qgr = CTCF_in_10a_overlaps_gr[1:5]
#' bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
#' raw_dt = ssvFetchBam(bam_file, qgr, return_unprocessed = TRUE)
#' expandCigar(raw_dt)
expandCigar = function(cigar_dt, op_2count = c("M", "D", "=", "X"), return_data.table = FALSE){
    if("GRanges" %in% class(cigar_dt)){
        cigar_dt = as.data.table(cigar_dt)
    }
    stopifnot(c("which_label", "seqnames", "strand", "start", "cigar") %in% colnames(cigar_dt))
    res = .expand_cigar_dt(cigar_dt, op_2count)
    if(!return_data.table){
        res = GRanges(res)
    }
    return(res)
}

#' Expand intermediate bam fetch by cigar codes
#'
#' see \href{https://samtools.github.io/hts-specs/SAMv1.pdf}{sam specs} for
#' cigar details
#'
#'
#' @param cigar_dt data.table with 5 required named columns in any order.
#'   c("which_label", "seqnames", "strand", "start", "cigar")
#' @param op_2count Cigar codes to count. Default is alignment (M), deletion
#'   (D), match (=), and mismatch (X).  Other useful codes may be skipped
#'   regions for RNA splicing (N).  The locations of any insterions (I) or
#'   clipping/padding (S, H, or P) will be a single bp immediately before the
#'   interval.
#'
#' @return data.table with cigar entries expanded
.expand_cigar_dt = function(cigar_dt, op_2count = c("M", "D", "=", "X")){
    cigar_type = NULL
    cigar_dt = copy(cigar_dt)
    cigar_dt[, end := start]
    exp_dt = .expand_cigar_dt_recursive(cigar_dt)
    exp_dt[cigar_type %in% op_2count]
}


#' Expand intermediate bam fetch by cigar codes
#'
#' see \href{https://samtools.github.io/hts-specs/SAMv1.pdf}{sam specs} for
#' cigar details
#'
#'
#' @param cigar_dt data.table with 5 required named columns in any order.
#'   c("which_label", "seqnames", "strand", "start", "cigar")
#'
#' @return data.table with cigar entries expanded
.expand_cigar_dt_recursive = function(cigar_dt){
    cigar_w = cigar = cigar_type = tmp = which_label = end_adj = NULL #dt bindings
    stopifnot(all(c("which_label", "seqnames",
                    "strand", "start",
                    "cigar") %in% colnames(cigar_dt)))
    reg1 = regexpr("[MIDNSHP=X]", cigar_dt$cigar)
    cigar_dt[, cigar_w := as.integer(substr(cigar, 1, reg1-1))]
    cigar_dt[, cigar_type :=
                 substr(cigar, reg1, reg1+attr(reg1, "match.length")-1L)]
    cigar_dt[, tmp := substring(cigar, reg1 + 1)]
    # these do not consume reference
    cigar_dt[cigar_type %in% c("I", "S", "H", "P"), cigar_w := 0L]

    cigar_dt$end_adj = -1L
    cigar_dt[cigar_type %in% c("I", "S", "H", "P"), end_adj := 0L]

    ass1 = cigar_dt[, list(which_label, seqnames, strand,
                           start, end  = start + cigar_w + end_adj, cigar_type)]


    next_dt = cigar_dt[tmp != "", list(which_label,
                                       seqnames,
                                       strand,
                                       start = end + cigar_w,
                                       end = end + cigar_w,
                                       cigar = tmp)]
    if(nrow(next_dt) > 0){
        return(rbind(ass1,
                     .expand_cigar_dt_recursive(next_dt)))
    }else{
        return(ass1)
    }
}

#' Remove duplicate reads based on stranded start position.  This is an
#' over-simplification.  For better duplicate handling, duplicates must be
#' marked in bam and flag passed to fetchBam() ... for ScanBamParam
#'
#' flag = scanBamFlag(isDuplicate = FALSE)
#'
#' @param reads_dt data.table of reads as loaded by fetchBam
#' @param max_dupes maximum allowed positional duplicates
#'
#' @return reads_dt with duplicated reads over max_dupes removed
.rm_dupes = function(reads_dt, max_dupes){
    ndupe = which_label = NULL
    reads_dt[, ndupe := 1L]
    reads_dt[strand == "+",
             ndupe := seq_len(.N)[order(width, decreasing = TRUE)],
             by = list(which_label, start)]
    reads_dt[strand == "-",
             ndupe := seq_len(.N)[order(width, decreasing = TRUE)],
             by = list(which_label, end)]
    reads_dt = reads_dt[ndupe <= max_dupes]
    reads_dt$ndupe = NULL
    reads_dt
}

testthat::context("splice_count")

library(seqsetvis)
library(testthat)

#### positive strand query_gr ####
query_gr = GenomicRanges::GRanges("chr6:151656691-152129619:+")
# se_rna_bam_file = system.file(package = "seqsetvis", "extdata/MCF7_R1.ESR1_RNA.bam", mustWork = TRUE)
se_rna_bam_file = system.file(package = "seqsetvis", "extdata/H1_ESC_PE_R1.ESR1_RNA.bam", mustWork = TRUE)


prof_dt.no_flip = ssvFetchBam(
    se_rna_bam_file,
    query_gr,
    return_data.table = TRUE,
    splice_strategy = "splice_count",
    flag = Rsamtools::scanBamFlag(isFirstMateRead = TRUE),
    target_strand = "both"
)



prof_dt.flip = ssvFetchBam(
    se_rna_bam_file,
    query_gr,
    return_data.table = TRUE,
    splice_strategy = "splice_count",
    flag = Rsamtools::scanBamFlag(isFirstMateRead = TRUE),
    target_strand = "both",
    flip_strand = TRUE
)

prof_dt.no_flip
prof_dt.flip

# prof_dt = ssvFetchBam(
#     se_rna_bam_file, query_gr, return_data.table = TRUE, fragLens = NA)
# ggplot(prof_dt, aes(x = (start + end)/2 , y = y)) +
#     geom_path() +
#     ggbio::geom_arch(data = prof_dt.no_flip, aes(x = start, xend = end, height = N, color = strand))


prof_dt.flip.neg = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = TRUE,
    target_strand = "-")
prof_dt.flip.pos = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = TRUE,
    target_strand = "+")
prof_dt.flip.both = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = TRUE,
    target_strand = "both")

prof_dt.no_flip.pos = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = FALSE,
    target_strand = "+")
prof_dt.no_flip.neg = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = FALSE,
    target_strand = "-")

prof_dt.no_flip.both = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = FALSE,
    target_strand = "both")


my_test_strand = function(dt, pos_exp, neg_exp, ns_exp){
    testthat::expect_equal(sum(dt$strand == "+"), pos_exp)
    testthat::expect_equal(sum(dt$strand == "-"), neg_exp)
    testthat::expect_equal(sum(dt$strand == "*"), ns_exp)
}

test_that("splice_count strand - default", {
    my_test_strand(prof_dt.no_flip, 1, 3, 0)
})

test_that("splice_count strand - flip", {
    my_test_strand(prof_dt.flip, 3, 1, 0)
})

test_that("splice_count strand - no flip and target_strand", {
my_test_strand(prof_dt.no_flip.both, 3, 4, 0)
    my_test_strand(prof_dt.no_flip.pos, 3, 0, 0)
    my_test_strand(prof_dt.no_flip.neg, 0, 4, 0)
})

test_that("splice_count strand - flip and target_strand", {
    my_test_strand(prof_dt.flip.both, 4, 3, 0)
    my_test_strand(prof_dt.flip.pos, 4, 0, 0)
    my_test_strand(prof_dt.flip.neg, 0, 3, 0)
})
#### star strand query_gr ####
query_gr = GenomicRanges::GRanges("chr6:151656691-152129619:*")

prof_dt.no_flip.pos = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = FALSE,
    target_strand = "+")
prof_dt.no_flip.neg = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = FALSE,
    target_strand = "-")

prof_dt.no_flip.both = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = FALSE,
    target_strand = "both")


test_that("splice_count strand - no flip and target_strand", {
    my_test_strand(prof_dt.no_flip.both, 3, 4, 0)
    my_test_strand(prof_dt.no_flip.pos, 3, 0, 0)
    my_test_strand(prof_dt.no_flip.neg, 0, 4, 0)
})



#### negative strand query_gr ####
query_gr = GenomicRanges::GRanges("chr6:151656691-152129619:-")

prof_dt.no_flip.pos = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = FALSE,
    target_strand = "+")
prof_dt.no_flip.neg = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = FALSE,
    target_strand = "-")

prof_dt.no_flip.both = ssvFetchBam(
    se_rna_bam_file, query_gr, return_data.table = TRUE, splice_strategy = "splice_count",
    flip_strand = FALSE,
    target_strand = "both")


test_that("splice_count strand - no flip and target_strand", {
    my_test_strand(prof_dt.no_flip.both, 3, 4, 0)
    my_test_strand(prof_dt.no_flip.pos, 3, 0, 0)
    my_test_strand(prof_dt.no_flip.neg, 0, 4, 0)
})

testthat::context("ssvFetchBamPE.RNA")
# this test both weak and overly complicated
library(seqsetvis)
library(GenomicRanges)
library(data.table)
library(testthat)

library(GenomicRanges)
pkg_dir = system.file(package = "seqsetvis", "extdata", mustWork = TRUE)
bam_files_esr1 = dir(pkg_dir, pattern = "H1.+R1.ESR1_RNA.+bam$", full.names = TRUE)
names(bam_files_esr1) = sub("_R.+", "", basename(bam_files_esr1))
query_gr = GenomicRanges::GRanges("chr6:151656691-152129619:+")
query_gr = GenomicRanges::GRanges("chr6:152121500-152123500:+")

strand(query_gr) = "+"

prof_dt.pe1 = ssvFetchBamPE.RNA(bam_files_esr1, query_gr, unique_names = "PE_noFlip", return_data.table = TRUE, win_size = 1)
prof_dt.pe2 = ssvFetchBamPE.RNA(bam_files_esr1, query_gr, unique_names = "PE_flip", return_data.table = TRUE, win_size = 1, flip_strand = TRUE)

prof_dt1 = ssvFetchBam(
    bam_files_esr1,
    unique_names = "pos_noFlip",
    fragLens = NA,
    query_gr,
    return_data.table = TRUE,
    win_size = 1,
    splice_strategy = "ignore",
    target_strand = "both")
prof_dt2 = ssvFetchBam(
    bam_files_esr1,
    unique_names = "pos_flip",
    fragLens = NA,
    query_gr,
    return_data.table = TRUE,
    win_size = 1,
    splice_strategy = "ignore",
    target_strand = "both",
    flip_strand = TRUE)

strand(query_gr) = "-"

prof_dt3 = ssvFetchBam(
    bam_files_esr1,
    unique_names = "neg_noFlip",
    fragLens = NA,
    query_gr,
    return_data.table = TRUE,
    win_size = 1,
    splice_strategy = "ignore",
    target_strand = "both")
prof_dt4 = ssvFetchBam(
    bam_files_esr1,
    unique_names = "neg_flip",
    fragLens = NA,
    query_gr,
    return_data.table = TRUE,
    win_size = 1,
    splice_strategy = "ignore",
    target_strand = "both",
    flip_strand = TRUE)

prof_dt = rbind(prof_dt1, prof_dt2, prof_dt3, prof_dt4, prof_dt.pe1, prof_dt.pe2)

prof_dt[, gen_x := (start + end) / 2]

ggplot(prof_dt, aes(x = gen_x, y = y, color = strand)) + geom_path() +
    facet_grid(sample~strand)

prof_dt.agg = prof_dt[, .(y = sum(y)), .(x, gen_x, sample)]
ggplot(prof_dt.agg, aes(x = gen_x, y = y)) + geom_path() +
    facet_grid(sample~.)
prof_dt.agg.w = dcast(prof_dt.agg, gen_x~sample, value.var = "y")
prof_dt.agg.w[pos_noFlip > 1]
prof_dt.agg.w[pos_noFlip == PE_noFlip]
test_that("all are equivalent", {
    expect_equal(prof_dt.agg.w$pos_noFlip, prof_dt.agg.w$pos_flip)
    expect_equal(prof_dt.agg.w$pos_noFlip, prof_dt.agg.w$neg_noFlip)
    expect_equal(prof_dt.agg.w$pos_noFlip, prof_dt.agg.w$neg_flip)
    expect_equal(prof_dt.agg.w$pos_noFlip, prof_dt.agg.w$PE_noFlip)
    expect_equal(prof_dt.agg.w$pos_noFlip, prof_dt.agg.w$PE_flip)
})

testthat::context("return_unprocessed")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
library(data.table)

qgr = CTCF_in_10a_overlaps_gr#[1:5]
k = unique(as.character(seqnames(qgr)))
seqlevels(qgr) = k
qgr = centerFixedSizeGRanges(qgr, 500)

exp_factor = 100
qgr = resize(qgr, seq_along(qgr)*exp_factor, fix = "center")
#bed used to intersect bam
# rtracklayer::export.bed(qgr, con = "ctcf_5.bed")
bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
bw_file = system.file("extdata/test_loading.bw", package = "seqsetvis", mustWork = TRUE)
win_size = 50

exp_cn = c("which_label", "seqnames",
          "qname", "strand", "start",
          "width", "cigar", "isize", "id",
          "sample", "flag", "mapq",
          "mrnm", "mpos", "seq", "qual")

test_that("ssvFetchBamPE - return_unprocessed", {
    pe_file = system.file("extdata/Bcell_PE.mm10.bam", package = "seqsetvis", mustWork = TRUE)
    data("Bcell_peaks")
    qgrPE = Bcell_peaks
    qgrPE$name = seq_along(qgrPE)
    qgrPE = resize(qgrPE, seq_along(qgrPE)*exp_factor, fix = "center")
    raw_dt = ssvFetchBamPE(pe_file, qgrPE, return_unprocessed = TRUE)
    expect_s3_class(raw_dt, "data.table")
    expect_setequal(colnames(raw_dt),
                    exp_cn)
})

test_that("ssvFetchBam - return_unprocessed", {
    raw_dt = ssvFetchBam(bam_file, qgr,
                         win_size = win_size,
                         force_skip_centerFix = TRUE,
                         return_data.table = TRUE,
                         return_unprocessed = TRUE)
    expect_s3_class(raw_dt, "data.table")
    expect_setequal(colnames(raw_dt),
                    exp_cn)
})

test_that("ssvFetchBamPE - expandCigar", {
    pe_file = system.file("extdata/Bcell_PE.mm10.bam", package = "seqsetvis", mustWork = TRUE)
    data("Bcell_peaks")
    qgrPE = Bcell_peaks
    qgrPE$name = seq_along(qgrPE)
    qgrPE = resize(qgrPE, seq_along(qgrPE)*exp_factor, fix = "center")
    raw_dt = ssvFetchBamPE(pe_file, qgrPE, return_unprocessed = TRUE)
    exp_gr = expandCigar(raw_dt)
    expect_s4_class(exp_gr, "GRanges")
    expect_setequal(colnames(mcols(exp_gr)), c("which_label", "cigar_type"))
    exp_dt = expandCigar(raw_dt, return_data.table = TRUE)
    expect_s3_class(exp_dt, "data.table")
    expect_setequal(colnames(exp_dt), c("which_label", "seqnames", "strand", "start", "end", "cigar_type"))
})

test_that("ssvFetchBam - expandCigar", {
    raw_dt = ssvFetchBam(bam_file, qgr,
                         win_size = win_size,
                         force_skip_centerFix = TRUE,
                         return_data.table = TRUE,
                         return_unprocessed = TRUE)
    exp_gr = expandCigar(raw_dt)
    expect_s4_class(exp_gr, "GRanges")
    expect_setequal(colnames(mcols(exp_gr)), c("which_label", "cigar_type"))
    exp_dt = expandCigar(raw_dt, return_data.table = TRUE)
    expect_s3_class(exp_dt, "data.table")
    expect_setequal(colnames(exp_dt), c("which_label", "seqnames", "strand", "start", "end", "cigar_type"))
})

testthat::context("FetchBam")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
library(data.table)

qgr = CTCF_in_10a_overlaps_gr[1:5]
qgr = centerFixedSizeGRanges(qgr, 500)
#bed used to intersect bam
# rtracklayer::export.bed(qgr, con = "ctcf_5.bed")
bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
xls_file = system.file("extdata/test_peaks.xls", package = "seqsetvis", mustWork = TRUE)

# ssvFetchBam(bam_file, qgr)

test_that("fragLen_fromMacs2Xls", {
    fl = fragLen_fromMacs2Xls(xls_file)
    expect_equal(fl, 187)
})

test_that("fetchBam auto fragLen", {
    expect_message(seqsetvis:::fetchBam(bam_file, qgr = qgr), "fragLen was calculated as:")
})

test_that("fragLen_calcStranded parameters matter", {
    # res2 = fragLen_calcStranded(bam_file, qgr = qgr, n_regions = 5, include_plot_in_output = TRUE)
    # res1 = fragLen_calcStranded(bam_file, qgr = qgr, n_regions = 1, include_plot_in_output = TRUE)
    #
    # f1 = res2[[1]]
    # f2 = res1[[1]]
    # expect_failure(expect_equal(f1, f2))

    res5 = fragLen_calcStranded(bam_file, qgr = qgr, test_fragLen = 180, include_plot_in_output = TRUE)
    res6 = fragLen_calcStranded(bam_file, qgr = qgr, test_fragLen = 200, include_plot_in_output = TRUE)

    f5 = res5[[1]]
    f6 = res6[[1]]
    expect_failure(expect_equal(f5, f6))
})

test_that("viewGRangesWinSample_dt strand and position functions", {
    bam_score = seqsetvis:::fetchBam(bam_file, qgr = qgr)
    score_gr = bam_score
    window_size = 50
    anchor = "center"
    qgr_stranded = qgr
    GenomicRanges::strand(qgr_stranded) = c(rep("+", 2), rep("-", 3))

    b_dt_center = viewGRangesWinSample_dt(bam_score, qgr_stranded, 50, anchor = "center")

    b_dt_center_uns = viewGRangesWinSample_dt(bam_score, qgr_stranded, 50, anchor = "center_unstranded")

    b_dt_left = viewGRangesWinSample_dt(bam_score, qgr_stranded, 50, anchor = "left")

    b_dt_left_uns = viewGRangesWinSample_dt(bam_score, qgr_stranded, 50, anchor = "left_unstranded")

    b_dt = rbindlist(list(center = b_dt_center,
                          center_unstranded = b_dt_center_uns,
                          left = b_dt_left,
                          left_unstranded = b_dt_left_uns), use.names = TRUE, idcol = "group")
    b_dt[, facet_label := paste(id, strand)]
    cowplot::plot_grid(
        ggplot(b_dt[grepl("center", group)]) + geom_path(aes(x = x, y = y, color = group)) + facet_wrap("facet_label") + labs(title = "centered"),
        ggplot(b_dt[!grepl("center", group)]) + geom_path(aes(x = x, y = y, color = group)) + facet_wrap("facet_label") + labs(title = "left")

    )


    #verify stranded equal for + and not equal for -
    expect_true(all(b_dt_center[strand == "+"]$x == b_dt_center_uns[strand == "+"]$x))
    expect_true(all(!b_dt_center[strand == "-"]$x == b_dt_center_uns[strand == "-"]$x))
    expect_true(all(b_dt_left[strand == "+"]$x == b_dt_left_uns[strand == "+"]$x))
    expect_true(all(!b_dt_left[strand == "-"]$x == b_dt_left_uns[strand == "-"]$x))
    #verify center and left not equal
    expect_true(all(!b_dt_center$x == b_dt_left$x))
    expect_true(all(!b_dt_center_uns$x == b_dt_left_uns$x))
    #verify center and left ARE equal if minus min
    expect_true(all(b_dt_center$x - min(b_dt_center$x) == b_dt_left$x - min(b_dt_left$x)))
    expect_true(all(b_dt_center_uns$x - min(b_dt_center_uns$x) == b_dt_left_uns$x - min(b_dt_left_uns$x)))
})

# #strand of qgr should be ignored for pileup in favor of target strand
# #strand of qgr should be used for feature orientation
# test_that("viewGRangesWinSample_dt strand and position functions", {
#     bam_score_unstranded = fetchBam(bam_file, qgr = qgr)
#     score_gr = bam_score
#     window_size = 50
#
#     qgr_stranded = qgr
#     GenomicRanges::strand(qgr_stranded) = c(rep("+", 2), rep("-", 3))
#
#     bam_score_stranded = fetchBam(bam_file, qgr = qgr_stranded)
#     plot(bam_score_unstranded$score, bam_score_stranded$score)
# })


test_that("ssvFetchBam query GRanges output id set", {
    skip_on_os("windows")
    test_gr = qgr
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true("id" %in% colnames(gr_sample))

    names(test_gr) = NULL
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true("id" %in% colnames(gr_sample))

    test_gr$id = seq_along(test_gr) #when id set, id in output is missing
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true("id" %in% colnames(gr_sample))

    test_gr = qgr
    test_gr$id = seq_along(test_gr) #when id set, id in output is missing
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true("id" %in% colnames(gr_sample))

})

test_that("ssvFetchBam query GRanges name gets used", {
    test_gr = qgr
    names(test_gr) = paste("test_region", seq_along(test_gr))
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_gr, return_data.table = TRUE)
    expect_true(all(unique(gr_sample$id) == unique(names(test_gr))))
})

test_that("ssvFetchBam bam_file as data.frame/table", {
    test_gr = CTCF_in_10a_narrowPeak_grs$MCF10A_CTCF
    config_df = data.frame(rep(bam_file,3),
                           colors = c("red", "green", "blue"),
                           sample = c("10a", "10b", "10c"))
    gr_sample = ssvFetchBam(
        config_df,
        # fragLens = 180,
        win_size = 5, qgr = test_gr,
        return_data.table = TRUE)
    expect_true(all(levels(gr_sample$colors) %in%
                        c("red", "green", "blue")))
    expect_true(all(levels(gr_sample$sample) %in%
                        c("10a", "10b", "10c")))
})

test_that("ssvFetchBam sample method correct bins", {
    skip_on_os("windows")
    test_qgr2 = qgr
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_qgr2)
    #for non-overlapping, all reduce granges width should be equal to input qgr
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    #for non-overlapping, all granges width should be equal to win_size
    expect_true(all(width(gr_sample) == 5))

    #same as above but width are auto adjusted to be divisible by win_size
    gr_sample = ssvFetchBam(bam_file, win_size = 3, qgr = test_qgr2)
    expect_true(all(width(reduce(gr_sample)) == 501))
    expect_true(all(width(gr_sample) == 3))

    ###varying widths
    end(test_qgr2) = end(test_qgr2) + seq_along(test_qgr2)*2
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_qgr2, win_method = "sample")
    #for non-overlapping, all reduce granges width should be equal to input qgr
    expect_true(all(width(reduce(gr_sample)) == 510))
    #for non-overlapping, all granges width should be equal to win_size
    expect_true(all(width(gr_sample) == 5))

    #same as above but individual widths now vary while total is unchanged
    gr_sample = ssvFetchBam(bam_file, win_size = 4, qgr = test_qgr2, win_method = "sample")
    expect_true(all(width(reduce(gr_sample)) == 508))
    expect_true(all(width(gr_sample) == 4))
})

test_that("ssvFetchBam summary method correct bins", {
    skip_on_os("windows")
    test_qgr2 = qgr
    width(test_qgr2)
    ###invariant widths
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_qgr2, win_method = "summary")
    #for non-overlapping, all reduce granges width should be equal to input qgr
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    #for non-overlapping, all granges width should be equal to win_size
    expect_true(all(width(gr_sample) == 500 / 5))

    #same as above but individual widths now vary while total is unchanged
    gr_sample = ssvFetchBam(bam_file, win_size = 3, qgr = test_qgr2, win_method = "summary")
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    expect_true(all(width(gr_sample) %in% 166:167))

    ###varying widths
    end(test_qgr2) = end(test_qgr2) + seq_along(test_qgr2)*2
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_qgr2, win_method = "summary")
    #for non-overlapping, all reduce granges width should be equal to input qgr
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    #for non-overlapping, all granges width should be equal to win_size
    expect_true(all(width(gr_sample) %in% 100:102))

    #same as above but individual widths now vary while total is unchanged
    gr_sample = ssvFetchBam(bam_file, win_size = 3, qgr = test_qgr2, win_method = "summary")
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    expect_true(all(width(gr_sample) %in% 167:170))
})


test_that("ssvFetchBam summary method correct bins", {
    skip_on_os("windows")
    test_qgr2 = qgr

    ###invariant widths
    gr_sample = ssvFetchBam(bam_file, win_size = 5,
                            qgr = test_qgr2, win_method = "summary")
    #for non-overlapping, all reduce granges width should be equal to input qgr
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    #for non-overlapping, all granges width should be equal to win_size
    expect_true(all(width(gr_sample) == 500 / 5))

    #same as above but individual widths now vary while total is unchanged
    gr_sample = ssvFetchBam(bam_file, win_size = 3, qgr = test_qgr2, win_method = "summary")
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    expect_true(all(width(gr_sample) %in% 166:167))

    ###varying widths
    end(test_qgr2) = end(test_qgr2) + seq_along(test_qgr2)*2
    gr_sample = ssvFetchBam(bam_file, win_size = 5, qgr = test_qgr2, win_method = "summary")
    #for non-overlapping, all reduce granges width should be equal to input qgr
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    #for non-overlapping, all granges width should be equal to win_size
    expect_true(all(width(gr_sample) %in% 100:102))

    #same as above but individual widths now vary while total is unchanged
    gr_sample = ssvFetchBam(bam_file, win_size = 3, qgr = test_qgr2, win_method = "summary")
    expect_true(all(width(reduce(gr_sample)) == width(test_qgr2)))
    expect_true(all(width(gr_sample) %in% 167:170))
})

test_that(".rm_dupes removes duplicates", {
    dt = data.table(which_label = 1:10, seqnames = "chr1", strand = c("+", "-"), start = 1:10, end = 1:10+10)
    dt[, width := end - start + 1]
    make_dupes = seq_len(nrow(dt))
    make_dupes = rep(make_dupes, make_dupes)
    dt = dt[make_dupes]
    dtl = lapply(1:10, function(x)seqsetvis:::.rm_dupes(dt, max_dupes = x))
    # max_dupes 1 should yield 10 unique entries
    expect_true(all(dtl[[1]]$which_label == 1:10))
    # max_dupes 10 should perform no dupe removal in this case
    expect_true(all(dtl[[10]]$which_label == make_dupes))
    lens = sapply(dtl, nrow)
    # as max_dupes increases, returned entries increases
    expect_true(all(lens[-length(lens)] - lens[-1] < 0))
})

test_that("ssvFetchBam removes duplicates if max_dupes set", {
    rFull = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, win_method = "summary")$y
    r1 = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, win_method = "summary", max_dupes = 1)$y
    r2 = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, win_method = "summary", max_dupes = 2)$y
    expect_true(all(r2 >= r1))
    expect_true(all(!r2 < r1))
    expect_true(all(rFull >= r1))
    expect_true(all(!rFull < r1))
})

test_that("ssvFetchBam strand of qgr does matter", {
    # * == + and * == rev(-)
    qres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 200)
    pgr = qgr
    strand(pgr) = "+"
    pres = ssvFetchBam(bam_file, win_size = 5, qgr = pgr[1], fragLens = 200)
    ngr = qgr
    strand(ngr) = "-"
    nres = ssvFetchBam(bam_file, win_size = 5, qgr = ngr[1], fragLens = 200)
    expect_equal(pres$y,  qres$y)
    expect_equal(nres$y,  rev(qres$y))
})

test_that("ssvFetchBam strand of target_strand does matter", {
    strand(qgr) = "*"
    qres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, fragLens = 200, target_strand = "*")
    pres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, fragLens = 200, target_strand = "+")
    nres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, fragLens = 200, target_strand = "-")

    expect_true(!all(pres$y == qres$y))
    expect_true(!all(nres$y == qres$y))
    expect_true(all(nres$y + pres$y == qres$y))

    strand(qgr) = "+"
    qres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, fragLens = 200, target_strand = "*")
    pres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, fragLens = 200, target_strand = "+")
    nres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, fragLens = 200, target_strand = "-")

    expect_true(!all(pres$y == qres$y))
    expect_true(!all(nres$y == qres$y))
    expect_true(all(nres$y + pres$y == qres$y))

    strand(qgr) = "-"
    qres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, fragLens = 200, target_strand = "*")
    pres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, fragLens = 200, target_strand = "+")
    nres = ssvFetchBam(bam_file, win_size = 5, qgr = qgr, fragLens = 200, target_strand = "-")

    expect_true(!all(pres$y == qres$y))
    expect_true(!all(nres$y == qres$y))
    expect_true(all(nres$y + pres$y == qres$y))
})

test_that("ssvFetchBam target_strand of both - sample", {
    strand(qgr) = "*"
    res_both = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 200,
                           target_strand = "both", return_data.table = TRUE)
    strand(qgr) = "+"
    res_pos = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 200,
                          target_strand = "both", return_data.table = TRUE)
    strand(qgr) = "-"
    res_neg = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 200,
                          target_strand = "both", return_data.table = TRUE)
    res_neg[, x := -x]
    res_neg = res_neg[order(x)][order(id)][order(strand)]

    expect_true(nrow(res_both[strand == "+"]) == nrow(res_both[strand == "-"]))
    expect_true(!all(res_both[strand == "+"]$y == res_both[strand == "-"]))
    expect_true(all(res_both[strand == "+"] == res_pos[strand == "+"]))
    expect_true(all(res_both[strand == "-"] == res_pos[strand == "-"]))
    expect_true(all(res_both[strand == "+"]$y == res_neg[strand == "-"]$y))
    expect_true(all(res_both[strand == "-"]$y == res_neg[strand == "+"]$y))

})

test_that("ssvFetchBam target_strand of both - summary", {
    strand(qgr) = "*"
    res_both = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 200,
                           target_strand = "both", return_data.table = TRUE,
                           win_method = "summary")
    strand(qgr) = "+"
    res_pos = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 200,
                          target_strand = "both", return_data.table = TRUE,
                          win_method = "summary")
    strand(qgr) = "-"
    res_neg = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 200,
                          target_strand = "both", return_data.table = TRUE,
                          win_method = "summary")
    res_neg[, x := -x]
    res_neg = res_neg[order(x)][order(id)][order(strand)]

    expect_true(nrow(res_both[strand == "+"]) == nrow(res_both[strand == "-"]))
    expect_true(!all(res_both[strand == "+"]$y == res_both[strand == "-"]))
    expect_true(all(res_both[strand == "+"] == res_pos[strand == "+"]))
    expect_true(all(res_both[strand == "-"] == res_pos[strand == "-"]))
    expect_true(all(res_both[strand == "+"]$y == res_neg[strand == "-"]$y))
    expect_true(all(res_both[strand == "-"]$y == res_neg[strand == "+"]$y))

})

test_that("ssvFetchBam single column data.table", {
    qdt = data.table(file = bam_file)
    res = ssvFetchBam(qdt,
                      qgr = qgr[1],
                      win_size = 5, fragLens = 200,
                      target_strand = "both", return_data.table = TRUE,
                      win_method = "summary")
    expect_equal(levels(res$sample)[1], bam_file)
})

test_that("ssvFetchBam sample target_strand of both - fragLens", {
    strand(qgr) = "*"
    res_100 = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 100,
                          target_strand = "both", return_data.table = TRUE)
    # strand(qgr) = "+"
    res_150 = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 150,
                          target_strand = "both", return_data.table = TRUE)
    # strand(qgr) = "-"
    res_200 = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 200,
                          target_strand = "both", return_data.table = TRUE)
    # ggplot(rbind(res_100[, group := "100"],
    #              res_150[, group := "150"],
    #              res_200[, group := "200"]),
    #        aes(x = x, y = y, color = strand, group = group)) +
    #     geom_path() + facet_wrap("strand")
    expect_gt(res_200[x == 2 & strand == "+"]$y, res_100[x == 2 & strand == "+"]$y)
    expect_gt(res_200[x == 2 & strand == "-"]$y, res_100[x == 2 & strand == "-"]$y)

    expect_equal(res_200[x == -248 & strand == "+"]$y, res_100[x == -248 & strand == "+"]$y)
    expect_equal(res_200[x == 247 & strand == "-"]$y, res_100[x == 247 & strand == "-"]$y)
})

test_that("ssvFetchBam summary target_strand of both - fragLens", {
    strand(qgr) = "*"
    res_100 = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 100, win_method = "summary",
                          target_strand = "both", return_data.table = TRUE)
    # strand(qgr) = "+"
    res_150 = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 150, win_method = "summary",
                          target_strand = "both", return_data.table = TRUE)
    # strand(qgr) = "-"
    res_200 = ssvFetchBam(bam_file, win_size = 5, qgr = qgr[1], fragLens = 200, win_method = "summary",
                          target_strand = "both", return_data.table = TRUE)
    # ggplot(rbind(res_100[, group := "100"],
    #              res_150[, group := "150"],
    #              res_200[, group := "200"]),
    #        aes(x = x, y = y, color = strand, group = group)) +
    #     geom_path() + facet_wrap("strand")
    expect_gt(res_200[x == 0 & strand == "+"]$y, res_100[x == 0 & strand == "+"]$y)
    expect_gt(res_200[x == 0 & strand == "-"]$y, res_100[x == 0 & strand == "-"]$y)
})

test_that("ssvFetchBam sample target_strand of both - fragLens", {
    strand(qgr) = "*"
    qdf = data.frame(file = bam_file, mark = "a")
    res_100 = ssvFetchBam(qdf, win_size = 5, qgr = qgr[1], fragLens = 100,
                          target_strand = "both", return_data.table = TRUE)
    # strand(qgr) = "+"
    res_150 = ssvFetchBam(qdf, win_size = 5, qgr = qgr[1], fragLens = 150,
                          target_strand = "both", return_data.table = TRUE)
    # strand(qgr) = "-"
    res_200 = ssvFetchBam(qdf, win_size = 5, qgr = qgr[1], fragLens = 200,
                          target_strand = "both", return_data.table = TRUE)
    # ggplot(rbind(res_100[, group := "100"],
    #              res_150[, group := "150"],
    #              res_200[, group := "200"]),
    #        aes(x = x, y = y, color = strand, group = group)) +
    #     geom_path() + facet_wrap("strand")
    expect_gt(res_200[x == 2 & strand == "+"]$y, res_100[x == 2 & strand == "+"]$y)
    expect_gt(res_200[x == 2 & strand == "-"]$y, res_100[x == 2 & strand == "-"]$y)

    expect_equal(res_200[x == -248 & strand == "+"]$y, res_100[x == -248 & strand == "+"]$y)
    expect_equal(res_200[x == 247 & strand == "-"]$y, res_100[x == 247 & strand == "-"]$y)
})

test_that("ssvFetchBam file_paths as data.table", {
    strand(qgr) = "*"
    qdf = data.frame(file = bam_file, mark = "a")
    res_qdf = ssvFetchBam(qdf, win_size = 5, qgr = qgr[1], fragLens = 100,
                          target_strand = "both", return_data.table = TRUE)

    qdt = as.data.table(qdf)
    res_qdt = ssvFetchBam(qdt, win_size = 5, qgr = qgr[1], fragLens = 100,
                          target_strand = "both", return_data.table = TRUE)

    expect_equal(res_qdf, res_qdt)
})

test_that("ssvFetchBam file_attribs as data.table", {
    strand(qgr) = "*"
    qdf = data.frame(file = bam_file, mark = "a")
    res_qdf = ssvFetchBam(qdf$file, file_attribs = qdf[,1, drop = FALSE], win_size = 5, qgr = qgr[1], fragLens = 100,
                          target_strand = "both", return_data.table = TRUE)

    qdt = as.data.table(qdf)
    res_qdt = ssvFetchBam(qdt$file, file_attribs = qdt[,1], win_size = 5, qgr = qgr[1], fragLens = 100,
                          target_strand = "both", return_data.table = TRUE)
    expect_equal(res_qdf, res_qdt)
})



test_that("ssvFetchBam strand sensitivity for overlapping features on opposite strands", {
    this_qgr = qgr[rep(1, 2)]
    strand(this_qgr) = c("+", "-")
    this_qgr$id = seq_along(this_qgr)
    res_qdf = ssvFetchBam(bam_file, win_size = 5, qgr = this_qgr, fragLens = 100,
                          target_strand = "both", return_data.table = TRUE)
    # ggplot(res_qdf, aes(x = x, y = y, color = strand)) +
    #     geom_path() +
    #     facet_wrap(~id)

    expect_equal(res_qdf[strand == "+"]$y, rev(res_qdf[strand == "-"]$y))
})

test_that("ssvFetchBam (target_strand '+', flip_strand TRUE) === (taget_strand '-', flip_strand FALSE", {
    this_qgr = qgr
    strand(this_qgr) = c("+", "+", "+", "+", "-")
    this_qgr = this_qgr[c(1, 4)]
    this_qgr$id = seq_along(this_qgr)
    res_both = ssvFetchBam(bam_file, win_size = 5, qgr = this_qgr, fragLens = 100,
                             target_strand = "both", flip_strand = FALSE,
                             return_data.table = TRUE)

    res_flip = ssvFetchBam(bam_file, win_size = 5, qgr = this_qgr, fragLens = 100,
                           target_strand = "+", flip_strand = TRUE,
                           return_data.table = TRUE)

    res_unflip = ssvFetchBam(bam_file, win_size = 5, qgr = this_qgr, fragLens = 100,
                           target_strand = "-", flip_strand = FALSE,
                           return_data.table = TRUE)

    res_flip_inv = ssvFetchBam(bam_file, win_size = 5, qgr = this_qgr, fragLens = 100,
                           target_strand = "-", flip_strand = TRUE,
                           return_data.table = TRUE)

    res_unflip_inv = ssvFetchBam(bam_file, win_size = 5, qgr = this_qgr, fragLens = 100,
                             target_strand = "+", flip_strand = FALSE,
                             return_data.table = TRUE)

    # ggplot() +
    #     geom_path(data = res_both, aes(x = x, y = y, group = paste(strand, id), color = strand)) +
    #     geom_path(data = res_flip, aes(x = x, y = y-2, group = id), color = "red") +
    #     geom_path(data = res_unflip, aes(x = x, y = y-4, group = id), color = "blue") +
    #     geom_path(data = res_flip_inv, aes(x = x, y = y-6, group = id), color = "orange") +
    #     geom_path(data = res_unflip_inv, aes(x = x, y = y-8, group = id), color = "green") +
    #     facet_wrap(~id)

    expect_equal(unique(as.character(res_flip$strand)), "+")
    expect_equal(unique(as.character(res_unflip$strand)), "-")
    expect_equal(unique(as.character(res_flip_inv$strand)), "-")
    expect_equal(unique(as.character(res_unflip_inv$strand)), "+")

    mismatch_val = sum(abs(res_flip$y - res_flip_inv$y))
    match_val = sum(abs(res_flip$y - res_unflip$y))
    expect_lt(match_val, mismatch_val*.1)

    mismatch_val2 = sum(abs(res_unflip$y - res_unflip_inv$y))
    match_val2 = sum(abs(res_flip_inv$y - res_unflip_inv$y))
    expect_lt(match_val2, mismatch_val2*.1)
})

valid_sort_strategies = c("hclust", "sort", "left", "right", "none", "reverse")

#' Clustering as for a heatmap.  This is used internally by
#' \code{\link{ssvSignalHeatmap}} but can also be run before calling
#' ssvSignalHeatmap for greater control and access to clustering results
#' directly.
#'
#' Clustering is via k-means by default.  The number of clusters is determined
#' by nclust. Optionally, k-means can be initialized with a data.frame provided
#' to k_centroids. As an alternative to k-means, a membership table from
#' \code{\link{ssvMakeMembTable}} can be provided to determine logical clusters.
#'
#' Within each cluster, items will either be sorted by decreasing average signal
#' or hierachically clustered; this is controlled via within_order_strategy.
#'
#' @export
#'
#' @param bw_data a GRanges or data.table of bigwig signal. As returned from
#'   \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}
#' @param nclust Number of clusters.  Defaults to 6 if nclust, k_centroids, and
#'   memb_table are not set.
#' @param k_centroids data.frame of centroids for k-means clusters. Incompatible
#'   with nclust or memb_table.
#' @param memb_table Membership table as from \code{\link{ssvMakeMembTable}}.
#'   Logical groups from membership table will be clusters. Incompatible with
#'   nclust or k_centroids.
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param column_ varaible mapped to column, likely bp position for ngs data.
#'   Default is "x" and works with ssvFetch* output.
#' @param fill_ numeric variable to map to fill. Default is "y" and works with
#'   ssvFetch* output.
#' @param facet_ variable name to facet horizontally by. Default is "sample" and
#'   works with ssvFetch* output. Set to "" if data is not facetted.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param max_rows for speed rows are sampled to 500 by default, use Inf to plot
#'   full data
#' @param max_cols for speed columns are sampled to 100 by default, use Inf to
#'   plot full data
#' @param clustering_col_min numeric minimum for col range considered when
#'   clustering, default in -Inf
#' @param clustering_col_max numeric maximum for col range considered when
#'   clustering, default in Inf
#' @param within_order_strategy one of "hclust", "sort", "right", "left",
#'   "none", "reverse".  If "hclust", hierarchical clustering will be used. If
#'   "sort", a simple decreasing sort of rosSums.  If "left", will attempt to
#'   put high signal on left ("right" is opposite).  If "none", existing order
#'   is preserved.  If "reverse" reverses existing order.
#' @param dcast_fill value to supply to dcast fill argument. default is NA.
#' @param iter.max Number of max iterations to allow for k-means. Default is 30.
#' @param fun.aggregate Function to aggregate when multiple values present for
#'   facet_, row_, and column_. The
#'   function should accept a single vector argument or be a character string
#'   naming such a function.
#' @rawNamespace import(data.table, except = c(shift, first, second, last))
#' @return data.table of signal profiles, ready for ssvSignalHeatmap
#' @examples
#' data(CTCF_in_10a_profiles_gr)
#' clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_gr)
#' ssvSignalHeatmap(clust_dt)
#'
#' clust_dt2 = ssvSignalClustering(CTCF_in_10a_profiles_gr, nclust = 2)
#' ssvSignalHeatmap(clust_dt2)
#'
#' #clustering can be targetted to a specific part of the region
#' clust_dt3 = ssvSignalClustering(CTCF_in_10a_profiles_gr, nclust = 2,
#'     clustering_col_min = -250, clustering_col_max = -150)
#' ssvSignalHeatmap(clust_dt3)
#'
#' # there are also multiple sorting strategies to apply within each cluster
#' clust_dt4 = ssvSignalClustering(
#'   CTCF_in_10a_profiles_gr,
#'   nclust = 2,
#'   within_order_strategy = "left"
#' )
#' ssvSignalHeatmap(clust_dt4)
#'
#' clust_dt5 = ssvSignalClustering(
#'   CTCF_in_10a_profiles_gr,
#'   nclust = 2,
#'   within_order_strategy = "sort"
#' )
#' ssvSignalHeatmap(clust_dt5)
ssvSignalClustering = function(bw_data,
                               nclust = NULL,
                               k_centroids = NULL,
                               memb_table = NULL,
                               row_ = "id",
                               column_ = "x",
                               fill_ = "y",
                               facet_ = "sample",
                               cluster_ = "cluster_id",
                               max_rows = 500,
                               max_cols = 100,
                               clustering_col_min = -Inf,
                               clustering_col_max = Inf,
                               within_order_strategy = valid_sort_strategies[2],
                               dcast_fill = NA,
                               iter.max = 30,
                               fun.aggregate = "mean"){
    message("clustering...")
    id__ = xbp = x = to_disp = y = hit = val = y = y_gap = group__ =  NULL#declare binding for data.table
    output_GRanges = FALSE
    if(is(bw_data, "GRanges")){
        bw_data = data.table::as.data.table(bw_data)
        output_GRanges = TRUE
    }
    stopifnot(is.data.table(bw_data))
    if(!is.null(k_centroids) & !is.null(memb_table)){
        stop("only one of k_centroids or memb_table is allowed.")
    }
    if(!is.null(k_centroids)){
        nclust = nrow(k_centroids)
    }
    if(!is.null(memb_table)){ #memb_table get handled to provide row_ to cluster_ mapping
        if(is.data.frame(memb_table)){
            if(!is.null(memb_table[[row_]]) & !is.null(memb_table[[cluster_]])){
                #in this case memb_table is already a valid row_ to cluster_ mapping
                #nothing needs done
            }else{
                if(is.null(memb_table[[row_]])){
                    memb_table[[row_]] = memb_table[["id"]]
                    memb_table[["id"]] = NULL
                }
                if(is.null(memb_table[[cluster_]])){
                    memb_table[[cluster_]] = memb_table[["group"]]
                    memb_table[["group"]] = NULL
                }
            }
        }else{
            memb_table = ssvFactorizeMembTable(memb_table)
            if(is.null(memb_table[[row_]])){
                memb_table[[row_]] = memb_table[["id"]]
                memb_table[["id"]] = NULL
            }
            if(is.null(memb_table[[cluster_]])){
                memb_table[[cluster_]] = memb_table[["group"]]
                memb_table[["group"]] = NULL
            }
        }
        if(is.null(memb_table[[row_]])){
            stop("Could not correctly handle memb_table.",
                 " It should look like output of ssvMakeMembTable or ",
                 "ssvFactorizeMembTable.")
        }
        if(is.null(memb_table[[cluster_]])){
            stop("Could not correctly handle memb_table. It should look like",
                 " output of ssvMakeMembTable or ssvFactorizeMembTable.")
        }
        nclust = length(unique(memb_table[[cluster_]]))
        cluster_assignment = memb_table[[cluster_]]
        names(cluster_assignment) = memb_table[[row_]]
    }else{
        cluster_assignment = NULL
    }
    if(is.null(nclust) & is.null(k_centroids) & is.null(memb_table)){
        nclust = 6
        # stop("one of nclust, k_centroids, or memb_table must be set.")
    }
    stopifnot(is.numeric(nclust))
    stopifnot(is.character(row_), is.character(column_), is.character(fill_),
              is.character(facet_), is.character(cluster_))
    stopifnot(row_ %in% colnames(bw_data), column_ %in% colnames(bw_data),
              fill_ %in% colnames(bw_data))
    stopifnot(facet_ %in% colnames(bw_data) || facet_ == "")
    stopifnot(is.numeric(max_rows), is.numeric(max_cols),
              is.numeric(clustering_col_min), is.numeric(clustering_col_max))

    plot_dt = data.table::copy(bw_data)
    dc_mat = make_clustering_matrix(
        plot_dt,
        row_ = row_,
        column_ = column_,
        fill_ = fill_,
        facet_ = facet_,
        max_rows = max_rows,
        max_cols = max_cols,
        clustering_col_min = clustering_col_min,
        clustering_col_max = clustering_col_max,
        dcast_fill = dcast_fill,
        fun.aggregate = fun.aggregate

    )

    if(!is.null(k_centroids)){
        if(!is.matrix(k_centroids)){
            if(!is.data.table(k_centroids)){
                stop("k_centroids must be a wide matrix of centroids or",
                     " a tidy data.table")
            }
            if(is.null(k_centroids[[cluster_]])){
                stop("k_centroids looks like a tidy data.table but is missing",
                     " the cluster variable ", cluster_)
            }
            k_centroids = make_clustering_matrix(
                k_centroids,
                row_ = cluster_,
                column_ = column_,
                fill_ = fill_, facet_ = facet_,
                max_rows = Inf,
                max_cols = max_cols,
                clustering_col_min = clustering_col_min,
                clustering_col_max = clustering_col_max,
                dcast_fill = dcast_fill
            )
        }
    }


    rclusters = clusteringKmeansNestedHclust(dc_mat,
                                             nclust = nclust,
                                             within_order_strategy,
                                             centroids = k_centroids,
                                             manual_mapping = cluster_assignment,
                                             iter.max = iter.max)
    rclusters = rclusters[rev(seq_len(nrow(rclusters))),]
    plot_dt = plot_dt[get(row_) %in% rclusters[["id__"]]]
    plot_dt[[row_]] = factor(plot_dt[[row_]], levels = rev(rclusters[["id__"]]))
    data.table::setkey(rclusters, id__)
    plot_dt[[cluster_]] = rclusters[list(plot_dt[[row_]]), group__]
    agg_cn = c(row_, column_, facet_)
    agg_cn = agg_cn[agg_cn != ""]
    agg_required = max(plot_dt[, .N, c(agg_cn)]$N) > 1
    if(agg_required){
        #apply fun.aggregate to plotted data
        constant_cn = c("seqnames", "start", "end", 'width',
                        "strand", "cluster_id")
        constant_cn = constant_cn[constant_cn %in% colnames(plot_dt)]
        if(is.character(fun.aggregate)){
            fun.aggregate = get(fun.aggregate)
        }
        plot_dt.agg = plot_dt[,
                              fun.aggregate(get(fill_)),
                              c(unique(c(agg_cn, constant_cn)))]
        setnames(plot_dt.agg, "V1", fill_)
        message("Aggregation has occured (multiple samples per facet),",
                " some variables have been dropped.")
    }else{
        plot_dt.agg = plot_dt
    }
    if(output_GRanges){
        plot_dt.agg = GRanges(plot_dt.agg)
    }
    return(plot_dt.agg)
}

#' perform kmeans clustering on matrix rows and return reordered matrix along
#' with order matched cluster assignments. clusters are sorted using hclust on
#' centers
#'
#' @param mat numeric matrix to cluster.
#' @param nclust the number of clusters.
#' @param centroids optional matrix with same columns as mat and one centroid
#'   per row to base clusters off of.  Overrides any setting to nclust. Default
#'   of NULL results in randomly initialized k-means.
#' @param iter.max Number of max iterations to allow for k-means. Default is 30.
#'
#' @return data.table with group__ variable indicating cluster membership and
#'   id__ variable that is a factor indicating order based on within cluster
#'   similarity
#' @export
#' @importFrom stats kmeans hclust dist
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' dt = data.table::copy(CTCF_in_10a_profiles_dt)
#' mat = data.table::dcast(dt, id ~ sample + x, value.var = "y" )
#' rn = mat$id
#' mat = as.matrix(mat[,-1])
#' rownames(mat) = rn
#' clust_dt = clusteringKmeans(mat, nclust = 3)
#' dt = merge(dt, clust_dt[, .(id = id__, group = group__)])
#' dt$id = factor(dt$id, levels = clust_dt$id)
#' dt[order(id)]
clusteringKmeans = function(mat, nclust, centroids = NULL, iter.max = 30) {
    if(!is.null(centroids)){
        if(!all(colnames(mat) == colnames(centroids))){
            stop("provided centroids matrix did not have identical",
                 " column names to mat matrix.")
        }
        nclust = nrow(centroids)
        centers_arg = centroids
    }else{
        centers_arg = nclust
    }
    stopifnot(is.numeric(nclust))
    cluster_ordered = mat_name = NULL#declare binding for data.table
    if(nrow(mat) <= nclust){
        nclust = nrow(mat)
        mat_kmclust = list(
            centers = mat,
            cluster = seq(nrow(mat))
        )
        names(mat_kmclust$cluster) = seq(nrow(mat))

    }else{
        if(is.null(centroids)){
            if(nrow(unique(mat)) < nclust){
                nclust = nrow(unique(mat))
                warning("Reducing nclust to ", nclust,
                        " - maximum number of clusters allowed due to ",
                        "low uniqueness.")
                centers_arg = nclust
            }

        }
        mat_kmclust = stats::kmeans(mat, centers = centers_arg,
                                    iter.max = iter.max)
    }

    if(nclust < 3 || !is.null(centroids)){
        center_o = seq(nclust)
    }else{
        center_o = stats::hclust(stats::dist(mat_kmclust$centers))$order
    }
    center_reo = seq_along(center_o)
    names(center_reo) = center_o
    center_reo[as.character(mat_kmclust$cluster)]
    mat_dt = data.table::data.table(
        mat_name = names(mat_kmclust$cluster),
        cluster = mat_kmclust$cluster,
        cluster_ordered = center_reo[as.character(mat_kmclust$cluster)])
    mat_dt = mat_dt[order(cluster_ordered),
                    list(id__ = mat_name, group__ = cluster_ordered)]
    return(mat_dt)
}


#' perform kmeans clustering on matrix rows and return reordered matrix along
#' with order matched cluster assignments clusters are sorted using hclust on
#' centers the contents of each cluster are sorted using hclust
#' @param mat A wide format matrix
#' @param nclust the number of clusters
#' @param within_order_strategy one of "hclust", "sort", "right", "left",
#'   "reverse".  If "hclust", hierarchical clustering will be used. If "sort", a
#'   simple decreasing sort of rosSums.  If "left", will atttempt to put high
#'   signal on left ("right" is opposite).  If "reverse" reverses existing order
#'   (should only be used after meaningful order imposed).
#' @param centroids optional matrix with same columns as mat and one centroid
#'   per row to base clusters off of.  Overrides any setting to nclust. Default
#'   of NULL results in randomly initialized k-means.
#' @param manual_mapping optional named vector manually specififying cluster
#'   assignments. names should be item ids and values should be cluster names
#'   the items are assigned to. Default of NULL allows clustering to proceed.
#' @param iter.max Number of max iterations to allow for k-means. Default is 30.
#'
#' @export
#' @importFrom stats  hclust dist
#' @return data.table with 2 columns of cluster info. id__ column corresponds
#'   with input matrix rownames and is sorted within each cluster using
#'   hierarchical clusering group__ column indicates cluster assignment
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' dt = data.table::copy(CTCF_in_10a_profiles_dt)
#' mat = data.table::dcast(dt, id ~ sample + x, value.var = "y" )
#' rn = mat$id
#' mat = as.matrix(mat[,-1])
#' rownames(mat) = rn
#' clust_dt = clusteringKmeansNestedHclust(mat, nclust = 3)
#' clust_dt
clusteringKmeansNestedHclust = function(
        mat,
        nclust,
        within_order_strategy = valid_sort_strategies[2],
        centroids = NULL,
        manual_mapping = NULL,
        iter.max = 30) {
    stopifnot(is.numeric(nclust))
    stopifnot(within_order_strategy %in% valid_sort_strategies)
    group__ = id__ = within_o = NULL#declare binding for data.table

    if(nclust > (nrow(mat)-1) & is.null(centroids)){
        nclust = nrow(mat)-1
        message("nclust too high for number of items. Reducing to ", nclust)
    }

    if(is.null(manual_mapping)){
        mat_dt = clusteringKmeans(mat, nclust, centroids, iter.max)
    }else{
        manual_mapping
        mat_dt = data.table(id__ = names(manual_mapping),
                            group__ = manual_mapping)
        mat_dt = mat_dt[order(group__)]
    }

    within_clust_sort.mat_dt(mat_dt = mat_dt,
                             mat = mat,
                             within_order_strategy = within_order_strategy)
}


within_clust_sort.mat_dt = function(mat_dt, mat, within_order_strategy){
    group__ = id__ = within_o = NULL
    stopifnot(within_order_strategy %in% valid_sort_strategies)
    mat_dt$within_o = as.numeric(-1)
    for (i in unique(mat_dt$group__)) {
        cmat = mat[mat_dt[group__ == i, id__], , drop = FALSE]
        if(within_order_strategy == "hclust"){
            if (nrow(cmat) > 2) {
                mat_dt[group__ == i, ]$within_o =
                    stats::hclust(stats::dist((cmat)))$order
            } else {
                mat_dt[group__ == i, ]$within_o = seq_len(nrow(cmat))
            }
        }else if(within_order_strategy == "sort"){
            mat_dt[group__ == i, ]$within_o = rank(-rowSums(cmat))
        }else if(within_order_strategy == "left"){
            #cmat must be column ordered by position
            o = order(as.numeric(
                vapply(
                    strsplit(colnames(cmat), "_"),
                    function(x)x[length(x)],
                    FUN.VALUE = "1")))
            mat_dt[group__ == i, ]$within_o = rank(
                apply(cmat[, o], 1,
                      function(x){which.max(x)[1]})
            )
        }else if(within_order_strategy == "right"){
            #cmat must be column ordered by position
            o = order(as.numeric(
                vapply(
                    strsplit(colnames(cmat), "_"),
                    function(x)x[length(x)],
                    FUN.VALUE = "1")))
            mat_dt[group__ == i, ]$within_o = rank(
                -apply(cmat[, o], 1,
                       function(x){which.max(x)[1]})
            )
        }else if(within_order_strategy == "none"){
            mat_dt[group__ == i, ]$within_o = seq_len(nrow(cmat))
        }else if(within_order_strategy == "reverse"){
            mat_dt[group__ == i, ]$within_o = rev(seq_len(nrow(cmat)))
        }
    }
    mat_dt = mat_dt[order(within_o), ][order(group__), ]
    data.table::set(
        mat_dt,
        j = "id__",
        value = factor(mat_dt$id__, levels = mat_dt$id__)
    )
    mat_dt$within_o = NULL
    if(!is.factor(mat_dt$group__)){
        data.table::set(
            mat_dt,
            j = "group__",
            value = factor(mat_dt$group__, levels = unique(mat_dt$group__))
        )
    }
    mat_dt
}

#' within_clust_sort
#'
#' Without modifying cluster assignments, modify the order of rows within each
#' cluster based on within_order_strategy.
#'
#' This is particularly useful when you want to sort within each cluster by a
#' different variable from cluster assignment. Also if you've imported cluster
#' assigments but want to sort within each for the new data for a prettier
#' heatmap.
#'
#' TODO refactor shared code with clusteringKmeansNestedHclust
#'
#' @param clust_dt data.table output from \code{\link{ssvSignalClustering}}
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param column_ varaible mapped to column, likely bp position for ngs data.
#'   Default is "x" and works with ssvFetch* output.
#' @param fill_ numeric variable to map to fill. Default is "y" and works with
#'   ssvFetch* output.
#' @param facet_ variable name to facet horizontally by. Default is "sample" and
#'   works with ssvFetch* output. Set to "" if data is not facetted.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param clustering_col_min numeric minimum for col range considered when
#'   clustering, default in -Inf
#' @param clustering_col_max numeric maximum for col range considered when
#'   clustering, default in Inf
#' @param within_order_strategy one of "hclust", "sort", "right", "left",
#'   "reverse".  If "hclust", hierarchical clustering will be used. If "sort", a
#'   simple decreasing sort of rosSums.  If "left", will atttempt to put high
#'   signal on left ("right" is opposite).  If "reverse" reverses existing order
#'   (should only be used after meaningful order imposed).
#' @param dcast_fill value to supply to dcast fill argument. default is NA.
#'
#' @return data.table matching input clust_dt save for the reassignment of
#'   levels of row_ variable.
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' #clustering by relative value per region does a good job highlighting changes
#' #when then plotting raw values the order within clusters is not smooth
#' #this is a good situation to apply a separate sort within clusters.
#' prof_dt = CTCF_in_10a_profiles_dt
#' prof_dt = append_ynorm(prof_dt)
#' prof_dt[, y_relative := y_norm / max(y_norm), list(id)]
#'
#' clust_dt = ssvSignalClustering(prof_dt, fill_ = "y_relative")
#' clust_dt.sort = within_clust_sort(clust_dt)
#'
#' cowplot::plot_grid(
#'   ssvSignalHeatmap(clust_dt) +
#'     labs(title = "clustered by relative, sorted by relative"),
#'   ssvSignalHeatmap(clust_dt.sort) +
#'     labs(title = "clustered by relative, sorted by raw value")
#' )
#'
within_clust_sort = function(
        clust_dt,
        row_ = "id",
        column_ = "x",
        fill_ = "y",
        facet_ = "sample",
        cluster_ = "cluster_id",
        within_order_strategy = c("hclust", "sort", "left",
                                  "right", "none", "reverse")[2],
        clustering_col_min = -Inf,
        clustering_col_max = Inf,
        dcast_fill = NA) {
    output_GRanges = FALSE
    if(is(clust_dt, "GRanges")){
        clust_dt = data.table::as.data.table(clust_dt)
        output_GRanges = TRUE
    }else{
        clust_dt = data.table::copy(clust_dt)
    }
    stopifnot(is.data.table(clust_dt))
    mat = make_clustering_matrix(
        clust_dt,
        row_ = row_,
        column_ = column_,
        fill_ = fill_,
        facet_ = facet_,
        max_rows = Inf,
        max_cols = Inf,
        clustering_col_min = clustering_col_min,
        clustering_col_max = clustering_col_max,
        dcast_fill = dcast_fill
    )


    group__ = id__ = within_o = NULL#declare binding for data.table

    mat_dt = unique(clust_dt[, list(id__ = get(row_), group__ = get(cluster_))])
    mat_dt = mat_dt[order(id__)]
    stopifnot(all(mat_dt$id__ == rownames(mat)))
    mat_dt = within_clust_sort.mat_dt(
        mat_dt, mat,
        within_order_strategy = within_order_strategy
    )
    #repair var names
    data.table::set(
        clust_dt,
        j = row_,
        value = factor(clust_dt[[row_]], levels = levels(mat_dt$id__))
    )
    return(clust_dt[])
}

#' copy_clust_info
#'
#' @param target A data.table or GRanges returned from ssvFetch*, the target to
#'   which cluster info will be added.
#' @param to_copy A data.table or GRanges returned from ssvSignalClustering,
#'   from which to copy cluster if.
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#'
#' @return data.table or GRanges (whichever target is) containing row order and
#'   cluster assignment derived from to_copy. Suitable for ssvSignalHeatmap and
#'   related functions.
#' @export
#'
#' @examples
#' data(CTCF_in_10a_narrowPeak_grs)
#' data(CTCF_in_10a_overlaps_gr)
#' data(CTCF_in_10a_profiles_dt)
#' #this takes cluster info from signal and applies to peak hits to
#' #create a heatmap of peak hits clustered by signal.
#' clust_dt1 = ssvSignalClustering(CTCF_in_10a_profiles_dt)
#' peak_hit_gr = ssvFetchGRanges(
#'   CTCF_in_10a_narrowPeak_grs,
#'   qgr = CTCF_in_10a_overlaps_gr
#' )
#' peak_hit_gr.clust = copy_clust_info(peak_hit_gr, clust_dt1)
#' peak_hit_gr.clust$hit = peak_hit_gr.clust$y > 0
#' ssvSignalHeatmap(peak_hit_gr.clust, fill_ = "hit") +
#'   scale_fill_manual(values = c("FALSE" = "gray90", "TRUE" = "black"))
copy_clust_info = function(target, to_copy,
                           row_ = "id", cluster_ = "cluster_id"){
    output_GRanges = FALSE
    if(is(target, "GRanges")){
        target = data.table::as.data.table(target)
        output_GRanges = TRUE
    }else{
        target = data.table::copy(target)
    }
    to_copy = data.table::as.data.table(to_copy)

    stopifnot(!is.null(target[[row_]]))
    stopifnot(!is.null(to_copy[[row_]]))
    stopifnot(!is.null(to_copy[[cluster_]]))
    stopifnot(length(intersect(target[[row_]], to_copy[[row_]])) > 0)

    if(!all(target[[row_]] %in% to_copy[[row_]])){
        warning("Some '", row_,
                "' items missing from to_copy, items will be lost from target.")
    }

    target[[cluster_]] = NULL
    target = merge(target,
                   unique(to_copy[, c(row_, cluster_), with = FALSE]),
                   by = row_
    )
    data.table::set(
        target,
        j = row_,
        value = factor(target[[row_]], levels = levels(to_copy[[row_]]))
    )
    if(output_GRanges){
        target = GenomicRanges::GRanges(target)
    }
    target
}


#' reorder_clusters_stepdown
#'
#' Attempts to reorder clusters so that rows with highest signal on the left
#' relative to the right appear at the top. Signal should have a roughly
#' diagonal pattern in a "stepdown" pattern.
#'
#' This can be down by column (step_by_column = TRUE) which averages across
#' facets.  By facet (step_by_column = FALSE, step_by_facet = TRUE) which
#' averages all columns per facet. Or both column and facet (step_by_column =
#' TRUE, step_by_facet = TRUE), which does no averaging so it looks at the full
#' matrix as plotted.
#'
#' @param clust_dt data.table output from \code{\link{ssvSignalClustering}}
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param column_ varaible mapped to column, likely bp position for ngs data.
#'   Default is "x" and works with ssvFetch* output.
#' @param fill_ numeric variable to map to fill. Default is "y" and works with
#'   ssvFetch* output.
#' @param facet_ variable name to facet horizontally by. Default is "sample" and
#'   works with ssvFetch* output. Set to "" if data is not facetted.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param reapply_cluster_names If TRUE, clusters will be renamed according to
#'   new order instead of their original names. Default is TRUE.
#' @param step_by_column If TRUE, column is considered for left-right cluster
#'   balance. Default is TRUE.
#' @param step_by_facet If TRUE, facet is considered for left-right cluster
#'   balance. Default is FALSE.
#'
#' @return data.table as output from \code{\link{ssvSignalClustering}}
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 10)
#' new_dt = reorder_clusters_stepdown(clust_dt)
#' cowplot::plot_grid(
#'     ssvSignalHeatmap(clust_dt),
#'     ssvSignalHeatmap(new_dt)
#' )
reorder_clusters_stepdown = function(clust_dt,
                                     row_ = "id",
                                     column_ = "x",
                                     fill_ = "y",
                                     facet_ = "sample",
                                     cluster_ = "cluster_id",
                                     reapply_cluster_names = TRUE,
                                     step_by_column = TRUE,
                                     step_by_facet = FALSE
){
    new_dt = copy(clust_dt)
    if(facet_ != ""){
        if(!is.factor(new_dt[[facet_]])){
            new_dt[[facet_]] = factor(new_dt[[facet_]],
                                      levels = unique(new_dt[[facet_]]))
        }
    }
    agg_dt = new_dt[, list(y = mean(get(fill_))), c(cluster_, column_, ifelse(facet_ == "", character(), facet_))]
    agg_dt = agg_dt[order(get(column_))][order(get(facet_))]

    if(step_by_column & step_by_facet){
        wide_dt = dcast(agg_dt,
                        paste0(cluster_, "~", ifelse(facet_ == "", character(), paste0(facet_, "+")), column_),
                        value.var = fill_,
                        fun.aggregate = mean)
    }else if(step_by_column){
        wide_dt = dcast(agg_dt,
                        paste0(cluster_, "~", column_),
                        value.var = fill_,
                        fun.aggregate = mean)
    }else if(step_by_facet){
        wide_dt = dcast(agg_dt,
                        paste0(cluster_, "~", ifelse(facet_ == "", character(), facet_)),
                        value.var = fill_,
                        fun.aggregate = mean)
    }


    mat = as.matrix(wide_dt[,-1])
    rownames(mat) = wide_dt[[cluster_]]

    wm_o = sort(apply(mat, 1, function(x){
        weighted.mean(seq_along(x), x)
    }))
    reorder_clusters_manual(new_dt,
                            manual_order = names(wm_o),
                            cluster_ = cluster_, row_ = row_)
}



#' reorder_clusters_hclust
#'
#' Applies hierarchical clustering to centroids of clusters to reorder.
#'
#' @param clust_dt data.table output from \code{\link{ssvSignalClustering}}
#' @param hclust_result hclust result returned by a previous call of this function with
#'   identical paramters when return_hclust = TRUE.
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param column_ varaible mapped to column, likely bp position for ngs data.
#'   Default is "x" and works with ssvFetch* output.
#' @param fill_ numeric variable to map to fill. Default is "y" and works with
#'   ssvFetch* output.
#' @param facet_ variable name to facet horizontally by. Default is "sample" and
#'   works with ssvFetch* output. Set to "" if data is not facetted.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param reapply_cluster_names If TRUE, clusters will be renamed according to
#'   new order instead of their original names. Default is TRUE.
#' @param return_hclust If TRUE, return the result of hclust instead of the
#'   reordered clustering data.table. Default is FALSE.  Ignored if
#'   hclust_result is supplied.
#'
#' @return data.table as output from \code{\link{ssvSignalClustering}}
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 10)
#' new_dt = reorder_clusters_hclust(clust_dt)
#' cowplot::plot_grid(
#'     ssvSignalHeatmap(clust_dt),
#'     ssvSignalHeatmap(new_dt)
#' )
reorder_clusters_hclust = function(clust_dt,
                                   hclust_result = NULL,
                                   row_ = "id",
                                   column_ = "x",
                                   fill_ = "y",
                                   facet_ = "sample",
                                   cluster_ = "cluster_id",
                                   reapply_cluster_names = TRUE,
                                   return_hclust = FALSE){
    new_dt = copy(clust_dt)
    if(is.null(hclust_result)){
        agg_dt = new_dt[, list(y = mean(get(fill_))), c(cluster_, column_, ifelse(facet_ == "", character(), facet_))]
        setnames(agg_dt, "y", fill_)
        wide_dt = dcast(agg_dt, paste0(cluster_, "~", column_, ifelse(facet_ == "", character(), paste0("+", facet_))), value.var = fill_)
        mat = as.matrix(wide_dt[,-1])
        rownames(mat) = wide_dt[[cluster_]]
        hclust_result = hclust(dist(mat))
        if(return_hclust){
            return(hclust_result)
        }
    }
    clust_lev = levels(new_dt[[cluster_]])[hclust_result$order]
    reorder_clusters_manual(clust_dt,
                            manual_order = clust_lev,
                            cluster_ = cluster_,
                            row_ = row_,
                            reapply_cluster_names = reapply_cluster_names)
}
#


#' reorder_clusters_manual
#'
#' Manually applies a new order (top to bottom) for cluster using the result of
#' ssvSignalClustering.
#'
#' @param clust_dt data.table output from \code{\link{ssvSignalClustering}}
#' @param manual_order New order for clusters  Does not need to include all
#'   clusters.  Any colors not included will be at the bottom in their original
#'   order.
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param reapply_cluster_names If TRUE, clusters will be renamed according to
#'   new order instead of their original names. Default is TRUE.
#'
#' @return data.table as output from \code{\link{ssvSignalClustering}}
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 3)
#' new_dt = reorder_clusters_manual(clust_dt = clust_dt, manual_order = 2)
#' cowplot::plot_grid(
#'     ssvSignalHeatmap(clust_dt),
#'     ssvSignalHeatmap(new_dt)
#' )
reorder_clusters_manual = function(clust_dt,
                                   manual_order,
                                   row_ = "id",
                                   cluster_ = "cluster_id",
                                   reapply_cluster_names = TRUE){
    stopifnot(manual_order %in% clust_dt[[cluster_]])
    new_dt = copy(clust_dt[order(get(row_))])
    if(!is.factor(new_dt[[cluster_]])){
        new_dt[[cluster_]] = factor(new_dt[[cluster_]], levels = rev(unique(new_dt[[cluster_]])))
    }
    manual_levels = c(manual_order, setdiff(levels(new_dt[[cluster_]]), manual_order))
    new_dt[[cluster_]] = factor(new_dt[[cluster_]], levels = manual_levels)
    new_dt = new_dt[order(get(cluster_))]
    new_dt[[row_]] = factor(new_dt[[row_]], levels = unique(new_dt[[row_]]))
    if(reapply_cluster_names){
        levels(new_dt[[cluster_]]) = seq_along(unique(new_dt[[cluster_]]))
    }
    new_dt
}



#' merge_clusters
#'
#' @param clust_dt data.table output from \code{\link{ssvSignalClustering}}
#' @param to_merge Clusters to merge. Must be items in clust_dt variable defined
#'   by cluster_ parameter.
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param reapply_cluster_names If TRUE, clusters will be renamed according to
#'   new order instead of their original names. Default is TRUE.
#'
#' @return data.table as output from \code{\link{ssvSignalClustering}}
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' set.seed(0)
#' clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 6)
#' ssvSignalHeatmap(clust_dt)
#' agg_dt = clust_dt[, list(y = mean(y)), list(x, cluster_id, sample)]
#' ggplot(agg_dt, aes(x = x, y = y, color = sample)) +
#'   geom_path() +
#'   facet_grid(cluster_id~.)
#'
#' to_merge = c(2, 3, 5)
#' # debug(merge_clusters)
#' new_dt = merge_clusters(clust_dt, c(2, 3, 5), reapply_cluster_names = FALSE)
#' new_dt.relabel = merge_clusters(clust_dt, c(2, 3, 5), reapply_cluster_names = TRUE)
#' new_dt.relabel.sort = within_clust_sort(new_dt.relabel, within_order_strategy = "sort")
#'
#' table(clust_dt$cluster_id)
#' table(new_dt$cluster_id)
#'
#' cowplot::plot_grid(
#'   ssvSignalHeatmap(clust_dt) + labs(title = "original"),
#'   ssvSignalHeatmap(new_dt) + labs(title = "2,3,5 merged"),
#'   ssvSignalHeatmap(new_dt.relabel) + labs(title = "2,3,5 merged, renumbered"),
#'   ssvSignalHeatmap(new_dt.relabel.sort) + labs(title = "2,3,5 merged, renumbered and sorted")
#' )
#'
#'
#'
#'
merge_clusters = function(clust_dt,
                          to_merge,
                          row_ = "id",
                          cluster_ = "cluster_id",
                          reapply_cluster_names = TRUE){
    to_merge = as.character(to_merge)
    clust_lev = levels(clust_dt[[cluster_]])
    stopifnot(all(to_merge %in% clust_lev))
    first_clust = min(which(clust_lev %in% to_merge))
    new_order = union(clust_lev[seq(1, first_clust)], to_merge)

    new_dt = reorder_clusters_manual(clust_dt, manual_order = new_order, row_ = row_, cluster_ = cluster_, reapply_cluster_names = FALSE)
    new_name = paste(to_merge, collapse = "-")

    data.table::set(new_dt, i = which(new_dt[[cluster_]] %in% to_merge), j = cluster_, value = new_name)
    # new_dt[[row_]] = factor(new_dt[[row_]], levels = unique(new_dt[[row_]]))
    new_dt[[cluster_]] = factor(new_dt[[cluster_]], levels = unique(new_dt[[cluster_]]))
    if(reapply_cluster_names){
        levels(new_dt[[cluster_]]) = seq_along(unique(new_dt[[cluster_]]))
    }
    new_dt

}

#' reverse_clusters
#'
#' @param clust_dt data.table output from \code{\link{ssvSignalClustering}}
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param column_ varaible mapped to column, likely bp position for ngs data.
#'   Default is "x" and works with ssvFetch* output.
#' @param fill_ numeric variable to map to fill. Default is "y" and works with
#'   ssvFetch* output.
#' @param facet_ variable name to facet horizontally by. Default is "sample" and
#'   works with ssvFetch* output. Set to "" if data is not facetted.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param reverse_rows_within If TRUE, rows within clusters will be reversed as well. Default is TRUE.
#' @param reapply_cluster_names If TRUE, clusters will be renamed according to
#'   new order instead of their original names. Default is TRUE.
#'
#' @return data.table as output from \code{\link{ssvSignalClustering}}
#'
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' set.seed(0)
#' clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 3)
#' rev_dt = reverse_clusters(clust_dt)
#' rev_dt.no_relabel = reverse_clusters(clust_dt, reapply_cluster_names = FALSE)
#' rev_dt.not_rows = reverse_clusters(clust_dt, reverse_rows_within = FALSE)
#' cowplot::plot_grid(nrow = 1,
#'   ssvSignalHeatmap(clust_dt) + labs(title = "original"),
#'   ssvSignalHeatmap(rev_dt) + labs(title = "reversed"),
#'   ssvSignalHeatmap(rev_dt.no_relabel) + labs(title = "reversed, no relabel"),
#'   ssvSignalHeatmap(rev_dt.not_rows) + labs(title = "reversed, not rows")
#' )
reverse_clusters = function(clust_dt,
                            row_ = "id",
                            column_ = "x",
                            fill_ = "y",
                            facet_ = "sample",
                            cluster_ = "cluster_id",
                            reverse_rows_within = TRUE,
                            reapply_cluster_names = TRUE){
    new_dt = reorder_clusters_manual(
        clust_dt,
        manual_order = rev(levels(clust_dt[[cluster_]])),
        row_ = row_,
        cluster_ = cluster_,
        reapply_cluster_names = reapply_cluster_names
    )
    if(reverse_rows_within){
        new_dt = within_clust_sort(
            new_dt,
            row_ = row_,
            column_ = column_,
            fill_ = fill_,
            facet_ = facet_,
            cluster_ = cluster_, within_order_strategy = "reverse"
        )
    }
    new_dt
}

#' split_cluster
#'
#' Splits one specified cluster in number of new clusters determined by nclust
#'
#' @param clust_dt data.table output from \code{\link{ssvSignalClustering}}
#' @param to_split Cluster to split.
#' @param nclust Number of new clusters to create.
#' @param row_ variable name mapped to row, likely id or gene name for ngs data.
#'   Default is "id" and works with ssvFetch* output.
#' @param column_ varaible mapped to column, likely bp position for ngs data.
#'   Default is "x" and works with ssvFetch* output.
#' @param fill_ numeric variable to map to fill. Default is "y" and works with
#'   ssvFetch* output.
#' @param facet_ variable name to facet horizontally by. Default is "sample" and
#'   works with ssvFetch* output. Set to "" if data is not facetted.
#' @param cluster_ variable name to use for cluster info. Default is
#'   "cluster_id".
#' @param reapply_cluster_names If TRUE, clusters will be renamed according to
#'   new order instead of their original names. Default is TRUE.
#'
#' @return data.table as output from \code{\link{ssvSignalClustering}}
#' @export
#'
#' @examples
#' data(CTCF_in_10a_profiles_dt)
#' set.seed(0)
#' clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 3)
#' split_dt = split_cluster(clust_dt, to_split = 2, nclust = 3)
#' split_dt.no_rename = split_cluster(
#'   clust_dt,
#'   to_split = 2,
#'   nclust = 3,
#'   reapply_cluster_names = FALSE
#' )
#' cowplot::plot_grid(nrow = 1,
#'   ssvSignalHeatmap(clust_dt),
#'   ssvSignalHeatmap(split_dt),
#'   ssvSignalHeatmap(split_dt.no_rename)
#' )
#'
split_cluster = function(clust_dt,
                         to_split,
                         nclust = 2,
                         row_ = "id",
                         column_ = "x",
                         fill_ = "y",
                         facet_ = "sample",
                         cluster_ = "cluster_id",
                         reapply_cluster_names = TRUE){
    to_split = as.character(to_split)
    stopifnot(to_split %in% clust_dt[[cluster_]])
    split_dt = clust_dt[clust_dt[[cluster_]] == to_split]
    split_dt = ssvSignalClustering(split_dt,
                                   nclust = nclust,
                                   row_ = row_,
                                   column_ = column_,
                                   fill_ = fill_,
                                   facet_ = facet_,
                                   cluster_ = cluster_,
                                   max_rows = Inf,
                                   max_cols = Inf)
    new_lev = paste0(to_split, letters[seq_len(nclust)])
    levels(split_dt[[cluster_]]) = new_lev
    insert_at = which(levels(clust_dt[[cluster_]]) == to_split)

    new_dt =rbind(split_dt, clust_dt[clust_dt[[cluster_]] != to_split])
    new_dt$cluster_id = droplevels(new_dt$cluster_id)
    c_lev = levels(clust_dt[[cluster_]])
    new_lev = c(c_lev[seq_along(c_lev) < insert_at], new_lev, c_lev[seq_along(c_lev) > insert_at])

    stopifnot(setequal(new_lev, levels(new_dt$cluster_id)))
    new_dt$cluster_id = factor(new_dt$cluster_id, levels = new_lev)
    reorder_clusters_manual(new_dt,
                            manual_order = new_lev,
                            row_ = row_,
                            cluster_ = cluster_,
                            reapply_cluster_names = reapply_cluster_names)
}

testthat::context("cluster sorting")

library(seqsetvis)
library(testthat)
library(GenomicRanges)

test_object = CTCF_in_10a_profiles_dt
set.seed(0)
clust_dt = ssvSignalClustering(test_object, nclust = 3)
show_plots = FALSE

test_that("within_clust_sort", {
    resort_dt = within_clust_sort(clust_dt, within_order_strategy = "hclust")
    resort_dt2 = within_clust_sort(clust_dt, within_order_strategy = "left")
    resort_dt3 = within_clust_sort(clust_dt, within_order_strategy = "right")

    if(show_plots){
        cowplot::plot_grid(ssvSignalHeatmap(clust_dt) + labs(title = "sort"),
                           ssvSignalHeatmap(resort_dt) + labs(title = "hclust"),
                           ssvSignalHeatmap(resort_dt2) + labs(title = "left"),
                           ssvSignalHeatmap(resort_dt3) + labs(title = "right"))
    }


    #test id levels still the same set
    expect_setequal(levels(clust_dt$id), levels(resort_dt$id))
    expect_setequal(levels(clust_dt$id), levels(resort_dt2$id))
    expect_setequal(levels(clust_dt$id), levels(resort_dt3$id))

    #test id levels are in different order
    expect_failure(expect_equal(levels(clust_dt$id), levels(resort_dt$id)))
    expect_failure(expect_equal(levels(clust_dt$id), levels(resort_dt2$id)))
    expect_failure(expect_equal(levels(clust_dt$id), levels(resort_dt3$id)))

    all_assign = lapply(list(clust_dt, resort_dt, resort_dt2, resort_dt3), function(x){
        unique(x[, .(id, cluster_id)])
    })
    all_membs = lapply(all_assign, function(x){
        split(as.character(x$id), x$cluster_id)
    })

    #test clusters haven't changed
    for(i in seq_along(all_membs[[1]])){
        expect_equal(all_membs[[1]][[i]], all_membs[[2]][[i]])
        expect_equal(all_membs[[1]][[i]], all_membs[[3]][[i]])
        expect_equal(all_membs[[1]][[i]], all_membs[[4]][[i]])
    }
})

test_that("reorder_clusters_manual", {
    reo_dt = reorder_clusters_manual(clust_dt, manual_order = c(2, 3))
    reo_dt.no_reapply = reorder_clusters_manual(clust_dt, manual_order = c(2, 3), reapply_cluster_names = FALSE)

    if(show_plots){
        cowplot::plot_grid(ssvSignalHeatmap(clust_dt) + labs(title = "original"),
                           ssvSignalHeatmap(reo_dt) + labs(title = "reordered"),
                           ssvSignalHeatmap(reo_dt.no_reapply) + labs(title = "no reapply"))
    }

    #test id levels still the same set
    expect_setequal(levels(clust_dt$id), levels(reo_dt$id))
    expect_setequal(levels(clust_dt$id), levels(reo_dt.no_reapply$id))

    #test id levels are in different order
    expect_failure(expect_equal(levels(clust_dt$id), levels(reo_dt$id)))
    expect_failure(expect_equal(levels(clust_dt$id), levels(reo_dt.no_reapply$id)))

    all_assign = lapply(list(clust_dt[order(id)], reo_dt[order(id)], reo_dt.no_reapply[order(id)]), function(x){
        unique(x[, .(id, cluster_id)])
    })
    all_membs = lapply(all_assign, function(x){
        split(as.character(x$id), x$cluster_id)
    })

    #test clusters haven't changed
    for(i in as.character(seq_along(all_membs[[1]]))){
        #this cluster should change
        expect_failure(expect_equal(all_membs[[1]][[i]], all_membs[[2]][[i]]))
        #this cluster should not
        expect_equal(all_membs[[1]][[i]], all_membs[[3]][[i]])
        # expect_setequal(all_membs[[1]][[i]], all_membs[[3]][[i]])
    }
})

test_that("reorder_clusters_hclust", {
    set.seed(0)
    clust_dt10.raw = ssvSignalClustering(test_object, nclust = 10)
    clust_dt10 = reorder_clusters_manual(clust_dt10.raw, c(3, 10), reapply_cluster_names = FALSE)
    clust_dt10.hclust = reorder_clusters_hclust(clust_dt10)
    clust_dt10.hclust.no_reapply = reorder_clusters_hclust(clust_dt10, reapply_cluster_names = FALSE)

    if(show_plots){
        cowplot::plot_grid(nrow = 1,
                           ssvSignalHeatmap(clust_dt10.raw) + labs(title = "raw"),
                           ssvSignalHeatmap(clust_dt10) + labs(title = "original"),
                           ssvSignalHeatmap(clust_dt10.hclust) + labs(title = "reordered"),
                           ssvSignalHeatmap(clust_dt10.hclust.no_reapply) + labs(title = "no reapply"))
    }

    #test id levels still the same set
    expect_setequal(levels(clust_dt10$id), levels(clust_dt10.hclust$id))
    expect_setequal(levels(clust_dt10$id), levels(clust_dt10.hclust.no_reapply$id))

    #test id levels are in different order
    expect_failure(expect_equal(levels(clust_dt10$id), levels(clust_dt10.hclust$id)))
    expect_failure(expect_equal(levels(clust_dt10$id), levels(clust_dt10.hclust.no_reapply$id)))

    all_assign = lapply(list(clust_dt10[order(id)], clust_dt10.hclust[order(id)], clust_dt10.hclust.no_reapply[order(id)]), function(x){
        unique(x[, .(id, cluster_id)])
    })
    all_membs = lapply(all_assign, function(x){
        split(as.character(x$id), x$cluster_id)
    })

    num_fail_1 = 0
    num_fail_2 = 0
    #test clusters haven't changed
    for(i in as.character(seq_along(all_membs[[1]]))){
        #this cluster should change
        if(length(all_membs[[1]][[i]]) != length(all_membs[[2]][[i]])){
            num_fail_1 = num_fail_1 + 1
        }else if(all(all_membs[[1]][[i]] != all_membs[[2]][[i]])){
            num_fail_1 = num_fail_1 + 1
        }
        if(length(all_membs[[1]][[i]]) != length(all_membs[[3]][[i]])){
            num_fail_2 = num_fail_2 + 1
        }else if(all(all_membs[[1]][[i]] != all_membs[[3]][[i]])){
            num_fail_2 = num_fail_2 + 1
        }
    }
    expect_gt(num_fail_1, 0)
    expect_equal(num_fail_2, 0)
})

test_that("reorder_clusters_hclust", {
    set.seed(0)
    rev_dt = reverse_clusters(clust_dt)
    rev_dt.no_relabel = reverse_clusters(clust_dt, reapply_cluster_names = FALSE)
    rev_dt.not_rows = reverse_clusters(clust_dt, reverse_rows_within = FALSE)
    if(show_plots){
        cowplot::plot_grid(nrow = 1,
                           ssvSignalHeatmap(clust_dt) + labs(title = "original"),
                           ssvSignalHeatmap(rev_dt) + labs(title = "reversed"),
                           ssvSignalHeatmap(rev_dt.no_relabel) + labs(title = "reversed, no relabel"),
                           ssvSignalHeatmap(rev_dt.not_rows) + labs(title = "reversed, not rows")
        )
    }
    expect_equal(levels(clust_dt$id), rev(levels(rev_dt$id)))
    expect_equal(levels(clust_dt$id), rev(levels(rev_dt.no_relabel$id)))
    expect_failure(expect_equal(levels(clust_dt$id), rev(levels(rev_dt.not_rows$id))))

    expect_setequal(clust_dt[cluster_id == 1]$id, rev_dt.no_relabel[cluster_id == 1]$id)
    expect_failure(expect_setequal(clust_dt[cluster_id == 1]$id, rev_dt[cluster_id == 1]$id))
    expect_failure(expect_setequal(clust_dt[cluster_id == 1]$id, rev_dt.not_rows[cluster_id == 1]$id))
})

test_that("split_cluster", {
    clust_dt = clust_dt[order(id)]
    set.seed(0)
    split_dt = split_cluster(clust_dt, to_split = 2, nclust = 3)
    set.seed(0)
    split_dt.no_rename = split_cluster(clust_dt, to_split = 2, nclust = 3, reapply_cluster_names = FALSE)
    if(show_plots){
        cowplot::plot_grid(nrow = 1,
                           ssvSignalHeatmap(clust_dt),
                           ssvSignalHeatmap(split_dt),
                           ssvSignalHeatmap(split_dt.no_rename)
        )
    }
    expect_failure(expect_equal(levels(clust_dt$id), levels(split_dt$id)))
    expect_failure(expect_equal(levels(clust_dt$id), levels(split_dt.no_rename$id)))
    expect_setequal(levels(clust_dt$id), levels(split_dt$id))
    expect_setequal(levels(clust_dt$id), levels(split_dt.no_rename$id))

    expect_equal(as.character(clust_dt[cluster_id == 1]$id), as.character(split_dt[cluster_id == 1]$id))
    expect_equal(as.character(clust_dt[cluster_id == 3]$id), as.character(split_dt[cluster_id == 5]$id))

    expect_failure(expect_equal(as.character(clust_dt[cluster_id == 2]$id), as.character(split_dt[cluster_id == 2]$id)))

    expect_equal(as.character(split_dt.no_rename[cluster_id == "2a"]$id), as.character(split_dt[cluster_id == "2"]$id))
    expect_equal(as.character(split_dt.no_rename[cluster_id == "2b"]$id), as.character(split_dt[cluster_id == "3"]$id))
})

testthat::context("Cigar")
library(data.table)
library(testthat)
library(GenomicRanges)
library(seqsetvis)



bam_dt = fread(system.file("extdata/test_cigar_dt.csv", package = "seqsetvis"))
bam_dt[, start := start - min(start) + 1]
bam_dt = bam_dt[order(start)]
expand_cigar_dt = seqsetvis:::.expand_cigar_dt

test_that("expand_cigar_dt simple alignment", {
    bdt = bam_dt[cigar == "82M"]
    exp_dt = expand_cigar_dt(bdt)
    expect_equal(nrow(exp_dt), 1)
    expect_equal(exp_dt$end, end(IRanges(start = bdt$start, width = bdt$width)))
    expect_equal(exp_dt$start, bdt$start)
    expect_equal(exp_dt$cigar_type, "M")
})

test_that("expand_cigar_dt deletion spanning alignment", {
    bdt = bam_dt[cigar == "46M2D45M",]
    exp_dt = expand_cigar_dt(bdt)
    expect_equal(nrow(exp_dt), 3)
    expect_equal(exp_dt$end, c(46, 46+2, 46+2+45))
    expect_equal(exp_dt$start, c(1, 1+46, 1+46+2))
    expect_equal(exp_dt$cigar_type, c("M", "D", "M"))
})

test_that("expand_cigar_dt skip region alignment", {
    bdt = bam_dt[cigar == "63M5005N28M"]
    exp_dt = expand_cigar_dt(bdt)
    expect_equal(nrow(exp_dt), 2)
    exp_dt = expand_cigar_dt(bdt, op_2count = c("M", "D", "=", "X", "N"))
    expect_equal(nrow(exp_dt), 3)
    expect_equal(exp_dt$end, c(21179, 21179 + 5005, 21179 + 5005+28))
    expect_equal(exp_dt$start, c(21117, 21117+63, 21117+63+5005))
    expect_equal(exp_dt$cigar_type, c("M", "N", "M"))
})

test_that("expand_cigar_dt skip region multiple", {
    #there happen to be 6 reads spanning the same gap
    bdt = bam_dt[grepl("6397N", cigar)]
    exp_dt = expand_cigar_dt(bdt)
    expect_equal(nrow(exp_dt), 12)
    exp_dt = expand_cigar_dt(bdt, op_2count = c("M", "D", "=", "X", "N"))
    expect_equal(nrow(exp_dt), 18)
    expect_equal(sum(exp_dt$end == 723), 6)
    expect_equal(sum(exp_dt$end == 7120), 6)
    expect_equal(sum(exp_dt$start == 724), 6)
    expect_equal(sum(exp_dt$start == 7121), 6)
    expect_equal(sum(exp_dt$cigar_type == "M"), 12)
    expect_equal(sum(exp_dt$cigar_type == "N"), 6)
})

testthat::context("MakeMembTable")
library(seqsetvis)
library(testthat)
library(GenomicRanges)
set_a = 1:8
set_b = 5:9
set_c = 8:10

gr_a = GRanges("chr1", IRanges(set_a*10, set_a*10+3))
gr_b = GRanges("chr1", IRanges(set_b*10-2, set_b*10+3))
gr_c = GRanges("chr1", IRanges(set_c*10+-4, set_c*10+1))

expected_table = as.matrix(data.frame("set_A" = 1:10 %in% set_a,
                                      "set_B" = 1:10 %in% set_b,
                                      "set_C" = 1:10 %in% set_c,
                                      row.names = 1:10))

expected_table_named = as.matrix(data.frame("named_set_A" = 1:10 %in% set_a,
                                            "named_set_B" = 1:10 %in% set_b,
                                            "named_set_C" = 1:10 %in% set_c,
                                            row.names = 1:10))

set_a = as.character(set_a)
set_b = as.character(set_b)
set_c = as.character(set_c)

test_that("ssvMakeMembTable unnamed numeric set list", {
    sets_list = list(set_a, set_b, set_c)
    mt = ssvMakeMembTable(sets_list)
    expect_is(mt, "matrix")
    expect_equal(as.logical(mt), as.logical(expected_table))
    expect_equal(rownames(mt), rownames(expected_table))
    expect_equal(colnames(mt), colnames(expected_table))
})

test_that("ssvMakeMembTable unnamed character set list", {
    sets_list = list(set_a, set_b, set_c)
    mt = ssvMakeMembTable(sets_list)
    expect_is(mt, "matrix")
    expect_equal(as.logical(mt), as.logical(expected_table))
    expect_equal(rownames(mt), as.character(1:nrow(expected_table)))
    expect_equal(colnames(mt), colnames(expected_table))
})

test_that("ssvMakeMembTable unnamed matrix", {
    sets_list = list(set_a, set_b, set_c)
    memb_table = ssvMakeMembTable(sets_list)
    colnames(memb_table) = NULL
    mt = ssvMakeMembTable(memb_table)
    expect_is(mt, "matrix")
    expect_equal(as.logical(mt), as.logical(expected_table))
    expect_equal(rownames(mt), rownames(expected_table))
    expect_equal(colnames(mt), colnames(expected_table))
})

test_that("ssvMakeMembTable unnamed data.frame", {
    sets_list = list(set_a, set_b, set_c)
    memb_table = ssvMakeMembTable(sets_list)
    colnames(memb_table) = NULL
    memb_table = as.data.frame(memb_table)
    mt = ssvMakeMembTable(memb_table)
    expect_is(mt, "matrix")
    expect_equal(as.logical(mt), as.logical(expected_table))
    expect_equal(rownames(mt), rownames(expected_table))
    expect_equal(colnames(mt), colnames(expected_table))
})

test_that("ssvMakeMembTable unnamed list of GRanges", {
    sets_gr_list = list(gr_a, gr_b, gr_c)

    suppressMessages({mt = ssvMakeMembTable(sets_gr_list)})
    expect_is(mt, "matrix")
    expect_equal(as.logical(mt), as.logical(expected_table))
    expect_equal(rownames(mt), rownames(expected_table))
    expect_equal(colnames(mt), colnames(expected_table))
})

test_that("ssvMakeMembTable NAMED list of GRanges", {
    sets_gr_list = list("named_set_A" = gr_a, "named_set_B" = gr_b, "named_set_C" = gr_c)

    suppressMessages({mt = ssvMakeMembTable(sets_gr_list)})
    expect_is(mt, "matrix")
    expect_equal(as.logical(mt), as.logical(expected_table_named))
    expect_equal(rownames(mt), rownames(expected_table_named))
    expect_equal(colnames(mt), colnames(expected_table_named))
})

test_that("ssvMakeMembTable tolerates NA values with a warning", {
    expect_warning(ssvMakeMembTable(list(c("a", "b", NA), "a")), regexp = "Removing NA values")
    expect_warning(ssvMakeMembTable(list(c("a", "b", NA), NA)), regexp = "Removing NA values")
})

testthat::context("DataFromPlots")
#test return_data functionality for plotting functions.  added 10/23/18, v1.1.6

library(seqsetvis)
library(testthat)
library(GenomicRanges)

test_signals = CTCF_in_10a_profiles_dt
test_features = CTCF_in_10a_overlaps_gr


for (ssvFeatureFUN_name in c("ssvFeatureBars", "ssvFeatureEuler", "ssvFeaturePie", "ssvFeatureBinaryHeatmap", "ssvFeatureVenn")){
    test_that(paste(ssvFeatureFUN_name, "return_data"), {
        ssvFeatureFUN = get(ssvFeatureFUN_name)
        p_default = ssvFeatureFUN(test_features)
        p_dataFALSE = ssvFeatureFUN(test_features, return_data = FALSE)
        p_dataTRUE = ssvFeatureFUN(test_features, return_data = TRUE)

        expect_s3_class(p_default, "ggplot")
        expect_s3_class(p_dataFALSE, "ggplot")
        expect_s3_class(p_dataTRUE, "data.table")
    })

}

for (ssvSignalFUN_name in c("ssvSignalBandedQuantiles", "ssvSignalHeatmap", "ssvSignalLineplot", "ssvSignalLineplotAgg")){
    test_that(paste(ssvFeatureFUN_name, "return_data"), {
        ssvSignalFUN = get(ssvSignalFUN_name)
        p_default = ssvSignalFUN(test_signals)
        p_dataFALSE = ssvSignalFUN(test_signals, return_data = FALSE)
        p_dataTRUE = ssvSignalFUN(test_signals, return_data = TRUE)

        expect_s3_class(p_default, "ggplot")
        expect_s3_class(p_dataFALSE, "ggplot")
        expect_s3_class(p_dataTRUE, "data.table")
    })
}

ssvSignalFUN_name = "ssvSignalScatterplot"
test_that(paste(ssvFeatureFUN_name, "return_data"), {
    ssvSignalFUN = get(ssvSignalFUN_name)
    p_default = ssvSignalFUN(test_signals, x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF")
    p_dataFALSE = ssvSignalFUN(test_signals, x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF", return_data = FALSE)
    p_dataTRUE = ssvSignalFUN(test_signals, x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF", return_data = TRUE)

    expect_s3_class(p_default, "ggplot")
    expect_s3_class(p_dataFALSE, "ggplot")
    expect_s3_class(p_dataTRUE, "data.table")
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{reverse_clusters}
\alias{reverse_clusters}
\title{reverse_clusters}
\usage{
reverse_clusters(
  clust_dt,
  row_ = "id",
  column_ = "x",
  fill_ = "y",
  facet_ = "sample",
  cluster_ = "cluster_id",
  reverse_rows_within = TRUE,
  reapply_cluster_names = TRUE
)
}
\arguments{
\item{clust_dt}{data.table output from \code{\link{ssvSignalClustering}}}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{column_}{varaible mapped to column, likely bp position for ngs data.
Default is "x" and works with ssvFetch* output.}

\item{fill_}{numeric variable to map to fill. Default is "y" and works with
ssvFetch* output.}

\item{facet_}{variable name to facet horizontally by. Default is "sample" and
works with ssvFetch* output. Set to "" if data is not facetted.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{reverse_rows_within}{If TRUE, rows within clusters will be reversed as well. Default is TRUE.}

\item{reapply_cluster_names}{If TRUE, clusters will be renamed according to
new order instead of their original names. Default is TRUE.}
}
\value{
data.table as output from \code{\link{ssvSignalClustering}}
}
\description{
reverse_clusters
}
\examples{
data(CTCF_in_10a_profiles_dt)
set.seed(0)
clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 3)
rev_dt = reverse_clusters(clust_dt)
rev_dt.no_relabel = reverse_clusters(clust_dt, reapply_cluster_names = FALSE)
rev_dt.not_rows = reverse_clusters(clust_dt, reverse_rows_within = FALSE)
cowplot::plot_grid(nrow = 1,
  ssvSignalHeatmap(clust_dt) + labs(title = "original"),
  ssvSignalHeatmap(rev_dt) + labs(title = "reversed"),
  ssvSignalHeatmap(rev_dt.no_relabel) + labs(title = "reversed, no relabel"),
  ssvSignalHeatmap(rev_dt.not_rows) + labs(title = "reversed, not rows")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_CTCF.R
\docType{data}
\name{CTCF_in_10a_data}
\alias{CTCF_in_10a_data}
\title{CTCF ChIP-seq in breast cancer cell lines}
\description{
Vignette data for seqsetvis was downloaded directly from GEO
series
\href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE98551}{GSE98551}.
This data is CTCF ChIP-seq from a model of breast cancer progression
derived from the MCF10A cell line.

Data from GEO series
\href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE98551}{GSE98551}
is from the publication
\href{https://www.ncbi.nlm.nih.gov/pubmed/28504305}{Fritz AJ et al. 2018}
}
\details{
Contains: \itemize{
\item \code{\link{CTCF_in_10a_overlaps_gr}}
\item \code{\link{CTCF_in_10a_profiles_dt}}
\item \code{\link{CTCF_in_10a_bigWig_urls}}
\item \code{\link{CTCF_in_10a_narrowPeak_urls}}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_chromHMM.R
\docType{data}
\name{chromHMM_demo_state_total_widths}
\alias{chromHMM_demo_state_total_widths}
\title{state name to total width mappings, hg38}
\format{
named numeric of total widths per state
}
\value{
named numeric of total widths per state
}
\description{
state name to total width mappings, hg38
}
\details{
part of \code{\link{chromHMM_demo_data}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{reorder_clusters_stepdown}
\alias{reorder_clusters_stepdown}
\title{reorder_clusters_stepdown}
\usage{
reorder_clusters_stepdown(
  clust_dt,
  row_ = "id",
  column_ = "x",
  fill_ = "y",
  facet_ = "sample",
  cluster_ = "cluster_id",
  reapply_cluster_names = TRUE,
  step_by_column = TRUE,
  step_by_facet = FALSE
)
}
\arguments{
\item{clust_dt}{data.table output from \code{\link{ssvSignalClustering}}}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{column_}{varaible mapped to column, likely bp position for ngs data.
Default is "x" and works with ssvFetch* output.}

\item{fill_}{numeric variable to map to fill. Default is "y" and works with
ssvFetch* output.}

\item{facet_}{variable name to facet horizontally by. Default is "sample" and
works with ssvFetch* output. Set to "" if data is not facetted.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{reapply_cluster_names}{If TRUE, clusters will be renamed according to
new order instead of their original names. Default is TRUE.}

\item{step_by_column}{If TRUE, column is considered for left-right cluster
balance. Default is TRUE.}

\item{step_by_facet}{If TRUE, facet is considered for left-right cluster
balance. Default is FALSE.}
}
\value{
data.table as output from \code{\link{ssvSignalClustering}}
}
\description{
Attempts to reorder clusters so that rows with highest signal on the left
relative to the right appear at the top. Signal should have a roughly
diagonal pattern in a "stepdown" pattern.
}
\details{
This can be down by column (step_by_column = TRUE) which averages across
facets.  By facet (step_by_column = FALSE, step_by_facet = TRUE) which
averages all columns per facet. Or both column and facet (step_by_column =
TRUE, step_by_facet = TRUE), which does no averaging so it looks at the full
matrix as plotted.
}
\examples{
data(CTCF_in_10a_profiles_dt)
clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 10)
new_dt = reorder_clusters_stepdown(clust_dt)
cowplot::plot_grid(
    ssvSignalHeatmap(clust_dt),
    ssvSignalHeatmap(new_dt)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{ssvSignalClustering}
\alias{ssvSignalClustering}
\title{Clustering as for a heatmap.  This is used internally by
\code{\link{ssvSignalHeatmap}} but can also be run before calling
ssvSignalHeatmap for greater control and access to clustering results
directly.}
\usage{
ssvSignalClustering(
  bw_data,
  nclust = NULL,
  k_centroids = NULL,
  memb_table = NULL,
  row_ = "id",
  column_ = "x",
  fill_ = "y",
  facet_ = "sample",
  cluster_ = "cluster_id",
  max_rows = 500,
  max_cols = 100,
  clustering_col_min = -Inf,
  clustering_col_max = Inf,
  within_order_strategy = valid_sort_strategies[2],
  dcast_fill = NA,
  iter.max = 30,
  fun.aggregate = "mean"
)
}
\arguments{
\item{bw_data}{a GRanges or data.table of bigwig signal. As returned from
\code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}}

\item{nclust}{Number of clusters.  Defaults to 6 if nclust, k_centroids, and
memb_table are not set.}

\item{k_centroids}{data.frame of centroids for k-means clusters. Incompatible
with nclust or memb_table.}

\item{memb_table}{Membership table as from \code{\link{ssvMakeMembTable}}.
Logical groups from membership table will be clusters. Incompatible with
nclust or k_centroids.}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{column_}{varaible mapped to column, likely bp position for ngs data.
Default is "x" and works with ssvFetch* output.}

\item{fill_}{numeric variable to map to fill. Default is "y" and works with
ssvFetch* output.}

\item{facet_}{variable name to facet horizontally by. Default is "sample" and
works with ssvFetch* output. Set to "" if data is not facetted.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{max_rows}{for speed rows are sampled to 500 by default, use Inf to plot
full data}

\item{max_cols}{for speed columns are sampled to 100 by default, use Inf to
plot full data}

\item{clustering_col_min}{numeric minimum for col range considered when
clustering, default in -Inf}

\item{clustering_col_max}{numeric maximum for col range considered when
clustering, default in Inf}

\item{within_order_strategy}{one of "hclust", "sort", "right", "left",
"none", "reverse".  If "hclust", hierarchical clustering will be used. If
"sort", a simple decreasing sort of rosSums.  If "left", will attempt to
put high signal on left ("right" is opposite).  If "none", existing order
is preserved.  If "reverse" reverses existing order.}

\item{dcast_fill}{value to supply to dcast fill argument. default is NA.}

\item{iter.max}{Number of max iterations to allow for k-means. Default is 30.}

\item{fun.aggregate}{Function to aggregate when multiple values present for
facet_, row_, and column_. The
function should accept a single vector argument or be a character string
naming such a function.}
}
\value{
data.table of signal profiles, ready for ssvSignalHeatmap
}
\description{
Clustering is via k-means by default.  The number of clusters is determined
by nclust. Optionally, k-means can be initialized with a data.frame provided
to k_centroids. As an alternative to k-means, a membership table from
\code{\link{ssvMakeMembTable}} can be provided to determine logical clusters.
}
\details{
Within each cluster, items will either be sorted by decreasing average signal
or hierachically clustered; this is controlled via within_order_strategy.
}
\examples{
data(CTCF_in_10a_profiles_gr)
clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_gr)
ssvSignalHeatmap(clust_dt)

clust_dt2 = ssvSignalClustering(CTCF_in_10a_profiles_gr, nclust = 2)
ssvSignalHeatmap(clust_dt2)

#clustering can be targetted to a specific part of the region
clust_dt3 = ssvSignalClustering(CTCF_in_10a_profiles_gr, nclust = 2,
    clustering_col_min = -250, clustering_col_max = -150)
ssvSignalHeatmap(clust_dt3)

# there are also multiple sorting strategies to apply within each cluster
clust_dt4 = ssvSignalClustering(
  CTCF_in_10a_profiles_gr,
  nclust = 2,
  within_order_strategy = "left"
)
ssvSignalHeatmap(clust_dt4)

clust_dt5 = ssvSignalClustering(
  CTCF_in_10a_profiles_gr,
  nclust = 2,
  within_order_strategy = "sort"
)
ssvSignalHeatmap(clust_dt5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_plots.R
\name{ssvSignalBandedQuantiles}
\alias{ssvSignalBandedQuantiles}
\title{plot profiles from bigwigs}
\usage{
ssvSignalBandedQuantiles(
  bw_data,
  y_ = "y",
  x_ = "x",
  by_ = "fake",
  hsv_reverse = FALSE,
  hsv_saturation = 1,
  hsv_value = 1,
  hsv_grayscale = FALSE,
  hsv_hue_min = 0,
  hsv_hue_max = 0.7,
  hsv_symmetric = FALSE,
  n_quantile = 18,
  quantile_min = 0.05,
  quantile_max = 0.95,
  return_data = FALSE
)
}
\arguments{
\item{bw_data}{a GRanges or data.table of bigwig signal.
As returned from \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}}

\item{y_}{the variable name in bw_data for y axis in plot}

\item{x_}{the variable name in bw_data for x axis in plot}

\item{by_}{the variable name in bw_data to facet on}

\item{hsv_reverse}{logical, should color scale be reversed? default FALSE}

\item{hsv_saturation}{numeric value from 0 to 1. Saturation for color scale. default 1}

\item{hsv_value}{numeric value from 0 to 1. Value for color scale. default 1}

\item{hsv_grayscale}{logical, if TRUE gray() is used instead of rainbow(). default FALSE}

\item{hsv_hue_min}{numeric [0, hsv_hue_max) hue min of color scale}

\item{hsv_hue_max}{numeric (hsv_hue_min, 1] hue max of color scale}

\item{hsv_symmetric}{if TRUE, colorscale is symmetrical, default FALSE.}

\item{n_quantile}{number of evenly size quantile bins}

\item{quantile_min}{the lowest quantile start}

\item{quantile_max}{the highest quantile end}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and
is instead the data used to generate that plot. Default is FALSE.}
}
\value{
ggplot object using ribbon plots to show quantile distributions
}
\description{
plot profiles from bigwigs
}
\examples{
data(CTCF_in_10a_profiles_gr)
#rainbow colors
qgr = CTCF_in_10a_profiles_gr
ssvSignalBandedQuantiles(qgr)
#grayscale
ssvSignalBandedQuantiles(qgr, hsv_grayscale = TRUE,
    hsv_symmetric = TRUE, hsv_reverse = TRUE)
#using "by_" per sample
ssvSignalBandedQuantiles(qgr, hsv_grayscale = TRUE,
    hsv_symmetric = TRUE, hsv_reverse = TRUE, by_ = "sample")
#adding spline smoothing
splined = applySpline(qgr, n = 10,
    by_ = c("id", "sample"))
ssvSignalBandedQuantiles(splined, n_quantile = 50,
    quantile_min = .25, quantile_max = .75,
    hsv_symmetric = TRUE, hsv_reverse = TRUE, by_ = "sample")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_CTCF.R
\docType{data}
\name{Bcell_peaks}
\alias{Bcell_peaks}
\title{4 random peaks for paired-end data}
\format{
GRanges length 4
}
\value{
GRanges length 4
}
\description{
matches \code{system.file("extdata/Bcell_PE.mm10.bam", package = "seqsetvis")}
}
\details{
this is included only for testing ssvFetchBamPE functions.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bamPE_RNA.R
\name{ssvFetchBamPE.RNA}
\alias{ssvFetchBamPE.RNA}
\title{ssvFetchBamPE.RNA}
\usage{
ssvFetchBamPE.RNA(
  file_paths,
  qgr,
  unique_names = NULL,
  win_size = 50,
  target_strand = "both",
  absolute_strand = FALSE,
  splice_strategy = "ignore",
  return_data.table = FALSE,
  win_method = "sample",
  max_dupes = Inf,
  flip_strand = FALSE,
  sum_reads = TRUE,
  n_cores = getOption("mc.cores", 1),
  force_skip_centerFix = TRUE,
  n_region_splits = 1
)
}
\arguments{
\item{file_paths}{character vector of file_paths to load from. Alternatively,
file_paths can be a data.frame or data.table whose first column is a
character vector of paths and additial columns will be used as metadata.}

\item{qgr}{Set of GRanges to query. For valid results the width of each
interval should be identical and evenly divisible by win_size.}

\item{unique_names}{names to use in final data.table to designate source
bigwig. Default is 'sample'}

\item{win_size}{The window size that evenly divides widths in qgr.}

\item{target_strand}{character. if one of "+" or "-", reads are filtered to
match. ignored if any other value.}

\item{absolute_strand}{If TRUE, strandedness of \code{qgr} will be ignored. This
is useful when creating tracks for similar.}

\item{splice_strategy}{character, one of c("none", "ignore", "add", "only",
"splice_count"). Default is "none" and spliced alignment are asssumed not
present. fragLen must be NA for any other value to be valid. "ignore" will
not count spliced regions. add" counts spliced regions along with others,
"only" will only count spliced regions and ignore others.}

\item{return_data.table}{logical. If TRUE the internal data.table is returned
instead of GRanges. Default is FALSE.}

\item{win_method}{character. one of c("sample", "summary"). "sample" selects
values at intervals and "summary" applies a weight mean function to all
values in window.}

\item{max_dupes}{numeric >= 1. duplicate reads by strandd start position over
this number are removed, Default is Inf.}

\item{flip_strand}{logical. if TRUE strands are flipped.}

\item{sum_reads}{logical. If true R1 and R2 reads are added together. If
FALSE they are returned separately, identified by the "read" attribute.}

\item{n_cores}{integer number of cores to use. Uses mc.cores option if not
supplied.}

\item{force_skip_centerFix}{boolean, if TRUE all query ranges will be used
"as is". This is already the case by default if win_method == "summary" but
may have applications where win_method == "sample".}

\item{n_region_splits}{integer number of splits to apply to qgr. The query
GRanges will be split into this many roughly equal parts for increased
parallelization. Default is 1, no split.}
}
\value{
A tidy formatted GRanges (or data.table if specified) containing
fetched values.
}
\description{
ssvFetchBamPE.RNA
}
\examples{
library(GenomicRanges)
pkg_dir = system.file(package = "seqsetvis", "extdata", mustWork = TRUE)
bam_files_esr1 = dir(pkg_dir, pattern = "H1.+R1.ESR1_RNA.+bam$", full.names = TRUE)
names(bam_files_esr1) = sub("_R.+", "", basename(bam_files_esr1))
query_gr = GenomicRanges::GRanges("chr6:151656691-152129619:+")
query_gr = GenomicRanges::GRanges("chr6:152116691-152129619:+")

strand(query_gr) = "+"

prof_dt = ssvFetchBamPE.RNA(bam_files_esr1, query_gr, return_data.table = TRUE, win_size = 1)
prof_dt
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_signal.R
\name{prepare_fetch_GRanges_width}
\alias{prepare_fetch_GRanges_width}
\title{prepares GRanges for windowed fetching.}
\usage{
prepare_fetch_GRanges_width(
  qgr,
  win_size,
  min_quantile = 0.75,
  target_size = NULL,
  skip_centerFix = FALSE
)
}
\arguments{
\item{qgr}{GRanges to prepare}

\item{win_size}{numeric window size for fetch}

\item{min_quantile}{numeric value from 0 to 1. Lowest possible quantile value.  Only
relevant if target_size is not specified.}

\item{target_size}{numeric final width of qgr if known. Default of NULL
leads to quantile based determination of target_size.}

\item{skip_centerFix}{boolean, if FALSE (default) all regions will be resized
GenomicRanges::resize(x, w, fix = "center") to a uniform size based on
min_quantile to a width divisible by win_size.}
}
\value{
GRanges, either identical to qgr or with suitable consistent width
applied.
}
\description{
output GRanges parallels input with consistent width evenly divisible by
win_size.  Has warning if GRanges needed resizing, otherwise no warning
and input GRanges is returned unchanged.
}
\examples{
data(CTCF_in_10a_overlaps_gr)
qgr = prepare_fetch_GRanges_width(CTCF_in_10a_overlaps_gr, win_size = 50)
#no warning if qgr is already valid for windowed fetching
prepare_fetch_GRanges_width(qgr, win_size = 50)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_easyLoad.R
\name{easyLoad_bed}
\alias{easyLoad_bed}
\title{easyLoad_bed takes a character vector of file paths to bed plus files and
returning named list of GRanges.}
\usage{
easyLoad_bed(
  file_paths,
  file_names = NULL,
  extraCols = character(),
  n_cores = getOption("mc.cores", 1)
)
}
\arguments{
\item{file_paths}{character vector of paths to narrowPeak files.  If named,
those names will be used in output unless overriden by providing file_names.}

\item{file_names}{character vector of names for output list.  If not NULL
will override any existing names for file_paths.  Default is NULL.}

\item{extraCols}{named character vector of classes.  passed to
rtracklayer::import for format = "BED". default is character().}

\item{n_cores}{number of cores to use, uses mc.cores option if set or 1.}
}
\value{
a named list of GRanges loaded from file_paths
}
\description{
Mainly a utility function for loading MACS2 narrowPeak and broadPeak.
}
\examples{
bed_f = system.file("extdata/test_loading.bed",
    package = "seqsetvis", mustWork = TRUE)
easyLoad_bed(bed_f, "my_bed")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_helpers.R
\name{ggellipse}
\alias{ggellipse}
\title{ggellipse}
\usage{
ggellipse(
  xcentres,
  ycentres,
  r,
  r2 = r,
  phi = rep(0, length(xcentres)),
  circle_colors = NULL,
  group_names = LETTERS[seq_along(xcentres)],
  line_alpha = 1,
  fill_alpha = 0.3,
  line_width = 2,
  n_points = 200
)
}
\arguments{
\item{xcentres}{numeric x-coord of centers of ellipses}

\item{ycentres}{numeric y-coord of centers of ellipses, must have same
length as xcentres}

\item{r}{numeric radius1 of ellipse, must have length of 1 or match length
of xcentres}

\item{r2}{numeric radius2 of ellipse, must have length of 1 or match length
of xcentres.  same as r by default.}

\item{phi}{numeric phi of ellipse, must have length of 1 or match length
of xcentres.  0 by default.}

\item{circle_colors}{character of rcolors or hex colors or NULL.  if
null safeBrew of Dark2 is used}

\item{group_names}{character/factor names of color/fill groups.  capital
letters by default.}

\item{line_alpha}{numeric value from 0 to 1. alpha of lines, 1 by default}

\item{fill_alpha}{numeric value from 0 to 1. alpha of fill, .3 by default.}

\item{line_width}{numeric > 0.  passed to size. 2 by default}

\item{n_points}{integer > 1.  number of points to approximate circle with.
200 by default}
}
\value{
a ggplot containing ellipses
}
\description{
returns a ggplot with ellipses drawn using specified parameters
used by ssvFeatureVenn and ssvFeatureEuler
}
\details{
uses eulerr's non-exported ellipse drawing coordinate function
}
\examples{
ggellipse(xcentres = c(1, 1, 2),
    ycentres = c(2, 1, 1),
    r = c(1, 2, 1))
ggellipse(xcentres = c(1, 1, 2),
    ycentres = c(2, 1, 1),
    r = c(1, 2, 1),
    fill_alpha = 0,
    group_names = paste("set", 1:3))
ggellipse(xcentres = c(1, 1, 2),
    ycentres = c(2, 1, 1),
    r = c(1, 2, 1),
    circle_colors = c("red", "orange", "yellow"),
    line_alpha = 0,
    group_names = paste("set", 1:3))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bamPE.R
\name{ssvFetchBamPE}
\alias{ssvFetchBamPE}
\title{ssvFetchBam for paired-end ChIP-seq files.
Only concordant reads are considered, but this has been minimally tested, please verify.}
\usage{
ssvFetchBamPE(
  file_paths,
  qgr,
  unique_names = NULL,
  win_size = 50,
  win_method = c("sample", "summary")[1],
  summary_FUN = stats::weighted.mean,
  fragLens = "not_used",
  anchor = c("left", "left_unstranded", "center", "center_unstranded")[3],
  names_variable = "sample",
  return_data.table = FALSE,
  max_dupes = Inf,
  n_cores = getOption("mc.cores", 1),
  n_region_splits = 1,
  min_isize = 1,
  max_isize = Inf,
  return_unprocessed = FALSE,
  return_fragSizes = FALSE,
  force_skip_centerFix = FALSE,
  ...
)
}
\arguments{
\item{file_paths}{character vector of file_paths to load from. Alternatively,
file_paths can be a data.frame or data.table whose first column is a
character vector of paths and additial columns will be used as metadata.}

\item{qgr}{Set of GRanges to query.  For valid results the width of each
interval should be identical and evenly divisible by \code{win_size}.}

\item{unique_names}{names to use in final data.table to designate source
bigwig. Default is 'sample'}

\item{win_size}{The window size that evenly divides widths in \code{qgr}.}

\item{win_method}{character.  one of c("sample", "summary").  Determines if
\code{\link{viewGRangesWinSample_dt}} or
\code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
qgr.}

\item{summary_FUN}{function.  only relevant if win_method is "summary".
passed to \code{\link{viewGRangesWinSummary_dt}}.}

\item{fragLens}{never used by ssvFetchBamPE Ignore.}

\item{anchor}{character, one of c("center", "center_unstranded", "left",
"left_unstranded")}

\item{names_variable}{The column name where unique_names are stored.}

\item{return_data.table}{logical. If TRUE the internal data.table is returned
instead of GRanges.  Default is FALSE.}

\item{max_dupes}{numeric >= 1.  duplicate reads by strandd start position
over this number are removed, Default is Inf.}

\item{n_cores}{integer number of cores to use.}

\item{n_region_splits}{integer number of splits to apply to qgr. The query
GRanges will be split into this many roughly equal parts for increased
parallelization. Default is 1, no split.}

\item{min_isize}{integer. Read pairs must have an isize greater than or equal to this value.  Default is 1.}

\item{max_isize}{integer. Read pairs must have an isize less than or equal to this value.  Default is Inf.}

\item{return_unprocessed}{boolean. if TRUE returns read alignment in data.table. Default is FALSE.}

\item{return_fragSizes}{boolean. if TRUE returns fragment sizes for all reads per region.}

\item{force_skip_centerFix}{boolean, if TRUE all query ranges will be
used "as is".  This is already the case by default if win_method == "summary"
but may have applications where win_method == "sample".}

\item{...}{passed to Rsamtools::ScanBamParam()
Uses mc.cores option if not supplied.}
}
\value{
A tidy formatted GRanges (or data.table if specified) containing
fetched values.
}
\description{
Iterates a character vector (ideally named) and calls
\code{ssvFetchBamPE.single} on each.  Appends grouping variable to each
resulting data.table and uses rbindlist to efficiently combine results
}
\details{
#' In contrast to ssvFetchBam, extension of reads to estimated fragment size is
not an issue as each read pair represents a fragment of exact size.

\code{ssvFetchBamPE} iteratively calls \code{fetchWindowedBam.single}. See
\code{\link{ssvFetchBamPE.single}} for more info.

if \code{qgr} contains the range chr1:1-100 and \code{win_size} is
10, values from positions chr1 5,15,25...85, and 95 will be retrieved from
\code{bw_file}
}
\examples{
if(Sys.info()['sysname'] != "Windows"){
library(GenomicRanges)
bam_f = system.file("extdata/Bcell_PE.mm10.bam",
    package = "seqsetvis", mustWork = TRUE)
bam_files = c("a" = bam_f, "b" = bam_f)
data("Bcell_peaks")
qgr = Bcell_peaks
bw_gr = ssvFetchBamPE(bam_files, qgr, win_size = 10)
bw_gr2 = ssvFetchBamPE(as.list(bam_files), qgr, win_size = 10)

bw_dt = ssvFetchBamPE(bam_files, qgr, win_size = 10,
    return_data.table = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bigwig.R
\name{ssvFetchBigwig}
\alias{ssvFetchBigwig}
\title{Iterates a character vector (ideally named) and calls
\code{ssvFetchBigwig.single} on each.  Appends grouping variable to each
resulting data.table and uses rbindlist to efficiently combine results.}
\usage{
ssvFetchBigwig(
  file_paths,
  qgr,
  unique_names = NULL,
  names_variable = "sample",
  win_size = 50,
  win_method = c("sample", "summary")[1],
  summary_FUN = stats::weighted.mean,
  fragLens = "not_used",
  anchor = c("left", "left_unstranded", "center", "center_unstranded")[3],
  return_data.table = FALSE,
  n_cores = getOption("mc.cores", 1),
  n_region_splits = 1,
  force_skip_centerFix = FALSE
)
}
\arguments{
\item{file_paths}{character vector of file_paths to load from. Alternatively,
file_paths can be a data.frame or data.table whose first column is a
character vector of paths and additial columns will be used as metadata.}

\item{qgr}{Set of GRanges to query.  For valid results the width of each
interval should be identical and evenly divisible by \code{win_size}.}

\item{unique_names}{names to use in final data.table to designate source
bigwig.}

\item{names_variable}{The column name where unique_names are stored.
Default is 'sample'}

\item{win_size}{The window size that evenly divides widths in \code{qgr}.}

\item{win_method}{character.  one of c("sample", "summary").  Determines
if \code{\link{viewGRangesWinSample_dt}} or
\code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
qgr.}

\item{summary_FUN}{function.  only relevant if win_method is "summary".
passed to \code{\link{viewGRangesWinSummary_dt}}.}

\item{fragLens}{never used by ssvFetchBigwig. Ignore.}

\item{anchor}{character, one of c("center", "center_unstranded",
"left", "left_unstranded")}

\item{return_data.table}{logical. If TRUE the internal data.table is
returned instead of GRanges.  Default is FALSE.}

\item{n_cores}{integer number of cores to use.
Uses mc.cores option if not supplied.}

\item{n_region_splits}{integer number of splits to apply to qgr. The query
GRanges will be split into this many roughly equal parts for increased
parallelization. Default is 1, no split.}

\item{force_skip_centerFix}{boolean, if TRUE all query ranges will be
used "as is".  This is already the case by default if win_method == "summary"
but may have applications where win_method == "sample".}
}
\value{
A tidy formatted GRanges (or data.table if specified) containing
fetched values.
}
\description{
\code{ssvFetchBigwig} iteratively calls \code{fetchWindowedBigwig.single}.
See \code{\link{ssvFetchBigwig.single}} for more info.
}
\details{
if \code{qgr} contains the range chr1:1-100 and \code{win_size} is
10, values from positions chr1 5,15,25...85, and 95 will be
retrieved from \code{bw_file}
}
\examples{
if(Sys.info()['sysname'] != "Windows"){
library(GenomicRanges)
bw_f = system.file("extdata/test_loading.bw",
    package = "seqsetvis", mustWork = TRUE)
bw_files = c("a" = bw_f, "b" = bw_f)
qgr = GRanges("chrTest", IRanges(1, 30))
bw_gr = ssvFetchBigwig(bw_files, qgr, win_size = 10)
bw_gr2 = ssvFetchBigwig(as.list(bw_files), qgr, win_size = 10)

bw_dt = ssvFetchBigwig(bw_files, qgr, win_size = 10,
    return_data.table = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_helpers.R
\name{set_list2memb}
\alias{set_list2memb}
\title{convert a list of sets, each list item should be a character vector
denoting items in sets}
\usage{
set_list2memb(set_list)
}
\arguments{
\item{set_list}{a list of character vectors.  default names will be added if
missing}
}
\value{
converts list of characters/numeric to membership table matrix
}
\description{
convert a list of sets, each list item should be a character vector
denoting items in sets
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{within_clust_sort}
\alias{within_clust_sort}
\title{within_clust_sort}
\usage{
within_clust_sort(
  clust_dt,
  row_ = "id",
  column_ = "x",
  fill_ = "y",
  facet_ = "sample",
  cluster_ = "cluster_id",
  within_order_strategy = c("hclust", "sort", "left", "right", "none", "reverse")[2],
  clustering_col_min = -Inf,
  clustering_col_max = Inf,
  dcast_fill = NA
)
}
\arguments{
\item{clust_dt}{data.table output from \code{\link{ssvSignalClustering}}}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{column_}{varaible mapped to column, likely bp position for ngs data.
Default is "x" and works with ssvFetch* output.}

\item{fill_}{numeric variable to map to fill. Default is "y" and works with
ssvFetch* output.}

\item{facet_}{variable name to facet horizontally by. Default is "sample" and
works with ssvFetch* output. Set to "" if data is not facetted.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{within_order_strategy}{one of "hclust", "sort", "right", "left",
"reverse".  If "hclust", hierarchical clustering will be used. If "sort", a
simple decreasing sort of rosSums.  If "left", will atttempt to put high
signal on left ("right" is opposite).  If "reverse" reverses existing order
(should only be used after meaningful order imposed).}

\item{clustering_col_min}{numeric minimum for col range considered when
clustering, default in -Inf}

\item{clustering_col_max}{numeric maximum for col range considered when
clustering, default in Inf}

\item{dcast_fill}{value to supply to dcast fill argument. default is NA.}
}
\value{
data.table matching input clust_dt save for the reassignment of
levels of row_ variable.
}
\description{
Without modifying cluster assignments, modify the order of rows within each
cluster based on within_order_strategy.
}
\details{
This is particularly useful when you want to sort within each cluster by a
different variable from cluster assignment. Also if you've imported cluster
assigments but want to sort within each for the new data for a prettier
heatmap.

TODO refactor shared code with clusteringKmeansNestedHclust
}
\examples{
data(CTCF_in_10a_profiles_dt)
#clustering by relative value per region does a good job highlighting changes
#when then plotting raw values the order within clusters is not smooth
#this is a good situation to apply a separate sort within clusters.
prof_dt = CTCF_in_10a_profiles_dt
prof_dt = append_ynorm(prof_dt)
prof_dt[, y_relative := y_norm / max(y_norm), list(id)]

clust_dt = ssvSignalClustering(prof_dt, fill_ = "y_relative")
clust_dt.sort = within_clust_sort(clust_dt)

cowplot::plot_grid(
  ssvSignalHeatmap(clust_dt) +
    labs(title = "clustered by relative, sorted by relative"),
  ssvSignalHeatmap(clust_dt.sort) +
    labs(title = "clustered by relative, sorted by raw value")
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fragLen.R
\name{crossCorrByRle}
\alias{crossCorrByRle}
\title{Calculate cross correlation by using shiftApply on read coverage Rle}
\usage{
crossCorrByRle(
  bam_file,
  query_gr,
  max_dupes = 1,
  fragment_sizes = 50:300,
  read_length = NULL,
  flip_strand = FALSE,
  ...
)
}
\arguments{
\item{bam_file}{character. Path to .bam file, must have index at .bam.bai.}

\item{query_gr}{GRanges.  Regions to calculate cross correlation for.}

\item{max_dupes}{integer.  Duplicate reads above this value will be removed.}

\item{fragment_sizes}{integer.  fragment size range to search for maximum
correlation.}

\item{read_length}{integer. Any values outside fragment_range that must be
searched.  If not supplied will be determined from bam_file.  Set as NA
to disable this behavior.}

\item{flip_strand}{boolean. if TRUE strands that reads align to are swapped.
This is typically only necessary if there was a mismatch between library
chemistry and aligner settings. Default is FALSE.}

\item{...}{arguments passed to ScanBamParam}
}
\value{
named list of results
}
\description{
Calculate cross correlation by using shiftApply on read coverage Rle
}
\examples{
data(CTCF_in_10a_overlaps_gr)
bam_f = system.file("extdata/test.bam",
    package = "seqsetvis", mustWork = TRUE)
query_gr = CTCF_in_10a_overlaps_gr[1:2]
crossCorrByRle(bam_f, query_gr[1:2], fragment_sizes = seq(50, 300, 50))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_helpers.R
\name{safeBrew}
\alias{safeBrew}
\title{safeBrew}
\usage{
safeBrew(n, pal = "Dark2")
}
\arguments{
\item{n}{integer value of number of colors to make palette for. Alternatively
a character or factor, in which case palette will be generated for each
unique item or factor level repsectively.}

\item{pal}{palette recognized by RColorBrewer}
}
\value{
a character vector of hex coded colors of length n from the color
brewer palette pal. If n is supplied as character or factor, output will be
named accordingly.
}
\description{
Allows RColorBrew to handle n values less than 3 and greater than 8 without
warnings and return expected number of colors.
}
\details{
For convenience, instead of the number n requested, n may be a character or
factor vector and outputs will be appropriately named for use with
scale_color/fill_manual.

Additionally, accepts pal as "gg", "ggplot", or "ggplot2" to reproduce
default ggplot colors in the same way.
}
\examples{
plot(1:2, rep(0, 2),  col = safeBrew(2, "dark2"), pch = 16, cex = 6)
plot(1:12, rep(0, 12),  col = safeBrew(12, "set1"), pch = 16, cex = 6)
plot(1:12, rep(0, 12),  col = safeBrew(12, "set2"), pch = 16, cex = 6)
plot(1:12, rep(0, 12),  col = safeBrew(12, "set3"), pch = 16, cex = 6)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_operations.R
\name{centerGRangesAtMax}
\alias{centerGRangesAtMax}
\title{Centers query GRanges at maximum signal in prof_dt.}
\usage{
centerGRangesAtMax(
  prof_dt,
  qgr,
  x_ = "x",
  y_ = "y",
  by_ = "id",
  width = 1,
  view_size = NULL
)
}
\arguments{
\item{prof_dt}{a GRanges or data.table as returned by ssvFetch*.}

\item{qgr}{the GRanges used to query ssvFetch* as the qgr argument.}

\item{x_}{positional variable.  Should almost always be the default, "x".}

\item{y_}{the signal value variable.  Likely the default value of "y" but
could be "y_norm" if append_ynorm was applied to data.}

\item{by_}{region identifier variable. Should almost always be the default,
"id".}

\item{width}{Desired width of final regions.  Default is 1.}

\item{view_size}{the size in \code{x_} to consider for finding the max
of \code{y_}.
if length(view_size) == 1, range will be c(-view_size, view_size).
if length(view_size) > 1, range will be range(view_size).
default value of NULL uses complete range of x.}
}
\value{
a GRanges with same mcols as qgr that has been centered based on
signal in prof_dt and with regions of specified width.
}
\description{
Centers query GRanges at maximum signal in prof_dt.
}
\examples{
data(CTCF_in_10a_overlaps_gr)
data(CTCF_in_10a_profiles_gr)
data(CTCF_in_10a_profiles_dt)
centerGRangesAtMax(CTCF_in_10a_profiles_dt, CTCF_in_10a_overlaps_gr)
centerGRangesAtMax(CTCF_in_10a_profiles_gr, CTCF_in_10a_overlaps_gr)

centerGRangesAtMax(CTCF_in_10a_profiles_gr, CTCF_in_10a_overlaps_gr, view_size = 100)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bam.R
\name{harmonize_seqlengths}
\alias{harmonize_seqlengths}
\title{harmonize_seqlengths}
\usage{
harmonize_seqlengths(query_gr, bam_file, force_fix = FALSE)
}
\arguments{
\item{query_gr}{GRanges, object to harmonize seqlengths for}

\item{bam_file}{character, a path to a valid bam file}

\item{force_fix}{Logical, if TRUE incompatible seqnames are removed from the query_gr.  Default is FALSE.}
}
\value{
GRanges with seqlengths matching bam_file
}
\description{
ensures compatibility between seqlength of gr and bam_file based on header
}
\examples{
library(GenomicRanges)
query_gr = GRanges("chr1", IRanges(1, 100))
#seqlengths has not been set
seqlengths(query_gr)
bam = system.file("extdata/test.bam", package = "seqsetvis")
gr2 = harmonize_seqlengths(query_gr, bam)
#seqlengths now set
seqlengths(gr2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{merge_clusters}
\alias{merge_clusters}
\title{merge_clusters}
\usage{
merge_clusters(
  clust_dt,
  to_merge,
  row_ = "id",
  cluster_ = "cluster_id",
  reapply_cluster_names = TRUE
)
}
\arguments{
\item{clust_dt}{data.table output from \code{\link{ssvSignalClustering}}}

\item{to_merge}{Clusters to merge. Must be items in clust_dt variable defined
by cluster_ parameter.}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{reapply_cluster_names}{If TRUE, clusters will be renamed according to
new order instead of their original names. Default is TRUE.}
}
\value{
data.table as output from \code{\link{ssvSignalClustering}}
}
\description{
merge_clusters
}
\examples{
data(CTCF_in_10a_profiles_dt)
set.seed(0)
clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 6)
ssvSignalHeatmap(clust_dt)
agg_dt = clust_dt[, list(y = mean(y)), list(x, cluster_id, sample)]
ggplot(agg_dt, aes(x = x, y = y, color = sample)) +
  geom_path() +
  facet_grid(cluster_id~.)

to_merge = c(2, 3, 5)
# debug(merge_clusters)
new_dt = merge_clusters(clust_dt, c(2, 3, 5), reapply_cluster_names = FALSE)
new_dt.relabel = merge_clusters(clust_dt, c(2, 3, 5), reapply_cluster_names = TRUE)
new_dt.relabel.sort = within_clust_sort(new_dt.relabel, within_order_strategy = "sort")

table(clust_dt$cluster_id)
table(new_dt$cluster_id)

cowplot::plot_grid(
  ssvSignalHeatmap(clust_dt) + labs(title = "original"),
  ssvSignalHeatmap(new_dt) + labs(title = "2,3,5 merged"),
  ssvSignalHeatmap(new_dt.relabel) + labs(title = "2,3,5 merged, renumbered"),
  ssvSignalHeatmap(new_dt.relabel.sort) + labs(title = "2,3,5 merged, renumbered and sorted")
)




}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_signal.R
\name{quantileGRangesWidth}
\alias{quantileGRangesWidth}
\title{Quantile width determination strategy}
\usage{
quantileGRangesWidth(qgr, min_quantile = 0.75, win_size = 1)
}
\arguments{
\item{qgr}{GRanges to calculate quantile width for}

\item{min_quantile}{numeric value from 0 to 1. The minimum quantile of width in qgr}

\item{win_size}{numeric/integer >=1, returned value will be a multiple of
this}
}
\value{
numeric that is >= min_quantile and evenly divisible by win_size
}
\description{
Returns the lowest multiple of win_size greater than
min_quantile quantile of width(qgr)
}
\examples{
data(CTCF_in_10a_overlaps_gr)
gr = CTCF_in_10a_overlaps_gr
quantileGRangesWidth(gr)
quantileGRangesWidth(gr, min_quantile = .5, win_size = 100)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_chromHMM.R
\docType{data}
\name{chromHMM_demo_segmentation_url}
\alias{chromHMM_demo_segmentation_url}
\title{URL to download hg19 MCF7 chromHMM segmentation}
\format{
a character containing a URL
}
\value{
a character containing a URL
}
\description{
URL to download hg19 MCF7 chromHMM segmentation
}
\details{
file is gzipped bed with name, score, itemRgb and thick meta columns

part of \code{\link{chromHMM_demo_data}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_CTCF.R
\docType{data}
\name{CTCF_in_10a_profiles_dt}
\alias{CTCF_in_10a_profiles_dt}
\title{Profiles for 100 randomly selected regions from overlapping CTCF peaks in
10a cell ChIP-seq
Results from fetching bigwigs with CTCF_in_10a_overlaps_gr.}
\format{
A tidy data.table of 2100 rows and 9 columns
}
\description{
A tidy data.table at window size 50 bp within 350 bp of peak center
The variables are as follows:
}
\details{
part of \code{\link{CTCF_in_10a_data}}
\enumerate{
\item seqnames. chromosome for GRanges compatibility
\item start. start of interval
\item end. end of interval
\item width. width of interval
\item strand. leftover from GRanges.
\item id. unique identifier
\item y. fold-enrichment over input.
\item x. bp relative to center
\item sample. name of originating sample
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_plots.R
\name{ssvSignalScatterplot}
\alias{ssvSignalScatterplot}
\title{maps signal from 2 sample profiles to the x and y axis. axes are standard or "volcano" min XY vs fold-change Y/X}
\usage{
ssvSignalScatterplot(
  bw_data,
  x_name,
  y_name,
  color_table = NULL,
  value_variable = "y",
  xy_variable = "sample",
  value_function = max,
  by_ = "id",
  plot_type = c("standard", "volcano")[1],
  show_help = FALSE,
  fixed_coords = TRUE,
  return_data = FALSE
)
}
\arguments{
\item{bw_data}{a GRanges or data.table of bigwig signal.
As returned from \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}}

\item{x_name}{sample name to map to x-axis, must be stored in variable specified in \code{xy_variable}}

\item{y_name}{sample name to map to y-axis, must be stored in variable specified in \code{xy_variable}}

\item{color_table}{data.frame with 2 columns, one of which must be named "group" and gets mapped to color.
The other column must be the same as by_ parameter and is used for merging.}

\item{value_variable}{variable name that stores numeric values for plotting, default is "y"}

\item{xy_variable}{variable name that stores sample, must contain entires for \code{x_name} and \code{y_name}}

\item{value_function}{a function to apply to \code{value_variable} in all combintations of \code{by_} per \code{x_name} and \code{y_name}}

\item{by_}{variables that store individual measurement ids}

\item{plot_type}{standard or volcano, default is "standard"}

\item{show_help}{if TRUE overlay labels to aid plot interpretation, default is FALSE}

\item{fixed_coords}{if TRUE coordinate system is 1:1 ratio, default is TRUE}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and
is instead the data used to generate that plot. Default is FALSE.}
}
\value{
ggplot of points comparing signal from 2 samples
}
\description{
maps signal from 2 sample profiles to the x and y axis. axes are standard or "volcano" min XY vs fold-change Y/X
}
\examples{
data(CTCF_in_10a_profiles_gr)
ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
    x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF")
ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
    x_name = "MCF10A_CTCF", y_name = "MCF10CA1_CTCF")

ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
    x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF",
    value_function = median) + labs(title = "median FE in regions")

ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
    x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF",
    plot_type = "volcano")

ssvSignalScatterplot(CTCF_in_10a_profiles_gr,
    x_name = "MCF10A_CTCF", y_name = "MCF10AT1_CTCF",
    plot_type = "volcano", show_help = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_easyLoad.R
\name{easyLoad_broadPeak}
\alias{easyLoad_broadPeak}
\title{easyLoad_broadPeak takes a character vector of file paths to narrowPeak
files from MACS2 and returns a named list of GRanges.}
\usage{
easyLoad_broadPeak(
  file_paths,
  file_names = NULL,
  n_cores = getOption("mc.cores", 1)
)
}
\arguments{
\item{file_paths}{character vector of paths to narrowPeak files.  If named,
those names will be used in output unless overriden by providing file_names.}

\item{file_names}{character vector of names for output list.  If not NULL
will override any existing names for file_paths.  Default is NULL.}

\item{n_cores}{number of cores to use, uses mc.cores option if set or 1.}
}
\value{
a named list of GRanges loaded from file_paths
}
\description{
easyLoad_broadPeak takes a character vector of file paths to narrowPeak
files from MACS2 and returns a named list of GRanges.
}
\examples{
bp_f = system.file("extdata/test_loading.broadPeak",
    package = "seqsetvis", mustWork = TRUE)
easyLoad_broadPeak(bp_f, "my_broadPeak")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_signal.R
\name{ssvFetchSignal}
\alias{ssvFetchSignal}
\title{signal loading framework}
\usage{
ssvFetchSignal(
  file_paths,
  qgr,
  unique_names = NULL,
  names_variable = "sample",
  file_attribs = NULL,
  win_size = 50,
  win_method = c("sample", "summary")[1],
  return_data.table = FALSE,
  load_signal = function(f, nam, qgr) {
     warning("nothing happened, ",
    "supply a function to", "load_signal parameter.")
 },
  n_cores = getOption("mc.cores", 1),
  n_region_splits = 1,
  force_skip_centerFix = FALSE
)
}
\arguments{
\item{file_paths}{character vector of file_paths to load from. Alternatively,
file_paths can be a data.frame or data.table whose first column is a
character vector of paths and additial columns will be used as metadata.}

\item{qgr}{GRanges of intervals to return from each file}

\item{unique_names}{unique file ids for each file in file_paths.  Default is
names of file_paths vector}

\item{names_variable}{character, variable name for column containing
unique_names entries.  Default is "sample"}

\item{file_attribs}{optional data.frame/data.table with one row per item in
file paths.  Each column will be a variable added to final tidy output.}

\item{win_size}{numeric/integer window size resolution to load signal at.
Default is 50.}

\item{win_method}{character.  one of c("sample", "summary").  Determines if
\code{\link{viewGRangesWinSample_dt}} or
\code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
qgr.}

\item{return_data.table}{logical. If TRUE data.table is returned instead of
GRanges, the default.}

\item{load_signal}{function taking f, nam, and qgr arguments.  f is from
file_paths, nam is from unique_names, and qgr is qgr. See details.}

\item{n_cores}{integer number of cores to use. Uses mc.cores option if not
supplied.}

\item{n_region_splits}{integer number of splits to apply to qgr. The query
GRanges will be split into this many roughly equal parts for increased
parallelization. Default is 1, no split.}

\item{force_skip_centerFix}{boolean, if TRUE all query ranges will be used
"as is".  This is already the case by default if win_method == "summary"
but may have applications where win_method == "sample".}
}
\value{
A GRanges with values read from file_paths at intervals of win_size.
Originating file is coded by unique_names and assigned to column of name
names_variable.  Output is data.table is return_data.table is TRUE.
}
\description{
Does nothing unless load_signal is overridden to carry out reading data from
file_paths (likely via the appropriate ssvFetch* function, ie.
\code{\link{ssvFetchBigwig}} or \code{\link{ssvFetchBam}}
}
\details{
load_signal is passed f, nam, and qgr and is executed in the
environment where load_signal is defined. See \code{\link{ssvFetchBigwig}}
and \code{\link{ssvFetchBam}} for examples.
}
\examples{
library(GenomicRanges)
data(CTCF_in_10a_overlaps_gr)
bam_f = system.file("extdata/test.bam",
    package = "seqsetvis", mustWork = TRUE)
bam_files = c("a" = bam_f, "b" = bam_f)
qgr = CTCF_in_10a_overlaps_gr[1:2]
qgr = resize(qgr, 500, "center")
load_bam = function(f, nam, qgr) {
    message("loading ", f, " ...")
    dt = seqsetvis:::ssvFetchBam.single(bam_f = f,
                      qgr = qgr,
                      win_size = 50,
                      fragLen = NULL,
                      target_strand = "*",
                      return_data.table = TRUE)

    data.table::set(dt, j = "sample", value = nam)
    message("finished loading ", nam, ".")
    dt
}
ssvFetchSignal(bam_files, qgr, load_signal = load_bam)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqsetvis.R
\docType{package}
\name{seqsetvis-package}
\alias{seqsetvis}
\alias{seqsetvis-package}
\title{easy awesome peak set vis
TESTING
seqsetvis allows you to...}
\description{
2 steps
\code{\link{ssvOverlapIntervalSets}}.
\code{\link{ssvFetchBigwig}}.
Otherwise refer to the vignettes to see
}
\author{
\strong{Maintainer}: Joseph R Boyd \email{jrboyd@uvm.edu} (\href{https://orcid.org/0000-0002-8969-9676}{ORCID})

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bigwig.R
\name{ssvFetchBigwig.single}
\alias{ssvFetchBigwig.single}
\title{Fetch values from a bigwig appropriate for heatmaps etc.}
\usage{
ssvFetchBigwig.single(
  bw_file,
  qgr,
  win_size = 50,
  win_method = c("sample", "summary")[1],
  summary_FUN = stats::weighted.mean,
  anchor = c("left", "left_unstranded", "center", "center_unstranded")[3],
  return_data.table = FALSE,
  force_skip_centerFix = FALSE
)
}
\arguments{
\item{bw_file}{The character vector path to bigwig files to read from.}

\item{qgr}{Set of GRanges to query.  For valid results the width of each
interval should be identical and evenly divisible by \code{win_size}.}

\item{win_size}{The window size that evenly divides widths in \code{qgr}.}

\item{win_method}{character.  one of c("sample", "summary").  Determines
if \code{\link{viewGRangesWinSample_dt}} or
\code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
qgr.}

\item{summary_FUN}{function.  only relevant if win_method is "summary".
passed to \code{\link{viewGRangesWinSummary_dt}}.}

\item{anchor}{character, one of c("center", "center_unstranded",
"left", "left_unstranded")}

\item{return_data.table}{logical. If TRUE the internal data.table is
returned instead of GRanges.  Default is FALSE.}

\item{force_skip_centerFix}{boolean, if TRUE all query ranges will be
used "as is".  This is already the case by default if win_method == "summary"
but may have applications where win_method == "sample".}
}
\value{
A GRanges (or data.table if specified) containing fetched values.
}
\description{
\code{ssvFetchBigwig.single} Gets values for each region of the query
GRanges (\code{qgr}). Values correspond to the center of each window of
size \code{win_size}.  A tidy formatted data.table
object is returned suitable for plotting using ggplots.
}
\details{
if \code{qgr} contains the range chr1:1-100 and \code{win_size} is
10, values from positions chr1 5,15,25...85, and 95 will be retrieved
from \code{bw_file}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_normalization.R
\name{append_ynorm}
\alias{append_ynorm}
\title{append_ynorm}
\usage{
append_ynorm(
  full_dt,
  value_ = "y",
  cap_value_ = "y_cap_value",
  norm_value_ = "y_norm",
  by1 = "id",
  by2 = "sample",
  aggFUN1 = max,
  aggFUN2 = function(x) quantile(x, 0.95),
  cap_dt = NULL,
  do_not_cap = FALSE,
  do_not_scaleTo1 = FALSE,
  force_append = FALSE
)
}
\arguments{
\item{full_dt}{a data.table, as returned by ssvFetch*(..., return_data.table
= TRUE).}

\item{value_}{character, attribute in full_dt to normalzie.}

\item{cap_value_}{character, new attribute name specifying values to cap to.}

\item{norm_value_}{character, new attribute name specifying normalized values.}

\item{by1}{character vector, specifies attributes relevant to step 1.}

\item{by2}{character vector, specifies attributes relevant to step 1 and 2.}

\item{aggFUN1}{function called on value_ with by = c(by1, by2) in step 1.}

\item{aggFUN2}{function called on result of aggFUN1 with by = by2 in step 2.}

\item{cap_dt}{optionally, provide user generated by2 to cap_value_ mapping}

\item{do_not_cap}{if TRUE, normalized values are not capped to 1. Default is FALSE.}

\item{do_not_scaleTo1}{if TRUE, normalized values are not scaled to 1. Default is FALSE.}

\item{force_append}{if TRUE, any previous cap_value or norm_value is overridden. Default is FALSE.}
}
\value{
data.table, full_dt with cap_value_ and norm_value_ values appended.
}
\description{
see \code{\link{calc_norm_factors}} for normalization details.
}
\examples{
data(CTCF_in_10a_profiles_dt)
append_ynorm(CTCF_in_10a_profiles_dt)
append_ynorm(CTCF_in_10a_profiles_dt,
  aggFUN1 = mean, aggFUN2 = function(x)quantile(x, .5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_easyLoad.R
\name{easyLoad_seacr}
\alias{easyLoad_seacr}
\title{easyLoad_seacr takes a character vector of file paths to seacr output bed files and returns a named list of GRanges.}
\usage{
easyLoad_seacr(
  file_paths,
  file_names = NULL,
  n_cores = getOption("mc.cores", 1)
)
}
\arguments{
\item{file_paths}{character vector of paths to seacr bed files.  If named,
those names will be used in output unless overriden by providing file_names.}

\item{file_names}{character vector of names for output list.  If not NULL
will override any existing names for file_paths.  Default is NULL.}

\item{n_cores}{number of cores to use, uses mc.cores option if set or 1.}
}
\value{
a named list of GRanges loaded from file_paths
}
\description{
easyLoad_seacr takes a character vector of file paths to seacr output bed files and returns a named list of GRanges.
}
\examples{
bed_f = system.file("extdata/test_loading.seacr.bed",
    package = "seqsetvis", mustWork = TRUE)
easyLoad_seacr(bed_f, "my_seacr")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_chromHMM.R
\docType{data}
\name{chromHMM_demo_state_colors}
\alias{chromHMM_demo_state_colors}
\title{original state name to color mappings stored in segmentation bed}
\format{
a named character vector mapping states to hex colors
}
\value{
a named character vector mapping states to hex colors
}
\description{
original state name to color mappings stored in segmentation bed
}
\details{
part of \code{\link{chromHMM_demo_data}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_signal.R
\name{viewGRangesWinSample_dt}
\alias{viewGRangesWinSample_dt}
\title{get a windowed sampling of score_gr}
\usage{
viewGRangesWinSample_dt(
  score_gr,
  qgr,
  window_size,
  attrib_var = "score",
  fill_value = 0,
  anchor = c("center", "center_unstranded", "left", "left_unstranded")[1]
)
}
\arguments{
\item{score_gr}{GRanges with a "score" metadata column.}

\item{qgr}{regions to view by window.}

\item{window_size}{qgr will be represented by value from score_gr every
window_size bp.}

\item{attrib_var}{character name of attribute to pull data from. Default is
"score", compatible with with bigWigs or bam coverage.}

\item{fill_value}{numeric or character value to use where queried regions are
empty.  Default is 0 and appropriate for both calculated coverage and
bedgraph/bigwig like files.  Will automatically switch to "MISSING" if data
is guessed to be qualitative.}

\item{anchor}{character. controls how x value is derived from position for
each region in qgr.  0 may be the left side or center.  If not unstranded,
x coordinates are flipped for (-) strand. One of c("center",
"center_unstranded", "left", "left_unstranded"). Default is "center".}
}
\value{
data.table that is GRanges compatible
}
\description{
This method is appropriate when all GRanges in qgr are identical width and
when it is practical to use a window_size smaller than features in genomic
signal.  For instance, when retrieving signal around peaks or promoters this
method maintains a fixed genomic scale across regions.  This allows meaingful
comparison of peak widths can be made.
}
\details{
Summarizes score_gr by grabbing value of "score" every window_size bp.
Columns in output data.table are: standard GRanges columns: seqnames, start,
end, width, strand id - matched to names(score_gr). if names(score_gr) is
missing, added as 1:length(score_gr). y - value of score from score_gr. x -
relative bp position.
}
\examples{
data(CTCF_in_10a_overlaps_gr)
bam_file = system.file("extdata/test.bam",
    package = "seqsetvis")
qgr = CTCF_in_10a_overlaps_gr[seq_len(5)]
qgr = GenomicRanges::resize(qgr, width = 500, fix = "center")
bam_gr = seqsetvis:::fetchBam(bam_file, qgr)
bam_dt = viewGRangesWinSample_dt(bam_gr, qgr, 50)

if(Sys.info()['sysname'] != "Windows"){
    bw_file = system.file("extdata/MCF10A_CTCF_FE_random100.bw",
        package = "seqsetvis")
    bw_gr = rtracklayer::import.bw(bw_file, which = qgr)
    bw_dt = viewGRangesWinSample_dt(bw_gr, qgr, 50)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_plots.R
\name{ssvSignalLineplot}
\alias{ssvSignalLineplot}
\title{construct line type plots where each region in each sample is represented}
\usage{
ssvSignalLineplot(
  bw_data,
  x_ = "x",
  y_ = "y",
  color_ = "sample",
  sample_ = "sample",
  region_ = "id",
  group_ = "auto_grp",
  line_alpha = 1,
  facet_ = "auto_facet",
  facet_method = facet_wrap,
  spline_n = NULL,
  return_data = FALSE
)
}
\arguments{
\item{bw_data}{a GRanges or data.table of bigwig signal.
As returned from \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}}

\item{x_}{variable name mapped to x aesthetic, x by default.}

\item{y_}{variable name mapped to y aesthetic, y by default.}

\item{color_}{variable name mapped to color aesthetic, sample by default.}

\item{sample_}{variable name, along with region_ used to group and
facet by default, change group_ or facet_ to override.}

\item{region_}{variable name, along with sample_ used to group and facet
by default, change group_ or facet_ to override.}

\item{group_}{group aesthetic keeps lines of geom_path from mis-connecting.
auto_grp by default which combines sample_ and region_.
probably shouldn't change.}

\item{line_alpha}{alpha value for lines. default is 1.}

\item{facet_}{facetting divides up plots.
auto_facet by default which combines sample_ and region_.
if overriding facet_method with facet_grid, make sure to include ~ between
two variables, ie. "a~b", ".~b", "a~."}

\item{facet_method}{ggplot2 facetting method or wrapper for same,
facet_wrap by default.}

\item{spline_n}{if not NULL, applySpline will be called with n = spline_n.
default is NULL.}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and
is instead the data used to generate that plot. Default is FALSE.}
}
\value{
ggplot of signal potentially facetted by region and sample
}
\description{
construct line type plots where each region in each sample is represented
}
\examples{
data(CTCF_in_10a_profiles_gr)

bw_gr = CTCF_in_10a_profiles_gr
ssvSignalLineplot(subset(bw_gr, bw_gr$id \%in\% seq_len(3)), facet_ = "sample")
ssvSignalLineplot(subset(bw_gr, bw_gr$id \%in\% seq_len(3)),
    facet_ = "sample~.",
    facet_method = facet_grid)
ssvSignalLineplot(subset(bw_gr, bw_gr$id \%in\% seq_len(3)),
    facet_ = paste("sample", "~", "id"), facet_method = facet_grid)
ssvSignalLineplot(subset(bw_gr, bw_gr$id \%in\% seq_len(3)))
ssvSignalLineplot(subset(bw_gr, bw_gr$id \%in\% seq_len(3)), facet_ = "id")
ssvSignalLineplot(subset(bw_gr, bw_gr$id \%in\% seq_len(3)),
    facet_ = "id", spline_n = 10)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bam.R
\name{ssvFetchBam.single}
\alias{ssvFetchBam.single}
\title{fetch a windowed version of a bam file, returns GRanges}
\usage{
ssvFetchBam.single(
  bam_f,
  qgr,
  win_size = 50,
  win_method = c("sample", "summary")[1],
  summary_FUN = stats::weighted.mean,
  fragLen = NULL,
  target_strand = c("*", "+", "-", "both")[1],
  anchor = c("left", "left_unstranded", "center", "center_unstranded")[3],
  return_data.table = FALSE,
  max_dupes = Inf,
  splice_strategy = c("none", "ignore", "add", "only", "splice_count")[1],
  flip_strand = FALSE,
  return_unprocessed = FALSE,
  force_skip_centerFix = FALSE,
  ...
)
}
\arguments{
\item{bam_f}{character or BamFile to load}

\item{qgr}{GRanges regions to fetchs}

\item{win_size}{numeric >=1.  pileup grabbed every win_size bp for win_method
sample.  If win_method is summary, this is the number of windows used
(confusing, sorry).}

\item{win_method}{character.  one of c("sample", "summary").  Determines if
\code{\link{viewGRangesWinSample_dt}} or
\code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
qgr.}

\item{summary_FUN}{function.  only relevant if win_method is "summary".
passed to \code{\link{viewGRangesWinSummary_dt}}.}

\item{fragLen}{numeric, NULL, or NA.  if numeric, supplied value is used. if
NULL, value is calculated with fragLen_calcStranded if NA, raw bam pileup
with no cross strand shift is returned.}

\item{target_strand}{character. if one of "+" or "-", reads are filtered
accordingly. ignored if any other value.}

\item{anchor}{character, one of c("center", "center_unstranded", "left",
"left_unstranded")}

\item{return_data.table}{logical. If TRUE the internal data.table is returned
instead of GRanges.  Default is FALSE.}

\item{max_dupes}{numeric >= 1.  duplicate reads by strandd start position
over this number are removed, Default is Inf.}

\item{splice_strategy}{character, one of c("none", "ignore", "add", "only",
"splice_count"). Default is "none" and spliced alignment are asssumed not
present. fragLen must be NA for any other value to be valid.  "ignore" will
not count spliced regions.  add" counts spliced regions along with others,
"only" will only count spliced regions and ignore others.}

\item{flip_strand}{if TRUE, strand alignment is flipped prior to fragLen
extension. Default is FALSE.}

\item{return_unprocessed}{boolean. if TRUE returns read alignment in data.table. Default is FALSE.}

\item{force_skip_centerFix}{boolean, if TRUE all query ranges will be
used "as is".  This is already the case by default if win_method == "summary"
but may have applications where win_method == "sample".}

\item{...}{passed to Rsamtools::ScanBamParam()}
}
\value{
tidy GRanges (or data.table if specified) with pileups from bam file.
pileup is calculated only every win_size bp.
}
\description{
fetch a windowed version of a bam file, returns GRanges
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_easyLoad.R
\name{easyLoad_FUN}
\alias{easyLoad_FUN}
\title{easyLoad_FUN takes a character vector of file paths run an arbitrary function defined in load_FUN}
\usage{
easyLoad_FUN(
  file_paths,
  load_FUN,
  file_names = NULL,
  n_cores = getOption("mc.cores", 1),
  ...
)
}
\arguments{
\item{file_paths}{character vector of paths to narrowPeak files.  If named,
those names will be used in output unless overriden by providing file_names.}

\item{load_FUN}{Arbitrary function that takes at least a file path as argument.  May take other arguments that should be set in call to easyLoad_FUN.}

\item{file_names}{character vector of names for output list.  If not NULL
will override any existing names for file_paths.  Default is NULL.}

\item{n_cores}{number of cores to use, uses mc.cores option if set or 1.}

\item{...}{extra parameters passed to load_FUN}
}
\value{
a named list of results from load_FUN
}
\description{
easyLoad_FUN takes a character vector of file paths run an arbitrary function defined in load_FUN
}
\examples{
bed_f = system.file("extdata/test_loading.bed",
    package = "seqsetvis", mustWork = TRUE)
easyLoad_bed(bed_f, "my_bed")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bam.R
\name{getReadLength}
\alias{getReadLength}
\title{determine the most common read length for input bam_file.  uses 50 randomly
selected regions from query_gr.  If fewer than 20 reads are present, loads
all of query_gr.}
\usage{
getReadLength(bam_file, query_gr)
}
\arguments{
\item{bam_file}{indexed bam file}

\item{query_gr}{GRanges to read from bam file}
}
\value{
numeric of most common read length.
}
\description{
determine the most common read length for input bam_file.  uses 50 randomly
selected regions from query_gr.  If fewer than 20 reads are present, loads
all of query_gr.
}
\examples{
data(CTCF_in_10a_overlaps_gr)
qgr = CTCF_in_10a_overlaps_gr[1:5]
bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
getReadLength(bam_file, qgr)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_helpers.R
\name{col2hex}
\alias{col2hex}
\title{converts a valid r color name ("black", "red", "white", etc.) to a hex value}
\usage{
col2hex(color_name)
}
\arguments{
\item{color_name}{character. one or more r color names.}
}
\value{
hex value of colors coded by colors()
}
\description{
converts a valid r color name ("black", "red", "white", etc.) to a hex value
}
\examples{
col2hex(c("red", "green", "blue"))
col2hex(c("lightgray", "gray", "darkgray"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bam.R
\name{.expand_cigar_dt}
\alias{.expand_cigar_dt}
\title{Expand intermediate bam fetch by cigar codes}
\usage{
.expand_cigar_dt(cigar_dt, op_2count = c("M", "D", "=", "X"))
}
\arguments{
\item{cigar_dt}{data.table with 5 required named columns in any order.
c("which_label", "seqnames", "strand", "start", "cigar")}

\item{op_2count}{Cigar codes to count. Default is alignment (M), deletion
(D), match (=), and mismatch (X).  Other useful codes may be skipped
regions for RNA splicing (N).  The locations of any insterions (I) or
clipping/padding (S, H, or P) will be a single bp immediately before the
interval.}
}
\value{
data.table with cigar entries expanded
}
\description{
see \href{https://samtools.github.io/hts-specs/SAMv1.pdf}{sam specs} for
cigar details
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_plots.R
\name{assemble_heatmap_cluster_bars}
\alias{assemble_heatmap_cluster_bars}
\title{assemble_heatmap_cluster_bars}
\usage{
assemble_heatmap_cluster_bars(plots, ...)
}
\arguments{
\item{plots}{list of plots as returned from ssvSignalHeatmap.ClusterBars when return_unassembled_plots = TRUE}

\item{...}{arguments passed to cowplot::plot_grid}
}
\value{
A grob produced by cowplot::plot_grid
}
\description{
assemble_heatmap_cluster_bars
}
\examples{
data(CTCF_in_10a_profiles_gr)
plots = ssvSignalHeatmap.ClusterBars(CTCF_in_10a_profiles_gr, return_unassembled_plots = TRUE)
assemble_heatmap_cluster_bars(plots)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_chromHMM.R
\docType{data}
\name{chromHMM_demo_chain_url}
\alias{chromHMM_demo_chain_url}
\title{URL to download hg19ToHg38 liftover chain from UCSC}
\format{
a character containing a URL
}
\value{
a character containing a URL
}
\description{
URL to download hg19ToHg38 liftover chain from UCSC
}
\details{
file is gzipped .txt

part of \code{\link{chromHMM_demo_data}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_feature_plots.R
\name{ssvFeatureBinaryHeatmap}
\alias{ssvFeatureBinaryHeatmap}
\title{ssvFeatureBinaryHeatmap}
\usage{
ssvFeatureBinaryHeatmap(
  object,
  raster_approximation = TRUE,
  true_color = "black",
  false_color = "#EFEFEF",
  raster_width_min = 1000,
  raster_height_min = 1000,
  return_data = FALSE
)
}
\arguments{
\item{object}{passed to ssvMakeMembTable}

\item{raster_approximation}{If TRUE, instead of standard ggplot, write
temporary raster png image and redraw that as plot background. default is
FALSE}

\item{true_color}{character. rcolor or hex color used for TRUE values.
default is "black".}

\item{false_color}{character. rcolor or hex color used for TRUE values.
default is "#EFEFEF", a gray.}

\item{raster_width_min}{raster width will be minimum multiple of number of
columns over this number.  ignored if raster_approximation is FALSE.}

\item{raster_height_min}{raster height will be minimum multiple of number of
rows over this number  ignored if raster_approximation is FALSE}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and
is instead the data used to generate that plot. Default is TRUE}
}
\value{
ggplot using geom_tile of membership table sorted from left to right.
}
\description{
Outputs a ggplot binary heatmap, where color indicates TRUE and the other
indicates FALSE in a membership table. The heatmap is sorted, TRUE at the
top, by column left to right. Changes to column order can reveal different
patterns.
}
\details{
As a svg output, the final plot can be unwieldy. The default of
\code{raster_approximation} = TRUE is easier to work with, especially for larger
membership tables.
}
\examples{
data(CTCF_in_10a_overlaps_gr)
ssvFeatureBinaryHeatmap(list(1:3, 2:6))
# horizontal version
ssvFeatureBinaryHeatmap(list(1:3, 2:6)) + coord_flip() +
  theme(axis.text.x = element_blank(), axis.text.y = element_text())
ssvFeatureBinaryHeatmap(CTCF_in_10a_overlaps_gr)
ssvFeatureBinaryHeatmap(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
ssvFeatureBinaryHeatmap(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,3:2])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_operations.R
\name{findMaxPos}
\alias{findMaxPos}
\title{findMaxPos}
\usage{
findMaxPos(prof_dt, qgr, x_ = "x", y_ = "y", by_ = "id", width = 1)
}
\arguments{
\item{prof_dt}{a GRanges or data.table as returned by ssvFetch*.}

\item{qgr}{the GRanges used to query ssvFetch* as the qgr argument.}

\item{x_}{positional variable.  Should almost always be the default, "x".}

\item{y_}{the signal value variable.  Likely the default value of "y" but
could be "y_norm" if append_ynorm was applied to data.}

\item{by_}{region identifier variable. Should almost always be the default,
"id".}

\item{width}{Desired width of final regions.  Default is 1.}
}
\value{
data.table of relative x position from center per id
}
\description{
findMaxPos
}
\examples{
data(CTCF_in_10a_overlaps_gr)
data(CTCF_in_10a_profiles_gr)
data(CTCF_in_10a_profiles_dt)
findMaxPos(CTCF_in_10a_profiles_dt, CTCF_in_10a_overlaps_gr)
findMaxPos(CTCF_in_10a_profiles_gr, CTCF_in_10a_overlaps_gr)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_exon_signal.R
\name{convert_collapsed_coord}
\alias{convert_collapsed_coord}
\title{convert_collapsed_coord}
\usage{
convert_collapsed_coord(genome_gr, x)
}
\arguments{
\item{genome_gr}{non-contiguous regions to collapse a la
\code{\link{collapse_gr}}}

\item{x}{numeric, positions within genome_gr to convert to collapsed
coordinates.}
}
\value{
numeric, positions of every value of x within collapse coordinates.
values outside of collapsed regions (an intron or outside range) will be NA.
}
\description{
(preliminary implementation, sub-optimal)
}
\details{
see \code{\link{collapse_gr}} for explanation of intended uses. this function
translates all values of x from original genomic coordinates to new
coordinate space created by \code{\link{collapse_gr}}.
}
\examples{
library(data.table)
library(GenomicRanges)
dev_dat = data.table(seqnames = "chrTest",
                     transcript_id = c(1, 1,  2, 2, 3, 3, 3),
                     start = c(5,  30,  8, 30, 2, 30, 40),
                     end = c(10, 35, 15, 38, 7, 35, 45),
                     strand = "+")

genome_gr = GRanges(dev_dat)
convert_collapsed_coord(genome_gr, start(genome_gr))
convert_collapsed_coord(genome_gr, end(genome_gr))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_chromHMM.R
\docType{data}
\name{chromHMM_demo_data}
\alias{chromHMM_demo_data}
\title{chromHMM state segmentation in the MCF7 cell line}
\description{
Vignette data for seqsetvis was downloaded directly from GEO
series
\href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE57498}{GSE57498}.
This data is the state segmentation by chromHMM in the MCF7 cell line.
chromHMM creates a hidden markov model by integrating several ChIP-seq
samples, in this case: \itemize{
\item MCF7_H3K27ac_ChIP-Seq
\item MCF7_H3K27me3_ChIP-Seq
\item MCF7_H3K4me1_ChIP-Seq
\item MCF7_H3K4me3_ChIP-Seq
\item MCF7_RNApolIIp_ChIP-Seq
}

Data from GEO series
\href{https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE57498}{GSE57498}
is from the publication
\href{https://www.ncbi.nlm.nih.gov/pubmed/24916973}{Taberlay PC et al. 2014}
}
\details{
Contains: \itemize{
\item \code{\link{chromHMM_demo_overlaps_gr}}
\item \code{\link{chromHMM_demo_bw_states_gr}}
\item \code{\link{chromHMM_demo_state_total_widths}}
\item \code{\link{chromHMM_demo_state_colors}}
\item \code{\link{chromHMM_demo_segmentation_url}}
\item \code{\link{chromHMM_demo_chain_url}}
}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overlapIntervalSets.R
\name{ssvConsensusIntervalSets}
\alias{ssvConsensusIntervalSets}
\title{Intersect a list of GRanges to create a single GRanges object of merged
ranges including metadata describing overlaps per input GRanges.}
\usage{
ssvConsensusIntervalSets(
  grs,
  ext = 0,
  min_number = 2,
  min_fraction = 0.5,
  preserve_mcols = FALSE,
  ...
)
}
\arguments{
\item{grs}{A list of GRanges}

\item{ext}{An integer specifying how far to extend ranges before merging. in
effect, ranges withing 2*ext of one another will be joined during the merge}

\item{min_number}{An integer number specifying the absloute minimum of input
grs that must overlap for a site to be considered consensus.}

\item{min_fraction}{A numeric between 0 and 1 specifying the fraction of grs
that must overlap to be considered consensus.}

\item{preserve_mcols}{Controls carrying forward mcols metadata from input
list of GRanges. If TRUE, all mcols will be carried forward with the item
name appended. If a character vector, only those attributes will be carried
and all must be present in all GRanges. The default of FALSE will carry
nothing forward and only membership table will be generated.
\link{ssvAnnotateSubjectGRanges} is used internally.}

\item{...}{arguments passed to IRanges::findOverlaps, i.e. maxgap, minoverlap, type, select, invert.}
}
\value{
GRanges with metadata columns describing consensus overlap of input
grs.
}
\description{
In constrast to ssvOverlapIntervalSets, only regions where a consensus of
input grs are present are preserved and annotated.
}
\details{
Only the most stringent of min_number or min_fraction will be
applied.
}
\examples{
library(GenomicRanges)
a = GRanges("chr1", IRanges(1:7*10, 1:7*10))
b = GRanges("chr1", IRanges(5:10*10, 5:10*10))
ssvConsensusIntervalSets(list(a, b))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_plots.R
\name{add_cluster_annotation}
\alias{add_cluster_annotation}
\title{add_cluster_annotation}
\usage{
add_cluster_annotation(
  cluster_ids,
  p = NULL,
  xleft = 0,
  xright = 1,
  rect_colors = c("black", "gray"),
  text_colors = rev(rect_colors),
  show_labels = TRUE,
  label_angle = 0,
  row_ = "id",
  cluster_ = "cluster_id"
)
}
\arguments{
\item{cluster_ids}{Vector of cluster ids for each item in heatmap. Should be
sorted by plot order for heatmap.}

\item{p}{Optionally an existing ggplot to add annotation to.}

\item{xleft}{left side of cluster annotation rectangles. Default is 0.}

\item{xright}{right side of cluster annotation rectangles. Default is 1.}

\item{rect_colors}{colors of rectangle fill, repeat to match number of
clusters. Default is c("black", "gray").}

\item{text_colors}{colors of text, repeat to match number of clusters.
Default is reverse of rect_colors.}

\item{show_labels}{logical, shoud rectangles be labelled with cluster
identity.  Default is TRUE.}

\item{label_angle}{angle to add clusters labels at.  Default is 0, which is
horizontal.}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data. Default is "id" and works with ssvFetch* outputs.}

\item{cluster_}{variable name to use for cluster info. Default is "cluster_id".}
}
\value{
A ggplot with cluster annotations added.
}
\description{
adds rectangle boxes proportional to cluster sizes of heatmap with optional
labels.
}
\examples{
data(CTCF_in_10a_profiles_dt)
#simplest uses
add_cluster_annotation(factor(c(rep("A", 3), "B")))
p = ggplot() + coord_cartesian(xlim = c(0,10))
add_cluster_annotation(factor(c(rep("A", 3), "B")), p)

#intended use with ssvSignalHeatmap
clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 3)
assign_dt = unique(clust_dt[, .(id, cluster_id)])[order(id)]
p_heat = ssvSignalHeatmap(clust_dt, show_cluster_bars = FALSE)
add_cluster_annotation(assign_dt$cluster_id, p_heat,
  xleft = -500, xright = -360, rect_colors = rainbow(3), text_colors = "gray")

#when colors are named, the names are used rather that just the order
rect_colors = safeBrew(assign_dt$cluster_id)
text_colors = safeBrew(assign_dt$cluster_id, "greys")
p_clusters = add_cluster_annotation(assign_dt$cluster_id,
  rect_colors = rect_colors, text_colors = text_colors)
#specialized use as plot outside of heatmap
p1 = assemble_heatmap_cluster_bars(plots = list(p_clusters, p_heat), rel_widths = c(1, 3))

#when colors are named, the names are used rather that just the order
#these plots will be identical even though order of colors changes.
rect_colors = rect_colors[c(2, 3, 1)]
text_colors = text_colors[c(3, 1, 2)]
p_clusters = add_cluster_annotation(assign_dt$cluster_id,
  rect_colors = rect_colors, text_colors = text_colors)
#specialized use as plot outside of heatmap
p2 = assemble_heatmap_cluster_bars(plots = list(p_clusters, p_heat), rel_widths = c(1, 3))

cowplot::plot_grid(p1, p2, ncol = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bam.R
\name{.rm_dupes}
\alias{.rm_dupes}
\title{Remove duplicate reads based on stranded start position.  This is an
over-simplification.  For better duplicate handling, duplicates must be
marked in bam and flag passed to fetchBam() ... for ScanBamParam}
\usage{
.rm_dupes(reads_dt, max_dupes)
}
\arguments{
\item{reads_dt}{data.table of reads as loaded by fetchBam}

\item{max_dupes}{maximum allowed positional duplicates}
}
\value{
reads_dt with duplicated reads over max_dupes removed
}
\description{
flag = scanBamFlag(isDuplicate = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_operations.R
\name{centerAtMax}
\alias{centerAtMax}
\title{centers profile of x and y.  default is to center by region but across all
samples.}
\usage{
centerAtMax(
  dt,
  x_ = "x",
  y_ = "y",
  by_ = "id",
  view_size = NULL,
  trim_to_valid = TRUE,
  check_by_dupes = TRUE,
  x_precision = 3,
  replace_x = TRUE
)
}
\arguments{
\item{dt}{data.table}

\item{x_}{the variable name of the x-values. default is 'x'}

\item{y_}{the variable name of the y-values default is 'y'}

\item{by_}{optionally, any variables that provide grouping to the data.
default is none.  see details.}

\item{view_size}{the size in \code{x_} to consider for finding the max
of \code{y_}.
if length(view_size) == 1, range will be c(-view_size, view_size).
if length(view_size) > 1, range will be range(view_size).
default value of NULL uses complete range of x.}

\item{trim_to_valid}{valid \code{x_} values are those with a set \code{y_}
value in all \code{by_} combinations}

\item{check_by_dupes}{default assumption is that there should be on set of
x_ for a by_ instance.
if this is not the case and you want to disable warnings about set this
to FALSE.}

\item{x_precision}{numerical precision of x, default is 3.}

\item{replace_x}{logical, default TRUE.
if TRUE x_ will be replaced with position relative to summit.
if FALSE x_ will be preserved and x_summitPosition added.}
}
\value{
data.table with x (or xnew if replace_x is FALSE) shifted such that
x = 0 matches the maximum y-value define by by_ grouping
}
\description{
\code{centerAtMax} locates the coordinate x of the maximum in y and shifts x
such that it is zero at max y.
}
\details{
character.  by_ controls at the level of the data centering is
applied.  If by_ is "" or NULL, a single max position will be determined
for the entire dataset.  If by is "id" (the default) then each region will be
centered individually across all samples.
}
\examples{
data(CTCF_in_10a_profiles_gr)
centerAtMax(CTCF_in_10a_profiles_gr, y_ = 'y', by_ = 'id',
  check_by_dupes = FALSE)
#it's a bit clearer what's happening with trimming disabled
#but results are less useful for heatmaps etc.
centerAtMax(CTCF_in_10a_profiles_gr, y_ = 'y', by_ = 'id',
  check_by_dupes = FALSE, trim_to_valid = FALSE)
#specify view_size to limit range of x values considered, prevents
#excessive data trimming.
centerAtMax(CTCF_in_10a_profiles_gr, y_ = 'y', view_size = 100, by_ = 'id',
check_by_dupes = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_feature_plots.R
\name{ssvFeatureBars}
\alias{ssvFeatureBars}
\title{bar plots of set sizes}
\usage{
ssvFeatureBars(
  object,
  show_counts = TRUE,
  bar_colors = NULL,
  counts_text_colors = NULL,
  return_data = FALSE,
  count_label_size = 8
)
}
\arguments{
\item{object}{passed to ssvMakeMembTable for conversion to membership table}

\item{show_counts}{logical.  should counts be displayed at the center of each
bar. default is TRUE}

\item{bar_colors}{character. rcolor or hex colors. default of NULL
uses RColorBrewer Dark2. Will repeat to match number of samples.}

\item{counts_text_colors}{character. rcolor or hex colors. default of NULL
uses RColorBrewer Dark2. Will repeat to match number of samples.}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and
is instead the data used to generate that plot. Default is FALSE.}

\item{count_label_size}{Font size bar count labels. Default is 8.}
}
\value{
ggplot of bar plot of set sizes
}
\description{
bar plots of set sizes
}
\examples{
data(CTCF_in_10a_overlaps_gr)
ssvFeatureBars(list(1:3, 2:6))
ssvFeatureBars(CTCF_in_10a_overlaps_gr, count_label_size = 10)
ssvFeatureBars(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_signal.R
\name{prepare_fetch_GRanges_names}
\alias{prepare_fetch_GRanges_names}
\title{Creates a named version of input GRanges using the same method seqsetvis uses internally to ensure consistency.}
\usage{
prepare_fetch_GRanges_names(qgr, include_id = FALSE)
}
\arguments{
\item{qgr}{input GRanges object the set/check names on}

\item{include_id}{if TRUE, $id is retained. Default is FALSE.}
}
\value{
and named GRanges based on input qgr.
}
\description{
If $id is set, that value is used as name and duplicates are checked for.
}
\examples{
data(CTCF_in_10a_overlaps_gr)
qgr = CTCF_in_10a_overlaps_gr
names(qgr) = NULL
#default is to paste "region_" and iteration along length of qgr
prepare_fetch_GRanges_names(qgr)
#id gets used is already set
qgr$id = paste0("peak_", rev(seq_along(qgr)), "_of_", length(qgr))
prepare_fetch_GRanges_names(qgr)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/overlapIntervalSets.R
\name{ssvOverlapIntervalSets}
\alias{ssvOverlapIntervalSets}
\title{Intersect a list of GRanges to create a single GRanges object of
merged ranges including metadata describing overlaps per input GRanges}
\usage{
ssvOverlapIntervalSets(
  grs,
  ext = 0,
  use_first = FALSE,
  preserve_mcols = FALSE,
  ...
)
}
\arguments{
\item{grs}{A list of GRanges}

\item{ext}{An integer specifying how far to extend ranges before merging.
in effect, ranges withing 2*ext of one another will be joined during the
merge}

\item{use_first}{A logical.  If True, instead of merging all grs, only use
first and add metadata logicals for others.}

\item{preserve_mcols}{Controls carrying forward mcols metadata from input
list of GRanges. If TRUE, all mcols will be carried forward with the item
name appended. If a character vector, only those attributes will be carried
and all must be present in all GRanges. The default of FALSE will carry
nothing forward and only membership table will be generated.
\link{ssvAnnotateSubjectGRanges} is used internally.}

\item{...}{arguments passed to IRanges::findOverlaps, i.e. maxgap, minoverlap, type, select, invert.}
}
\value{
GRanges with metadata columns describing overlap of input grs.
}
\description{
Intersect a list of GRanges to create a single GRanges object of
merged ranges including metadata describing overlaps per input GRanges
}
\examples{
library(GenomicRanges)
a = GRanges("chr1", IRanges(1:7*10, 1:7*10))
b = GRanges("chr1", IRanges(5:10*10, 5:10*10))
ssvOverlapIntervalSets(list(a, b))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_CTCF.R
\docType{data}
\name{CTCF_in_10a_narrowPeak_urls}
\alias{CTCF_in_10a_narrowPeak_urls}
\title{FTP URL path for vignette data.
from}
\format{
named character vector of length 3
}
\description{
macs2 peak calls for CTCF ChIP-seq in a MCF10A progression model.
See  GEO series GSE98551 for details.
}
\details{
part of \code{\link{CTCF_in_10a_data}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ssvMakeMembTable.R
\name{ssvMakeMembTable}
\alias{ssvMakeMembTable}
\alias{ssvMakeMembTable,list-method}
\alias{ssvMakeMembTable,GRangesList-method}
\alias{ssvMakeMembTable,GRanges-method}
\alias{ssvMakeMembTable,DataFrame-method}
\alias{ssvMakeMembTable,matrix-method}
\alias{ssvMakeMembTable,data.frame-method}
\title{generic for methods to convert various objects to a logical matrix indicating
membership of items (rows) in sets (columns)}
\usage{
ssvMakeMembTable(object)

\S4method{ssvMakeMembTable}{list}(object)

\S4method{ssvMakeMembTable}{GRangesList}(object)

\S4method{ssvMakeMembTable}{GRanges}(object)

\S4method{ssvMakeMembTable}{DataFrame}(object)

\S4method{ssvMakeMembTable}{matrix}(object)

\S4method{ssvMakeMembTable}{data.frame}(object)
}
\arguments{
\item{object}{the object to convert. Supported types: list (of character
or GRanges), GRanges with membership table metadata, GRangesList,
data.frame/matrix/DataFrame of membership table}
}
\value{
a logical matrix indicating membership of items (rows) in sets
(columns)
}
\description{
generic for methods to convert various objects to a logical matrix indicating
membership of items (rows) in sets (columns)

list of character vectors input

GRangesList input

GRanges with mcols input

DataFrame input

matrix of logicals, membership table

data.frame input, final output
The final method for all inputs, checks column names and returns logical
matrix
}
\examples{
char_list = list(letters[1:3], letters[2:4])
ssvMakeMembTable(char_list)
library(GenomicRanges)
gr_list = list(GRanges("chr1", IRanges(1:3*2, 1:3*2)),
    GRanges("chr1", IRanges(2:4*2, 2:4*2)))
ssvMakeMembTable(gr_list)
library(GenomicRanges)
gr_list = list(GRanges("chr1", IRanges(1:3*2, 1:3*2)),
    GRanges("chr1", IRanges(2:4*2, 2:4*2)))
ssvMakeMembTable(GRangesList(gr_list))
gr = GRanges("chr1", IRanges(1:3*2, 1:3*2))
gr$set_a = c(TRUE, TRUE, FALSE)
gr$set_b = c(FALSE, TRUE, TRUE)
ssvMakeMembTable(gr)
gr = GRanges("chr1", IRanges(1:3*2, 1:3*2))
gr$set_a = c(TRUE, TRUE, FALSE)
gr$set_b = c(FALSE, TRUE, TRUE)
ssvMakeMembTable(mcols(gr))
memb_mat =  matrix(c(TRUE, TRUE, FALSE, FALSE, TRUE, FALSE, TRUE, FALSE),
    ncol = 2, byrow = FALSE)
ssvMakeMembTable(memb_mat)
memb_df = data.frame(a = c(TRUE, TRUE, FALSE, FALSE),
    b = c(TRUE, FALSE, TRUE, FALSE))
ssvMakeMembTable(memb_df)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_feature_plots.R
\name{ssvFeatureUpset}
\alias{ssvFeatureUpset}
\title{ssvFeatureUpset}
\usage{
ssvFeatureUpset(
  object,
  return_UpSetR = FALSE,
  nsets = NULL,
  nintersects = 15,
  order.by = "freq",
  ...
)
}
\arguments{
\item{object}{will be passed to \code{\link{ssvMakeMembTable}} for conversion
to membership matrix}

\item{return_UpSetR}{If TRUE, return the UpSetR object, The default is FALSE
and results in a ggplotified version compatible with cowplot etc.}

\item{nsets}{Number of sets to look at}

\item{nintersects}{Number of intersections to plot. If set to NA, all
intersections will be plotted.}

\item{order.by}{How the intersections in the matrix should be ordered by.
Options include frequency (entered as "freq"), degree, or both in any
order.}

\item{...}{Additional parameters passed to \code{\link[UpSetR]{upset}} in the
UpSetR  package.}
}
\value{
ggplot version of UpSetR plot
}
\description{
Uses the UpSetR package to create an \link[UpSetR:upset]{UpSetR::upset} plot of region overlaps.
}
\examples{
data(CTCF_in_10a_overlaps_gr)
ssvFeatureUpset(list(1:3, 2:6))
ssvFeatureUpset(CTCF_in_10a_overlaps_gr)
ssvFeatureUpset(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_plots.R
\name{ssvSignalHeatmap.ClusterBars}
\alias{ssvSignalHeatmap.ClusterBars}
\title{heatmap style representation of membership table. instead of clustering, each
column is sorted starting from the left.}
\usage{
ssvSignalHeatmap.ClusterBars(
  bw_data,
  nclust = 6,
  perform_clustering = c("auto", "yes", "no")[1],
  row_ = "id",
  column_ = "x",
  fill_ = "y",
  facet_ = "sample",
  cluster_ = "cluster_id",
  FUN_format_heatmap = NULL,
  max_rows = 500,
  max_cols = 100,
  fill_limits = NULL,
  clustering_col_min = -Inf,
  clustering_col_max = Inf,
  within_order_strategy = c("hclust", "sort")[2],
  dcast_fill = NA,
  return_data = FALSE,
  return_unassembled_plots = FALSE,
  rel_widths = c(1, 9),
  rect_colors = c("black", "gray"),
  text_colors = rev(rect_colors),
  show_labels = TRUE,
  label_angle = 0,
  fun.aggregate = "mean",
  ...
)
}
\arguments{
\item{bw_data}{a GRanges or data.table of bigwig signal. As returned from
\code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}}

\item{nclust}{number of clusters}

\item{perform_clustering}{should clustering be done? default is auto. auto
considers if row_ has been ordered by being a factor and if cluster_ is a
numeric.}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{column_}{varaible mapped to column, likely bp position for ngs data.
Default is "x" and works with ssvFetch* output.}

\item{fill_}{numeric variable to map to fill. Default is "y" and works with
ssvFetch* output.}

\item{facet_}{variable name to facet horizontally by. Default is "sample" and
works with ssvFetch* output. Set to "" if data is not facetted.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{FUN_format_heatmap}{optional function to modify main ggplot (labels,
themes, scales, etc.).  Take a ggplot and returns a ggplot. Default is
NULL.}

\item{max_rows}{for speed rows are sampled to 500 by default, use Inf to plot
full data}

\item{max_cols}{for speed columns are sampled to 100 by default, use Inf to
plot full data}

\item{fill_limits}{limits for fill legend.  values will be cropped to this
range if set.  Default of NULL uses natural range of fill_.}

\item{clustering_col_min}{numeric minimum for col range considered when
clustering, default in -Inf}

\item{clustering_col_max}{numeric maximum for col range considered when
clustering, default in Inf}

\item{within_order_strategy}{one of "hclust" or "sort".  if hclust,
hierarchical clustering will be used. if sort, a simple decreasing sort of
rosSums.}

\item{dcast_fill}{value to supply to dcast fill argument. default is NA.}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and is
instead the data used to generate that plot. Default is FALSE.}

\item{return_unassembled_plots}{logical. If TRUE, return list of heatmap and
cluster-bar ggplots.  Can be customized and passed to
\code{\link{assemble_heatmap_cluster_bars}}}

\item{rel_widths}{numeric of length 2.  Passed to cowplot::plot_grid. Default
is c(1, 9).}

\item{rect_colors}{colors of rectangle fill, repeat to match number of
clusters. Default is c("black", "gray").}

\item{text_colors}{colors of text, repeat to match number of clusters.
Default is reverse of rect_colors.}

\item{show_labels}{logical, shoud rectangles be labelled with cluster
identity.  Default is TRUE.}

\item{label_angle}{angle to add clusters labels at.  Default is 0, which is
horizontal.}

\item{fun.aggregate}{Function to aggregate when multiple values present for
facet_, row_, and column_. Affects both clustering and plotting. The
function should accept a single vector argument or be a character string
naming such a function.}

\item{...}{addtional arguments passed to cowplot::plot_grid}
}
\value{
ggplot heatmap of signal profiles, facetted by sample
}
\description{
Compared to ssvSignalHeatmap, cluster_bars are displayed on the left once
instead of for each facet
}
\examples{
data(CTCF_in_10a_profiles_gr)

#the simplest use
ssvSignalHeatmap.ClusterBars(CTCF_in_10a_profiles_gr)
ssvSignalHeatmap.ClusterBars(CTCF_in_10a_profiles_gr, rel_widths = c(1, 5))

#clustering can be done manually beforehand
clust_dt = ssvSignalClustering(data.table::as.data.table(CTCF_in_10a_profiles_gr), nclust = 3)
ssvSignalHeatmap.ClusterBars(clust_dt)

# aggregation occurs when facet_ is shared by multiple samples
prof_gr = CTCF_in_10a_profiles_gr
prof_gr$mark = "CTCF"

# by default, aggregation is done by mean
ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark")

# Using fun.aggregate we can control aggregation behavior
# 'fun.aggregate = mean' preserves the default behavior
ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark",
  fun.aggregate = mean)


# maximum and minimum may be useful to plot
ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark",
  fun.aggregate = max)
ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark",
  fun.aggregate = min)

# How many samples have at least a signal value of 10 at each position?
ssvSignalHeatmap.ClusterBars(prof_gr, facet_ = "mark",
  fun.aggregate = function(x)sum(as.numeric(x > 10)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{split_cluster}
\alias{split_cluster}
\title{split_cluster}
\usage{
split_cluster(
  clust_dt,
  to_split,
  nclust = 2,
  row_ = "id",
  column_ = "x",
  fill_ = "y",
  facet_ = "sample",
  cluster_ = "cluster_id",
  reapply_cluster_names = TRUE
)
}
\arguments{
\item{clust_dt}{data.table output from \code{\link{ssvSignalClustering}}}

\item{to_split}{Cluster to split.}

\item{nclust}{Number of new clusters to create.}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{column_}{varaible mapped to column, likely bp position for ngs data.
Default is "x" and works with ssvFetch* output.}

\item{fill_}{numeric variable to map to fill. Default is "y" and works with
ssvFetch* output.}

\item{facet_}{variable name to facet horizontally by. Default is "sample" and
works with ssvFetch* output. Set to "" if data is not facetted.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{reapply_cluster_names}{If TRUE, clusters will be renamed according to
new order instead of their original names. Default is TRUE.}
}
\value{
data.table as output from \code{\link{ssvSignalClustering}}
}
\description{
Splits one specified cluster in number of new clusters determined by nclust
}
\examples{
data(CTCF_in_10a_profiles_dt)
set.seed(0)
clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 3)
split_dt = split_cluster(clust_dt, to_split = 2, nclust = 3)
split_dt.no_rename = split_cluster(
  clust_dt,
  to_split = 2,
  nclust = 3,
  reapply_cluster_names = FALSE
)
cowplot::plot_grid(nrow = 1,
  ssvSignalHeatmap(clust_dt),
  ssvSignalHeatmap(split_dt),
  ssvSignalHeatmap(split_dt.no_rename)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bam.R
\name{fetchBam}
\alias{fetchBam}
\title{fetch a bam file pileup with the ability to consider read extension to
fragment size (fragLen)}
\usage{
fetchBam(
  bam_f,
  qgr,
  fragLen = NULL,
  target_strand = c("*", "+", "-")[1],
  max_dupes = Inf,
  splice_strategy = c("none", "ignore", "add", "only", "splice_count")[1],
  flip_strand = FALSE,
  return_unprocessed = FALSE,
  ...
)
}
\arguments{
\item{bam_f}{character or BamFile to load}

\item{qgr}{GRanges regions to fetchs}

\item{fragLen}{numeric, NULL, or NA.  if numeric, supplied value is used. if
NULL, value is calculated with fragLen_calcStranded (default) if NA, raw
bam pileup with no cross strand shift is returned.}

\item{target_strand}{character. if one of "+" or "-", reads are filtered to
match. ignored if any other value.}

\item{max_dupes}{numeric >= 1.  duplicate reads by strandd start position
over this number are removed, Default is Inf.}

\item{splice_strategy}{character, one of c("none", "ignore", "add", "only").
Default is "none" and split read alignments are asssumed not present.
fragLen must be NA for any other value to be valid.  "ignore" will not
count spliced regions.  "add" counts spliced regions along with others,
"only" will only count spliced regions and ignore others.}

\item{flip_strand}{if TRUE, strand alignment is flipped prior to fragLen
extension. Default is FALSE.}

\item{return_unprocessed}{boolean. if TRUE returns read alignment in data.table. Default is FALSE.}

\item{...}{passed to ScanBamParam(), can't be which or what.}
}
\value{
GRanges containing tag pileup values in score meta column.  tags are
optionally extended to fragment length (fragLen) prior to pile up.
}
\description{
fetch a bam file pileup with the ability to consider read extension to
fragment size (fragLen)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_GRanges.R
\name{ssvFetchGRanges}
\alias{ssvFetchGRanges}
\title{Fetch coverage values for a list of GRanges.}
\usage{
ssvFetchGRanges(
  grs,
  qgr,
  file_attribs = data.frame(matrix(0, nrow = length(grs), ncol = 0)),
  unique_names = names(grs),
  names_variable = "sample",
  win_size = 50,
  win_method = c("sample", "summary")[1],
  summary_FUN = function(x, w) max(x),
  target_strand = c("*", "+", "-", "both")[1],
  use_coverage = NULL,
  attrib_var = "score",
  fill_value = 0,
  anchor = c("left", "left_unstranded", "center", "center_unstranded")[3],
  return_data.table = FALSE,
  n_cores = getOption("mc.cores", 1),
  force_skip_centerFix = FALSE
)
}
\arguments{
\item{grs}{a list of GRanges for which to calculate coverage.}

\item{qgr}{Set of GRanges to query.  For valid results the width of each
interval should be identical and evenly divisible by \code{win_size}.}

\item{file_attribs}{data.frame (1 row per item in grs) containing attributes
to append to results.}

\item{unique_names}{The column name where unique_names are stored. Default is
'sample'}

\item{names_variable}{The column name where unique_names are stored. Default
is 'sample'}

\item{win_size}{The window size that evenly divides widths in \code{qgr}.}

\item{win_method}{character.  one of c("sample", "summary").  Determines if
\code{\link{viewGRangesWinSample_dt}} or
\code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
qgr.}

\item{summary_FUN}{function.  only relevant if win_method is "summary".
passed to \code{\link{viewGRangesWinSummary_dt}}.}

\item{target_strand}{character. if one of "+" or "-", reads are filtered to
match. ignored if any other value.}

\item{use_coverage}{boolean or NULL, if TRUE, query regions are scored by the
number of intervals overlapping.  Default of NULL checks if attrib_var is
"score" and uses coverage if so.}

\item{attrib_var}{character, column in mcols of GRanges to pull values from.
Default of "score" is compatible with internal coverage calculation or
bedgraph-like files.}

\item{fill_value}{numeric or character value to use where queried regions are
empty.  Default is 0 and appropriate for both calculated coverage and
bedgraph/bigwig like files.  Will automatically switch to "MISSING" if data
is guessed to be qualitative.}

\item{anchor}{character, one of c("center", "center_unstranded", "left",
"left_unstranded")}

\item{return_data.table}{logical. If TRUE the internal data.table is returned
instead of GRanges.  Default is FALSE.}

\item{n_cores}{integer number of cores to use. Uses mc.cores option if not
supplied.}

\item{force_skip_centerFix}{boolean, if TRUE all query ranges will be
used "as is".  This is already the case by default if win_method == "summary"
but may have applications where win_method == "sample".}
}
\value{
A tidy formatted GRanges (or data.table if specified) containing
fetched values.
}
\description{
\code{ssvFetchGRanges} Gets coverage values for each region of the query
GRanges (\code{qgr}). Values correspond to the center of each window of size
\code{win_size}.  A tidy formatted data.table object is returned suitable for
plotting using ggplots.
}
\examples{
data(CTCF_in_10a_narrowPeak_grs)
data(CTCF_in_10a_overlaps_gr)
ssvFetchGRanges(CTCF_in_10a_narrowPeak_grs, CTCF_in_10a_overlaps_gr, win_size = 200)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{clusteringKmeansNestedHclust}
\alias{clusteringKmeansNestedHclust}
\title{perform kmeans clustering on matrix rows and return reordered matrix along
with order matched cluster assignments clusters are sorted using hclust on
centers the contents of each cluster are sorted using hclust}
\usage{
clusteringKmeansNestedHclust(
  mat,
  nclust,
  within_order_strategy = valid_sort_strategies[2],
  centroids = NULL,
  manual_mapping = NULL,
  iter.max = 30
)
}
\arguments{
\item{mat}{A wide format matrix}

\item{nclust}{the number of clusters}

\item{within_order_strategy}{one of "hclust", "sort", "right", "left",
"reverse".  If "hclust", hierarchical clustering will be used. If "sort", a
simple decreasing sort of rosSums.  If "left", will atttempt to put high
signal on left ("right" is opposite).  If "reverse" reverses existing order
(should only be used after meaningful order imposed).}

\item{centroids}{optional matrix with same columns as mat and one centroid
per row to base clusters off of.  Overrides any setting to nclust. Default
of NULL results in randomly initialized k-means.}

\item{manual_mapping}{optional named vector manually specififying cluster
assignments. names should be item ids and values should be cluster names
the items are assigned to. Default of NULL allows clustering to proceed.}

\item{iter.max}{Number of max iterations to allow for k-means. Default is 30.}
}
\value{
data.table with 2 columns of cluster info. id__ column corresponds
with input matrix rownames and is sorted within each cluster using
hierarchical clusering group__ column indicates cluster assignment
}
\description{
perform kmeans clustering on matrix rows and return reordered matrix along
with order matched cluster assignments clusters are sorted using hclust on
centers the contents of each cluster are sorted using hclust
}
\examples{
data(CTCF_in_10a_profiles_dt)
dt = data.table::copy(CTCF_in_10a_profiles_dt)
mat = data.table::dcast(dt, id ~ sample + x, value.var = "y" )
rn = mat$id
mat = as.matrix(mat[,-1])
rownames(mat) = rn
clust_dt = clusteringKmeansNestedHclust(mat, nclust = 3)
clust_dt
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bam.R
\name{expandCigar}
\alias{expandCigar}
\title{Expand cigar codes to GRanges}
\usage{
expandCigar(
  cigar_dt,
  op_2count = c("M", "D", "=", "X"),
  return_data.table = FALSE
)
}
\arguments{
\item{cigar_dt}{data.table with 5 required named columns in any order.
c("which_label", "seqnames", "strand", "start", "cigar")}

\item{op_2count}{Cigar codes to count. Default is alignment (M), deletion
(D), match (=), and mismatch (X).  Other useful codes may be skipped
regions for RNA splicing (N).  The locations of any insterions (I) or
clipping/padding (S, H, or P) will be a single bp immediately before the
interval.}

\item{return_data.table}{if TRUE, a data.table is returned, else a GRanges.
Default is FALSE.}
}
\value{
data.table with cigar entries expanded
}
\description{
see \href{https://samtools.github.io/hts-specs/SAMv1.pdf}{sam specs} for
cigar details
}
\examples{
data(CTCF_in_10a_overlaps_gr)
qgr = CTCF_in_10a_overlaps_gr[1:5]
bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
raw_dt = ssvFetchBam(bam_file, qgr, return_unprocessed = TRUE)
expandCigar(raw_dt)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bam.R
\name{ssvFetchBam}
\alias{ssvFetchBam}
\title{Iterates a character vector (ideally named) and calls
\code{ssvFetchBam.single} on each.  Appends grouping variable to each
resulting data.table and uses rbindlist to efficiently combine results}
\usage{
ssvFetchBam(
  file_paths,
  qgr,
  unique_names = NULL,
  names_variable = "sample",
  file_attribs = NULL,
  win_size = 50,
  win_method = c("sample", "summary")[1],
  summary_FUN = stats::weighted.mean,
  fragLens = "auto",
  target_strand = c("*", "+", "-", "both")[1],
  flip_strand = FALSE,
  anchor = c("left", "left_unstranded", "center", "center_unstranded")[3],
  return_data.table = FALSE,
  max_dupes = Inf,
  splice_strategy = c("none", "ignore", "add", "only", "splice_count")[1],
  n_cores = getOption("mc.cores", 1),
  n_region_splits = 1,
  return_unprocessed = FALSE,
  force_skip_centerFix = FALSE,
  ...
)
}
\arguments{
\item{file_paths}{character vector of file_paths to load from. Alternatively,
file_paths can be a data.frame or data.table whose first column is a
character vector of paths and additial columns will be used as metadata.}

\item{qgr}{Set of GRanges to query.  For valid results the width of each
interval should be identical and evenly divisible by \code{win_size}.}

\item{unique_names}{names to use in final data.table to designate source
bigwig. Default is 'sample'}

\item{names_variable}{The column name where unique_names are stored.}

\item{file_attribs}{optional data.frame/data.table with one row per item in
file paths.  Each column will be a variable added to final tidy output.}

\item{win_size}{The window size that evenly divides widths in \code{qgr}.}

\item{win_method}{character.  one of c("sample", "summary").  Determines if
\code{\link{viewGRangesWinSample_dt}} or
\code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
qgr.}

\item{summary_FUN}{function.  only relevant if win_method is "summary".
passed to \code{\link{viewGRangesWinSummary_dt}}.}

\item{fragLens}{numeric. The fragment length to use to extend reads.  The
default value "auto" causes an automatic calculation from 100 regions in
qgr.  NA causes no extension of reads to fragment size.}

\item{target_strand}{character. One of c("\emph{", "+", "-"). Controls filtering
of reads by strand.  Default of "}" combines both strands.}

\item{flip_strand}{boolean. if TRUE strands are flipped.}

\item{anchor}{character, one of c("center", "center_unstranded", "left",
"left_unstranded")}

\item{return_data.table}{logical. If TRUE the internal data.table is returned
instead of GRanges.  Default is FALSE.}

\item{max_dupes}{numeric >= 1.  duplicate reads by strandd start position
over this number are removed, Default is Inf.}

\item{splice_strategy}{character, one of c("none", "ignore", "add", "only",
"splice_count"). Default is "none" and spliced alignment are asssumed not
present. fragLen will be forced to be NA for any other value.  "ignore" will
not count spliced regions.  add" counts spliced regions along with others,
"only" will only count spliced regions and ignore others.}

\item{n_cores}{integer number of cores to use.
Uses mc.cores option if not supplied.}

\item{n_region_splits}{integer number of splits to apply to qgr. The query
GRanges will be split into this many roughly equal parts for increased
parallelization. Default is 1, no split.}

\item{return_unprocessed}{boolean. if TRUE returns read alignment in data.table. Default is FALSE.}

\item{force_skip_centerFix}{boolean, if TRUE all query ranges will be
used "as is".  This is already the case by default if win_method == "summary"
but may have applications where win_method == "sample".}

\item{...}{passed to Rsamtools::ScanBamParam()}
}
\value{
A tidy formatted GRanges (or data.table if specified) containing
fetched values.
}
\description{
\code{ssvFetchBam} iteratively calls \code{fetchWindowedBam.single}. See
\code{\link{ssvFetchBam.single}} for more info.
}
\details{
if \code{qgr} contains the range chr1:1-100 and \code{win_size} is
10, values from positions chr1 5,15,25...85, and 95 will be retrieved from
\code{bw_file}
}
\examples{
if(Sys.info()['sysname'] != "Windows"){
data(CTCF_in_10a_overlaps_gr)
library(GenomicRanges)
bam_f = system.file("extdata/test.bam",
    package = "seqsetvis", mustWork = TRUE)
bam_files = c("a" = bam_f, "b" = bam_f)
qgr = CTCF_in_10a_overlaps_gr[1:5]
bw_gr = ssvFetchBam(bam_files, qgr, win_size = 10)
bw_gr2 = ssvFetchBam(as.list(bam_files), qgr, win_size = 10)

bw_dt = ssvFetchBam(bam_files, qgr, win_size = 10,
    return_data.table = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_normalization.R
\name{calc_norm_factors}
\alias{calc_norm_factors}
\title{calc_norm_factors}
\usage{
calc_norm_factors(
  full_dt,
  value_ = "y",
  cap_value_ = "y_cap_value",
  by1 = "id",
  by2 = "sample",
  aggFUN1 = max,
  aggFUN2 = function(x) quantile(x, 0.95)
)
}
\arguments{
\item{full_dt}{a data.table, as returned by ssvFetch*(..., return_data.table.
= TRUE)}

\item{value_}{character, attribute in full_dt to normalzie.}

\item{cap_value_}{character, new attribute name specifying values to cap to.}

\item{by1}{character vector, specifies attributes relevant to step 1.}

\item{by2}{character vector, specifies attributes relevant to step 1 and 2.}

\item{aggFUN1}{function called on value_ with by = c(by1, by2) in step 1.}

\item{aggFUN2}{function called on result of aggFUN1 with by = by2 in step 2.}
}
\value{
data.table mapping by2 to cap_value_.
}
\description{
Calculate normalization factors in a two step process:
}
\details{
\enumerate{
\item summarize every region for each sample (default summary function is max)
\item caclulate a value to cap each sample to based on regions (default is 95th
quantile).
}

The uderlying assumption here is that meaningful enrichment is present at the
majority of regions provided.  If prevalence varies by a specific factor, say
ChIP-seq targets with different characteristics - ie. when analyzing TSSes
for H3K4me3 and an infrequent transcription factor it is more appropriate to
specify appropriate quantile cutoffs per factor.
}
\examples{
data(CTCF_in_10a_profiles_dt)
calc_norm_factors(CTCF_in_10a_profiles_dt)
calc_norm_factors(CTCF_in_10a_profiles_dt,
  aggFUN1 = mean, aggFUN2 = function(x)quantile(x, .5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fragLen.R
\name{fragLen_calcStranded}
\alias{fragLen_calcStranded}
\title{calculate fragLen from a bam file for specified regions}
\usage{
fragLen_calcStranded(
  bam_f,
  qgr,
  n_regions = 100,
  include_plot_in_output = FALSE,
  test_fragLen = seq(100, 400, 5),
  flip_strand = FALSE,
  ...
)
}
\arguments{
\item{bam_f}{character or BamFile. bam file to read from.
.bai index file must be in same directory}

\item{qgr}{GRanges.  used as which for ScanBamParam. Can be NULL if it's
REALLY important to load the entire bam, force_no_which = TRUE also required.}

\item{n_regions}{numeric (integer) it's generally overkill to pull all
regions at this stage and will slow calculation down.  Default is 100.}

\item{include_plot_in_output}{if TRUE ouptut is a list of fragLen and a
ggplot showing values considered by calculation. Default is FALSE.}

\item{test_fragLen}{numeric.  The set of fragment lenghts to gather
strand cross correlation for.}

\item{flip_strand}{boolean. if TRUE strands that reads align to are swapped.
This is typically only necessary if there was a mismatch between library
chemistry and aligner settings. Default is FALSE.}

\item{...}{passed to Rsamtools::ScanBamParam, can't be which or what.}
}
\value{
numeric fragment length
}
\description{
calculate fragLen from a bam file for specified regions
}
\examples{
data(CTCF_in_10a_overlaps_gr)
bam_file = system.file("extdata/test.bam",
    package = "seqsetvis")
qgr = CTCF_in_10a_overlaps_gr[1:5]
fragLen_calcStranded(bam_file, qgr)
#if plot is included, a list is returned, item 2 is the plot
fragLen_calcStranded(bam_file, qgr,
  include_plot_in_output = TRUE)[[2]]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_CTCF.R
\docType{data}
\name{CTCF_in_10a_narrowPeak_grs}
\alias{CTCF_in_10a_narrowPeak_grs}
\title{list of GRanges that results in 100 random subset when overlapped}
\format{
named list of GRanges of length 3
}
\value{
named list of GRanges of length 3
}
\description{
list of GRanges that results in 100 random subset when overlapped
}
\details{
part of \code{\link{CTCF_in_10a_data}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_CTCF.R
\docType{data}
\name{CTCF_in_10a_overlaps_gr}
\alias{CTCF_in_10a_overlaps_gr}
\title{100 randomly selected regions from overlapping CTCF peaks in 10a cell
ChIP-seq}
\format{
GenomicRanges with 3 metadata columns of membership table
}
\description{
MACS2 narrowPeak calls on pooled biological replicates at pval 1e-5 and
then 0.05 IDR filtered. IDR cutoffs determined  by comparing top 150,000
pvalue sorted peak in replicates.
}
\details{
See  GEO series GSE98551 for details.

part of \code{\link{CTCF_in_10a_data}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_helpers.R
\name{ssv_mclapply}
\alias{ssv_mclapply}
\title{ssv_mclapply}
\usage{
ssv_mclapply(X, FUN, mc.cores = getOption("mc.cores", 1), ...)
}
\arguments{
\item{X}{For pbsapply and pblapply, a vector (atomic or list) or an
expressions vector (other objects including classed objects will be coerced
by as.list.) For pbapply an array, including a matrix. For pbtapply an R
object for which a split method exists. Typically vector-like, allowing
subsetting with "[".}

\item{FUN}{The function to be applied to each element of X: see apply,
sapply, and lapply. In the case of functions like +, '\%*\%', etc., the
function name must be backquoted or quoted. If FUN is NULL, pbtapply
returns a vector which can be used to subscript the multi-way array
pbtapply normally produces.}

\item{mc.cores}{Number of cores to use for pbmclapply. Defaults to option
mc.cores.}

\item{...}{passed to pbapply::pblapply or pbmcapply::pbmclapply}
}
\value{
result of either pblapply or pbmclapply
}
\description{
ssv_mclapply
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_easyLoad.R
\name{easyLoad_IDRmerged}
\alias{easyLoad_IDRmerged}
\title{easyLoad_IDRmerged loads "overlapped-peaks.txt" from IDR.}
\usage{
easyLoad_IDRmerged(
  file_paths,
  file_names = NULL,
  n_cores = getOption("mc.cores", 1),
  max_idr = 0.05
)
}
\arguments{
\item{file_paths}{character vector of paths to narrowPeak files.  If named,
those names will be used in output unless overriden by providing file_names.}

\item{file_names}{character vector of names for output list.  If not NULL
will override any existing names for file_paths.  Default is NULL.}

\item{n_cores}{number of cores to use, uses mc.cores option if set or 1.}

\item{max_idr}{maximum IDR value allowed}
}
\value{
named list of GRanges
}
\description{
easyLoad_IDRmerged loads "overlapped-peaks.txt" from IDR.
}
\examples{
idr_file = system.file("extdata/test_idr.overlapped-peaks.txt",
    package = "seqsetvis", mustWork = TRUE)
easyLoad_IDRmerged(idr_file)
easyLoad_IDRmerged(idr_file, max_idr = .01)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bamPE.R
\name{ssvFetchBamPE.single}
\alias{ssvFetchBamPE.single}
\title{fetch a windowed version of a paired-end bam file, returns GRanges
In contrast to ssvFetchBam, extension of reads to estimated fragment size is
not an issue as each read pair represents a fragment of exact size.}
\usage{
ssvFetchBamPE.single(
  bam_f,
  qgr,
  win_size = 50,
  win_method = c("sample", "summary")[1],
  summary_FUN = stats::weighted.mean,
  anchor = c("left", "left_unstranded", "center", "center_unstranded")[3],
  return_data.table = FALSE,
  max_dupes = Inf,
  min_isize = 1,
  max_isize = Inf,
  return_unprocessed = FALSE,
  return_fragSizes = FALSE,
  force_skip_centerFix = FALSE,
  ...
)
}
\arguments{
\item{bam_f}{character or BamFile to load}

\item{qgr}{GRanges regions to fetchs}

\item{win_size}{numeric >=1.  pileup grabbed every win_size bp for win_method
sample.  If win_method is summary, this is the number of windows used
(confusing, sorry).}

\item{win_method}{character.  one of c("sample", "summary").  Determines if
\code{\link{viewGRangesWinSample_dt}} or
\code{\link{viewGRangesWinSummary_dt}} is used to represent each region in
qgr.}

\item{summary_FUN}{function.  only relevant if win_method is "summary".
passed to \code{\link{viewGRangesWinSummary_dt}}.}

\item{anchor}{character, one of c("center", "center_unstranded", "left",
"left_unstranded")}

\item{return_data.table}{logical. If TRUE the internal data.table is returned
instead of GRanges.  Default is FALSE.}

\item{max_dupes}{numeric >= 1.  duplicate reads by strandd start position
over this number are removed, Default is Inf.}

\item{min_isize}{integer. Read pairs must have an isize greater than or equal to this value.  Default is 1.}

\item{max_isize}{integer. Read pairs must have an isize less than or equal to this value.  Default is Inf.}

\item{return_unprocessed}{boolean. if TRUE returns read alignment in data.table. Default is FALSE.}

\item{return_fragSizes}{boolean. if TRUE returns fragment sizes for all reads per region.}

\item{force_skip_centerFix}{boolean, if TRUE all query ranges will be
used "as is".  This is already the case by default if win_method == "summary"
but may have applications where win_method == "sample".}

\item{...}{passed to Rsamtools::ScanBamParam()}
}
\value{
tidy GRanges (or data.table if specified) with pileups from bam file.
pileup is calculated only every win_size bp.
}
\description{
fetch a windowed version of a paired-end bam file, returns GRanges
In contrast to ssvFetchBam, extension of reads to estimated fragment size is
not an issue as each read pair represents a fragment of exact size.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_feature_plots.R
\name{ssvFeatureVenn}
\alias{ssvFeatureVenn}
\title{ssvFeatureVenn}
\usage{
ssvFeatureVenn(
  object,
  group_names = NULL,
  counts_txt_size = 5,
  counts_as_labels = FALSE,
  show_outside_count = FALSE,
  line_width = 3,
  circle_colors = NULL,
  fill_alpha = 0.3,
  line_alpha = 1,
  counts_color = NULL,
  counts_as_percent = FALSE,
  percentage_digits = 1,
  percentage_suffix = "\%",
  n_points = 200,
  return_data = FALSE
)
}
\arguments{
\item{object}{will be passed to \link{ssvMakeMembTable} for
conversion to membership matrix}

\item{group_names}{useful if names weren't provided or were lost in
creating membership matrix}

\item{counts_txt_size}{font size for count numbers}

\item{counts_as_labels}{if TRUE, geom_label is used instead of geom_text.
can be easier to read.}

\item{show_outside_count}{if TRUE, items outside of all sets are counted
outside. can be confusing.}

\item{line_width}{uses size aesthetic to control line width of circles.}

\item{circle_colors}{colors to use for circle line colors. Uses Dark2 set
from RColorBrewer by default.}

\item{fill_alpha}{alpha value to use for fill, defaults to .3.}

\item{line_alpha}{numeric value from 0 to 1. Alpha value for circle line}

\item{counts_color}{character. single color to use for displaying counts}

\item{counts_as_percent}{if TRUE, convert counts to percentages in plots.}

\item{percentage_digits}{The number of digits to round percentages to, default is 1.}

\item{percentage_suffix}{The character to append to percentages, default is "\%".}

\item{n_points}{integer.  number of points to approximate circle with.
default is 200.}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and
is instead the data used to generate that plot. Default is FALSE.}
}
\value{
ggplot venn diagram
}
\description{
ggplot implementation of vennDiagram from limma package.  Currently limited
at 3 sets.  \link{ssvFeatureUpset} and \link{ssvFeatureBinaryHeatmap} are good options for
more than 3 sets. \link{ssvFeatureEuler} can work too but can take a very long time
to run for more than 5 or so.
}
\examples{
data(CTCF_in_10a_overlaps_gr)
ssvFeatureVenn(list(1:3, 2:6))
ssvFeatureVenn(CTCF_in_10a_overlaps_gr)
ssvFeatureVenn(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])

ssvFeatureVenn(list(1:3, 2:6),
  counts_as_percent = TRUE,
  percentage_digits = 2)

ssvFeatureVenn(list(1:3, 2:6),
  counts_as_percent = TRUE,
  percentage_digits = 0,
  percentage_suffix = " \%",
  counts_txt_size = 12)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ssvMakeMembTable.R
\name{ssvFactorizeMembTable}
\alias{ssvFactorizeMembTable}
\title{Convert any object accepted by ssvMakeMembTable to a factor
To avoid ambiguity,}
\usage{
ssvFactorizeMembTable(object)
}
\arguments{
\item{object}{a valid object for conversion to a membership table and then
factor}
}
\value{
a 2 column ("id" and "group") data.frame.
"id" is factor of item names if any or simply order of items.
"group" is a factor of set combinations
}
\description{
see \code{\link{ssvMakeMembTable}}
}
\examples{
data(CTCF_in_10a_overlaps_gr)
ssvFactorizeMembTable(CTCF_in_10a_overlaps_gr)
ssvFactorizeMembTable(list(1:4, 2:3, 4:6))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/annotate_subject_GR.R
\name{ssvAnnotateSubjectGRanges}
\alias{ssvAnnotateSubjectGRanges}
\alias{ssvAnnotateSubjectGRanges,GRanges-method}
\alias{ssvAnnotateSubjectGRanges,list-method}
\alias{ssvAnnotateSubjectGRanges,GRangesList-method}
\title{ssvAnnotateSubjectGRanges}
\usage{
ssvAnnotateSubjectGRanges(
  annotation_source,
  subject_gr,
  annotation_name = NULL,
  multi_resolver_FUN = "default"
)

\S4method{ssvAnnotateSubjectGRanges}{GRanges}(
  annotation_source,
  subject_gr,
  annotation_name = NULL,
  multi_resolver_FUN = "default"
)

\S4method{ssvAnnotateSubjectGRanges}{list}(
  annotation_source,
  subject_gr,
  annotation_name = NULL,
  multi_resolver_FUN = "default"
)

\S4method{ssvAnnotateSubjectGRanges}{GRangesList}(
  annotation_source,
  subject_gr,
  annotation_name = NULL,
  multi_resolver_FUN = "default"
)
}
\arguments{
\item{annotation_source}{A single GRanges, a list of GRanges, or a
GRangesList}

\item{subject_gr}{The base GRanges to add annotation mcols to.}

\item{annotation_name}{Optional name for single GRanges. Required for list
inputs if list does not have names.}

\item{multi_resolver_FUN}{Optional function to resolve multiple overlapping
annotation source regions per subject region. This function must accept 2
arguments. \code{x} is the values in a single mcol attribute and \code{variable.name}
is the name of variable. A single value must be returned or an error will
be generated. The default of "default" can handle numeric, logical,
character, and factor types.}
}
\value{
GRanges with the same regions as \code{subject_gr} but with addtional
mcols added from annotation_source.
}
\description{
ssvAnnotateSubjectGRanges
}
\examples{
library(GenomicRanges)
data(CTCF_in_10a_narrowPeak_grs)
np_grs = CTCF_in_10a_narrowPeak_grs
olap_gr = ssvOverlapIntervalSets(np_grs)
# annotating with a signle GRanges is OK
ssvAnnotateSubjectGRanges(np_grs$MCF10A_CTCF, olap_gr)
# provide a name if that's useful
ssvAnnotateSubjectGRanges(np_grs$MCF10A_CTCF, olap_gr,
  annotation_name = "MCF10A")
# a named list adds each annotation
ssvAnnotateSubjectGRanges(np_grs, olap_gr)
# overriding list names is an option
ssvAnnotateSubjectGRanges(np_grs, olap_gr, LETTERS[1:3])
# GRangeList are handled like a standard list
ssvAnnotateSubjectGRanges(GRangesList(np_grs), olap_gr, LETTERS[1:3])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{reorder_clusters_hclust}
\alias{reorder_clusters_hclust}
\title{reorder_clusters_hclust}
\usage{
reorder_clusters_hclust(
  clust_dt,
  hclust_result = NULL,
  row_ = "id",
  column_ = "x",
  fill_ = "y",
  facet_ = "sample",
  cluster_ = "cluster_id",
  reapply_cluster_names = TRUE,
  return_hclust = FALSE
)
}
\arguments{
\item{clust_dt}{data.table output from \code{\link{ssvSignalClustering}}}

\item{hclust_result}{hclust result returned by a previous call of this function with
identical paramters when return_hclust = TRUE.}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{column_}{varaible mapped to column, likely bp position for ngs data.
Default is "x" and works with ssvFetch* output.}

\item{fill_}{numeric variable to map to fill. Default is "y" and works with
ssvFetch* output.}

\item{facet_}{variable name to facet horizontally by. Default is "sample" and
works with ssvFetch* output. Set to "" if data is not facetted.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{reapply_cluster_names}{If TRUE, clusters will be renamed according to
new order instead of their original names. Default is TRUE.}

\item{return_hclust}{If TRUE, return the result of hclust instead of the
reordered clustering data.table. Default is FALSE.  Ignored if
hclust_result is supplied.}
}
\value{
data.table as output from \code{\link{ssvSignalClustering}}
}
\description{
Applies hierarchical clustering to centroids of clusters to reorder.
}
\examples{
data(CTCF_in_10a_profiles_dt)
clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 10)
new_dt = reorder_clusters_hclust(clust_dt)
cowplot::plot_grid(
    ssvSignalHeatmap(clust_dt),
    ssvSignalHeatmap(new_dt)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bam.R
\name{.expand_cigar_dt_recursive}
\alias{.expand_cigar_dt_recursive}
\title{Expand intermediate bam fetch by cigar codes}
\usage{
.expand_cigar_dt_recursive(cigar_dt)
}
\arguments{
\item{cigar_dt}{data.table with 5 required named columns in any order.
c("which_label", "seqnames", "strand", "start", "cigar")}
}
\value{
data.table with cigar entries expanded
}
\description{
see \href{https://samtools.github.io/hts-specs/SAMv1.pdf}{sam specs} for
cigar details
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_operations.R
\name{applySpline}
\alias{applySpline}
\title{applies a spline smoothing to a tidy data.table containing x and y values.}
\usage{
applySpline(
  dt,
  n,
  x_ = "x",
  y_ = "y",
  by_ = c("id", "sample"),
  splineFun = stats::spline
)
}
\arguments{
\item{dt}{a tidy data.table containing two-dimensional data}

\item{n}{the number of interpolation points to use per input point, see
\code{?spline}.  n must be > 1.}

\item{x_}{the variable name of the x-values}

\item{y_}{the variable name of the y-values}

\item{by_}{optionally, any variables that provide grouping to the data.
default is none. see details.}

\item{splineFun}{a function that accepts x, y, and n as arguments and
returns a list of length 2 with named elements x and y.
\code{stats::spline} by default.
see \code{stats::spline} for details.}
}
\value{
a newly derived data.table that is \code{n} times longer than
original.
}
\description{
\code{applySpline} Is intended for two-dimensional tidy data.tables, as
retured by \code{ssvFetchBigwig}
}
\details{
by_ is quite powerful.  If \code{by_ = c('gene_id', 'sample_id')},
splines
will be calculated individually for each gene in each sample. alternatively
if \code{by_ = c('gene_id')}
}
\examples{
data(CTCF_in_10a_profiles_dt)
#data may be blockier than we'd like
ggplot(CTCF_in_10a_profiles_dt[, list(y = mean(y)), by = list(sample, x)]) +
    geom_line(aes(x = x, y = y, color = sample))

#can be smoothed by applying a spline  (think twice about doing so,
#it may look prettier but may also be deceptive or misleading)

splined_smooth = applySpline(CTCF_in_10a_profiles_dt, n = 10,
    y_ = 'y', by_ = c('id', 'sample'))
ggplot(splined_smooth[, list(y = mean(y)), by = list(sample, x)]) +
    geom_line(aes(x = x, y = y, color = sample))
}
\seealso{
\code{\link{ssvFetchBigwig}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_signal.R
\name{viewGRangesWinSummary_dt}
\alias{viewGRangesWinSummary_dt}
\title{Summarizes signal in bins.  The same number of bins per region in qgr is used
and widths can vary in qgr, in contrast to
\code{\link{viewGRangesWinSample_dt}} where width must be constant across
regions.}
\usage{
viewGRangesWinSummary_dt(
  score_gr,
  qgr,
  n_tiles = 100,
  attrib_var = "score",
  attrib_type = NULL,
  fill_value = 0,
  anchor = c("center", "center_unstranded", "left", "left_unstranded")[1],
  summary_FUN = stats::weighted.mean
)
}
\arguments{
\item{score_gr}{GRanges with a "score" metadata column.}

\item{qgr}{regions to view by window.}

\item{n_tiles}{numeric >= 1, the number of tiles to use for every region in
qgr.}

\item{attrib_var}{character name of attribute to pull data from. Default is
"score", compatible with with bigWigs or bam coverage.}

\item{attrib_type}{one of NULL, qualitative or quantitative.  If NULL will
attempt to guess by casting attrib_var attribute to character or factor.
Default is NULL.}

\item{fill_value}{numeric or character value to use where queried regions are
empty.  Default is 0 and appropriate for both calculated coverage and
bedgraph/bigwig like files.  Will automatically switch to "MISSING" if data
is guessed to be qualitative.}

\item{anchor}{character. controls how x value is derived from position for
each region in qgr.  0 may be the left side or center.  If not unstranded,
x coordinates are flipped for (-) strand. One of c("center",
"center_unstranded", "left", "left_unstranded"). Default is "center".}

\item{summary_FUN}{function. used to aggregate score by tile.  must accept
x=score and w=width numeric vectors as only arguments. default is
weighted.mean.  limma::weighted.median is a good alternative.}
}
\value{
data.table that is GRanges compatible
}
\description{
This function is most appropriate where features are expected to vary greatly
in size and feature boundaries are important, ie. gene bodies, enhancers or
TADs.
}
\details{
Columns in output data.table are: standard GRanges columns: seqnames, start,
end, width, strand id - matched to names(score_gr). if names(score_gr) is
missing, added as seq_along(score_gr). y - value of score from score_gr x -
relative bp position
}
\examples{
data(CTCF_in_10a_overlaps_gr)
bam_file = system.file("extdata/test.bam",
    package = "seqsetvis")
qgr = CTCF_in_10a_overlaps_gr[1:5]
# unlike viewGRangesWinSample_dt, width is not fixed
# qgr = GenomicRanges::resize(qgr, width = 500, fix = "center")
bam_gr = seqsetvis:::fetchBam(bam_file, qgr)
bam_dt = viewGRangesWinSummary_dt(bam_gr, qgr, 50)

if(Sys.info()['sysname'] != "Windows"){
    bw_file = system.file("extdata/MCF10A_CTCF_FE_random100.bw",
        package = "seqsetvis")
    bw_gr = rtracklayer::import.bw(bw_file, which = qgr)
    bw_dt = viewGRangesWinSummary_dt(bw_gr, qgr, 50)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_feature_plots.R
\name{ssvFeatureEuler}
\alias{ssvFeatureEuler}
\title{Try to load a bed-like file and convert it to a GRanges object}
\usage{
ssvFeatureEuler(
  object,
  line_width = 2,
  shape = c("circle", "ellipse")[1],
  n_points = 200,
  fill_alpha = 0.3,
  line_alpha = 1,
  circle_colors = NULL,
  return_data = FALSE
)
}
\arguments{
\item{object}{A membership table}

\item{line_width}{numeric, passed to size aesthetic to control line width}

\item{shape}{shape argument passed to eulerr::euler}

\item{n_points}{number of points to use for drawing ellipses, passed to
eulerr:::ellipse}

\item{fill_alpha}{numeric value from 0 to 1. Alpha value for circle fill}

\item{line_alpha}{numeric value from 0 to 1. Alpha value for circle line}

\item{circle_colors}{colors to choose from for circles.  passed to ggplot2
color scales.}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and
is instead the data used to generate that plot. Default is FALSE.}
}
\value{
ggplot of venneuler results
}
\description{
Try to load a bed-like file and convert it to a GRanges object
}
\examples{
data(CTCF_in_10a_overlaps_gr)
ssvFeatureEuler(list(1:3, 2:6))
ssvFeatureEuler(CTCF_in_10a_overlaps_gr)
ssvFeatureEuler(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_CTCF.R
\docType{data}
\name{test_peaks}
\alias{test_peaks}
\title{4 random peaks for single-end data and 4 control regions 30kb downstream from
each peak.}
\format{
GRanges length 8
}
\value{
GRanges length 8
}
\description{
matches \code{system.file("extdata/test_peaks.bam", package =
"seqsetvis")}
}
\details{
this is included only for testing ssvFetchBam functions.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_exon_signal.R
\name{collapse_gr}
\alias{collapse_gr}
\title{collapse_gr}
\usage{
collapse_gr(genome_gr)
}
\arguments{
\item{genome_gr}{a GRanges of regions on a single chromosome.  Regions
are intended to be non-contiguous and may even overlap.}
}
\value{
a new GRanges object with same mcols as input with all intervals
starting at 1 and no empty space between syntenic regions.
}
\description{
collapse non-contiguous regions (i.e. exons) into a contiguous coordinate
starting at 1.  this is strand sensitive and intended for use with
all exons of a single gene.
}
\examples{
library(data.table)
library(GenomicRanges)
dev_dat = data.table(seqnames = "chrTest",
                     transcript_id = c(1, 1,  2, 2, 3, 3, 3),
                     start = c(5,  30,  8, 30, 2, 30, 40),
                     end = c(10, 35, 15, 38, 7, 35, 45),
                     strand = "+")

genome_gr = GRanges(dev_dat)
collapse_gr(genome_gr)

neg_gr = genome_gr
strand(neg_gr) = "-"
collapse_gr(neg_gr)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fragLen.R
\name{fragLen_fromMacs2Xls}
\alias{fragLen_fromMacs2Xls}
\title{parse fragLen from MACS2 output}
\usage{
fragLen_fromMacs2Xls(macs2xls_file)
}
\arguments{
\item{macs2xls_file}{character.  an xls file output by MACS2 to parse
frag length from}
}
\value{
numeric fragment length
}
\description{
parse fragLen from MACS2 output
}
\examples{
xls_file = system.file("extdata/test_peaks.xls",
    package = "seqsetvis")
fragLen_fromMacs2Xls(xls_file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_helpers.R
\name{applyMovingAverage}
\alias{applyMovingAverage}
\title{applyMovingAverage}
\usage{
applyMovingAverage(
  dt,
  n,
  centered = TRUE,
  x_ = "x",
  y_ = "y",
  by_ = c("id", "sample"),
  maFun = movingAverage
)
}
\arguments{
\item{dt}{a tidy data.table containing two-dimensional data}

\item{n}{the number of samples centered: if FALSE, then average}

\item{centered}{current sample and previous (n-1) samples if TRUE, then
average symmetrically in past and future. (If n is even, use one more
sample from future.)}

\item{x_}{the variable name of the x-values}

\item{y_}{the variable name of the y-values}

\item{by_}{optionally, any variables that provide grouping to the data.
default is none. see details.}

\item{maFun}{a function that accepts x, y, and n as arguments and
returns a list of length 2 with named elements x and y.}
}
\value{
a newly derived data.table where a movingAverage has been applied.
}
\description{
http://www.cookbook-r.com/Manipulating_data/Calculating_a_moving_average/
}
\examples{
data(CTCF_in_10a_profiles_dt)
agg_dt = CTCF_in_10a_profiles_dt[, list(y = mean(y)), by = list(sample, x)]
ggplot(agg_dt) +
    geom_line(aes(x = x, y = y, color = sample))

ma_smooth = applyMovingAverage(agg_dt, n = 5,
    y_ = 'y', by_ = c('sample'))
ggplot(ma_smooth) +
    geom_line(aes(x = x, y = y, color = sample))

ma_smooth$method = "moving_average"
agg_dt$method = "none"
ggplot(rbind(ma_smooth, agg_dt)) +
    geom_line(aes(x = x, y = y, color = method)) +
    facet_wrap(~sample)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_chromHMM.R
\docType{data}
\name{chromHMM_demo_overlaps_gr}
\alias{chromHMM_demo_overlaps_gr}
\title{overlap of MCF10A CTCF with MCF7 chromHMM states, hg38.}
\format{
a GRanges object of length 98
with 10 logical metadata columns, 1 per state.
}
\value{
a GRanges object of length 98
with 10 logical metadata columns, 1 per state.
}
\description{
overlap of MCF10A CTCF with MCF7 chromHMM states, hg38.
}
\details{
part of \code{\link{chromHMM_demo_data}}

the result of ssvOverlapIntervalSets() on MCF10A CTCF peaks and
MCF7 chromHMM states with use_first = TRUE

first (the MCF10A peaks) and no_hit columns have been removed
each remaining column represents MCF10A peaks overlapping with a state.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_plots.R
\name{make_clustering_matrix}
\alias{make_clustering_matrix}
\title{make_clustering_matrix}
\usage{
make_clustering_matrix(
  tidy_dt,
  row_ = "id",
  column_ = "x",
  fill_ = "y",
  facet_ = "sample",
  max_rows = 500,
  max_cols = 100,
  clustering_col_min = -Inf,
  clustering_col_max = Inf,
  dcast_fill = NA,
  fun.aggregate = "mean"
)
}
\arguments{
\item{tidy_dt}{the tidy data.table to covert to a wide matrix.  Must have
entries for variables specified by row_, column_, fill_, and facet_.}

\item{row_}{variable name mapped to row, likely peak id or gene name for ngs
data}

\item{column_}{varaible mapped to column, likely bp position for ngs data}

\item{fill_}{numeric variable to map to fill}

\item{facet_}{variable name to facet horizontally by}

\item{max_rows}{for speed rows are sampled to 500 by default, use Inf to plot
full data}

\item{max_cols}{for speed columns are sampled to 100 by default, use Inf to
plot full data}

\item{clustering_col_min}{numeric minimum for col range considered when
clustering, default in -Inf}

\item{clustering_col_max}{numeric maximum for col range considered when
clustering, default in Inf}

\item{dcast_fill}{value to supply to dcast fill argument. default is NA.}

\item{fun.aggregate}{Function to aggregate when multiple values present for
facet_, row_, and column_. The
function should accept a single vector argument or be a character string
naming such a function.}
}
\value{
A wide matrix version of input tidy data.table
}
\description{
Create a wide matrix from a tidy data.table more suitable for clustering
methods
}
\examples{
data(CTCF_in_10a_profiles_dt)
mat = make_clustering_matrix(CTCF_in_10a_profiles_dt)
mat[1:5, 1:5]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{clusteringKmeans}
\alias{clusteringKmeans}
\title{perform kmeans clustering on matrix rows and return reordered matrix along
with order matched cluster assignments. clusters are sorted using hclust on
centers}
\usage{
clusteringKmeans(mat, nclust, centroids = NULL, iter.max = 30)
}
\arguments{
\item{mat}{numeric matrix to cluster.}

\item{nclust}{the number of clusters.}

\item{centroids}{optional matrix with same columns as mat and one centroid
per row to base clusters off of.  Overrides any setting to nclust. Default
of NULL results in randomly initialized k-means.}

\item{iter.max}{Number of max iterations to allow for k-means. Default is 30.}
}
\value{
data.table with group__ variable indicating cluster membership and
id__ variable that is a factor indicating order based on within cluster
similarity
}
\description{
perform kmeans clustering on matrix rows and return reordered matrix along
with order matched cluster assignments. clusters are sorted using hclust on
centers
}
\examples{
data(CTCF_in_10a_profiles_dt)
dt = data.table::copy(CTCF_in_10a_profiles_dt)
mat = data.table::dcast(dt, id ~ sample + x, value.var = "y" )
rn = mat$id
mat = as.matrix(mat[,-1])
rownames(mat) = rn
clust_dt = clusteringKmeans(mat, nclust = 3)
dt = merge(dt, clust_dt[, .(id = id__, group = group__)])
dt$id = factor(dt$id, levels = clust_dt$id)
dt[order(id)]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_info_chromHMM.R
\docType{data}
\name{chromHMM_demo_bw_states_gr}
\alias{chromHMM_demo_bw_states_gr}
\title{MCF10A CTCF profiles at 20 windows per chromHMM state, hg38.}
\format{
a GRanges object of length 4000
with 5 metadata columns sufficient for use with ggplot2
}
\value{
a GRanges object of length 4000
with 5 metadata columns sufficient for use with ggplot2
}
\description{
MCF10A CTCF profiles at 20 windows per chromHMM state, hg38.
}
\details{
part of \code{\link{chromHMM_demo_data}}

the result of ssvFetchBigwig() on the MCF10A_CTCF_FE.bw
near 20 randomly selected windows per chromHMM state.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{copy_clust_info}
\alias{copy_clust_info}
\title{copy_clust_info}
\usage{
copy_clust_info(target, to_copy, row_ = "id", cluster_ = "cluster_id")
}
\arguments{
\item{target}{A data.table or GRanges returned from ssvFetch*, the target to
which cluster info will be added.}

\item{to_copy}{A data.table or GRanges returned from ssvSignalClustering,
from which to copy cluster if.}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}
}
\value{
data.table or GRanges (whichever target is) containing row order and
cluster assignment derived from to_copy. Suitable for ssvSignalHeatmap and
related functions.
}
\description{
copy_clust_info
}
\examples{
data(CTCF_in_10a_narrowPeak_grs)
data(CTCF_in_10a_overlaps_gr)
data(CTCF_in_10a_profiles_dt)
#this takes cluster info from signal and applies to peak hits to
#create a heatmap of peak hits clustered by signal.
clust_dt1 = ssvSignalClustering(CTCF_in_10a_profiles_dt)
peak_hit_gr = ssvFetchGRanges(
  CTCF_in_10a_narrowPeak_grs,
  qgr = CTCF_in_10a_overlaps_gr
)
peak_hit_gr.clust = copy_clust_info(peak_hit_gr, clust_dt1)
peak_hit_gr.clust$hit = peak_hit_gr.clust$y > 0
ssvSignalHeatmap(peak_hit_gr.clust, fill_ = "hit") +
  scale_fill_manual(values = c("FALSE" = "gray90", "TRUE" = "black"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/misc_helpers.R
\name{get_mapped_reads}
\alias{get_mapped_reads}
\title{get_mapped_reads}
\usage{
get_mapped_reads(bam_files)
}
\arguments{
\item{bam_files}{Path to 1 or more bam files.  Must be indexed.}
}
\value{
the total mapped reads in each bam file as a named numeric vector.
}
\description{
get_mapped_reads
}
\examples{
bam_file = system.file("extdata/test.bam", package = "seqsetvis", mustWork = TRUE)
get_mapped_reads(bam_file)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_bamPE.R
\name{.rm_dupesPE}
\alias{.rm_dupesPE}
\title{Remove duplicate paired-end reads based on start and end position.  This is
an over-simplification.  For better duplicate handling, duplicates must be
marked in bam and flag passed to fetchBamPE() ... for ScanBamParam}
\usage{
.rm_dupesPE(reads_dt, max_dupes)
}
\arguments{
\item{reads_dt}{data.table of reads as loaded by fetchBamPE}

\item{max_dupes}{maximum allowed positional duplicates}
}
\value{
reads_dt with duplicated reads over max_dupes removed
}
\description{
flag = scanBamFlag(isDuplicate = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_plots.R
\name{ssvSignalLineplotAgg}
\alias{ssvSignalLineplotAgg}
\title{aggregate line signals in a single line plot}
\usage{
ssvSignalLineplotAgg(
  bw_data,
  x_ = "x",
  y_ = "y",
  sample_ = "sample",
  color_ = sample_,
  group_ = sample_,
  agg_fun = mean,
  spline_n = NULL,
  return_data = FALSE
)
}
\arguments{
\item{bw_data}{a GRanges or data.table of bigwig signal.
As returned from \code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}}

\item{x_}{variable name mapped to x aesthetic, x by default.}

\item{y_}{variable name mapped to y aesthetic, y by default.}

\item{sample_}{variable name, along with region_ used to group by default,}

\item{color_}{variable name mapped to color aesthetic, sample_ by default.
change group_ to override.}

\item{group_}{group aesthetic keeps lines of geom_path from mis-connecting.
Most useful if you need to supply a
variable to later facet upon. Defaults to value of sample_.}

\item{agg_fun}{the aggregation function to apply by sample_ and x_,
default is mean}

\item{spline_n}{if not NULL, applySpline will be called with n = spline_n.
default is NULL.}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and
is instead the data used to generate that plot. Default is FALSE.}
}
\value{
ggplot of signal aggregated with agg_fun() by sample.
}
\description{
aggregate line signals in a single line plot
}
\examples{
data(CTCF_in_10a_profiles_gr)
bw_gr = CTCF_in_10a_profiles_gr
ssvSignalLineplotAgg(bw_gr) +
    labs(title = "agg regions by sample.")
ssvSignalLineplotAgg(CTCF_in_10a_profiles_gr, spline_n = 10) +
    labs(title = "agg regions by sample, with spline smoothing.")
ssvSignalLineplotAgg(subset(bw_gr, bw_gr$id \%in\% seq_len(10)),
    sample_ = "id", color_ = "id") +
    labs(title = "agg samples by region id (weird)")
ssvSignalLineplotAgg(subset(bw_gr, bw_gr$id \%in\% seq_len(10)), sample_ = "id",
    color_ = "id", spline_n = 10) +
    labs(title = "agg samples by region id (weird), with spline smoothing")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_fetch_signal.R
\name{shift_anchor}
\alias{shift_anchor}
\title{orients the relative position of x's zero value and
extends ranges to be contiguous}
\usage{
shift_anchor(score_dt, window_size, anchor)
}
\arguments{
\item{score_dt}{data.table, GRanges() sufficient}

\item{window_size}{numeric, window size used to generate score_dt}

\item{anchor}{character, one of c("center", "center_unstranded",
"left", "left_unstranded")}
}
\value{
score_dt with x values shifted appropriately and start and end
extended to make ranges contiguous
}
\description{
orients the relative position of x's zero value and
extends ranges to be contiguous
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_operations.R
\name{centerFixedSizeGRanges}
\alias{centerFixedSizeGRanges}
\title{Transforms set of GRanges to all have the same size.}
\usage{
centerFixedSizeGRanges(grs, fixed_size = 2000)
}
\arguments{
\item{grs}{Set of GRanges with incosistent and/or incorrect size}

\item{fixed_size}{The final width of each GRange returned.}
}
\value{
Set of GRanges after resizing all input GRanges, either shortened
or lengthened as required to match \code{fixed_size}
}
\description{
\code{centerFixedSizeGRanges} First calculates the central coordinate of each
GRange in \code{grs} and extends in both direction by half of
\code{fixed_size}
}
\examples{
library(GenomicRanges)
grs = GRanges("chr1", IRanges(1:10+100, 1:10*3+100))
centered_grs = centerFixedSizeGRanges(grs, 10)
width(centered_grs)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_clusteringKmeans.R
\name{reorder_clusters_manual}
\alias{reorder_clusters_manual}
\title{reorder_clusters_manual}
\usage{
reorder_clusters_manual(
  clust_dt,
  manual_order,
  row_ = "id",
  cluster_ = "cluster_id",
  reapply_cluster_names = TRUE
)
}
\arguments{
\item{clust_dt}{data.table output from \code{\link{ssvSignalClustering}}}

\item{manual_order}{New order for clusters  Does not need to include all
clusters.  Any colors not included will be at the bottom in their original
order.}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{reapply_cluster_names}{If TRUE, clusters will be renamed according to
new order instead of their original names. Default is TRUE.}
}
\value{
data.table as output from \code{\link{ssvSignalClustering}}
}
\description{
Manually applies a new order (top to bottom) for cluster using the result of
ssvSignalClustering.
}
\examples{
data(CTCF_in_10a_profiles_dt)
clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_dt, nclust = 3)
new_dt = reorder_clusters_manual(clust_dt = clust_dt, manual_order = 2)
cowplot::plot_grid(
    ssvSignalHeatmap(clust_dt),
    ssvSignalHeatmap(new_dt)
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_signal_plots.R
\name{ssvSignalHeatmap}
\alias{ssvSignalHeatmap}
\title{heatmap style representation of membership table. instead of clustering, each
column is sorted starting from the left.}
\usage{
ssvSignalHeatmap(
  bw_data,
  nclust = 6,
  perform_clustering = c("auto", "yes", "no")[1],
  row_ = "id",
  column_ = "x",
  fill_ = "y",
  facet_ = "sample",
  cluster_ = "cluster_id",
  max_rows = 500,
  max_cols = 100,
  fill_limits = NULL,
  clustering_col_min = -Inf,
  clustering_col_max = Inf,
  within_order_strategy = c("hclust", "sort")[2],
  dcast_fill = NA,
  return_data = FALSE,
  show_cluster_bars = TRUE,
  rect_colors = c("black", "gray"),
  text_colors = rev(rect_colors),
  show_labels = TRUE,
  label_angle = 0,
  fun.aggregate = "mean"
)
}
\arguments{
\item{bw_data}{a GRanges or data.table of bigwig signal. As returned from
\code{\link{ssvFetchBam}} and \code{\link{ssvFetchBigwig}}}

\item{nclust}{number of clusters}

\item{perform_clustering}{should clustering be done? default is auto. auto
considers if row_ has been ordered by being a factor and if cluster_ is a
numeric.}

\item{row_}{variable name mapped to row, likely id or gene name for ngs data.
Default is "id" and works with ssvFetch* output.}

\item{column_}{varaible mapped to column, likely bp position for ngs data.
Default is "x" and works with ssvFetch* output.}

\item{fill_}{numeric variable to map to fill. Default is "y" and works with
ssvFetch* output.}

\item{facet_}{variable name to facet horizontally by. Default is "sample" and
works with ssvFetch* output. Set to "" if data is not facetted.}

\item{cluster_}{variable name to use for cluster info. Default is
"cluster_id".}

\item{max_rows}{for speed rows are sampled to 500 by default, use Inf to plot
full data}

\item{max_cols}{for speed columns are sampled to 100 by default, use Inf to
plot full data}

\item{fill_limits}{limits for fill legend.  values will be cropped to this
range if set.  Default of NULL uses natural range of fill_.}

\item{clustering_col_min}{numeric minimum for col range considered when
clustering, default in -Inf}

\item{clustering_col_max}{numeric maximum for col range considered when
clustering, default in Inf}

\item{within_order_strategy}{one of "hclust" or "sort".  if hclust,
hierarchical clustering will be used. if sort, a simple decreasing sort of
rosSums.}

\item{dcast_fill}{value to supply to dcast fill argument. default is NA.}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and is
instead the data used to generate that plot. Default is FALSE.}

\item{show_cluster_bars}{if TRUE, show bars indicating cluster membership.}

\item{rect_colors}{colors of rectangle fill, repeat to match number of
clusters. Default is c("black", "gray").}

\item{text_colors}{colors of text, repeat to match number of clusters.
Default is reverse of rect_colors.}

\item{show_labels}{logical, shoud rectangles be labelled with cluster
identity.  Default is TRUE.}

\item{label_angle}{angle to add clusters labels at.  Default is 0, which is
horizontal.}

\item{fun.aggregate}{Function to aggregate when multiple values present for
facet_, row_, and column_. Affects both clustering and plotting. The
function should accept a single vector argument or be a character string
naming such a function.}
}
\value{
ggplot heatmap of signal profiles, facetted by sample
}
\description{
See \code{\link{ssvSignalHeatmap.ClusterBars}} for an alternative with more
control over where the cluster bars appear.
}
\examples{
data(CTCF_in_10a_profiles_gr)

#the simplest use
ssvSignalHeatmap(CTCF_in_10a_profiles_gr)
ssvSignalHeatmap(CTCF_in_10a_profiles_gr, show_cluster_bars = FALSE)

#clustering can be done manually beforehand
clust_dt = ssvSignalClustering(CTCF_in_10a_profiles_gr, nclust = 3)
ssvSignalHeatmap(clust_dt)

ssvSignalHeatmap(clust_dt, max_rows = 20, max_cols = 7)

# aggregation occurs when facet_ is shared by multiple samples
prof_gr = CTCF_in_10a_profiles_gr
prof_gr$mark = "CTCF"

# by default, aggregation is done by mean
ssvSignalHeatmap(prof_gr, facet_ = "mark")

# Using fun.aggregate we can control aggregation behavior
# 'fun.aggregate = mean' preserves the default behavior
ssvSignalHeatmap(prof_gr, facet_ = "mark",
  fun.aggregate = mean)


# maximum and minimum may be useful to plot
ssvSignalHeatmap(prof_gr, facet_ = "mark",
  fun.aggregate = max)
ssvSignalHeatmap(prof_gr, facet_ = "mark",
  fun.aggregate = min)

# How many samples have at least a signal value of 10 at each position?
ssvSignalHeatmap(prof_gr, facet_ = "mark",
  fun.aggregate = function(x)sum(as.numeric(x > 10)))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/functions_feature_plots.R
\name{ssvFeaturePie}
\alias{ssvFeaturePie}
\title{ssvFeaturePie}
\usage{
ssvFeaturePie(object, slice_colors = NULL, return_data = FALSE)
}
\arguments{
\item{object}{object that ssvMakeMembTable can convert to logical matrix
membership}

\item{slice_colors}{colors to use for pie slices}

\item{return_data}{logical.  If TRUE, return value is no longer ggplot and is
instead the data used to generate that plot. Default is FALSE.}
}
\value{
ggplot pie graph of set sizes
}
\description{
Generate a ggplot pie plot of set sizes.
}
\examples{
data(CTCF_in_10a_overlaps_gr)
ssvFeaturePie(list(1:3, 2:6))
ssvFeaturePie(CTCF_in_10a_overlaps_gr)
ssvFeaturePie(S4Vectors::mcols(CTCF_in_10a_overlaps_gr)[,2:3])
}
